/////////////////////////////////////////////////////////////////////
//
//            X   X           X
//           XXX XX         XX
//          XXXXXXXX      XXX
//         XX X XXXXXXXXXXX
//        XXXXX XXXXXXXXX
//       XXXXX XXXXXXXXXX
//            XXX XXX XXX
//           XXX XX   XX
//           X   X     X
//
//    Copyright (C) 2003-2023  Ron Jakl
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
/////////////////////////////////////////////////////////////////////

#include <QStringList>
#include <QRegExp>
#include <iostream>
#include <string>
#include <stack>
#include <math.h>
#include <assert.h>

#include "expression.h"

TExpression::TExpression(void)
{
}

TExpression::~TExpression(void)
{
}

QString TExpression::Variable_Name(
	const int							index) const
{
	assert(static_cast<unsigned int>(index) < d_variables.size());
	
	return d_variables[index].name;
}

double TExpression::Variable_Value(
	const int							index) const
{
	assert(static_cast<unsigned int>(index) < d_variables.size());
	
	return d_variables[index].value;
}

double TExpression::Variable_Value(
	const QString						&name,
	bool								*exists) const
{
	std::vector<TExpression::TVariable>::const_iterator iter;
	
	for(iter = d_variables.begin();iter != d_variables.end();++iter)
	{
		if(name == (*iter).name)
		{
			if(exists)
			{
				*exists = true;
			}
			
			return (*iter).value;
		}
	}
	
	if(exists)
	{
		*exists = false;
	}
	
	return 0.0;
}

TExpression::TVariable TExpression::Variable(
	const int							index) const
{
	assert(static_cast<unsigned int>(index) < d_variables.size());
		
	return d_variables[index];
}

double TExpression::Evaluate(
	const QString						&expression,
	bool								*valid) const
{
	QString								expanded_expression(expression);
	QRegExp								regular_expression;
	std::vector<TExpression::TVariable>::const_iterator iter;

	
	if(valid)
	{
		*valid = false;
	}
	
	if(expression.size() == 0)
	{
		return 0.0;
	}
	
	expanded_expression = expression.simplified().toUpper();
//	expanded_expression = expanded_expression.trimmed();
	expanded_expression.remove(' ');
	
	// substitue variable values
	// using regexp to prevent sub-word replacement (i.e. variable 'sq' replaced in 'sqrt')
	
	for(iter = d_variables.begin();iter != d_variables.end();++iter)
	{
		regular_expression.setPattern(QString("(\\b%1\\b)").arg((*iter).name.toUpper()));
		
		if((*iter).value < 0)
		{
			expanded_expression.replace(regular_expression,QString("(%1)").arg((*iter).value,0,'f',16));
		}
		else
		{
			expanded_expression.replace(regular_expression,QString("%1").arg((*iter).value,0,'f',16));
		}
	}
		
	return this->Process_Expression(expanded_expression,valid);
}

bool TExpression::Add_Variable(
	const TExpression::TVariable		&variable)
{
	std::vector<TExpression::TVariable>::iterator iter;
	
	if(variable.name.size() == 0)
	{
		return false;
	}
	
	for(iter = d_variables.begin();iter != d_variables.end();++iter)
	{
		if(variable.name == (*iter).name)
		{
			(*iter).value = variable.value;			// allow updating of existing variables
			return true;
		}
	}
	
	// add variables so that longest names are first
	for(iter = d_variables.begin();iter != d_variables.end();++iter)
	{
		if(variable.name.length() > (*iter).name.length())
		{
			d_variables.insert(iter,variable);
			return true;
		}
	}
	
	d_variables.push_back(variable);
	
	return true;
}

bool TExpression::Add_Variable(
	const QString						&name,
	const double						&value)
{
	TExpression::TVariable				var;
	
	var.name = name;
	var.value = value;
	
	return this->Add_Variable(var);
}

// Processing done by converting InFix expression to PostFix
// Parenthesis and functions handled recursively
double TExpression::Process_Expression(
	const QString						&input,
	bool								*valid) const
{
	std::stack<double>					numbers;
	std::stack<int>						operators;
	int									character;
	int									cntr;
	int									index;
	int									index_end;
	double								value;
	QString								sub_string;
	bool								conversion_valid;
	bool								use_parenthesis_value;
	double								parenthesis_value(0.0);
	
	if(valid)
	{
		*valid = false;
	}
		
	index = 0;
	use_parenthesis_value = false;
	
	for(cntr = 0;cntr < input.length();++cntr)
	{
		character = input[cntr].toLatin1();
		
		switch(character)
		{
			case '*':
			case '/':
			case '+':
			case '-':
			case '^':	
				
				if(cntr != 0)
				{					
					if(use_parenthesis_value)
					{
						value = parenthesis_value;
						use_parenthesis_value = false;
					}
					else
					{
						value = input.mid(index,cntr-index).toDouble(&conversion_valid);
				
						if(!conversion_valid)
						{
							return 0.0;
						}
					}
					
					index = cntr+1;
					numbers.push(value);
					
					while(operators.size() && this->Operator_Precidence(operators.top()) >= this->Operator_Precidence(character))
					{
						value = this->Apply_Expression(&numbers,&operators);
						numbers.push(value);
					}
					
					operators.push(character);
				}
				
				break;				
				
			case '(':
				index_end = this->Find_Closing_Brace(input,cntr);
				
				sub_string = input.mid(cntr+1,index_end-cntr-1);
				
				parenthesis_value = Process_Expression(sub_string,&conversion_valid);
				
				if(!conversion_valid)
				{
					return 0.0;
				}
				
				use_parenthesis_value = true;
				cntr = index_end;
				
				break;
				
			case ')':
				return 0.0;
				
			case '0':
			case '1':
			case '2':
			case '3':
			case '4':
			case '5':
			case '6':
			case '7':
			case '8':
			case '9':
			case '.':
				break;
				
			default:
				// check for function
				sub_string = input.mid(cntr);
				
				parenthesis_value = Process_Function(sub_string,&index_end,&conversion_valid);
								
				if(!conversion_valid)
				{
					return 0.0;
				}
				
				use_parenthesis_value = true;
				cntr += index_end;

				break;
		}
	}
	
	if(use_parenthesis_value)
	{
		numbers.push(parenthesis_value);
	}
	else
	{
		if(index < input.length())
		{
			numbers.push(input.mid(index,input.length()-index).toDouble(&conversion_valid));
			
			if(!conversion_valid)
			{
				return 0.0;
			}
		}
	}
	
	if(numbers.size() != (operators.size()+1))
	{
		return 0.0;
	}
	
	value = numbers.top();	// default value if no operators are found
	
	while(operators.size())
	{
		value = Apply_Expression(&numbers,&operators);
		numbers.push(value);
	}
	
	if(valid)
	{
		*valid = true;
	}
	
	return value;
}

double TExpression::Apply_Expression(
	std::stack<double>					*number_stack,
	std::stack<int>						*oper_stack) const
{
	int									oper;
	double								number_r;
	double								number_l;
	double								result(0);
		
	assert(number_stack->size() > 1);
	assert(oper_stack->size());
	
	oper = oper_stack->top();
	oper_stack->pop();
	
	number_r = number_stack->top();
	number_stack->pop();
	
	number_l = number_stack->top();
	number_stack->pop();
	
	switch(oper)
	{
		case '*':			
			result = number_l * number_r;
			break;
			
		case '/':
			result = number_l / number_r;
			break;
			
		case '+':
			result = number_l + number_r;
			break;
			
		case '-':
			result = number_l - number_r;
			break;

		case '^':
			result = pow(number_l,number_r);
			break;
	}
	return result;
}

int TExpression::Operator_Precidence(
	const int							oper) const
{
	switch(oper)
	{
		case '^':
			return 3;
			
		case '*':
		case '/':
			return 2;
			
		case '+':
		case '-':
			return 1;
			
		default:
			break;
	}
	
	return 0;
}

double TExpression::Process_Function(
	const QString						&string,
	int									*function_length,
	bool								*valid) const
{
	QString								fn_string;
	QStringList							fn_values;
	int									fn_name_length;
	double								value;
	double								value_sum;
	int									end;
	int									cntr;
	
	assert(function_length);
	assert(valid);
		
	*valid = false;
	
	if(string.startsWith("SUMSQ(",Qt::CaseInsensitive))
	{
		fn_name_length = QString("SUMSQ(").length();
		
		end = this->Find_Closing_Brace(string,fn_name_length);
		
		if(end < 0)
		{
			return 0.0;
		}
		
		*function_length = end;
		
		fn_string = string.mid(fn_name_length,end - fn_name_length);
		
		fn_values = fn_string.split(',');
		
		value_sum = 0.0;
		for(cntr = 0;cntr < fn_values.size();++cntr)
		{
			value = this->Process_Expression(fn_values[cntr],valid);
			
			if(!valid)
			{
				return 0.0;
			}
			
			value_sum += (value * value);
		}
				
		value_sum = sqrt(value_sum);
		return value_sum;
	}
	if(string.startsWith("SQRT(",Qt::CaseInsensitive))
	{
		fn_name_length = QString("SQRT(").length();

		end = this->Find_Closing_Brace(string,fn_name_length);
		
		if(end < 0)
		{
			return 0.0;
		}
		
		*function_length = end;
		
		fn_string = string.mid(fn_name_length,end - fn_name_length);
		
		value = this->Process_Expression(fn_string,valid);
		
		if(!valid)
		{
			return 0.0;
		}
		
		value = sqrt(value);
		return value;
	}
	
	return 0.0;
}

int TExpression::Find_Closing_Brace(
	const QString						&input,
	const int							open_brace_index) const
{
	int									brace_cntr;
	int									character;
	int									cntr;
	
	if(input.size())
	{
		if(open_brace_index < input.size())
		{
			brace_cntr = 0;
			for(cntr = open_brace_index;cntr < input.size();++cntr)
			{
				character = input[cntr].toLatin1();
				
				switch(character)
				{
					case '(':
						if(cntr != open_brace_index)
						{
							brace_cntr++;
						}
						break;
						
					case ')':
						brace_cntr--;
						
						if(brace_cntr < 0)
						{
							return cntr;
						}
						break;
				}
			}
		}
	}
	
	return -1;
}

