/////////////////////////////////////////////////////////////////////
//
//            X   X           X
//           XXX XX         XX
//          XXXXXXXX      XXX
//         XX X XXXXXXXXXXX
//        XXXXX XXXXXXXXX
//       XXXXX XXXXXXXXXX
//            XXX XXX XXX
//           XXX XX   XX
//           X   X     X
//
//    Copyright (C) 2003-2023  Ron Jakl
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
/////////////////////////////////////////////////////////////////////

#include <QApplication>
#include <QFont>
#include <QGridLayout>
#include <QHBoxLayout>
#include <QLabel>
#include <QListWidget>
#include <QPlainTextEdit>
#include <QPushButton>
#include <QSpacerItem>
#include <QTabWidget>
#include <QVBoxLayout>
#include <QWidget>
#include <QLineEdit>
#include <QTextCursor>
#include <QFile>
#include <QFileDialog>
#include <cmath>
#include <assert.h>

#include "measurement.h"
#include "opengldisplaywidget.h"
#include "displaywidget.h"

TDisplayWidget::TDisplayWidget(
	const QWidget						*parent,
	const Qt::WindowFlags				flags)
:QWidget(const_cast<QWidget*>(parent),flags)
{
	QFont								font;
	QGridLayout							*widget_layout;
	QHBoxLayout							*measurement_button_hlayout;
	QHBoxLayout							*map_options_hlayout;
	QVBoxLayout							*graphic_vlayout;
	QVBoxLayout							*map_vlayout;
	QVBoxLayout							*measurement_vlayout;
	QLabel								*map_increment_x_label;
	QLabel								*map_increment_y_label;
	QLabel								*map_increment_z_label;
	QSpacerItem							*measurement_hspacer;
	QSpacerItem							*map_hspacer;
	QWidget								*graphic_widget;
	QWidget								*map_widget;
	QWidget								*measurement_widget;

	font.setFamily(QStringLiteral("Courier New"));
	font.setFixedPitch(true);

	widget_layout = new QGridLayout(this);
	widget_layout->setContentsMargins(0,0,0,0);
	
	d_display_tab = new QTabWidget(this);

	graphic_widget = new QWidget();
	graphic_vlayout = new QVBoxLayout(graphic_widget);
	graphic_vlayout->setContentsMargins(0,0,0,0);

	d_opengl_display_widget = new TOpenGLDisplayWidget(graphic_widget);
	graphic_vlayout->addWidget(d_opengl_display_widget);
	
	d_selection_edit = new QPlainTextEdit(this);
	d_selection_edit->setMaximumSize(16777215,100);
	d_selection_edit->setFont(font);
	d_selection_edit->setUndoRedoEnabled(false);
	d_selection_edit->setLineWrapMode(QPlainTextEdit::NoWrap);
	d_selection_edit->setReadOnly(true);
	graphic_vlayout->addWidget(d_selection_edit);

	d_display_tab->addTab(graphic_widget,QStringLiteral("Graphical View"));

	measurement_widget = new QWidget();

	measurement_vlayout = new QVBoxLayout(measurement_widget);

	d_measurement_edit = new QPlainTextEdit(measurement_widget);
	d_measurement_edit->setFont(font);
	d_measurement_edit->setUndoRedoEnabled(false);
	d_measurement_edit->setLineWrapMode(QPlainTextEdit::NoWrap);
	d_measurement_edit->setReadOnly(true);
	measurement_vlayout->addWidget(d_measurement_edit);

	measurement_button_hlayout = new QHBoxLayout();
	
	measurement_hspacer = new QSpacerItem(0,0,QSizePolicy::Expanding,QSizePolicy::Minimum);
	measurement_button_hlayout->addItem(measurement_hspacer);
	
	d_evaluate_button = new QPushButton(measurement_widget);
	d_evaluate_button->setAutoDefault(false);
	measurement_button_hlayout->addWidget(d_evaluate_button);

	d_export_button = new QPushButton(measurement_widget);
	d_export_button->setAutoDefault(false);
	measurement_button_hlayout->addWidget(d_export_button);

	d_save_text_button = new QPushButton(measurement_widget);
	d_save_text_button->setAutoDefault(false);
	measurement_button_hlayout->addWidget(d_save_text_button);
	measurement_vlayout->addLayout(measurement_button_hlayout);

	d_display_tab->addTab(measurement_widget,QStringLiteral("Measurement Results"));

	map_widget = new QWidget();

	map_vlayout = new QVBoxLayout(map_widget);
	
	map_options_hlayout = new QHBoxLayout();
	
	map_increment_x_label = new QLabel(map_widget);
	map_options_hlayout->addWidget(map_increment_x_label);
	
	d_map_increment_x_edit = new QLineEdit(map_widget);
	map_options_hlayout->addWidget(d_map_increment_x_edit);
	
	map_increment_y_label = new QLabel(map_widget);
	map_options_hlayout->addWidget(map_increment_y_label);
	
	d_map_increment_y_edit = new QLineEdit(map_widget);
	map_options_hlayout->addWidget(d_map_increment_y_edit);
	
	map_increment_z_label = new QLabel(map_widget);
	map_options_hlayout->addWidget(map_increment_z_label);
	
	d_map_increment_z_edit = new QLineEdit(map_widget);
	map_options_hlayout->addWidget(d_map_increment_z_edit);
	
	map_hspacer = new QSpacerItem(0,0,QSizePolicy::Expanding,QSizePolicy::Minimum);
	map_options_hlayout->addItem(map_hspacer);
	
	d_update_map_button = new QPushButton(map_widget);
	d_update_map_button->setAutoDefault(false);
	map_options_hlayout->addWidget(d_update_map_button);
	
	map_vlayout->addLayout(map_options_hlayout);
	
	d_map_data_edit = new QPlainTextEdit(map_widget);
	d_map_data_edit->setFont(font);
	d_map_data_edit->setUndoRedoEnabled(false);
	d_map_data_edit->setLineWrapMode(QPlainTextEdit::NoWrap);
	d_map_data_edit->setReadOnly(true);
	map_vlayout->addWidget(d_map_data_edit);

	d_display_tab->addTab(map_widget,QStringLiteral("Machine Errors"));
	
	widget_layout->addWidget(d_display_tab,0,1,1,1);
	
	d_display_tab->setTabPosition(QTabWidget::North);
	d_display_tab->setCurrentIndex(0);
	d_save_text_button->setText(QStringLiteral("Save Text"));
	d_export_button->setText(QStringLiteral("Export Measurements"));
	d_evaluate_button->setText(QStringLiteral("Evaluation Report"));
	d_update_map_button->setText(QStringLiteral("Update"));
	map_increment_x_label->setText(QStringLiteral("Increment X:"));
	map_increment_y_label->setText(QStringLiteral("Increment Y:"));
	map_increment_z_label->setText(QStringLiteral("Increment Z:"));
	
	// defaults
	d_map_increment_x_edit->setText(QStringLiteral("100.000"));
	d_map_increment_y_edit->setText(QStringLiteral("100.000"));
	d_map_increment_z_edit->setText(QStringLiteral("100.000"));
	
	this->Enable_Output_Options(false);
	
	connect(d_opengl_display_widget,&TOpenGLDisplayWidget::Model_Item_Clicked,this,&TDisplayWidget::Measurement_Item_Clicked);
	
	connect(d_save_text_button,&QPushButton::clicked,this,&TDisplayWidget::Save_Measurement_Text);
	connect(d_export_button,&QPushButton::clicked,this,&TDisplayWidget::Export_Measurements);
	connect(d_evaluate_button,&QPushButton::clicked,this,&TDisplayWidget::Evaluate_Measurements);
	connect(d_update_map_button,&QPushButton::clicked,this,&TDisplayWidget::Update_Map_Display);
}

TDisplayWidget::~TDisplayWidget(void)
{
}

double TDisplayWidget::Map_Data_Increment_X(void) const
{
	return fabs(d_map_increment_x_edit->text().toDouble());
}

double TDisplayWidget::Map_Data_Increment_Y(void) const
{
	return fabs(d_map_increment_y_edit->text().toDouble());
}

double TDisplayWidget::Map_Data_Increment_Z(void) const
{
	return fabs(d_map_increment_z_edit->text().toDouble());
}

void TDisplayWidget::Clear_Map_Data(void)
{
	d_map_data_edit->clear();
}

void TDisplayWidget::Add_Map_Data_Line(
	const QString						&line)
{
	d_map_data_edit->appendPlainText(line);
}

void TDisplayWidget::Add_Map_Data_End(void)
{
	QTextCursor							text_cursor;
	
	text_cursor = d_map_data_edit->textCursor();
	text_cursor.setPosition(0);	// beginning of document
	
	d_map_data_edit->setTextCursor(text_cursor);
}

void TDisplayWidget::Clear_Measurement_Data(void)
{
	d_measurement_edit->clear();
}

void TDisplayWidget::Add_Measurement_Data_Line(
	const QString						&line)
{
	d_measurement_edit->appendPlainText(line);
}

void TDisplayWidget::Add_Measurement_Data_End(void)
{
	QTextCursor							text_cursor;
	
	text_cursor = d_measurement_edit->textCursor();
	text_cursor.setPosition(0);	// beginning of document
	
	d_measurement_edit->setTextCursor(text_cursor);
}

void TDisplayWidget::Clear_Graphical_Data(void)
{
	d_opengl_display_widget->Clear_Render_Measurements();
}

void TDisplayWidget::Add_Graphical_Data(
	TMeasurement						* const measurement)
{
	assert(measurement);
	
	d_opengl_display_widget->Render_Measurements(measurement);
}

void TDisplayWidget::Add_Graphical_Data_End(void)
{
	d_opengl_display_widget->Update(false);
}

void TDisplayWidget::Clear_Selection_Data(void)
{
	d_selection_edit->clear();
}

void TDisplayWidget::Add_Selection_Data_Line(
	const QString						&line)
{
	d_selection_edit->appendPlainText(line);
}

void TDisplayWidget::Add_Selection_Data_End(void)
{
	QTextCursor							text_cursor;
	
	text_cursor = d_selection_edit->textCursor();
	text_cursor.setPosition(0);	// beginning of document
	
	d_selection_edit->setTextCursor(text_cursor);
}

void TDisplayWidget::Enable_Output_Options(const bool state)
{
	d_save_text_button->setEnabled(state);
	d_export_button->setEnabled(state);
	d_evaluate_button->setEnabled(state);
}

void TDisplayWidget::Define_Machine_Volume(
	const TVector3						&min_corner,
	const TVector3						&max_corner)
{
	d_opengl_display_widget->Define_Machine_Volume(min_corner,max_corner);
	d_opengl_display_widget->Update(true);
}

void TDisplayWidget::Clear_Machine_Volume(void)
{
	d_opengl_display_widget->Clear_Render_Volume();
	d_opengl_display_widget->Update(true);
}

void TDisplayWidget::Measurement_Item_Clicked(
	const double						&x,
	const double						&y,
	const double						&z,
	const int							id)
{
	Q_UNUSED(x);
	Q_UNUSED(y);
	Q_UNUSED(z);
	
	int									measurement_id;
	int									measurement_item_id;

	if(id < 0)
	{
		d_selection_edit->clear();
		d_opengl_display_widget->Clear_Selection();
		return;
	}
	
	// selection information is split into two parts,
	// the major is in units of 1000 and the minor is in units up to 1000
	measurement_id = id;
	measurement_id /= 1000;
	measurement_id *= 1000;
	
	measurement_item_id = id - measurement_id;
	
	d_opengl_display_widget->Set_Selection(id);
	
	emit Item_Selected(measurement_id,measurement_item_id);
}

void TDisplayWidget::Save_Measurement_Text(void)
{
	QString								file_name;
	QFile								file;
	QString								text;
	
	file_name = QFileDialog::getSaveFileName(this,
								QStringLiteral("Save Measurement Text"),
								QString(),
								QStringLiteral("Text Files (*.txt *.*)"));
	
	if(!file_name.length())
	{
		return;
	}
	
	file.setFileName(file_name);
	
	if(file.open(QIODevice::WriteOnly | QIODevice::Text))
	{
		text = d_measurement_edit->toPlainText();
		file.write(text.toLatin1());
		
		file.close();
	}
}
