/////////////////////////////////////////////////////////////////////
//
//            X   X           X
//           XXX XX         XX
//          XXXXXXXX      XXX
//         XX X XXXXXXXXXXX
//        XXXXX XXXXXXXXX
//       XXXXX XXXXXXXXXX
//            XXX XXX XXX
//           XXX XX   XX
//           X   X     X
//
//    Copyright (C) 2003-2023  Ron Jakl
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
/////////////////////////////////////////////////////////////////////

#include <QApplication>
#include <QAbstractItemDelegate>
#include <QAction>
#include <QDialog>
#include <QDockWidget>
#include <QGridLayout>
#include <QVBoxLayout>
#include <QHBoxLayout>
#include <QList>
#include <QListWidget>
#include <QMenu>
#include <QMenuBar>
#include <QSpacerItem>
#include <QToolBar>
#include <QToolButton>
#include <QWidget>
#include <QLabel>
#include <QLineEdit>
#include <QDir>
#include <QFile>
#include <QFileDialog>
#include <QCloseEvent>
#include <QFont>
#include <assert.h>

#include "../../core/expression.h"
#include "../../core/mat4.h"
#include "../../core/referencesystem.h"
#include "../../core/xmlfile.h"
#include "../../core/messagebox.h"

#include "displaywidget.h"
#include "measurementdialog.h"
#include "measurement_laser.h"
#include "measurement_iso.h"
#include "measurement_mcg.h"
#include "measurement_ballbar.h"
#include "progressdialog.h"
#include "adddialog.h"
#include "types.h"

#include "errorsimulator.h"

TErrorSimulator::TErrorSimulator(
	const QWidget						*parent,
	const Qt::WindowFlags				flags)
:QMainWindow(const_cast<QWidget*>(parent),flags)
{
	QAction								*about_action;
	QToolBar							*toolbar;
	QWidget								*central_widget;
	QWidget								*machine_dock_contents_widget;
	QWidget								*measurement_dock_contents_widget;
	QDockWidget							*machine_dock;
	QDockWidget							*measurement_dock;
	QGridLayout							*central_widget_layout;
	QVBoxLayout							*machine_dock_vlayout;
	QVBoxLayout							*measurement_dock_vlayout;
	QHBoxLayout							*machine_button_hlayout;
	QHBoxLayout							*measurement_button_hlayout;
	QHBoxLayout							*copyright_hlayout;
	QMenuBar							*main_menu;
	QMenu								*file_menu;
	QMenu								*edit_menu;
	QMenu								*view_menu;
	QMenu								*help_menu;
	QSpacerItem							*machine_hspacer;
	QSpacerItem							*measurement_hspacer;
	QSpacerItem							*copyright_hspacer;
	QLabel								*logo_label;
	QLabel								*copyright_label;
	QDir								data_location;
	QFont								copyright_font;
	QStringList							list;
	static const int					TOOL_SPACER_DIMENSION(4);
	
	copyright_font.setFamily(QStringLiteral("Verdana"));
	copyright_font.setPointSize(8);
	copyright_font.setItalic(true);

	this->resize(1101,750);

	data_location = QDir::home();
	
	if(!data_location.exists(".errorsimulator"))
	{
		data_location.mkdir(".errorsimulator");
	}
	
	data_location.cd(".errorsimulator");
	
	d_data_path = data_location.absolutePath();
	if(d_data_path.endsWith('/'))
	{
		d_data_path.chop(1);
	}
	
	central_widget = new QWidget(this);
	this->setCentralWidget(central_widget);

	central_widget_layout = new QGridLayout(central_widget);

	d_display_widget = new TDisplayWidget(central_widget);
	central_widget_layout->addWidget(d_display_widget,0,0,1,1);
	
	d_machine_dialog = new TMachineDialog(this);	
	d_measurement_dialog = new TMeasurementDialog(this);	
	
	copyright_hlayout = new QHBoxLayout();
	
	copyright_label = new QLabel(this);
	copyright_label->setFont(copyright_font);
	copyright_label->setAlignment(Qt::AlignBottom|Qt::AlignLeading|Qt::AlignLeft);
	copyright_hlayout->addWidget(copyright_label);

	copyright_hspacer = new QSpacerItem(0,0,QSizePolicy::Expanding,QSizePolicy::Minimum);
	copyright_hlayout->addItem(copyright_hspacer);
	
	logo_label = new QLabel(this);
	logo_label->setMinimumSize(100,40);
	logo_label->setMaximumSize(100,40);
	copyright_hlayout->addWidget(logo_label);

	central_widget_layout->addLayout(copyright_hlayout,1,0,1,1);
	
	d_machine_dialog->Set_Data_Path(d_data_path);
	d_measurement_dialog->Set_Data_Path(d_data_path);

	d_add_machine_action = new QAction(this);
	d_add_machine_action->setIcon(QIcon(":/toolbar/add_machine_32x32.png"));
	
	d_remove_machine_action = new QAction(this);
	d_remove_machine_action->setIcon(QIcon(":/toolbar/remove_machine_32x32.png"));
	
	d_edit_machine_action = new QAction(this);
	d_edit_machine_action->setIcon(QIcon(":/toolbar/edit_machine_32x32.png"));

	d_add_measurement_action = new QAction(this);
	d_add_measurement_action->setIcon(QIcon(":/toolbar/add_measurement_32x32.png"));
	
	d_remove_measurement_action = new QAction(this);
	d_remove_measurement_action->setIcon(QIcon(":/toolbar/remove_measurement_32x32.png"));
	
	d_edit_measurement_action = new QAction(this);
	d_edit_measurement_action->setIcon(QIcon(":/toolbar/edit_measurement_32x32.png"));
	
	about_action = new QAction(this);
	about_action->setIcon(QIcon(":/toolbar/about_32x32.png"));

	d_quit_action = new QAction(this);

	main_menu = new QMenuBar(this);
	this->setMenuBar(main_menu);

	file_menu = new QMenu(main_menu);
	main_menu->addMenu(file_menu);

	edit_menu = new QMenu(main_menu);
	main_menu->addMenu(edit_menu);

	view_menu = new QMenu(main_menu);
	main_menu->addMenu(view_menu);
	
	help_menu = new QMenu(main_menu);
	main_menu->addMenu(help_menu);
	
	main_menu->addAction(file_menu->menuAction());
	main_menu->addAction(edit_menu->menuAction());
	main_menu->addAction(view_menu->menuAction());
	main_menu->addAction(help_menu->menuAction());

	file_menu->addAction(d_add_machine_action);
	file_menu->addAction(d_add_measurement_action);
	file_menu->addSeparator();
	file_menu->addAction(d_quit_action);

	edit_menu->addAction(d_edit_machine_action);
	edit_menu->addAction(d_edit_measurement_action);
	edit_menu->addSeparator();
	edit_menu->addAction(d_remove_machine_action);
	edit_menu->addAction(d_remove_measurement_action);
	
	help_menu->addAction(about_action);

	machine_dock = new QDockWidget(this);
	machine_dock->setAllowedAreas(Qt::LeftDockWidgetArea|Qt::RightDockWidgetArea);
	this->addDockWidget(Qt::LeftDockWidgetArea,machine_dock);
	
	machine_dock_contents_widget = new QWidget();
	machine_dock->setWidget(machine_dock_contents_widget);

	view_menu->addAction(machine_dock->toggleViewAction());

	machine_dock_vlayout = new QVBoxLayout(machine_dock_contents_widget);
	machine_dock_vlayout->setSpacing(TOOL_SPACER_DIMENSION);
	
	measurement_dock = new QDockWidget(this);
	measurement_dock->setAllowedAreas(Qt::LeftDockWidgetArea|Qt::RightDockWidgetArea);
	this->addDockWidget(Qt::LeftDockWidgetArea,measurement_dock);
	
	measurement_dock_contents_widget = new QWidget();
	measurement_dock->setWidget(measurement_dock_contents_widget);

	view_menu->addAction(measurement_dock->toggleViewAction());

	measurement_dock_vlayout = new QVBoxLayout(measurement_dock_contents_widget);
	measurement_dock_vlayout->setSpacing(TOOL_SPACER_DIMENSION);

	toolbar = new QToolBar(this);
	toolbar->setIconSize(QSize(32,32));
	this->addToolBar(Qt::TopToolBarArea,toolbar);
	
	toolbar->addAction(d_add_machine_action);
	toolbar->addAction(d_remove_machine_action);
	toolbar->addAction(d_edit_machine_action);
	toolbar->addSeparator();
	
	toolbar->addAction(d_add_measurement_action);
	toolbar->addAction(d_remove_measurement_action);
	toolbar->addAction(d_edit_measurement_action);
	toolbar->addSeparator();
	
	toolbar->addAction(about_action);

	d_machines_list = new QListWidget(machine_dock_contents_widget);
	d_machines_list->setSelectionMode(QAbstractItemView::ExtendedSelection);
	machine_dock_vlayout->addWidget(d_machines_list);
	
	machine_button_hlayout = new QHBoxLayout();
	machine_button_hlayout->setSpacing(TOOL_SPACER_DIMENSION);
	
	machine_hspacer = new QSpacerItem(0, 0, QSizePolicy::Expanding, QSizePolicy::Minimum);
	machine_button_hlayout->addItem(machine_hspacer);
	
	d_add_machine_button = new QToolButton(machine_dock_contents_widget);
	d_add_machine_button->setFixedSize(16,16);
	machine_button_hlayout->addWidget(d_add_machine_button);
	
	d_remove_machine_button = new QToolButton(machine_dock_contents_widget);
	d_remove_machine_button->setFixedSize(16,16);
	machine_button_hlayout->addWidget(d_remove_machine_button);
	
	d_edit_machine_button = new QToolButton(machine_dock_contents_widget);
	d_edit_machine_button->setFixedSize(16,16);
	machine_button_hlayout->addWidget(d_edit_machine_button);

	machine_dock_vlayout->addLayout(machine_button_hlayout);

	d_measurements_list = new QListWidget(measurement_dock_contents_widget);
	d_measurements_list->setSelectionMode(QAbstractItemView::ExtendedSelection);
	measurement_dock_vlayout->addWidget(d_measurements_list);

	measurement_button_hlayout = new QHBoxLayout();
	measurement_button_hlayout->setSpacing(TOOL_SPACER_DIMENSION);

	measurement_hspacer = new QSpacerItem(0, 0, QSizePolicy::Expanding, QSizePolicy::Minimum);
	measurement_button_hlayout->addItem(measurement_hspacer);
	
	d_add_measurement_button = new QToolButton(machine_dock_contents_widget);
	d_add_measurement_button->setFixedSize(16,16);
	measurement_button_hlayout->addWidget(d_add_measurement_button);
	
	d_remove_measurement_button = new QToolButton(machine_dock_contents_widget);
	d_remove_measurement_button->setFixedSize(16,16);
	measurement_button_hlayout->addWidget(d_remove_measurement_button);
	
	d_edit_measurement_button = new QToolButton(machine_dock_contents_widget);
	d_edit_measurement_button->setFixedSize(16,16);
	measurement_button_hlayout->addWidget(d_edit_measurement_button);
	
	measurement_dock_vlayout->addLayout(measurement_button_hlayout);
	
	d_add_machine_button->setIcon(QIcon(":/gui/add_16x16.png"));
	d_remove_machine_button->setIcon(QIcon(":/gui/remove_16x16.png"));
	d_edit_machine_button->setIcon(QIcon(":/gui/edit_16x16.png"));
	
	d_add_measurement_button->setIcon(QIcon(":/gui/add_16x16.png"));
	d_remove_measurement_button->setIcon(QIcon(":/gui/remove_16x16.png"));
	d_edit_measurement_button->setIcon(QIcon(":/gui/edit_16x16.png"));

	d_machine_dialog->Load_Data();
	d_measurement_dialog->Load_Data();
	
	this->setWindowTitle(QStringLiteral("Error Simulator"));
	this->setWindowIcon(QIcon(":/images/errorsimulator.png"));
	
	file_menu->setTitle(QStringLiteral("File"));
	edit_menu->setTitle(QStringLiteral("Edit"));
	view_menu->setTitle(QStringLiteral("View"));
	help_menu->setTitle(QStringLiteral("Help"));

	machine_dock->setWindowTitle(QStringLiteral("Machines"));
	measurement_dock->setWindowTitle(QStringLiteral("Measurements"));
	
	d_add_machine_action->setText(QStringLiteral("Add Machine"));
	d_remove_machine_action->setText(QStringLiteral("Remove Machine"));
	d_edit_machine_action->setText(QStringLiteral("Edit Machine"));
	d_add_measurement_action->setText(QStringLiteral("Add Measurement"));
	d_remove_measurement_action->setText(QStringLiteral("Remove Measurement"));
	d_edit_measurement_action->setText(QStringLiteral("Edit Measurement"));
	about_action->setText(QStringLiteral("About"));
	d_quit_action->setText(QStringLiteral("Quit"));

	d_add_machine_action->setToolTip(QStringLiteral("Add a new machine"));
	d_remove_machine_action->setToolTip(QStringLiteral("Remove the selected machines"));
	d_edit_machine_action->setToolTip(QStringLiteral("Edit the selected machine"));
	d_add_measurement_action->setToolTip(QStringLiteral("Add a new measurement"));
	d_remove_measurement_action->setToolTip(QStringLiteral("Remove the selected measurement"));
	d_edit_measurement_action->setToolTip(QStringLiteral("Edit the selected measurement"));

	logo_label->setPixmap(QPixmap(QStringLiteral(":/images/logo.png")));
	copyright_label->setText(QStringLiteral("Copyright @2023 Select Calibration Incorporated"));

	// default values
	list = d_machine_dialog->Machine_Names();
	this->Update_Machine_Names();
	
	d_remove_machine_button->setEnabled(false);
	d_edit_machine_button->setEnabled(false);
	d_remove_measurement_button->setEnabled(false);
	d_edit_measurement_button->setEnabled(false);
	d_remove_machine_action->setEnabled(false);
	d_edit_machine_action->setEnabled(false);
	d_remove_measurement_action->setEnabled(false);
	d_edit_measurement_action->setEnabled(false);
	
	d_msg_box = new TMessageBox(this);
	
	d_progress_dialog = 0;
	d_add_dialog = 0;
	
	list = d_measurement_dialog->Measurement_Names();
	this->Update_Measurement_Names();
	
	connect(d_add_measurement_action,&QAction::triggered,this,&TErrorSimulator::Add_Measurement);
	connect(d_add_machine_action,&QAction::triggered,this,&TErrorSimulator::Add_Machine);
	connect(d_remove_machine_action,&QAction::triggered,this,&TErrorSimulator::Remove_Machine);
	connect(d_edit_measurement_action,&QAction::triggered,this,&TErrorSimulator::Edit_Measurement);
	connect(d_remove_measurement_action,&QAction::triggered,this,&TErrorSimulator::Remove_Measurement);
	connect(d_quit_action,&QAction::triggered,this,&TErrorSimulator::close);
	connect(d_edit_machine_action,&QAction::triggered,this,&TErrorSimulator::Edit_Machine);
	connect(about_action,&QAction::triggered,this,&TErrorSimulator::Help_About);

	connect(d_machines_list,&QListWidget::itemSelectionChanged,this,&TErrorSimulator::Machine_Selection_Changed);
	connect(d_measurements_list,&QListWidget::itemSelectionChanged,this,&TErrorSimulator::Measurement_Selection_Changed);
	connect(d_add_machine_button,&QToolButton::clicked,this,&TErrorSimulator::Add_Machine);
	connect(d_remove_machine_button,&QToolButton::clicked,this,&TErrorSimulator::Remove_Machine);
	connect(d_add_measurement_button,&QToolButton::clicked,this,&TErrorSimulator::Add_Measurement);
	connect(d_remove_measurement_button,&QToolButton::clicked,this,&TErrorSimulator::Remove_Measurement);
	connect(d_edit_machine_button,&QToolButton::clicked,this,&TErrorSimulator::Edit_Machine);
	connect(d_edit_measurement_button,&QToolButton::clicked,this,&TErrorSimulator::Edit_Measurement);
	
	connect(d_machines_list->itemDelegate(),&QAbstractItemDelegate::commitData,this,&TErrorSimulator::Rename_Machine);
	connect(d_measurements_list->itemDelegate(),&QAbstractItemDelegate::commitData,this,&TErrorSimulator::Rename_Measurement);

	connect(d_measurement_dialog,&TMeasurementDialog::Update_Measurement_Display,this,&TErrorSimulator::Update_Measurement_Display);
	
	connect(d_display_widget,&TDisplayWidget::Item_Selected,this,&TErrorSimulator::Display_Selected_Item);
	connect(d_display_widget,&TDisplayWidget::Export_Measurements,this,&TErrorSimulator::Export_Measurements);
	connect(d_display_widget,&TDisplayWidget::Evaluate_Measurements,this,&TErrorSimulator::Evaluate_Measurements);
	connect(d_display_widget,&TDisplayWidget::Update_Map_Display,this,&TErrorSimulator::Update_Map_Display);
}

TErrorSimulator::~TErrorSimulator(void)
{
}

void TErrorSimulator::Machine_Selection_Changed(void)
{
	QList<QListWidgetItem*>				items;
		
	items = d_machines_list->selectedItems();

	d_remove_machine_button->setEnabled(false);
	d_edit_machine_button->setEnabled(false);
	
	d_remove_machine_action->setEnabled(false);
	d_edit_machine_action->setEnabled(false);

	if(items.size() == 1)
	{		
		this->Activate_Machine(items[0]->text());
		
		d_remove_machine_button->setEnabled(true);
		d_edit_machine_button->setEnabled(true);
		
		d_remove_machine_action->setEnabled(true);
		d_edit_machine_action->setEnabled(true);
	}
	else if(items.size() > 1)
	{
		d_display_widget->Clear_Map_Data();
		d_display_widget->Clear_Measurement_Data();
		
		this->Activate_Machine(QString());

		d_remove_machine_button->setEnabled(true);
		d_remove_machine_action->setEnabled(true);
	}
	else
	{
		d_display_widget->Clear_Map_Data();
		d_display_widget->Clear_Measurement_Data();
		d_display_widget->Clear_Graphical_Data();
		d_display_widget->Clear_Selection_Data();

		this->Activate_Machine(QString());
	}
}

void TErrorSimulator::Measurement_Selection_Changed(void)
{
	QString								machine;
	QStringList							list;
	QString								line;
	std::vector<TMeasurement*>			measurements;
	QList<QString>::const_iterator		str_iter;
	std::vector<TMeasurement*>::iterator meas_iter;
	double								length(0),min_length(0),max_length(0),average_length(0);
	int									ballbar_count;
		
	d_remove_measurement_button->setEnabled(false);
	d_edit_measurement_button->setEnabled(false);

	d_remove_measurement_action->setEnabled(false);
	d_edit_measurement_action->setEnabled(false);

	machine = this->Active_Machine_Name();
	
	d_display_widget->Clear_Measurement_Data();
	d_display_widget->Clear_Graphical_Data();
	d_display_widget->Clear_Selection_Data();
	
	list = this->Selected_Measurements();

	if(list.size() == 1)
	{
		d_remove_measurement_button->setEnabled(true);
		d_edit_measurement_button->setEnabled(true);
		
		d_remove_measurement_action->setEnabled(true);
		d_edit_measurement_action->setEnabled(true);
	}
	else if(list.size() > 1)
	{
		d_remove_measurement_button->setEnabled(true);
		d_remove_measurement_action->setEnabled(true);
	}
	
	d_display_widget->Enable_Output_Options(false);

	if(machine.length())
	{
		measurements = d_measurement_dialog->Measurements(list);
		
		if(!this->Evaluate_Measurement_Expressions(&d_machine_comp,&measurements,&list))
		{
			d_display_widget->Add_Measurement_Data_Line(QStringLiteral("Measurement Expression Errors"));
			d_display_widget->Add_Measurement_Data_Line(QStringLiteral("---------------------------------"));
			
			d_display_widget->Add_Selection_Data_Line(QStringLiteral("Measurement Expression Errors"));
			d_display_widget->Add_Selection_Data_Line(QStringLiteral("---------------------------------"));

			for(str_iter = list.begin();str_iter != list.end();++str_iter)
			{
				d_display_widget->Add_Measurement_Data_Line(*str_iter);
				d_display_widget->Add_Selection_Data_Line(*str_iter);
			}
			
			d_display_widget->Add_Measurement_Data_End();
			d_display_widget->Add_Selection_Data_End();
			return;
		}
		
		this->Update_Measurement_Items(&d_machine_comp,&measurements);
		
		d_display_widget->Enable_Output_Options(true);
		
		ballbar_count = 0;
		average_length = 0.0;
		
		for(meas_iter = measurements.begin();meas_iter != measurements.end();++meas_iter)
		{
			d_display_widget->Add_Graphical_Data(*meas_iter);
			
			if((*meas_iter)->Type() == Types::MEASUREMENT_BALLBAR)
			{
				if((*meas_iter)->Measurement_Item_Count() == 1)	// should only be one ball bar measurement
				{
					length = (*meas_iter)->Measurement_Item(0).length_actual;
					
					average_length += length;
					
					if(ballbar_count == 0)
					{
						min_length = max_length = length;
					}
					else
					{
						if(length < min_length)
						{
							min_length = length;
						}
						else if(length > max_length)
						{
							max_length = length;
						}
					}
					
					ballbar_count++;
				}
			}
			else
			{
				list = this->Report_Measurement_Item(*meas_iter);
				
				for(str_iter = list.begin();str_iter != list.end();++str_iter)
				{
					d_display_widget->Add_Measurement_Data_Line(*str_iter);
				}
			}
		}
		
		if(ballbar_count)
		{
			average_length /= static_cast<double>(ballbar_count);
			
			line = QStringLiteral("ASME B89 Ballbar Measurement");
			d_display_widget->Add_Measurement_Data_Line(line);

			line = QStringLiteral("--------------------------------------------------------------------------------------------------------------------------------");
			d_display_widget->Add_Measurement_Data_Line(line);
			
			line = QString("Length Min:     %1").arg(min_length,9,'f',4);
			d_display_widget->Add_Measurement_Data_Line(line);
			
			line = QString("Length Max:     %1").arg(max_length,9,'f',4);
			d_display_widget->Add_Measurement_Data_Line(line);
			
			line = QString("Length Mean:    %1").arg(average_length,9,'f',4);
			d_display_widget->Add_Measurement_Data_Line(line);
			
			line = QString("Length Range:   %1").arg(max_length - min_length,9,'f',4);
			d_display_widget->Add_Measurement_Data_Line(line);
			
			d_display_widget->Add_Measurement_Data_Line(QString());

			
			
//                                 0         1         2         3         4         5         6         7         8         9          0         1         2         3
//                                 012345678901234567890123456789012345678901234567890123456789012345678901234567890123456789001234567890123456789012345678901234567890
//                                       |          |          |          |          |          |          |          |          |          |          |          |
			line = QStringLiteral("    Name                 Length    Sph1. X    Sph1. Y   Sph1. Z     Sph2. X    Sph2. Y   Sph2. Z     Probe X    Probe Y   Probe Z");
			d_display_widget->Add_Measurement_Data_Line(line);
			
			line = QStringLiteral("------------------------------------------------------------------------------------------------------------------------------------------");
			d_display_widget->Add_Measurement_Data_Line(line);

	
			for(meas_iter = measurements.begin();meas_iter != measurements.end();++meas_iter)
			{
				if((*meas_iter)->Type() == Types::MEASUREMENT_BALLBAR && (*meas_iter)->Measurement_Item_Count() == 1)
				{
					list = this->Report_Measurement_Item(*meas_iter);
					
					for(str_iter = list.begin();str_iter != list.end();++str_iter)
					{
						d_display_widget->Add_Measurement_Data_Line(*str_iter);
					}
				}
			}
		}
		
		d_display_widget->Add_Measurement_Data_End();
		d_display_widget->Add_Graphical_Data_End();
	}
	else
	{
		d_display_widget->Add_Measurement_Data_Line(QStringLiteral("Select a machine to display the measurement."));
		d_display_widget->Add_Selection_Data_Line(QStringLiteral("Select a machine to display the measurement."));
		
		d_display_widget->Add_Measurement_Data_End();
		d_display_widget->Add_Selection_Data_End();
	}
}

void TErrorSimulator::Update_Measurement_Display(void)
{
	this->Update_Measurement_Names();
}

void TErrorSimulator::Update_Map_Display(void)
{
	QList<QListWidgetItem*>				items;
	double								increment_x;
	double								increment_y;
	double								increment_z;
	QStringList							list;
	QList<QString>::const_iterator		str_iter;
	QString								machine_name;
	
	d_display_widget->Clear_Map_Data();

	items = d_machines_list->selectedItems();
	
	if(items.size() == 1)
	{
		machine_name = items[0]->text();
	}
	else
	{
		return;
	}
	
	increment_x = d_display_widget->Map_Data_Increment_X();
	increment_y = d_display_widget->Map_Data_Increment_Y();
	increment_z = d_display_widget->Map_Data_Increment_Z();
	
	if(increment_x < 10 || increment_y < 10 || increment_z < 10)
	{
		d_display_widget->Add_Map_Data_Line(QStringLiteral("ERR:  Map increment too small for X,Y, or Z axis"));
		d_display_widget->Add_Map_Data_End();
		
		return;
	}
	
	d_display_widget->Add_Map_Data_Line(QString("Selected Machine: %1").arg(machine_name));
	d_display_widget->Add_Map_Data_Line(QString());
	
	list = d_machine_comp.Map_Data_Quad();
	
	for(str_iter = list.begin();str_iter != list.end();++str_iter)
	{
		d_display_widget->Add_Map_Data_Line(*str_iter);
	}
	d_display_widget->Add_Map_Data_Line(QString());
	
	list = d_machine_comp.Map_Data_X(increment_x);
	
	for(str_iter = list.begin();str_iter != list.end();++str_iter)
	{
		d_display_widget->Add_Map_Data_Line(*str_iter);
	}
	d_display_widget->Add_Map_Data_Line(QString());
	
	list = d_machine_comp.Map_Data_Y(increment_y);
	
	for(str_iter = list.begin();str_iter != list.end();++str_iter)
	{
		d_display_widget->Add_Map_Data_Line(*str_iter);
	}
	d_display_widget->Add_Map_Data_Line(QString());
	
	list = d_machine_comp.Map_Data_Z(increment_z);
	
	for(str_iter = list.begin();str_iter != list.end();++str_iter)
	{
		d_display_widget->Add_Map_Data_Line(*str_iter);
	}
	d_display_widget->Add_Map_Data_Line(QString());
	
	list = d_machine_comp.Map_Data_D(increment_z);
	
	for(str_iter = list.begin();str_iter != list.end();++str_iter)
	{
		d_display_widget->Add_Map_Data_Line(*str_iter);
	}
	
	d_display_widget->Add_Map_Data_End();
}

void TErrorSimulator::Add_Machine(void)
{
	QStringList							existing_machines;
	
	if(!d_add_dialog)
	{
		d_add_dialog = new TAddDialog(this);
	}
	
	existing_machines = d_machine_dialog->Machine_Names();
	
	d_add_dialog->Reset(TAddDialog::MACHINE,existing_machines);
	
	if(QDialog::Accepted == d_add_dialog->exec())
	{
		d_machine_dialog->Create_Machine(d_add_dialog->Name());
		
		this->Update_Machine_Names();
	}
}

void TErrorSimulator::Remove_Machine(void)
{
	QList<QListWidgetItem*>				items;
	QList<QListWidgetItem*>::iterator	iter;
	int									machine_id;
	
	items = d_machines_list->selectedItems();
	
	for(iter = items.begin();iter != items.end();++iter)
	{
		machine_id = d_machine_dialog->Machine_Id((*iter)->text());
		d_machine_dialog->Remove_Machine(machine_id);
	}
	
	this->Update_Machine_Names();
}

void TErrorSimulator::Edit_Machine(void)
{
	QString								machine_name;
	
	machine_name = this->Active_Machine_Name();
	
	d_machine_dialog->Reset(machine_name);
	if(QDialog::Accepted == d_machine_dialog->exec())
	{
		this->Machine_Selection_Changed();	// force update of machine data
	}
}

void TErrorSimulator::Rename_Machine(
	QWidget 							*editor)
{
	QString 							text;
	int									id;
	
	text = reinterpret_cast<QLineEdit*>(editor)->text();
	id = d_machines_list->currentItem()->data(Qt::UserRole).toInt();
	
	text = d_machine_dialog->Rename_Machine(id,text);
	d_machines_list->currentItem()->setText(text);	// name may be different than what was entered
}

void TErrorSimulator::Add_Measurement(void)
{
	QStringList							list;
	QString								measurement_copy_name;
	
	list = this->Selected_Measurements();
	
	if(list.size() == 1)
	{
		measurement_copy_name = list[0];
	}

	d_measurement_dialog->Reset(QString(),measurement_copy_name);
	d_measurement_dialog->exec();
}

void TErrorSimulator::Remove_Measurement(void)
{
	QStringList							existing_measurements;
	QStringList::const_iterator			iter;
	int									measurement_id;
	
	existing_measurements = this->Selected_Measurements();
	
	for(iter = existing_measurements.begin();iter != existing_measurements.end();++iter)
	{
		measurement_id = d_measurement_dialog->Measurement_Id(*iter);
		d_measurement_dialog->Remove_Measurement(measurement_id);
	}
}

void TErrorSimulator::Edit_Measurement(void)
{
	QStringList							list;
	
	list = this->Selected_Measurements();
	
	if(list.size() == 1)
	{
		d_measurement_dialog->Reset(list[0],QString());
		if(QDialog::Accepted == d_measurement_dialog->exec())
		{
			this->Measurement_Selection_Changed();	// force update of measurement data
		}
	}
}

void TErrorSimulator::Rename_Measurement(
	QWidget 							*editor)
{
	QString 							text;
	int									id;
	
	text = reinterpret_cast<QLineEdit*>(editor)->text();
    id = d_measurements_list->currentItem()->data(Qt::UserRole).toInt();
	
	text = d_measurement_dialog->Rename_Measurement(id,text);
	d_measurements_list->currentItem()->setText(text);	// name may be different than what was entered
}

void TErrorSimulator::Display_Selected_Item(
	const int							class_id,
	const int							item_index)
{
	TMeasurement						*measurement(0);
	TMeasurement::TMeasurementItem		meas_item;
	double								deviation;
	
	d_display_widget->Clear_Selection_Data();

	if(item_index < 0 || class_id < 0)
	{
		return;
	}
	
	measurement = d_measurement_dialog->Measurement(class_id);
	
	if(measurement)
	{
		meas_item = measurement->Measurement_Item(item_index);
		
		if(measurement->Type() == Types::MEASUREMENT_BALLBAR)
		{
			d_display_widget->Add_Selection_Data_Line(QString("Name:           %1").arg(measurement->Name()));
		}
		else
		{
			d_display_widget->Add_Selection_Data_Line(QString("Name:           %1").arg(measurement->Name()));
			d_display_widget->Add_Selection_Data_Line(QString("Index:          %1").arg(item_index + 1));
		}
		
		d_display_widget->Add_Selection_Data_Line(QString("Nominal Length: %1").arg(meas_item.length_nominal,0,'f',4));
		d_display_widget->Add_Selection_Data_Line(QString("Actual Length:  %1").arg(meas_item.length_actual,0,'f',4));
		
		deviation = meas_item.length_actual - meas_item.length_nominal;
		
		d_display_widget->Add_Selection_Data_Line(QString("Deviation:       %1").arg(deviation,0,'f',4));
		
		if(measurement->Error_State())
		{
			d_display_widget->Add_Selection_Data_Line(QStringLiteral("Note:  Feature outside of the machine volume."));
		}
		
		d_display_widget->Add_Selection_Data_End();
	}
}

void TErrorSimulator::Export_Measurements(void)
{
	QString								machine;
	QStringList							list;
	std::vector<TMeasurement*>			measurements;
	std::vector<TMeasurement*>::iterator meas_iter;
	QString								dir;
	
	dir = QFileDialog::getExistingDirectory(this,
											QStringLiteral("Export Into Directory"),
											QString(),
											QFileDialog::ShowDirsOnly | QFileDialog::DontResolveSymlinks);
	
	if(!dir.length())
	{
		return;
	}
	
	machine = this->Active_Machine_Name();
	
	if(machine.length())
	{
		list = this->Selected_Measurements();
		
		measurements = d_measurement_dialog->Measurements(list);
		
		if(!this->Evaluate_Measurement_Expressions(&d_machine_comp,&measurements))
		{
			return;
		}
		
		this->Update_Measurement_Items(&d_machine_comp,&measurements);
		
		this->Export_Ballbar_Measurement(dir,measurements);
		
		for(meas_iter = measurements.begin();meas_iter != measurements.end();++meas_iter)
		{
			if((*meas_iter)->Type() == Types::MEASUREMENT_LASER)
			{
				this->Export_Laser_Measurement(dir,*meas_iter);
			}
			else if((*meas_iter)->Type() == Types::MEASUREMENT_MCG)
			{
				this->Export_MCG_Measurement(dir,*meas_iter);
			}
			else if((*meas_iter)->Type() == Types::MEASUREMENT_ISO)
			{
				this->Export_ISO_Measurement(dir,*meas_iter);
			}
		}
	}
}

void TErrorSimulator::Evaluate_Measurements(void)
{
	std::vector<TMeasurement*>			measurements;
	std::vector<TMachineDialog::TMapErrorItem> error_items;
	std::vector<std::vector<TMachineDialog::TMapErrorItem> >error_patterns;
	std::vector<TMachineDialog::TMapErrorItem>::const_iterator error_item_iter;
	std::vector<std::vector<TMachineDialog::TMapErrorItem> >::iterator error_pattern_iter;
	TMachineComp::TInitParameters		init_param;
	TMachineComp						test_machine;
	QString								machine_name;
	int									machine_id;
	QStringList							measurement_list;
	QString								line;
	QString								file_name;
	QFile								file;
	int									progress_value;
	double								min_error,max_error;
	double								min_error_length,max_error_length;
	QString								axis_text;
	
	machine_name = this->Active_Machine_Name();
	
	if(!machine_name.length())
	{
		return;
	}
	
	measurement_list = this->Selected_Measurements();
	
	if(!measurement_list.size())
	{
		return;
	}

	init_param = d_machine_comp.Init_Parameters();
	
	file_name = QFileDialog::getSaveFileName(this,
											 QStringLiteral("Create Evaluation Results File"),
											 QString(),
											 QStringLiteral("Text Files (*.txt)"));
	
	if(!file_name.length())
	{
		return;
	}
	
	machine_id = d_machine_dialog->Machine_Id(machine_name);
	
	if(machine_id < 0)
	{
		return;
	}
	
	file.setFileName(file_name);
	
	if(!file.open(QIODevice::WriteOnly | QIODevice::Text))
	{
		return;
	}
	
	if(!d_progress_dialog)
	{
		d_progress_dialog = new TProgressDialog(this,Qt::FramelessWindowHint|Qt::WindowStaysOnTopHint);
		
		connect(d_progress_dialog,SIGNAL(Cancel(void)),this,SLOT(Progress_Cancel(void)));

	}
	
	line = QString("Test Machine: %1\n").arg(machine_name);
	file.write(line.toLatin1());
	
	line = QStringLiteral("Machine Errors\n");
	file.write(line.toLatin1());

	measurements = d_measurement_dialog->Measurements(measurement_list);
	
	error_items = d_machine_dialog->Map_Error_Items(machine_id);
	for(error_item_iter = error_items.begin();error_item_iter != error_items.end();++error_item_iter)
	{
		switch((*error_item_iter).axis)
		{
			case Types::AXIS_X:
				
				switch((*error_item_iter).parameter)
				{
					case Types::PARAMETER_RX:
						line = QStringLiteral("X Axis Rx");
						break;
						
					case Types::PARAMETER_RY:
						line = QStringLiteral("X Axis Ry");
						break;
						
					case Types::PARAMETER_RZ:
						line = QStringLiteral("X Axis Rz");
						break;
						
					case Types::PARAMETER_SCALE:
						line = QStringLiteral("X Axis Scale");
						break;
				}
				
				break;
				
			case Types::AXIS_Y:
				
				switch((*error_item_iter).parameter)
				{
					case Types::PARAMETER_RX:
						line = QStringLiteral("Y Axis Rx");
						break;
						
					case Types::PARAMETER_RY:
						line = QStringLiteral("Y Axis Ry");
						break;
						
					case Types::PARAMETER_RZ:
						line = QStringLiteral("Y Axis Rz");
						break;
						
					case Types::PARAMETER_SCALE:
						line = QStringLiteral("Y Axis Scale");
						break;
				}
				
				break;
				
			case Types::AXIS_Z:
				
				switch((*error_item_iter).parameter)
				{
					case Types::PARAMETER_RX:
						line = QStringLiteral("Z Axis Rx");
						break;
						
					case Types::PARAMETER_RY:
						line = QStringLiteral("Z Axis Ry");
						break;
						
					case Types::PARAMETER_RZ:
						line = QStringLiteral("Z Axis Rz");
						break;
						
					case Types::PARAMETER_SCALE:
						line = QStringLiteral("Z Axis Scale");
						break;
				}
				
				break;
				
			case Types::AXIS_D:
				
				if((*error_item_iter).parameter == Types::PARAMETER_DEFLECTION)
				{
					line = QStringLiteral("D Axis Deflection");
				}
				
				break;
				
			case Types::AXIS_XYZ:
				
				switch((*error_item_iter).parameter)
				{
					case Types::PARAMETER_QXY:
						line = QStringLiteral("QXY");
						break;
						
					case Types::PARAMETER_QYZ:
						line = QStringLiteral("QYZ");
						break;
						
					case Types::PARAMETER_QZX:
						line = QStringLiteral("QZX");
						break;
				}
				
				break;
		}
		
		if((*error_item_iter).axis == Types::AXIS_D || (*error_item_iter).axis == Types::AXIS_XYZ)
		{
			line += QString(": %1\n").arg((*error_item_iter).value,0,'f',4);
		}
		else
		{
			line += QString(": %1\n").arg((*error_item_iter).expression);
		}
		
		file.write(line.toLatin1());
	}
	
	
	line = QStringLiteral("\n\n");
	file.write(line.toLatin1());
		
//             0         1         2         3         4         5         6         7         8         9         0         1         2         3         4         5         6         7         8         9
//             0123456789012345678901234567890123456789012345678901234567890123456789012345678901234567890123456789012345678901234567890123456789012345678901234567890123456789012345678901234567890
//                    |          |          |          |          |          |          |          |          |          |          |          |          |          |          |          |          |          |          |          |          |          |          |          |          |
	file.write("                                                                  |                  Ball Bar                   |                ISO 10360-2                |                    MCG                    |                   Laser                   |\n");
	file.write(" Active Compensation Parameters                                       Min Error  Max Error   Min Len    Max Len   Min Error  Max Error   Min Len    Max Len   Min Error  Max Error   Min Len    Max Len   Min Error  Max Error   Min Len    Max Len\n");
	file.write("------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------\n");


	this->setEnabled(false);
	d_progress_dialog->show();
	d_progress_dialog->setEnabled(true);	// automatically disabled with parent
	d_progress_cancel = false;

	error_patterns = this->Get_Evaluation_Error_Patterns(error_items);
	d_progress_dialog->Reset_Progress(0,error_patterns.size());
	
	progress_value = 0;
	for(error_pattern_iter = error_patterns.begin(); error_pattern_iter != error_patterns.end();++error_pattern_iter)
	{
		d_progress_dialog->Set_Progress(++progress_value);
		
		if(d_progress_cancel)
		{
			break;
		}
		
		test_machine.Initialize_Begin(init_param);
		axis_text = QStringLiteral("|");
		
		for(error_item_iter = (*error_pattern_iter).begin();error_item_iter != (*error_pattern_iter).end();++error_item_iter)
		{
			switch((*error_item_iter).parameter)
			{
				case Types::PARAMETER_RX:
					
					if((*error_item_iter).axis == Types::AXIS_X)
					{
						axis_text += QString("RXX|");
						test_machine.Set_Rx((*error_item_iter).axis,(*error_item_iter).expression,(*error_item_iter).sign);
					}
					
					if((*error_item_iter).axis == Types::AXIS_Y)
					{
						axis_text += QString("RYX|");
						test_machine.Set_Rx((*error_item_iter).axis,(*error_item_iter).expression,(*error_item_iter).sign);
					}
					
					if((*error_item_iter).axis == Types::AXIS_Z)
					{
						axis_text += QString("RZX|");
						test_machine.Set_Rx((*error_item_iter).axis,(*error_item_iter).expression,(*error_item_iter).sign);
					}
					break;
					
				case Types::PARAMETER_RY:
					
					if((*error_item_iter).axis == Types::AXIS_X)
					{
						axis_text += QString("RXY|");
						test_machine.Set_Ry((*error_item_iter).axis,(*error_item_iter).expression,(*error_item_iter).sign);
					}
					
					if((*error_item_iter).axis == Types::AXIS_Y)
					{
						axis_text += QString("RYY|");
						test_machine.Set_Ry((*error_item_iter).axis,(*error_item_iter).expression,(*error_item_iter).sign);
					}
					
					if((*error_item_iter).axis == Types::AXIS_Z)
					{
						axis_text += QString("RZY|");
						test_machine.Set_Ry((*error_item_iter).axis,(*error_item_iter).expression,(*error_item_iter).sign);
					}
					break;
					
				case Types::PARAMETER_RZ:
					
					if((*error_item_iter).axis == Types::AXIS_X)
					{
						axis_text += QString("RXZ|");
						test_machine.Set_Rz((*error_item_iter).axis,(*error_item_iter).expression,(*error_item_iter).sign);
					}
					
					if((*error_item_iter).axis == Types::AXIS_Y)
					{
						axis_text += QString("RYZ|");
						test_machine.Set_Rz((*error_item_iter).axis,(*error_item_iter).expression,(*error_item_iter).sign);
					}
					
					if((*error_item_iter).axis == Types::AXIS_Z)
					{
						axis_text += QString("RZZ|");
						test_machine.Set_Rz((*error_item_iter).axis,(*error_item_iter).expression,(*error_item_iter).sign);
					}
					break;
					
				case Types::PARAMETER_SCALE:
					
					if((*error_item_iter).axis == Types::AXIS_X)
					{
						axis_text += QString("LXX|");
						test_machine.Set_Scale((*error_item_iter).axis,(*error_item_iter).expression,(*error_item_iter).sign);
					}
					
					if((*error_item_iter).axis == Types::AXIS_Y)
					{
						axis_text += QString("LYY|");
						test_machine.Set_Scale((*error_item_iter).axis,(*error_item_iter).expression,(*error_item_iter).sign);
					}
					
					if((*error_item_iter).axis == Types::AXIS_Z)
					{
						axis_text += QString("LZZ|");
						test_machine.Set_Scale((*error_item_iter).axis,(*error_item_iter).expression,(*error_item_iter).sign);
					}
					break;
					
				case Types::PARAMETER_DEFLECTION:
					
					if((init_param.kinematic & Types::KINEMATIC_XZY) == Types::KINEMATIC_XZY)
					{
						if((*error_item_iter).axis == Types::AXIS_D)
						{
							axis_text += QString("RXD|");
							test_machine.Set_Deflection(Types::AXIS_D,(*error_item_iter).value,(*error_item_iter).sign);
						}
					}
					else if((init_param.kinematic & Types::KINEMATIC_YZX) == Types::KINEMATIC_YZX)
					{
						if((*error_item_iter).axis == Types::AXIS_D)
						{
							axis_text += QString("RYD|");
							test_machine.Set_Deflection(Types::AXIS_D,(*error_item_iter).value,(*error_item_iter).sign);
						}
					}
					break;
					
				case Types::PARAMETER_QXY:
					
					test_machine.Set_Quad(Types::AXIS_X,(*error_item_iter).value,(*error_item_iter).sign);
					axis_text += QString("QXY|");
					
					break;
					
				case Types::PARAMETER_QYZ:
					
					test_machine.Set_Quad(Types::AXIS_Y,(*error_item_iter).value,(*error_item_iter).sign);
					axis_text += QString("QYZ|");

					break;
					
				case Types::PARAMETER_QZX:
					
					test_machine.Set_Quad(Types::AXIS_Z,(*error_item_iter).value,(*error_item_iter).sign);
					axis_text += QString("QZX|");
					
					break;
			}
		}
		
		test_machine.Initialize_End();
		
		if(axis_text.length() > 1)	// must have at least one compensation parameter active
		{
			this->Evaluate_Measurement_Expressions(&test_machine,&measurements);
			this->Update_Measurement_Items(&test_machine,&measurements);
			
			line = this->Pad_String(axis_text,66,false);
			line += QStringLiteral("|");
			
			if(this->Get_Min_Max_Ballbar_Error(&test_machine,&measurements,&min_error,&max_error,&min_error_length,&max_error_length))
			{
				line += QString(" %1 %2 %3 %4")
								.arg(min_error,10,'f',6)
								.arg(max_error,10,'f',6)
								.arg(min_error_length,10,'f',4)
								.arg(max_error_length,10,'f',4);
				
			}
			else
			{
				line += QStringLiteral("   --------   --------  ---------  ---------");
			}
			
			if(this->Get_Min_Max_ISO_Error(&test_machine,&measurements,&min_error,&max_error,&min_error_length,&max_error_length))
			{
				line += QString(" %1 %2 %3 %4")
								.arg(min_error,10,'f',6)
								.arg(max_error,10,'f',6)
								.arg(min_error_length,10,'f',4)
								.arg(max_error_length,10,'f',4);
			}
			else
			{
				line += QStringLiteral("   --------   --------  ---------  ---------");
			}
			
			if(this->Get_Min_Max_MCG_Error(&test_machine,&measurements,&min_error,&max_error,&min_error_length,&max_error_length))
			{
				line += QString(" %1 %2 %3 %4")
								.arg(min_error,10,'f',6)
								.arg(max_error,10,'f',6)
								.arg(min_error_length,10,'f',4)
								.arg(max_error_length,10,'f',4);
				
			}
			else
			{
				line += QStringLiteral("   --------   --------  ---------  ---------");
			}
			
			if(this->Get_Min_Max_Laser_Error(&test_machine,&measurements,&min_error,&max_error,&min_error_length,&max_error_length))
			{
				line += QString(" %1 %2 %3 %4")
								.arg(min_error,10,'f',6)
								.arg(max_error,10,'f',6)
								.arg(min_error_length,10,'f',4)
								.arg(max_error_length,10,'f',4);
				
			}
			else
			{
				line += QStringLiteral("   --------   --------  ---------  ---------");
			}
					
			line += QStringLiteral("\n");
			file.write(line.toLatin1());
		}
		qApp->processEvents();
	}
	
	this->setEnabled(true);

	d_progress_dialog->End_Progress();
	d_progress_dialog->hide();
	
	file.close();
}

void TErrorSimulator::Progress_Cancel(void)
{
	d_progress_cancel = true;
}

void TErrorSimulator::Help_About(void)
{
	d_msg_box->setText(QStringLiteral("Error Simulator"));
	d_msg_box->setInformativeText(QStringLiteral("Error Simulator - Version 3.0"));
	d_msg_box->setDetailedText(QString());
	d_msg_box->setStandardButtons(QMessageBox::Ok);
	d_msg_box->setDefaultButton(QMessageBox::Ok);
	d_msg_box->setIcon(QMessageBox::Information);
	
	d_msg_box->exec();
}

void TErrorSimulator::closeEvent(
	QCloseEvent							*event)
{
	d_machine_dialog->Save_Data();
	d_measurement_dialog->Save_Data();
	
	event->accept();
}

QString TErrorSimulator::Active_Machine_Name(void) const
{
	QList<QListWidgetItem*>				items;
	QString								name;
	
	items = d_machines_list->selectedItems();
	
	if(items.size() == 1)
	{
		name = items[0]->text();
	}
	
	return name;
}

QStringList TErrorSimulator::Selected_Measurements(void) const
{
	QList<QListWidgetItem*>				items;
	QList<QListWidgetItem*>::const_iterator iter;
	QStringList							list;
	
	items = d_measurements_list->selectedItems();

	for(iter = items.begin();iter != items.end();++iter)
	{
		list.push_back((*iter)->text());
	}
	
	return list;
}

QString TErrorSimulator::Pad_String(
	const QString						&s,
	const int							width,
	const bool							add_comma) const
{
	QString								pad_string(s);
	
	if(add_comma)
	{
		pad_string += QStringLiteral(",");
	}
	
	if(pad_string.length() < width)
	{
		pad_string += QString(width - pad_string.length(),' ');
	}
	
	return pad_string;
}

void TErrorSimulator::Activate_Machine(
	const QString						&machine_name)
{
	QStringList							list;
	QStringList::const_iterator			iter;
	
	d_display_widget->Clear_Selection_Data();
	d_display_widget->Clear_Graphical_Data();
	
	if(!machine_name.length())
	{
		d_display_widget->Clear_Machine_Volume();
	}
	else
	{
		if(!this->Update_Machine_Comp(machine_name,&list))
		{
			d_display_widget->Add_Map_Data_Line(QStringLiteral("Map Initialization Errors"));
			d_display_widget->Add_Map_Data_Line(QStringLiteral("-------------------------"));
			
			d_display_widget->Add_Selection_Data_Line(QStringLiteral("Map Initialization Errors"));
			d_display_widget->Add_Selection_Data_Line(QStringLiteral("-------------------------"));
			
			for(iter = list.begin();iter != list.end();++iter)
			{
				d_display_widget->Add_Map_Data_Line(*iter);
				d_display_widget->Add_Selection_Data_Line(*iter);
				
			}
			
			d_display_widget->Add_Map_Data_End();
			d_display_widget->Add_Selection_Data_End();
			
			return;
		}
		
		this->Update_Map_Display();
		this->Measurement_Selection_Changed();	// force update of measurement data
	}
}

bool TErrorSimulator::Evaluate_Measurement_Expressions(
	TMachineComp						* const active_machine,
	std::vector<TMeasurement*>			* const measurements,
	QStringList							* const error_list)
{
	TMeasurement_Ballbar				*measurement_ballbar;
	TMeasurement_Laser					*measurement_laser;
	TMeasurement_ISO					*measurement_iso;
	TMeasurement_MCG					*measurement_mcg;
	std::vector<TMeasurement*>::iterator iter;
	TExpression							expression_evaluator;
	TVector3							pmin,pmax,pmid;
	TVector3							pnt;
	double								lmin,lmax,len;
	double								dval;
	int									ival;
	bool								valid;
	bool								return_state(true);
	static const double					ZER0_EPSILON(0.000001);
	
	assert(measurements);
	assert(active_machine);
	
	if(error_list)
	{
		error_list->clear();
	}
	
	pmin = active_machine->Map_Data_Min();
	
	expression_evaluator.Add_Variable(QStringLiteral("XMin"),pmin.x);
	expression_evaluator.Add_Variable(QStringLiteral("YMin"),pmin.y);
	expression_evaluator.Add_Variable(QStringLiteral("ZMin"),pmin.z);
	
	pmax = active_machine->Map_Data_Max();

	expression_evaluator.Add_Variable(QStringLiteral("XMax"),pmax.x);
	expression_evaluator.Add_Variable(QStringLiteral("YMax"),pmax.y);
	expression_evaluator.Add_Variable(QStringLiteral("ZMax"),pmax.z);
	
	pmid = (pmin + pmax)/2.0;
	
	expression_evaluator.Add_Variable(QStringLiteral("XMid"),pmid.x);
	expression_evaluator.Add_Variable(QStringLiteral("YMid"),pmid.y);
	expression_evaluator.Add_Variable(QStringLiteral("ZMid"),pmid.z);
	
	lmin = lmax = pmax.x - pmin.x;
	
	len = pmax.y - pmin.y;
	if(len < lmin)
	{
		lmin = len;
	}
	else if(len > lmax)
	{
		lmax = len;
	}
	
	len = pmax.z - pmin.z;
	if(len < lmin)
	{
		lmin = len;
	}
	else if(len > lmax)
	{
		lmax = len;
	}
	
	expression_evaluator.Add_Variable(QStringLiteral("AxisMin"),lmin);
	expression_evaluator.Add_Variable(QStringLiteral("AxisMax"),lmax);

	for(iter = measurements->begin();iter != measurements->end();++iter)
	{
		
		switch((*iter)->Type())
		{
			case Types::MEASUREMENT_LASER:
				measurement_laser = reinterpret_cast<TMeasurement_Laser*>((*iter));
				
				pnt.x = expression_evaluator.Evaluate(measurement_laser->Start_Position_X(),&valid);
				
				if(!valid)
				{
					return_state = false;
					
					if(error_list)
						error_list->push_back(QString("%1: Laser_Measurement Expression: Start_Position_X invalid").arg((*iter)->Name()));
				}
				
				pnt.y = expression_evaluator.Evaluate(measurement_laser->Start_Position_Y(),&valid);
				
				if(!valid)
				{
					return_state = false;
					
					if(error_list)
						error_list->push_back(QString("%1: Laser_Measurement Expression: Start_Position_Y invalid").arg((*iter)->Name()));
				}
				
				pnt.z = expression_evaluator.Evaluate(measurement_laser->Start_Position_Z(),&valid);
				
				
				if(!valid)
				{
					return_state = false;
					
					if(error_list)
						error_list->push_back(QString("%1: Laser_Measurement Expression: Start_Position_Z invalid").arg((*iter)->Name()));
				}
				
				measurement_laser->Set_Calc_Start_Position(pnt);
				
				pnt.x = expression_evaluator.Evaluate(measurement_laser->End_Position_X(),&valid);
				
				
				if(!valid)
				{
					return_state = false;
					
					if(error_list)
						error_list->push_back(QString("%1: Laser_Measurement Expression: End_Position_X invalid").arg((*iter)->Name()));
				}
				
				pnt.y = expression_evaluator.Evaluate(measurement_laser->End_Position_Y(),&valid);
				
				
				if(!valid)
				{
					return_state = false;
					
					if(error_list)
						error_list->push_back(QString("%1: Laser_Measurement Expression: End_Position_Y invalid").arg((*iter)->Name()));
				}
				
				pnt.z = expression_evaluator.Evaluate(measurement_laser->End_Position_Z(),&valid);
				
				
				if(!valid)
				{
					return_state = false;
					
					if(error_list)
						error_list->push_back(QString("%1: Laser_Measurement Expression: End_Position_Z invalid").arg((*iter)->Name()));
				}
				
				measurement_laser->Set_Calc_End_Position(pnt);
				
				dval = expression_evaluator.Evaluate(measurement_laser->Increment(),&valid);
				
				if(!valid)
				{
					return_state = false;
					
					if(error_list)
						error_list->push_back(QString("%1: Laser_Measurement Expression: Increment invalid").arg((*iter)->Name()));
				}
				
				if(dval < 20)
				{
					return_state = false;
					
					if(error_list)
						error_list->push_back(QString("%1: Laser_Measurement Expression: Increment too short").arg((*iter)->Name()));
				}

				measurement_laser->Set_Calc_Increment(dval);


				break;
				
			case Types::MEASUREMENT_MCG:
				measurement_mcg = reinterpret_cast<TMeasurement_MCG*>((*iter));
				
				// calculate length first as it may be used in other expressions
				dval = expression_evaluator.Evaluate(measurement_mcg->Length(),&valid);
				
				if(!valid)
				{
					return_state = false;
					
					if(error_list)
						error_list->push_back(QString("%1: MCG_Measurement Expression: Length invalid").arg((*iter)->Name()));
				}
				
				if(dval < 50)
				{
					return_state = false;
					
					if(error_list)
						error_list->push_back(QString("%1: MCG_Measurement Expression: Length too short").arg((*iter)->Name()));
				}

				measurement_mcg->Set_Calc_Length(dval);
				
				expression_evaluator.Add_Variable(QStringLiteral("L"),dval);
				expression_evaluator.Add_Variable(QStringLiteral("R"),dval/2.0);

				
				pnt.x = expression_evaluator.Evaluate(measurement_mcg->Position_X(),&valid);
				
				
				if(!valid)
				{
					return_state = false;
					
					if(error_list)
						error_list->push_back(QString("%1: MCG_Measurement Expression: Position_X invalid").arg((*iter)->Name()));
				}
				
				pnt.y = expression_evaluator.Evaluate(measurement_mcg->Position_Y(),&valid);
				
				
				if(!valid)
				{
					return_state = false;
					
					if(error_list)
						error_list->push_back(QString("%1: MCG_Measurement Expression: Position_Y invalid").arg((*iter)->Name()));
				}
				
				pnt.z = expression_evaluator.Evaluate(measurement_mcg->Position_Z(),&valid);
				
				
				if(!valid)
				{
					return_state = false;
					
					if(error_list)
						error_list->push_back(QString("%1: MCG_Measurement Expression: Position_Z invalid").arg((*iter)->Name()));
				}
				
				measurement_mcg->Set_Calc_Position(pnt);
				
				ival = expression_evaluator.Evaluate(measurement_mcg->Azimuth_Count(),&valid);
				
				if(!valid)
				{
					return_state = false;
					
					if(error_list)
						error_list->push_back(QString("%1: MCG_Measurement Expression: Azimuth_Count invalid").arg((*iter)->Name()));
				}
				
				if(ival > 36)
				{
					return_state = false;
					
					if(error_list)
						error_list->push_back(QString("%1: MCG_Measurement Expression: Azimuth_Count too large").arg((*iter)->Name()));
				}

				
				measurement_mcg->Set_Calc_Azimuth_Count(ival);
				
				
				ival = expression_evaluator.Evaluate(measurement_mcg->Elevation_Count(),&valid);
				
				
				if(!valid)
				{
					return_state = false;
					
					if(error_list)
						error_list->push_back(QString("%1: MCG_Measurement Expression: Elevation_Count invalid").arg((*iter)->Name()));
				}
				
				if(ival > 144)
				{
					return_state = false;
					
					if(error_list)
						error_list->push_back(QString("%1: MCG_Measurement Expression: Elevation_Count too large").arg((*iter)->Name()));
				}

				measurement_mcg->Set_Calc_Elevation_Count(ival);
				
				
				dval = expression_evaluator.Evaluate(measurement_mcg->Start_Elevation(),&valid);
				
				
				if(!valid)
				{
					return_state = false;
					
					if(error_list)
						error_list->push_back(QString("%1: MCG_Measurement Expression: Start_Elevation invalid").arg((*iter)->Name()));
				}
				
				
				if(fabs(dval) > 45)
				{
					return_state = false;
					
					if(error_list)
						error_list->push_back(QString("%1: MCG_Measurement Expression: Start_Elevation out of range").arg((*iter)->Name()));
				}

				measurement_mcg->Set_Calc_Start_Elevation(dval);
				
				
				dval = expression_evaluator.Evaluate(measurement_mcg->End_Elevation(),&valid);
				
				
				if(!valid)
				{
					return_state = false;
					
					if(error_list)
						error_list->push_back(QString("%1: MCG_Measurement Expression: End_Elevation invalid").arg((*iter)->Name()));
				}
				
				if(fabs(dval) > 45)
				{
					return_state = false;
					
					if(error_list)
						error_list->push_back(QString("%1: MCG_Measurement Expression: End_Elevation out of range").arg((*iter)->Name()));
				}

				measurement_mcg->Set_Calc_End_Elevation(dval);


				break;
				
			case Types::MEASUREMENT_ISO:
				measurement_iso = reinterpret_cast<TMeasurement_ISO*>((*iter));
				
				
				pnt.x = expression_evaluator.Evaluate(measurement_iso->Start_Position_X(),&valid);
				
				
				if(!valid)
				{
					return_state = false;
					
					if(error_list)
						error_list->push_back(QString("%1: ISO_Measurement Expression: Start_Position_X invalid").arg((*iter)->Name()));
				}
				
				pnt.y = expression_evaluator.Evaluate(measurement_iso->Start_Position_Y(),&valid);
				
				
				if(!valid)
				{
					return_state = false;
					
					if(error_list)
						error_list->push_back(QString("%1: ISO_Measurement Expression: Start_Position_Y invalid").arg((*iter)->Name()));
				}
				
				pnt.z = expression_evaluator.Evaluate(measurement_iso->Start_Position_Z(),&valid);
				
				
				if(!valid)
				{
					return_state = false;
					
					if(error_list)
						error_list->push_back(QString("%1: ISO_Measurement Expression: Start_Position_Z invalid").arg((*iter)->Name()));
				}
				
				measurement_iso->Set_Calc_Start_Position(pnt);
				
				
				pnt.x = expression_evaluator.Evaluate(measurement_iso->End_Position_X(),&valid);
				
				
				if(!valid)
				{
					return_state = false;
					
					if(error_list)
						error_list->push_back(QString("%1: ISO_Measurement Expression: End_Position_X invalid").arg((*iter)->Name()));
				}
				
				pnt.y = expression_evaluator.Evaluate(measurement_iso->End_Position_Y(),&valid);
				
				
				if(!valid)
				{
					return_state = false;
					
					if(error_list)
						error_list->push_back(QString("%1: ISO_Measurement Expression: End_Position_Y invalid").arg((*iter)->Name()));
				}
				
				pnt.z = expression_evaluator.Evaluate(measurement_iso->End_Position_Z(),&valid);
				
				
				if(!valid)
				{
					return_state = false;
					
					if(error_list)
						error_list->push_back(QString("%1: ISO_Measurement Expression: End_Position_Z invalid").arg((*iter)->Name()));
				}
				
				measurement_iso->Set_Calc_End_Position(pnt);
				
				break;
				
			case Types::MEASUREMENT_BALLBAR:
				measurement_ballbar = reinterpret_cast<TMeasurement_Ballbar*>((*iter));				

				// calculate length first as it may be used in other expressions
				dval = expression_evaluator.Evaluate(measurement_ballbar->Length(),&valid);
				
				
				if(!valid)
				{
					return_state = false;
					
					if(error_list)
						error_list->push_back(QString("%1: BallBar_Measurement Expression: Length invalid").arg((*iter)->Name()));
				}
				
				if(dval < 50)
				{
					return_state = false;
					
					if(error_list)
						error_list->push_back(QString("%1: BallBar_Measurement Expression: Length too short").arg((*iter)->Name()));
				}

				
				measurement_ballbar->Set_Calc_Length(dval);
				
				expression_evaluator.Add_Variable(QStringLiteral("L"),dval);
				expression_evaluator.Add_Variable(QStringLiteral("R"),dval/2.0);
				
				pnt.x = expression_evaluator.Evaluate(measurement_ballbar->Direction_I(),&valid);
				
				
				if(!valid)
				{
					return_state = false;
					
					if(error_list)
						error_list->push_back(QString("%1: BallBar_Measurement Expression: Direction_I invalid").arg((*iter)->Name()));
				}
				
				pnt.y = expression_evaluator.Evaluate(measurement_ballbar->Direction_J(),&valid);
				
				
				if(!valid)
				{
					return_state = false;
					
					if(error_list)
						error_list->push_back(QString("%1: BallBar_Measurement Expression: Direction_J invalid").arg((*iter)->Name()));
				}
				
				pnt.z = expression_evaluator.Evaluate(measurement_ballbar->Direction_K(),&valid);
				
				
				if(!valid)
				{
					return_state = false;
					
					if(error_list)
						error_list->push_back(QString("%1: BallBar_Measurement Expression: Direction_K invalid").arg((*iter)->Name()));
				}
				
				if(pnt.Length() < ZER0_EPSILON)
				{
					if(error_list)
						error_list->push_back(QString("%1: BallBar_Measurement Expression: Direction_IJK invalid.  Using <1,0,0>").arg((*iter)->Name()));
					
					pnt.Set(1,0,0);
				}
				
				pnt.Normal();
				
				measurement_ballbar->Set_Calc_Direction(pnt);

				expression_evaluator.Add_Variable(QStringLiteral("I"),pnt.x);
				expression_evaluator.Add_Variable(QStringLiteral("J"),pnt.y);
				expression_evaluator.Add_Variable(QStringLiteral("K"),pnt.z);
				
				pnt.x = expression_evaluator.Evaluate(measurement_ballbar->Position_X(),&valid);
				
				
				if(!valid)
				{
					return_state = false;
					
					if(error_list)
						error_list->push_back(QString("%1: BallBar_Measurement Expression: Position_X invalid").arg((*iter)->Name()));
				}
				
				pnt.y = expression_evaluator.Evaluate(measurement_ballbar->Position_Y(),&valid);
				
				
				if(!valid)
				{
					return_state = false;
					
					if(error_list)
						error_list->push_back(QString("%1: BallBar_Measurement Expression: Position_Y invalid").arg((*iter)->Name()));
				}
				
				pnt.z = expression_evaluator.Evaluate(measurement_ballbar->Position_Z(),&valid);
				
				
				if(!valid)
				{
					return_state = false;
					
					if(error_list)
						error_list->push_back(QString("%1: BallBar_Measurement Expression: Position_Z invalid").arg((*iter)->Name()));
				}
				
				measurement_ballbar->Set_Calc_Position(pnt);

				break;
				
			default:
				return false;
		}
	}
	
	return return_state;
}

void TErrorSimulator::Update_Measurement_Items(
	TMachineComp						* const machine,
	std::vector<TMeasurement*>			* const measurements)
{
	TMeasurement_Ballbar				*measurement_ballbar(0);
	TMeasurement_Laser					*measurement_laser(0);
	TMeasurement_ISO					*measurement_iso(0);
	TMeasurement_MCG					*measurement_mcg(0);
	TMeasurement::TMeasurementItem		meas_item;
	std::vector<TMeasurement*>::iterator iter;
	TVector3							v1,p1,p2,pnt;
	TVector3							probe_axis;
	double								length;
	double								increment,increment_elevation;
	double								dval,az,ax;
	double								start,end;
	int									cntr,azimuth_count,azimuth_cntr;
	int									elevation_count,elevation_cntr;
	bool								valid;
	static const double					PI2(6.28318530717959);
	static const double					ZERO_EPSILON(0.000001);
	
	assert(machine);
	assert(measurements);
	
	for(iter = measurements->begin();iter != measurements->end();++iter)
	{
		switch((*iter)->Type())
		{
			case Types::MEASUREMENT_LASER:

				measurement_laser = reinterpret_cast<TMeasurement_Laser*>((*iter));
				
				measurement_laser->Clear_Measurement_Items();
				measurement_laser->Set_Error_State(false);
				
				meas_item.probe_offset = measurement_laser->Probe_Offset();
				
				p1 = measurement_laser->Calc_Start_Position();
				p2 = measurement_laser->Calc_End_Position();
							
				v1 = p2 - p1;
				length = v1.Length();
				
				if(length < ZERO_EPSILON)
				{
					v1.Set(1,0,0);
				}
				
				v1.Normal();
				meas_item.axis_nominal = v1;
				
				increment = measurement_laser->Calc_Increment();
				
				if(increment < 1.0)
				{
					increment = 10;
				}
				
				meas_item.start_nominal = p1;
				
				meas_item.start_actual = machine->Evaluate(p1,meas_item.probe_offset,&valid);
				if(!valid) measurement_laser->Set_Error_State(true);

				meas_item.axis_actual = machine->Evaluate(p2,meas_item.probe_offset,&valid);
				if(!valid) measurement_laser->Set_Error_State(true);
				
				meas_item.axis_actual -= meas_item.start_actual;
				meas_item.axis_actual.Normal();
				
				// set angular data for laser measurements
				probe_axis = meas_item.probe_offset;
				probe_axis.z = 0;
				
				if(probe_axis.Length() < ZERO_EPSILON)
				{
					// assume probe is pointing down
					probe_axis.Set(0,0,-1);
				}
				else
				{
					probe_axis.Normal();
				}
				
				meas_item.start_map_abc = machine->Get_Angles_RA_RB_RC(meas_item.axis_nominal,probe_axis,meas_item.start_nominal);
				
				dval = 0;
				while(!(dval > length))
				{
					pnt = p1 + dval * v1;
					
					meas_item.end_nominal = pnt;
					meas_item.end_actual = machine->Evaluate(pnt,meas_item.probe_offset,&valid);
					if(!valid) measurement_laser->Set_Error_State(true);

					pnt = meas_item.end_nominal - meas_item.start_nominal;
					meas_item.length_nominal = pnt.Length();
					
					pnt = meas_item.end_actual - meas_item.start_actual;
					meas_item.length_actual = pnt.Length();
					
					meas_item.end_map_abc = machine->Get_Angles_RA_RB_RC(meas_item.axis_nominal,probe_axis,meas_item.end_nominal);
					
					measurement_laser->Add_Measurement_Item(meas_item);
					
					dval += increment;
				};
				
				break;
				
			case Types::MEASUREMENT_MCG:

				measurement_mcg = reinterpret_cast<TMeasurement_MCG*>((*iter));
				
				measurement_mcg->Clear_Measurement_Items();
				measurement_mcg->Set_Error_State(false);
				
				meas_item.probe_offset = measurement_mcg->Probe_Offset();
				
				p1 = measurement_mcg->Calc_Position();
				length = measurement_mcg->Calc_Length();
				
				azimuth_count = measurement_mcg->Calc_Azimuth_Count();
				elevation_count = measurement_mcg->Calc_Elevation_Count();
				
				if(azimuth_count < 2)
				{
					azimuth_count = 2;
				}
				
				if(elevation_count < 2)
				{
					elevation_count = 2;
				}
				
				start = measurement_mcg->Calc_Start_Elevation() * PI2 / 360.0;	// angle converted from degrees
				end = measurement_mcg->Calc_End_Elevation() * PI2 / 360.0;	// angle converted from degrees
				
				if(fabs(end - start) < 1.0)
				{
					// +/- 1 deg.
					start = -0.017453292519943;
					end = 0.017453292519943;
				}
				
				increment_elevation = (end - start) / static_cast<double>(elevation_count - 1);
				increment = PI2 / static_cast<double>(azimuth_count);
				
				meas_item.start_nominal = p1;
				meas_item.start_actual = machine->Evaluate(p1,meas_item.probe_offset,&valid);
				if(!valid) measurement_mcg->Set_Error_State(true);

				for(elevation_cntr = 0;elevation_cntr < elevation_count;++elevation_cntr)
				{
					az = start + static_cast<double>(elevation_cntr) * increment_elevation;

					for(azimuth_cntr = 0;azimuth_cntr < azimuth_count;++azimuth_cntr)
					{
						ax = static_cast<double>(azimuth_cntr) * increment;
						
						v1.x = cos(ax) * cos(az);
						v1.y = sin(ax) * cos(az);
						v1.z = sin(az);
						
						meas_item.axis_nominal = v1;
						
						p2 = p1 + v1 * length;
						
						meas_item.end_nominal = p2;
						meas_item.end_actual = machine->Evaluate(p2,meas_item.probe_offset,&valid);
						if(!valid) measurement_mcg->Set_Error_State(true);

						pnt = meas_item.end_nominal - meas_item.start_nominal;
						meas_item.length_nominal = pnt.Length();
						
						pnt = meas_item.end_actual - meas_item.start_actual;
						meas_item.length_actual = pnt.Length();
						
						v1 = meas_item.end_actual - meas_item.start_actual;
						if(v1.Length() > ZERO_EPSILON)
						{
							v1.Normal();
						}
						else
						{
							v1.Set(1,0,0);
						}
						
						meas_item.axis_actual = v1;
						
						measurement_mcg->Add_Measurement_Item(meas_item);
					}
				}
				break;
				
			case Types::MEASUREMENT_ISO:

				measurement_iso = reinterpret_cast<TMeasurement_ISO*>((*iter));
				
				measurement_iso->Clear_Measurement_Items();
				measurement_iso->Set_Error_State(false);

				meas_item.probe_offset = measurement_iso->Probe_Offset();

				p1 = measurement_iso->Calc_Start_Position();
				p2 = measurement_iso->Calc_End_Position();
								
				v1 = p2 - p1;
				length = v1.Length();
				
				if(length < ZERO_EPSILON)
				{
					v1.Set(1,0,0);
				}
				
				v1.Normal();
				meas_item.axis_nominal = v1;
				
				increment = length / 5.0;	// 5 lengths required starting at zero and ending at max
				
				// round increment to nearest 1 mm
				increment = static_cast<int>(increment + 0.5);

				meas_item.start_nominal = p1;
				meas_item.start_actual = machine->Evaluate(p1,meas_item.probe_offset,&valid);
				if(!valid) measurement_iso->Set_Error_State(true);

				meas_item.axis_actual = machine->Evaluate(p2,meas_item.probe_offset,&valid);
				if(!valid) measurement_iso->Set_Error_State(true);
				
				meas_item.axis_actual -= meas_item.start_actual;
				meas_item.axis_actual.Normal();
				
				// set angular data for laser measurements
				probe_axis = meas_item.probe_offset;
				probe_axis.z = 0;
				
				if(probe_axis.Length() < ZERO_EPSILON)
				{
					// assume probe is pointing down
					probe_axis.Set(0,0,-1);
				}
				else
				{
					probe_axis.Normal();
				}
				
				meas_item.start_map_abc = machine->Get_Angles_RA_RB_RC(meas_item.axis_nominal,probe_axis,meas_item.start_nominal);

				for(cntr = 0,dval = increment;cntr < 5;++cntr,dval += increment)
				{
					pnt = p1 + dval * v1;
					
					meas_item.end_nominal = pnt;
					meas_item.end_actual = machine->Evaluate(pnt,meas_item.probe_offset,&valid);
					if(!valid) measurement_iso->Set_Error_State(true);

					pnt = meas_item.end_nominal - meas_item.start_nominal;
					meas_item.length_nominal = pnt.Length();
					
					pnt = meas_item.end_actual - meas_item.start_actual;
					meas_item.length_actual = pnt.Length();
					
					meas_item.end_map_abc = machine->Get_Angles_RA_RB_RC(meas_item.axis_nominal,probe_axis,meas_item.end_nominal);
					
					measurement_iso->Add_Measurement_Item(meas_item);
				}
				break;
				
			case Types::MEASUREMENT_BALLBAR:

				measurement_ballbar = reinterpret_cast<TMeasurement_Ballbar*>((*iter));
				
				measurement_ballbar->Clear_Measurement_Items();
				measurement_ballbar->Set_Error_State(false);

				meas_item.probe_offset = measurement_ballbar->Probe_Offset();

				p1 = measurement_ballbar->Calc_Position();
				v1 = measurement_ballbar->Calc_Direction();
							
				length = measurement_ballbar->Calc_Length();
				
				pnt = p1 - v1 * length/2.0;
				
				meas_item.start_nominal = pnt;
				meas_item.start_actual = machine->Evaluate(pnt,meas_item.probe_offset,&valid);
				if(!valid) measurement_ballbar->Set_Error_State(true);

				pnt = p1 + v1 * length/2.0;
				
				meas_item.end_nominal = pnt;
				meas_item.end_actual = machine->Evaluate(pnt,meas_item.probe_offset,&valid);
				if(!valid) measurement_ballbar->Set_Error_State(true);

				pnt = meas_item.end_nominal - meas_item.start_nominal;
				meas_item.length_nominal = pnt.Length();
				
				pnt = meas_item.end_actual - meas_item.start_actual;
				meas_item.length_actual = pnt.Length();
				
				meas_item.axis_actual = meas_item.end_actual - meas_item.start_actual;
				if(meas_item.axis_actual.Length() > ZERO_EPSILON)
				{
					meas_item.axis_actual.Normal();
				}
				else
				{
					meas_item.axis_actual.Set(1,0,0);
				}
				
				meas_item.axis_nominal = meas_item.end_nominal - meas_item.start_nominal;
				if(meas_item.axis_nominal.Length() > ZERO_EPSILON)
				{
					meas_item.axis_nominal.Normal();
				}
				else
				{
					meas_item.axis_nominal.Set(1,0,0);
				}
				
				measurement_ballbar->Add_Measurement_Item(meas_item);

				break;
		}
	}
}

QStringList TErrorSimulator::Report_Measurement_Item(
	TMeasurement						* const measurement) const
{
	TMeasurement::TMeasurementItem		meas_item;
	QStringList							lines;
	QString								line;
	TVector3							v1;
	TVector3							p1,p2,p3,p4;
	TVector3							pzero;
	TVector3							probe_axis;
	TVector3							probe_offset;
	TVector3							scale_xx_yy_error,scale_xx_yy_error0;
	int									cntr;
	double								nominal,actual,error;
	double								nominal0(0),actual0(0);
	double								min_error(0),max_error(0);
	double								azimuth,elevation;
	TReferenceSystem					refsys;
	static const double					RAD2DEGREE(57.29577951308233);
	static const double					ZERO_EPSILON(0.000001);

	assert(measurement);
	
	switch(measurement->Type())
	{
		case Types::MEASUREMENT_LASER:
			line = QStringLiteral("Laser Measurement");
			lines.push_back(line);
			
			line = QStringLiteral("--------------------------------------------------------------------------------------------------------------------------------");
			lines.push_back(line);
			
			for(cntr = 0;cntr < measurement->Measurement_Item_Count();++cntr)
			{
				meas_item = measurement->Measurement_Item(cntr);
				
				if(cntr == 0)	// initialize output information
				{
					if(measurement->Error_State())
					{
						line = QString("Name:           %1  Note:  outside of machine volume").arg(measurement->Name());
					}
					else
					{
						line = QString("Name:           %1").arg(measurement->Name());
					}
					lines.push_back(line);
					
					probe_offset = meas_item.probe_offset;
					
					line = QString("Probe Offset:   %1, %2, %3")
									.arg(probe_offset.x,0,'f',4)
									.arg(probe_offset.y,0,'f',4)
									.arg(probe_offset.z,0,'f',4);
					lines.push_back(line);
					
					line = QString("Start Position: %1, %2, %3")
									.arg(meas_item.start_nominal.x + probe_offset.x,0,'f',4)
									.arg(meas_item.start_nominal.y + probe_offset.y,0,'f',4)
									.arg(meas_item.start_nominal.z + probe_offset.z,0,'f',4);
					lines.push_back(line);
					
					line = QString("Test Axis:      %1, %2, %3")
									.arg(meas_item.axis_nominal.x,0,'f',9)
									.arg(meas_item.axis_nominal.y,0,'f',9)
									.arg(meas_item.axis_nominal.z,0,'f',9);
					lines.push_back(line);

					lines.push_back(QString());
					
					probe_axis = probe_offset;
					probe_axis.z = 0;
					
					if(probe_axis.Length() > ZERO_EPSILON)
					{
						probe_axis.Normal();
					}
					else
					{
						probe_axis.Set(0,0,-1);
					}
					
					nominal0 = meas_item.start_nominal ^ meas_item.axis_nominal;
					actual0 = meas_item.start_actual ^ meas_item.axis_nominal;
					scale_xx_yy_error0 = d_machine_comp.Calculate_Scale_XX_YY_Error(meas_item.axis_nominal,probe_axis,meas_item.start_nominal,meas_item.start_actual);
					
//                                         0         1         2         3         4         5         6         7         8         9
//                                         0123456789012345678901234567890123456789012345678901234567890123456789012345678901234567890
//                                               |          |          |          |          |          |          |          |
					line = QStringLiteral("  Scale Nom  Scale Act  Scale Dev      XX         YY         Ra         Rb         Rc");
					lines.push_back(line);
					
					line = QStringLiteral("--------------------------------------------------------------------------------------------------------------------------------");
					lines.push_back(line);
				}

				scale_xx_yy_error = d_machine_comp.Calculate_Scale_XX_YY_Error(meas_item.axis_nominal,probe_axis,meas_item.end_nominal,meas_item.end_actual);
				scale_xx_yy_error -= scale_xx_yy_error0;
								
				nominal = meas_item.end_nominal ^ meas_item.axis_nominal;
				actual = meas_item.end_actual ^ meas_item.axis_nominal;
				
				nominal -= nominal0;
				actual -= actual0;
				
				error = actual - nominal;
				
				if(cntr == 0)
				{
					min_error = max_error = error;
				}
				else
				{
					if(error > max_error)
					{
						max_error = error;
					}
					else if(error < min_error)
					{
						min_error = error;
					}
				}
				
				line = QString("%1 %2 %3 %4 %5 %6 %7 %8")
						.arg(nominal,10,'f',4)
						.arg(actual,10,'f',4)
						.arg(error,10,'f',4)
						.arg(scale_xx_yy_error.y,10,'f',4)
						.arg(scale_xx_yy_error.z,10,'f',4)
						.arg(meas_item.end_map_abc.x - meas_item.start_map_abc.x,10,'f',4)
						.arg(meas_item.end_map_abc.y - meas_item.start_map_abc.y,10,'f',4)
						.arg(meas_item.end_map_abc.z - meas_item.start_map_abc.z,10,'f',4);
				
				lines.push_back(line);
			}
			
			lines.push_back(QString());
			
			line = QString("Max Scale Error: %1").arg(max_error,7,'f',4);
			lines.push_back(line);
			
			line = QString("Min Scale Error: %1").arg(min_error,7,'f',4);
			lines.push_back(line);

			lines.push_back(QString());

			break;
			
		case Types::MEASUREMENT_MCG:
			line = QStringLiteral("MCG Measurement");
			lines.push_back(line);
			
			line = QStringLiteral("--------------------------------------------------------------------------------------------------------------------------------");
			lines.push_back(line);
			
			for(cntr = 0;cntr < measurement->Measurement_Item_Count();++cntr)
			{
				meas_item = measurement->Measurement_Item(cntr);
				
				if(cntr == 0)	// initialize output information
				{
					if(measurement->Error_State())
					{
						line = QString("Name:           %1  Note:  outside of machine volume").arg(measurement->Name());
					}
					else
					{
						line = QString("Name:           %1").arg(measurement->Name());
					}
					lines.push_back(line);
					
					probe_offset = meas_item.probe_offset;
					
					line = QString("Probe Offset:    %1, %2, %3")
								.arg(probe_offset.x,0,'f',4)
								.arg(probe_offset.y,0,'f',4)
								.arg(probe_offset.z,0,'f',4);
					lines.push_back(line);
					
					line = QString("Center Position: %1, %2, %3")
								.arg(meas_item.start_nominal.x + probe_offset.x,0,'f',4)
								.arg(meas_item.start_nominal.y + probe_offset.y,0,'f',4)
								.arg(meas_item.start_nominal.z + probe_offset.z,0,'f',4);
					lines.push_back(line);
					
					lines.push_back(QString());
					
//                                         0         1         2         3         4         5         6         7         8         9
//                                         0123456789012345678901234567890123456789012345678901234567890123456789012345678901234567890
//                                               |          |          |          |          |          |          |          |
					line = QStringLiteral("  Elevation   Azimuth     Length     Error");
					lines.push_back(line);
					
					line = QStringLiteral("--------------------------------------------------------------------------------------------------------------------------------");
					lines.push_back(line);
				}
				
				elevation = asin(meas_item.axis_actual.z) * RAD2DEGREE;
				
				p1 = meas_item.axis_actual;
				p1.z = 0;
				p1.Normal();
				
				azimuth = acos(p1.x) * RAD2DEGREE;
				
				if(p1.y < 0)
				{
					azimuth = 360 - azimuth;
				}

				nominal = meas_item.length_nominal;
				actual = meas_item.length_actual;
				
				error = actual - nominal;
				
				if(cntr == 0)
				{
					min_error = max_error = error;
					
				}
				else
				{
					if(error > max_error)
					{
						max_error = error;
					}
					else if(error < min_error)
					{
						min_error = error;
					}
				}

				line = QString("%1 %2 %3 %4")
								.arg(elevation,10,'f',4)
								.arg(azimuth,10,'f',4)
								.arg(actual,10,'f',4)
								.arg(error,10,'f',4);
				lines.push_back(line);
			}
			
			lines.push_back(QString());
			
			line = QString("Max Length Error: %1").arg(max_error,7,'f',4);
			lines.push_back(line);
			
			line = QString("Min Length Error: %1").arg(min_error,7,'f',4);
			lines.push_back(line);
			
			lines.push_back(QString());

			break;
			
		case Types::MEASUREMENT_ISO:
			line = QStringLiteral("ISO 10360-2 Measurement");
			lines.push_back(line);
			
			line = QStringLiteral("--------------------------------------------------------------------------------------------------------------------------------");
			lines.push_back(line);
			
			for(cntr = 0;cntr < measurement->Measurement_Item_Count();++cntr)
			{
				meas_item = measurement->Measurement_Item(cntr);
				
				if(cntr == 0)	// initialize output information
				{
					if(measurement->Error_State())
					{
						line = QString("Name:           %1  Note:  outside of machine volume").arg(measurement->Name());
					}
					else
					{
						line = QString("Name:           %1").arg(measurement->Name());
					}
					lines.push_back(line);
					
					probe_offset = meas_item.probe_offset;
					
					line = QString("Probe Offset:   %1, %2, %3")
								.arg(probe_offset.x,0,'f',4)
								.arg(probe_offset.y,0,'f',4)
								.arg(probe_offset.z,0,'f',4);
					lines.push_back(line);
					
					line = QString("Start Position: %1, %2, %3")
								.arg(meas_item.start_nominal.x + probe_offset.x,0,'f',4)
								.arg(meas_item.start_nominal.y + probe_offset.y,0,'f',4)
								.arg(meas_item.start_nominal.z + probe_offset.z,0,'f',4);
					lines.push_back(line);
					
					line = QString("Test Axis:      %1, %2, %3")
								.arg(meas_item.axis_nominal.x,0,'f',9)
								.arg(meas_item.axis_nominal.y,0,'f',9)
								.arg(meas_item.axis_nominal.z,0,'f',9);
					lines.push_back(line);
					
					lines.push_back(QString());
					
					
					nominal0 = meas_item.start_nominal ^ meas_item.axis_nominal;
					actual0 = meas_item.start_actual ^ meas_item.axis_nominal;

					
//                                         0         1         2         3         4         5         6         7         8         9
//                                         0123456789012345678901234567890123456789012345678901234567890123456789012345678901234567890
//                                               |          |          |          |          |          |          |          |
					line = QStringLiteral("   Nominal     Actual      Dev");
					lines.push_back(line);
					
					line = QStringLiteral("--------------------------------------------------------------------------------------------------------------------------------");
					lines.push_back(line);
				}
				
				nominal = meas_item.end_nominal ^ meas_item.axis_nominal;
				actual = meas_item.end_actual ^ meas_item.axis_nominal;
				
				nominal -= nominal0;
				actual -= actual0;
				error = actual - nominal;
				
				
				if(cntr == 0)
				{
					min_error = max_error = error;
				}
				else
				{
					if(error > max_error)
					{
						max_error = error;
					}
					else if(error < min_error)
					{
						min_error = error;
					}
				}

				line = QString("%1 %2 %3")
								.arg(nominal,10,'f',4)
								.arg(actual,10,'f',4)
								.arg(error,10,'f',4);
				lines.push_back(line);
			}
			
			lines.push_back(QString());
			
			line = QString("Max Error: %1").arg(max_error,7,'f',4);
			lines.push_back(line);
			
			line = QString("Min Error: %1").arg(min_error,7,'f',4);
			lines.push_back(line);
			
			lines.push_back(QString());
			
			break;
			
		case Types::MEASUREMENT_BALLBAR:
			
			for(cntr = 0;cntr < measurement->Measurement_Item_Count();++cntr)
			{
				meas_item = measurement->Measurement_Item(cntr);
				
				probe_offset = meas_item.probe_offset;
				
				line = QString("%1").arg(this->Pad_String(measurement->Name(),20,false));
				
				line += QString(" %1").arg(meas_item.length_actual,10,'f',4);
				
				line += QString(" %1 %2 %3")
								.arg(meas_item.start_actual.x + probe_offset.x,10,'f',4)
								.arg(meas_item.start_actual.y + probe_offset.y,10,'f',4)
								.arg(meas_item.start_actual.z + probe_offset.z,10,'f',4);
				
				line += QString(" %1 %2 %3")
								.arg(meas_item.end_actual.x + probe_offset.x,10,'f',4)
								.arg(meas_item.end_actual.y + probe_offset.y,10,'f',4)
								.arg(meas_item.end_actual.z + probe_offset.z,10,'f',4);
				
				line += QString(" %1 %2 %3")
								.arg(probe_offset.x,10,'f',4)
								.arg(probe_offset.y,10,'f',4)
								.arg(probe_offset.z,10,'f',4);

				if(measurement->Error_State())
				{
					line += QStringLiteral("  Note:  outside of machine volume");
				}
				
				lines.push_back(line);
			}
			
			break;
	}

	return lines;
}

bool TErrorSimulator::Get_Min_Max_Ballbar_Error(
	TMachineComp						* const machine,
	std::vector<TMeasurement*>			* const measurements,
	double								* const min,
	double								* const max,
	double								* const min_length,
	double								* const max_length) const
{
	double								error_min(0);
	double								error_max(0);
	double								length_error_min(0);
	double								length_error_max(0);
	double								nominal,actual,error;
	bool								init;
	int									cntr;
	TMeasurement::TMeasurementItem		meas_item;
	
	std::vector<TMeasurement*>::const_iterator iter;

	assert(machine);
	assert(measurements);
	assert(min);
	assert(max);
	assert(min_length);
	assert(max_length);
	
	// length is always the same. Reporting worst error without consideration of length
	
	init = true;
	for(iter = measurements->begin();iter != measurements->end();++iter)
	{
		if((*iter)->Type() == Types::MEASUREMENT_BALLBAR)
		{
			for(cntr = 0;cntr < (*iter)->Measurement_Item_Count();++cntr)
			{
				meas_item = (*iter)->Measurement_Item(cntr);
				
				nominal = meas_item.length_nominal;
				actual = meas_item.length_actual;
				error = actual - nominal;
				
				if(init)
				{
					error_min = error;
					error_max = error;
					length_error_min = actual;
					length_error_max = actual;
					
					init = false;
				}
				else
				{
					if(error > error_max)
					{
						error_max = error;
						length_error_max = actual;
					}
					else if(error < error_min)
					{
						error_min = error;
						length_error_min = actual;
					}
				}
			}
		}
	}
	
	*min = error_min;
	*max = error_max;
	*min_length = length_error_min;
	*max_length = length_error_max;
	
	return (!init);	// only return true if something was found
}

bool TErrorSimulator::Get_Min_Max_Laser_Error(
	TMachineComp						* const machine,
	std::vector<TMeasurement*>			* const measurements,
	double								* const min,
	double								* const max,
	double								* const min_length,
	double								* const max_length) const
{
	double								error_min(0);
	double								error_max(0);
	double								length_error_min(0);
	double								length_error_max(0);
	double								length_error_nominal_min(0);
	double								length_error_nominal_max(0);
	double								nominal,actual,error;
	bool								init;
	int									cntr;
	TMeasurement::TMeasurementItem		meas_item;
	static const double					MINIMUM_LENGTH(10);
	
	std::vector<TMeasurement*>::const_iterator iter;
	
	assert(machine);
	assert(measurements);
	assert(min);
	assert(max);
	assert(min_length);
	assert(max_length);
	
	// length varies.  Reporting worst error per nominal unit length
	
	init = true;
	for(iter = measurements->begin();iter != measurements->end();++iter)
	{
		if((*iter)->Type() == Types::MEASUREMENT_LASER)
		{
			for(cntr = 0;cntr < (*iter)->Measurement_Item_Count();++cntr)
			{
				meas_item = (*iter)->Measurement_Item(cntr);
				
				nominal = meas_item.length_nominal;
				actual = meas_item.length_actual;
				error = actual - nominal;
				
				if(nominal > MINIMUM_LENGTH)
				{
					error /= nominal;
				}
				
				if(init)
				{
					error_min = error;
					error_max = error;
					length_error_min = actual;
					length_error_max = actual;
					
					length_error_nominal_min = nominal;
					length_error_nominal_max = nominal;
					
					init = false;
				}
				else
				{
					if(error > error_max)
					{
						error_max = error;
						length_error_max = actual;
						length_error_nominal_max = nominal;
					}
					else if(error < error_min)
					{
						error_min = error;
						length_error_min = actual;
						length_error_nominal_min = nominal;
					}
				}
			}
		}
	}
	
	if(length_error_nominal_max > MINIMUM_LENGTH)
	{
		error_max *= length_error_nominal_max;
	}
	
	if(length_error_nominal_min > MINIMUM_LENGTH)
	{
		error_min *= length_error_nominal_min;
	}

	
	*min = error_min;
	*max = error_max;
	*min_length = length_error_min;
	*max_length = length_error_max;
	
	return (!init);	// only return true if something was found
}

bool TErrorSimulator::Get_Min_Max_ISO_Error(
	TMachineComp						* const machine,
	std::vector<TMeasurement*>			* const measurements,
	double								* const min,
	double								* const max,
	double								* const min_length,
	double								* const max_length) const
{
	double								error_min(0);
	double								error_max(0);
	double								length_error_min(0);
	double								length_error_max(0);
	double								length_error_nominal_min(0);
	double								length_error_nominal_max(0);
	double								nominal,actual,error;
	bool								init;
	int									cntr;
	TMeasurement::TMeasurementItem		meas_item;
	static const double					MINIMUM_LENGTH(1);

	std::vector<TMeasurement*>::const_iterator iter;
	
	assert(machine);
	assert(measurements);
	assert(min);
	assert(max);
	assert(min_length);
	assert(max_length);
	
	// length varies.  Reporting worst error per nominal unit length
	
	init = true;
	for(iter = measurements->begin();iter != measurements->end();++iter)
	{
		if((*iter)->Type() == Types::MEASUREMENT_ISO)
		{
			for(cntr = 0;cntr < (*iter)->Measurement_Item_Count();++cntr)
			{
				meas_item = (*iter)->Measurement_Item(cntr);
				
				nominal = meas_item.length_nominal;
				actual = meas_item.length_actual;
				error = actual - nominal;
				
				if(nominal > MINIMUM_LENGTH)
				{
					error /= nominal;
				}

				if(init)
				{
					error_min = error;
					error_max = error;
					length_error_min = actual;
					length_error_max = actual;
					
					length_error_nominal_min = nominal;
					length_error_nominal_max = nominal;

					init = false;
				}
				else
				{
					if(error > error_max)
					{
						error_max = error;
						length_error_max = actual;
						length_error_nominal_max = nominal;
					}
					else if(error < error_min)
					{
						error_min = error;
						length_error_min = actual;
						length_error_nominal_min = nominal;
					}
				}
			}
		}
	}
	
	if(length_error_nominal_max > MINIMUM_LENGTH)
	{
		error_max *= length_error_nominal_max;
	}
	
	if(length_error_nominal_min > MINIMUM_LENGTH)
	{
		error_min *= length_error_nominal_min;
	}

	*min = error_min;
	*max = error_max;
	*min_length = length_error_min;
	*max_length = length_error_max;
	
	return (!init);	// only return true if something was found
}

bool TErrorSimulator::Get_Min_Max_MCG_Error(
	TMachineComp						* const machine,
	std::vector<TMeasurement*>			* const measurements,
	double								* const min,
	double								* const max,
	double								* const min_length,
	double								* const max_length) const
{
	double								error_min(0);
	double								error_max(0);
	double								length_error_min(0);
	double								length_error_max(0);
	double								nominal,actual,error;
	bool								init;
	int									cntr;
	TMeasurement::TMeasurementItem		meas_item;
	
	std::vector<TMeasurement*>::const_iterator iter;
	
	assert(machine);
	assert(measurements);
	assert(min);
	assert(max);
	assert(min_length);
	assert(max_length);
	
	// length is always the same. Reporting worst error without consideration of length

	init = true;
	for(iter = measurements->begin();iter != measurements->end();++iter)
	{
		if((*iter)->Type() == Types::MEASUREMENT_MCG)
		{
			for(cntr = 0;cntr < (*iter)->Measurement_Item_Count();++cntr)
			{
				meas_item = (*iter)->Measurement_Item(cntr);
				
				nominal = meas_item.length_nominal;
				actual = meas_item.length_actual;
				error = actual - nominal;
				
				if(init)
				{
					error_min = error;
					error_max = error;
					length_error_min = actual;
					length_error_max = actual;
					
					init = false;
				}
				else
				{
					if(error > error_max)
					{
						error_max = error;
						length_error_max = actual;
					}
					else if(error < error_min)
					{
						error_min = error;
						length_error_min = actual;
					}
				}
			}
		}
	}
	
	*min = error_min;
	*max = error_max;
	*min_length = length_error_min;
	*max_length = length_error_max;
	
	return (!init);	// only return true if something was found
}

void TErrorSimulator::Export_ISO_Measurement(
	const QString						&dir,
	TMeasurement						* const measurement) const
{
	QFile								ac_file;
	QFile								pm_file;
	TMeasurement::TMeasurementItem		meas_item;
	int									cntr;
	int									pnt_cntr;
	TReferenceSystem					refsys;
	TVector3							probe_offset,probe_vector;
	TVector3							scale_xx_yy_error,scale_xx_yy_error0;
	double								nominal0(0),actual0(0);
	double								nominal,actual;
	QString								line;
	
	assert(measurement);
	
	ac_file.setFileName(QString("%1/AutoCapture %2.dat").arg(dir).arg(measurement->Name()));
	pm_file.setFileName(QString("%1/PMove %2.dat").arg(dir).arg(measurement->Name()));
	
	if(!ac_file.open(QIODevice::WriteOnly | QIODevice::Text))
	{
		d_display_widget->Clear_Selection_Data();
		d_display_widget->Add_Selection_Data_Line(QStringLiteral("Export ISO:  Cannot open AutoCapture file for writing."));
		d_display_widget->Add_Selection_Data_End();
		return;
	}
	
	if(!pm_file.open(QIODevice::WriteOnly | QIODevice::Text))
	{
		d_display_widget->Clear_Selection_Data();
		d_display_widget->Add_Selection_Data_Line(QStringLiteral("Export ISO:  Cannot open PMove file for writing."));
		d_display_widget->Add_Selection_Data_End();
		
		ac_file.close();
		return;
	}
	
	for(cntr = 0;cntr < measurement->Measurement_Item_Count();++cntr)
	{
		meas_item = measurement->Measurement_Item(cntr);
		
		if(cntr == 0)
		{
			refsys.Skew1(meas_item.axis_nominal,TReferenceSystem::XPLUS);
			
			probe_offset = probe_vector = meas_item.probe_offset;
			probe_vector.z = 0;
			
			if(probe_vector.Length())
			{
				probe_vector.Normal();
			}
			else
			{
				// no probe offset.  Assume Z
				probe_vector.Set(0,0,-1);
			}
			
			refsys.Skew2(probe_vector,TReferenceSystem::ZMINUS,TReferenceSystem::XPLUS);
			refsys.Set_Origin(meas_item.start_nominal);
			
			// pmove start
			pm_file.write("B89.4.10360_Raw_Measurement:Version=1:Type=Five_Laser\n\n");
			
			pm_file.write("# Note: Alignment translation includes probe offset.\n\n");
			
			pm_file.write("Scales_Begin:\n");
			pm_file.write("X_Expansion_Coefficient:11.5\n");
			pm_file.write("Y_Expansion_Coefficient:11.5\n");
			pm_file.write("Z_Expansion_Coefficient:11.5\n");
			pm_file.write("Scales_End:\n\n");
			
			pm_file.write("Alignment_Begin:\n");
			pm_file.write(QString("X_Axis:%1,%2,%3\n")
						  .arg(refsys.Rotation_Mat().X().x,0,'f',9)
						  .arg(refsys.Rotation_Mat().X().y,0,'f',9)
						  .arg(refsys.Rotation_Mat().X().z,0,'f',9)
						  .toLatin1());
			pm_file.write(QString("Y_Axis:%1,%2,%3\n")
						  .arg(refsys.Rotation_Mat().Y().x,0,'f',9)
						  .arg(refsys.Rotation_Mat().Y().y,0,'f',9)
						  .arg(refsys.Rotation_Mat().Y().z,0,'f',9)
						  .toLatin1());
			pm_file.write(QString("Z_Axis:%1,%2,%3\n")
						  .arg(refsys.Rotation_Mat().Z().x,0,'f',9)
						  .arg(refsys.Rotation_Mat().Z().y,0,'f',9)
						  .arg(refsys.Rotation_Mat().Z().z,0,'f',9)
						  .toLatin1());
			
			pm_file.write(QString("Translation:%1,%2,%3\n")
						  .arg(meas_item.start_nominal.x + probe_offset.x,0,'f',4)
						  .arg(meas_item.start_nominal.y + probe_offset.y,0,'f',4)
						  .arg(meas_item.start_nominal.z + probe_offset.z,0,'f',4)
						  .toLatin1());
			pm_file.write("Alignment_End:\n\n");
			
			pm_file.write("Temperature_Begin:\n");
			pm_file.write("X_Scale:20.00\n");
			pm_file.write("Y_Scale:20.00\n");
			pm_file.write("Z_Scale:20.00\n");
			pm_file.write("Temperature_End:\n\n");
			
			pm_file.write(QString("Probe_Offset:%1,%2,%3\n")
						  .arg(probe_offset.x,0,'f',4)
						  .arg(probe_offset.y,0,'f',4)
						  .arg(probe_offset.z,0,'f',4)
						  .toLatin1());
			pm_file.write(QString("Probe_Vector:%1,%2,%3\n\n")
						  .arg(probe_vector.x,0,'f',9)
						  .arg(probe_vector.y,0,'f',9)
						  .arg(probe_vector.z,0,'f',9)
						  .toLatin1());
			
			pm_file.write("Measurement_Begin:\n\n");
			
			pm_file.write(QString("Description:%1\n\n")
						  .arg(measurement->Name())
						  .toLatin1());
			
			
			// autocapture start
			ac_file.write("AutoCapture_Data:Version=1:Type=Laser_10360\n\n");
			
			ac_file.write("# Note: Straightness and angular measurements are zero for non-axis aligned measurements.\n\n");
			
			ac_file.write("       Scale         XX         YY         RA         RB         RC      AIR    MAT\n");
			ac_file.write("-------------------------------------------------------------------------------------\n");
			ac_file.write(":Data_Begin\n");
			
			nominal0 = meas_item.start_nominal ^ meas_item.axis_nominal;
			actual0 = meas_item.start_actual ^ meas_item.axis_nominal;
			scale_xx_yy_error0 = d_machine_comp.Calculate_Scale_XX_YY_Error(meas_item.axis_nominal,probe_vector,meas_item.start_nominal,meas_item.start_actual);
			
		}
		
		scale_xx_yy_error = d_machine_comp.Calculate_Scale_XX_YY_Error(meas_item.axis_nominal,probe_vector,meas_item.end_nominal,meas_item.end_actual);
		scale_xx_yy_error -= scale_xx_yy_error0;
		
		nominal = meas_item.end_nominal ^ meas_item.axis_nominal;
		actual = meas_item.end_actual ^ meas_item.axis_nominal;
		
		nominal -= nominal0;
		actual -= actual0;
		
		for(pnt_cntr = 0;pnt_cntr < 3;++pnt_cntr)
		{
			pm_file.write(QString("Point:%1,A,%2,%3,%4\n")
						  .arg(cntr + 1)
						  .arg(0.0,0,'f',5)
						  .arg(0.0,0,'f',5)
						  .arg(0.0,0,'f',5)
						  .toLatin1());
			
			pm_file.write(QString("Point:%1,B,%2,%3,%4\n")
						  .arg(cntr + 1)
						  .arg(nominal,0,'f',5)
						  .arg(0.0,0,'f',5)
						  .arg(0.0,0,'f',5)
						  .toLatin1());
			
			line = QString("%1,%2,%3,%4,%5,%6,  20.00, 20.00\n")
						.arg(0.0,14,'f',5)
						.arg(0.0,10,'f',5)
						.arg(0.0,10,'f',5)
						.arg(0.0,10,'f',5)
						.arg(0.0,10,'f',5)
						.arg(0.0,10,'f',5);
			
			ac_file.write(line.toLatin1());
			
			line = QString("%1,%2,%3,%4,%5,%6,  20.00, 20.00\n")
						.arg(actual,14,'f',5)
						.arg(scale_xx_yy_error.y,10,'f',5)
						.arg(scale_xx_yy_error.z,10,'f',5)
						.arg(meas_item.end_map_abc.x - meas_item.start_map_abc.x,10,'f',5)
						.arg(meas_item.end_map_abc.y - meas_item.start_map_abc.y,10,'f',5)
						.arg(meas_item.end_map_abc.z - meas_item.start_map_abc.z,10,'f',5);
			
			ac_file.write(line.toLatin1());
			
		}
	}
	
	pm_file.write("Measurement_End:\n");
	ac_file.write(":Data_End\n");
	
	pm_file.close();
	ac_file.close();
}

void TErrorSimulator::Export_MCG_Measurement(
	const QString						&dir,
	TMeasurement						* const measurement) const
{
	TXmlFile							xmlfile;
	TMeasurement::TMeasurementItem		meas_item;
	int									cntr;
	QDomElement							en1;
	QDomElement							en2;
	QDomElement							en3;
	TVector3							pnt;
	QString								text;
	
	assert(measurement);
	
	try
	{
		xmlfile.Set_File_Name(QString("%1/MCG %2.xml").arg(dir).arg(measurement->Name()));
		xmlfile.Reset(QStringLiteral("MCG_Measurement_Data"),1);
		
		xmlfile.Write_Attribute(xmlfile.Root_Node(),QStringLiteral("Units"),QStringLiteral("MM"));
		
		en1 = xmlfile.Create_Node(xmlfile.Root_Node(),QStringLiteral("Environment"));
		
		en2 = xmlfile.Create_Node(&en1,QStringLiteral("Temperature"));
		xmlfile.Write_Text_Node(&en2,20.0,2);
		
		en1 = xmlfile.Create_Node(xmlfile.Root_Node(),QStringLiteral("Gauge_Points"));
		
		for(cntr = 0;cntr < measurement->Measurement_Item_Count();++cntr)
		{
			meas_item = measurement->Measurement_Item(cntr);
			
			if(cntr == 0)
			{
				en2 = xmlfile.Create_Node(xmlfile.Root_Node(),QStringLiteral("Probe"));
				
				en3 = xmlfile.Create_Node(&en2,QStringLiteral("Probe_Offset_XYZ"));
				
				text = QString("%1,%2,%3")
							.arg(meas_item.probe_offset.x,0,'f',4)
							.arg(meas_item.probe_offset.y,0,'f',4)
							.arg(meas_item.probe_offset.z,0,'f',4);
				
				xmlfile.Write_Text_Node(&en3,text);
				
				en3 = xmlfile.Create_Node(&en2,QStringLiteral("Probe_Error"));
				xmlfile.Write_Text_Node(&en3,0.0,5);
				
				en2 = xmlfile.Create_Node(xmlfile.Root_Node(),QStringLiteral("Pivot_Sphere_Initial"));
				en3 = xmlfile.Create_Node(&en2,QStringLiteral("Location_XYZ"));
				
				text = QString("%1,%2,%3")
							.arg(meas_item.start_nominal.x + meas_item.probe_offset.x,0,'f',4)
							.arg(meas_item.start_nominal.y + meas_item.probe_offset.y,0,'f',4)
							.arg(meas_item.start_nominal.z + meas_item.probe_offset.z,0,'f',4);
				
				xmlfile.Write_Text_Node(&en3,text);
			}
			
			pnt = meas_item.end_actual - meas_item.start_nominal;
			
			en2 = xmlfile.Create_Node(&en1,QStringLiteral("Point_XYZ"));
			
			text = QString("%1,%2,%3")
						.arg(pnt.x,0,'f',6)
						.arg(pnt.y,0,'f',6)
						.arg(pnt.z,0,'f',6);
			
			xmlfile.Write_Text_Node(&en2,text);
		}
		
		en1 = xmlfile.Create_Node(xmlfile.Root_Node(),QStringLiteral("Pivot_Sphere_Final"));
		
		en2 = xmlfile.Create_Node(&en1,QStringLiteral("Location_XYZ"));
		
		text = QString("%1,%2,%3")
					.arg(meas_item.start_nominal.x + meas_item.probe_offset.x,0,'f',4)
					.arg(meas_item.start_nominal.y + meas_item.probe_offset.y,0,'f',4)
					.arg(meas_item.start_nominal.z + meas_item.probe_offset.z,0,'f',4);
		
		xmlfile.Write_Text_Node(&en2,text);
		
		
		xmlfile.Write_File();
	}
	catch(TXmlFileException xml_exception)
	{
		d_display_widget->Clear_Selection_Data();
		d_display_widget->Add_Selection_Data_Line(QString("Export MCG: %1").arg(xml_exception.ErrorText()));
		d_display_widget->Add_Selection_Data_End();
	}
}

void TErrorSimulator::Export_Laser_Measurement(
	const QString						&dir,
	TMeasurement						* const measurement) const
{
	QFile								ac_file;
	QFile								pm_file;
	TMeasurement::TMeasurementItem		meas_item;
	int									cntr;
	TReferenceSystem					refsys;
	TVector3							probe_offset,probe_vector;
	double								actual,actual0(0);
	TVector3							scale_xx_yy_error,scale_xx_yy_error0;
	QString								line;
	
	assert(measurement);
	
	ac_file.setFileName(QString("%1/AutoCapture %2.dat").arg(dir).arg(measurement->Name()));
	pm_file.setFileName(QString("%1/PMove %2.dat").arg(dir).arg(measurement->Name()));
	
	if(!ac_file.open(QIODevice::WriteOnly | QIODevice::Text))
	{
		d_display_widget->Clear_Selection_Data();
		d_display_widget->Add_Selection_Data_Line(QStringLiteral("Export Laser:  Cannot open AutoCapture file for writing."));
		d_display_widget->Add_Selection_Data_End();
		return;
	}
	
	if(!pm_file.open(QIODevice::WriteOnly | QIODevice::Text))
	{
		d_display_widget->Clear_Selection_Data();
		d_display_widget->Add_Selection_Data_Line(QStringLiteral("Export Laser:  Cannot open PMove file for writing."));
		d_display_widget->Add_Selection_Data_End();
		
		ac_file.close();
		return;
	}
	
	for(cntr = 0;cntr < measurement->Measurement_Item_Count();++cntr)
	{
		meas_item = measurement->Measurement_Item(cntr);
		
		if(cntr == 0)
		{
			refsys.Skew1(meas_item.axis_nominal,TReferenceSystem::XPLUS);
			
			probe_offset = probe_vector = meas_item.probe_offset;
			probe_vector.z = 0;
			
			if(probe_vector.Length())
			{
				probe_vector.Normal();
			}
			else
			{
				// no probe offset.  Assume Z
				probe_vector.Set(0,0,-1);
			}
			
			refsys.Skew2(probe_vector,TReferenceSystem::ZMINUS,TReferenceSystem::XPLUS);
			refsys.Set_Origin(meas_item.start_nominal);
			
			// pmove start
			pm_file.write("PMove_Raw_Measurement:Version=1:Type=Laser\n\n");
			
			pm_file.write("# Note: Alignment translation and all target points include the probe offset.\n\n");
			
			pm_file.write("Alignment_Begin:\n");
			pm_file.write(QString("X_Axis:%1,%2,%3\n")
						  .arg(refsys.Rotation_Mat().X().x,0,'f',9)
						  .arg(refsys.Rotation_Mat().X().y,0,'f',9)
						  .arg(refsys.Rotation_Mat().X().z,0,'f',9)
						  .toLatin1());
			pm_file.write(QString("Y_Axis:%1,%2,%3\n")
						  .arg(refsys.Rotation_Mat().Y().x,0,'f',9)
						  .arg(refsys.Rotation_Mat().Y().y,0,'f',9)
						  .arg(refsys.Rotation_Mat().Y().z,0,'f',9)
						  .toLatin1());
			pm_file.write(QString("Z_Axis:%1,%2,%3\n")
						  .arg(refsys.Rotation_Mat().Z().x,0,'f',9)
						  .arg(refsys.Rotation_Mat().Z().y,0,'f',9)
						  .arg(refsys.Rotation_Mat().Z().z,0,'f',9)
						  .toLatin1());
			
			pm_file.write(QString("Translation:%1,%2,%3\n")
						  .arg(meas_item.start_nominal.x + probe_offset.x,0,'f',4)
						  .arg(meas_item.start_nominal.y + probe_offset.y,0,'f',4)
						  .arg(meas_item.start_nominal.z + probe_offset.z,0,'f',4)
						  .toLatin1());
			pm_file.write("Alignment_End:\n\n");
			
			pm_file.write("Temperature_Begin:\n");
			pm_file.write("X_Scale:20.00\n");
			pm_file.write("Y_Scale:20.00\n");
			pm_file.write("Z_Scale:20.00\n");
			pm_file.write("Temperature_End:\n\n");
			
			pm_file.write(QString("Probe_Offset:%1,%2,%3\n")
						  .arg(probe_offset.x,0,'f',4)
						  .arg(probe_offset.y,0,'f',4)
						  .arg(probe_offset.z,0,'f',4)
						  .toLatin1());
			pm_file.write(QString("Probe_Vector:%1,%2,%3\n\n")
						  .arg(probe_vector.x,0,'f',9)
						  .arg(probe_vector.y,0,'f',9)
						  .arg(probe_vector.z,0,'f',9)
						  .toLatin1());
			
			pm_file.write("Measurement_Begin:\n");
			
			
			// autocapture start
			ac_file.write("AutoCapture_Data:Version=1:Type=Laser\n\n");
			
			ac_file.write("# Note: Straightness and angular measurements are zero for non-axis aligned measurements.\n\n");
			
			ac_file.write("       Scale         XX         YY         RA         RB         RC      AIR    MAT\n");
			ac_file.write("-------------------------------------------------------------------------------------\n");
			ac_file.write(":Data_Begin\n");
			
			actual0 = meas_item.start_actual ^ meas_item.axis_nominal;
			scale_xx_yy_error0 = d_machine_comp.Calculate_Scale_XX_YY_Error(meas_item.axis_nominal,probe_vector,meas_item.start_nominal,meas_item.start_actual);
		}
		
		scale_xx_yy_error = d_machine_comp.Calculate_Scale_XX_YY_Error(meas_item.axis_nominal,probe_vector,meas_item.end_nominal,meas_item.end_actual);
		scale_xx_yy_error -= scale_xx_yy_error0;
		
		actual = meas_item.end_actual ^ meas_item.axis_nominal;
		actual -= actual0;
		
		pm_file.write(QString("Point:1,F,%1,%2,%3\n")
					.arg(meas_item.end_nominal.x + probe_offset.x,0,'f',5)
					.arg(meas_item.end_nominal.y + probe_offset.y,0,'f',5)
					.arg(meas_item.end_nominal.z + probe_offset.z,0,'f',5)
					.toLatin1());
		
		line = QString("%1,%2,%3,%4,%5,%6,  20.00, 20.00\n")
					.arg(actual,14,'f',5)
					.arg(scale_xx_yy_error.y,10,'f',5)
					.arg(scale_xx_yy_error.z,10,'f',5)
					.arg(meas_item.end_map_abc.x - meas_item.start_map_abc.x,10,'f',5)
					.arg(meas_item.end_map_abc.y - meas_item.start_map_abc.y,10,'f',5)
					.arg(meas_item.end_map_abc.z - meas_item.start_map_abc.z,10,'f',5);
		
		ac_file.write(line.toLatin1());
	}
	
	// write reverse direction measurements
	for(cntr = measurement->Measurement_Item_Count() - 1;!(cntr < 0);--cntr)
	{
		meas_item = measurement->Measurement_Item(cntr);
		
		scale_xx_yy_error = d_machine_comp.Calculate_Scale_XX_YY_Error(meas_item.axis_nominal,probe_vector,meas_item.end_nominal,meas_item.end_actual);
		scale_xx_yy_error -= scale_xx_yy_error0;
		
		actual = meas_item.end_actual ^ meas_item.axis_nominal;
		actual -= actual0;
		
		pm_file.write(QString("Point:1,B,%1,%2,%3\n")
					.arg(meas_item.end_nominal.x + probe_offset.x,0,'f',5)
					.arg(meas_item.end_nominal.y + probe_offset.y,0,'f',5)
					.arg(meas_item.end_nominal.z + probe_offset.z,0,'f',5)
					.toLatin1());
		
		line = QString("%1,%2,%3,%4,%5,%6,  20.00, 20.00\n")
					.arg(actual,14,'f',5)
					.arg(scale_xx_yy_error.y,10,'f',5)
					.arg(scale_xx_yy_error.z,10,'f',5)
					.arg(meas_item.end_map_abc.x - meas_item.start_map_abc.x,10,'f',5)
					.arg(meas_item.end_map_abc.y - meas_item.start_map_abc.y,10,'f',5)
					.arg(meas_item.end_map_abc.z - meas_item.start_map_abc.z,10,'f',5);
		
		ac_file.write(line.toLatin1());
		
	}
	
	pm_file.write("Measurement_End:\n");
	ac_file.write(":Data_End\n");
	
	pm_file.close();
	ac_file.close();
}


void TErrorSimulator::Export_Ballbar_Measurement(
	const QString						&dir,
	const std::vector<TMeasurement*>	&measurements) const
{
	QString								line;
	TMeasurement::TMeasurementItem		meas_item;
	std::vector<TMeasurement*>::const_iterator meas_iter;
	QFile								file;
	TVector3							pos;
	
	for(meas_iter = measurements.begin();meas_iter != measurements.end();++meas_iter)
	{
		if((*meas_iter)->Type() == Types::MEASUREMENT_BALLBAR)
		{
			if((*meas_iter)->Measurement_Item_Count() == 1)	// should only be one ball bar measurement
			{
				meas_item = (*meas_iter)->Measurement_Item(0);
				
				if(!file.isOpen())
				{
					file.setFileName(QString("%1/Ballbar Measurements.dat").arg(dir));
					
					if(!file.open(QIODevice::WriteOnly | QIODevice::Text))
					{
						d_display_widget->Clear_Selection_Data();
						d_display_widget->Add_Selection_Data_Line(QStringLiteral("Export Ballbar:  Cannot open 'Ballbar Measurements.dat' file for writing."));
						d_display_widget->Add_Selection_Data_End();
						return;
					}
					
					file.write("B89.4.1_Data:Version=1:Type=BallBar\n\n");
					file.write("# Note:  Sphere locations include the probe offset.\n\n");
					
					file.write("   Name                 Length      Sph1 X      Sph1 Y      Sph1 Z      Sph2 X      Sph2 Y      Sph2 Z     Sph1 Form   Sph2 Form    Probe X     Probe Y     Probe Z\n");
					file.write("-----------------------------------------------------------------------------------------------------------------------------------------------------------------------\n");
				}
				
				line = QString("%1 ").arg(this->Pad_String((*meas_iter)->Name(),20,true));
				line += QString("%1").arg(meas_item.length_actual,11,'f',5);
				
				pos = meas_item.start_actual;
				pos += meas_item.probe_offset;
				
				line += QString(",%1,%2,%3")
							.arg(pos.x,11,'f',5)
							.arg(pos.y,11,'f',5)
							.arg(pos.z,11,'f',5);
				
				pos = meas_item.end_actual;
				pos += meas_item.probe_offset;
				
				line += QString(",%1,%2,%3")
							.arg(pos.x,11,'f',5)
							.arg(pos.y,11,'f',5)
							.arg(pos.z,11,'f',5);
				
				line += QString(",%1,%2").arg(0.0,11,'f',5).arg(0.0,11,'f',5);
				
				line += QString(",%1,%2,%3\n")
							.arg(meas_item.probe_offset.x,11,'f',5)
							.arg(meas_item.probe_offset.y,11,'f',5)
							.arg(meas_item.probe_offset.z,11,'f',5);
				
				file.write(line.toLatin1());
			}
		}
	}
	
	if(file.isOpen())
	{
		file.close();
	}
}


std::vector<std::vector<TMachineDialog::TMapErrorItem> > TErrorSimulator::Get_Evaluation_Error_Patterns(
	const std::vector<TMachineDialog::TMapErrorItem> &input_error_items) const
{
	std::vector<std::vector<TMachineDialog::TMapErrorItem> > patterns;
	std::vector<std::vector<TMachineDialog::TMapErrorItem> >::iterator iter;
	std::vector<TMachineDialog::TMapErrorItem>::const_iterator item_iter;
	int									count;
	int									cntr;
	int									mask;

	if(!input_error_items.size())
	{
		return patterns;
	}
	
	// calculate total number of arrays
	count = (0b1 << input_error_items.size());
	patterns.resize(count);

	for(iter = patterns.begin(),cntr = 0b1;iter != patterns.end();++iter,++cntr)
	{
		for(item_iter = input_error_items.begin(),mask = 0b1;item_iter != input_error_items.end();++item_iter,mask <<= 1)
		{
			if((mask & cntr) != 0)
			{
				(*iter).push_back(*item_iter);
			}
		}
	}
	
	return patterns;
}

bool TErrorSimulator::Update_Machine_Comp(
	const QString						&machine_name,
	QStringList * const					error_list)
{
	TMachineComp::TInitParameters		init_param;
	TVector3							machine_start;
	TVector3							machine_end;
	TVector3							machine_math_origin;
	TVector3							machine_mechanical_origin;
	std::vector<TMachineDialog::TMapErrorItem> error_items;
	std::vector<TMachineDialog::TMapErrorItem>::const_iterator error_item_iter;
	int									id;
	bool								return_state(true);
	
	id = d_machine_dialog->Machine_Id(machine_name);
	
	if(id < 0)
	{
		if(error_list)
		{
			error_list->push_back(QStringLiteral("Machine not found."));
		}
		
		d_display_widget->Clear_Machine_Volume();
		
		return false;
	}
	
	init_param.kinematic = d_machine_dialog->Machine_Kinematic(id);
		
	machine_start = d_machine_dialog->Machine_Axis_Start(id);
	machine_end = d_machine_dialog->Machine_Axis_End(id);
	machine_math_origin = d_machine_dialog->Machine_Math_Origin(id);
	machine_mechanical_origin = d_machine_dialog->Machine_Mechanical_Origin(id);
	
	switch(init_param.kinematic)
	{
		case Types::KINEMATIC_XYZ:
			
			init_param.first_axis_id = Types::AXIS_X;
			init_param.second_axis_id = Types::AXIS_Y;
			init_param.third_axis_id = Types::AXIS_Z;
			
			init_param.first_axis_start = machine_start.x;
			init_param.second_axis_start = machine_start.y;
			init_param.third_axis_start = machine_start.z;
			
			init_param.first_axis_end = machine_end.x;
			init_param.second_axis_end = machine_end.y;
			init_param.third_axis_end = machine_end.z;
			
			init_param.first_axis_math_origin = machine_math_origin.x;
			init_param.second_axis_math_origin = machine_math_origin.y;
			init_param.third_axis_math_origin = machine_math_origin.z;
			
			init_param.first_axis_mechanical_origin = machine_mechanical_origin.x;
			init_param.second_axis_mechanical_origin = machine_mechanical_origin.y;
			init_param.third_axis_mechanical_origin = machine_mechanical_origin.z;
			
			break;
			
		case Types::KINEMATIC_YXZ:
			
			init_param.first_axis_id = Types::AXIS_Y;
			init_param.second_axis_id = Types::AXIS_X;
			init_param.third_axis_id = Types::AXIS_Z;
			
			init_param.first_axis_start = machine_start.y;
			init_param.second_axis_start = machine_start.x;
			init_param.third_axis_start = machine_start.z;
			
			init_param.first_axis_end = machine_end.y;
			init_param.second_axis_end = machine_end.x;
			init_param.third_axis_end = machine_end.z;
			
			init_param.first_axis_math_origin = machine_math_origin.y;
			init_param.second_axis_math_origin = machine_math_origin.x;
			init_param.third_axis_math_origin = machine_math_origin.z;
			
			init_param.first_axis_mechanical_origin = machine_mechanical_origin.y;
			init_param.second_axis_mechanical_origin = machine_mechanical_origin.x;
			init_param.third_axis_mechanical_origin = machine_mechanical_origin.z;
			break;
			
		case Types::KINEMATIC_XZY:
			
			init_param.first_axis_id = Types::AXIS_X;
			init_param.second_axis_id = Types::AXIS_Z;
			init_param.third_axis_id = Types::AXIS_Y;
			
			init_param.first_axis_start = machine_start.x;
			init_param.second_axis_start = machine_start.z;
			init_param.third_axis_start = machine_start.y;
			
			init_param.first_axis_end = machine_end.x;
			init_param.second_axis_end = machine_end.z;
			init_param.third_axis_end = machine_end.y;
			
			init_param.first_axis_math_origin = machine_math_origin.x;
			init_param.second_axis_math_origin = machine_math_origin.z;
			init_param.third_axis_math_origin = machine_math_origin.y;
			
			init_param.first_axis_mechanical_origin = machine_mechanical_origin.x;
			init_param.second_axis_mechanical_origin = machine_mechanical_origin.z;
			init_param.third_axis_mechanical_origin = machine_mechanical_origin.y;
			break;
			
		case Types::KINEMATIC_YZX:
			
			init_param.first_axis_id = Types::AXIS_Y;
			init_param.second_axis_id = Types::AXIS_Z;
			init_param.third_axis_id = Types::AXIS_X;
			
			init_param.first_axis_start = machine_start.y;
			init_param.second_axis_start = machine_start.z;
			init_param.third_axis_start = machine_start.x;
			
			init_param.first_axis_end = machine_end.y;
			init_param.second_axis_end = machine_end.z;
			init_param.third_axis_end = machine_end.x;
			
			init_param.first_axis_math_origin = machine_math_origin.y;
			init_param.second_axis_math_origin = machine_math_origin.z;
			init_param.third_axis_math_origin = machine_math_origin.x;
			
			init_param.first_axis_mechanical_origin = machine_mechanical_origin.y;
			init_param.second_axis_mechanical_origin = machine_mechanical_origin.z;
			init_param.third_axis_mechanical_origin = machine_mechanical_origin.x;
			break;
			
		default:
			return false;
	}
	
	if(!d_machine_comp.Initialize_Begin(init_param))
	{
		if(error_list)
		{
			error_list->push_back(QString(d_machine_comp.Last_Error()));
		}
		
		return false;
	}
	
	d_display_widget->Define_Machine_Volume(d_machine_comp.Map_Data_Min(),d_machine_comp.Map_Data_Max());
	
	error_items = d_machine_dialog->Map_Error_Items(id);
	for(error_item_iter = error_items.begin();error_item_iter != error_items.end();++error_item_iter)
	{
		switch((*error_item_iter).parameter)
		{
			case Types::PARAMETER_RX:
				if(!d_machine_comp.Set_Rx((*error_item_iter).axis,(*error_item_iter).expression,(*error_item_iter).sign))
				{
					if(error_list)
					{
						error_list->push_back(d_machine_comp.Last_Error());
					}
					
					return_state = false;
				}
				break;
				
			case Types::PARAMETER_RY:
				if(!d_machine_comp.Set_Ry((*error_item_iter).axis,(*error_item_iter).expression,(*error_item_iter).sign))
				{
					if(error_list)
					{
						error_list->push_back(d_machine_comp.Last_Error());
					}
					
					return_state = false;
				}
				break;
				
			case Types::PARAMETER_RZ:
				if(!d_machine_comp.Set_Rz((*error_item_iter).axis,(*error_item_iter).expression,(*error_item_iter).sign))
				{
					if(error_list)
					{
						error_list->push_back(d_machine_comp.Last_Error());
					}
					
					return_state = false;
				}
				break;
				
			case Types::PARAMETER_SCALE:
				if(!d_machine_comp.Set_Scale((*error_item_iter).axis,(*error_item_iter).expression,(*error_item_iter).sign))
				{
					if(error_list)
					{
						error_list->push_back(d_machine_comp.Last_Error());
					}
					
					return_state = false;
				}
				break;
				
			case Types::PARAMETER_DEFLECTION:
				if(!d_machine_comp.Set_Deflection((*error_item_iter).axis,(*error_item_iter).value,(*error_item_iter).sign))
				{
					if(error_list)
					{
						error_list->push_back(d_machine_comp.Last_Error());
					}
					
					return_state = false;
				}
				break;
				
			case Types::PARAMETER_QXY:
				if(!d_machine_comp.Set_Quad(Types::AXIS_X,(*error_item_iter).value,(*error_item_iter).sign))
				{
					if(error_list)
					{
						error_list->push_back(d_machine_comp.Last_Error());
					}
					
					return_state = false;
				}
				break;
				
			case Types::PARAMETER_QYZ:
				if(!d_machine_comp.Set_Quad(Types::AXIS_Y,(*error_item_iter).value,(*error_item_iter).sign))
				{
					if(error_list)
					{
						error_list->push_back(d_machine_comp.Last_Error());
					}
					
					return_state = false;
				}
				break;
				
			case Types::PARAMETER_QZX:
				if(!d_machine_comp.Set_Quad(Types::AXIS_Z,(*error_item_iter).value,(*error_item_iter).sign))
				{
					if(error_list)
					{
						error_list->push_back(d_machine_comp.Last_Error());
					}
					
					return_state = false;
				}
				break;
		}
	}
			
	if(!d_machine_comp.Initialize_End())
	{
		return false;
	}
	
	return return_state;
}


void TErrorSimulator::Update_Machine_Names(void)
{
	QListWidgetItem						*item;
	QStringList							list;
	QStringList::const_iterator			iter;
	int									id;
	
	list = d_machine_dialog->Machine_Names();
	
	d_machines_list->clear();

	for(iter = list.begin();iter != list.end();++iter)
	{
		id = d_machine_dialog->Machine_Id(*iter);
		
		item = new QListWidgetItem(d_machines_list);
		item->setFlags(item->flags() | Qt::ItemIsEditable);
		item->setText(*iter);
		item->setData(Qt::UserRole,id);
	}
}

void TErrorSimulator::Update_Measurement_Names(void)
{
	QListWidgetItem						*item;
	QStringList							list;
	QStringList::const_iterator			iter;
	int									id;

	list = d_measurement_dialog->Measurement_Names();
	
	d_measurements_list->clear();

	for(iter = list.begin();iter != list.end();++iter)
	{
		id = d_measurement_dialog->Measurement_Id(*iter);

		item = new QListWidgetItem(d_measurements_list);
		item->setFlags(item->flags() | Qt::ItemIsEditable);
		item->setText(*iter);
		item->setData(Qt::UserRole,id);
	}
}
