/////////////////////////////////////////////////////////////////////
//
//            X   X           X
//           XXX XX         XX
//          XXXXXXXX      XXX
//         XX X XXXXXXXXXXX
//        XXXXX XXXXXXXXX
//       XXXXX XXXXXXXXXX
//            XXX XXX XXX
//           XXX XX   XX
//           X   X     X
//
//    Copyright (C) 2003-2023  Ron Jakl
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
/////////////////////////////////////////////////////////////////////

#include <QApplication>
#include <QFrame>
#include <QGridLayout>
#include <QHBoxLayout>
#include <QLabel>
#include <QLineEdit>
#include <QPushButton>
#include <QSpacerItem>
#include <QCloseEvent>
#include <vector>
#include <cmath>

#include "../../core/expression.h"
#include "../../core/interpolation.h"
#include "../../core/vector2.h"

#include "expressiondialog.h"

TExpressionDialog::TExpressionDialog(
	const QWidget						*parent,
	const Qt::WindowFlags				flags)
:QDialog(const_cast<QWidget*>(parent),flags)
{
	QFrame								*separator_hline;
	QGridLayout							*dialog_layout;
	QHBoxLayout							*button_hlayout;
	QLabel								*position_label;
	QLabel								*error_label;
	QSpacerItem							*button_hspacer;
	QPushButton							*cancel_button;


	this->resize(301,201);

	dialog_layout = new QGridLayout(this);

	position_label = new QLabel(this);
	position_label->setAlignment(Qt::AlignCenter);
	dialog_layout->addWidget(position_label,0,0,1,1);

	error_label = new QLabel(this);
	error_label->setAlignment(Qt::AlignCenter);
	dialog_layout->addWidget(error_label,0,1,1,1);

	d_pos1_edit = new QLineEdit(this);
	d_pos1_edit->setAlignment(Qt::AlignCenter);
	dialog_layout->addWidget(d_pos1_edit,1,0,1,1);

	d_err1_edit = new QLineEdit(this);
	d_err1_edit->setAlignment(Qt::AlignCenter);
	dialog_layout->addWidget(d_err1_edit,1,1,1,1);

	d_pos2_edit = new QLineEdit(this);
	d_pos2_edit->setAlignment(Qt::AlignCenter);
	dialog_layout->addWidget(d_pos2_edit,2,0,1,1);

	d_err2_edit = new QLineEdit(this);
	d_err2_edit->setAlignment(Qt::AlignCenter);
	dialog_layout->addWidget(d_err2_edit,2,1,1,1);

	d_pos3_edit = new QLineEdit(this);
	d_pos3_edit->setAlignment(Qt::AlignCenter);
	dialog_layout->addWidget(d_pos3_edit,3,0,1,1);

	d_err3_edit = new QLineEdit(this);
	d_err3_edit->setAlignment(Qt::AlignCenter);
	dialog_layout->addWidget(d_err3_edit,3,1,1,1);

	d_pos4_edit = new QLineEdit(this);
	d_pos4_edit->setAlignment(Qt::AlignCenter);
	dialog_layout->addWidget(d_pos4_edit,4,0,1,1);

	d_err4_edit = new QLineEdit(this);
	d_err4_edit->setAlignment(Qt::AlignCenter);
	dialog_layout->addWidget(d_err4_edit,4,1,1,1);

	d_pos5_edit = new QLineEdit(this);
	d_pos5_edit->setAlignment(Qt::AlignCenter);
	dialog_layout->addWidget(d_pos5_edit,5,0,1,1);

	d_err5_edit = new QLineEdit(this);
	d_err5_edit->setAlignment(Qt::AlignCenter);
	dialog_layout->addWidget(d_err5_edit,5,1,1,1);
	
	separator_hline = new QFrame(this);
	separator_hline->setFrameShape(QFrame::HLine);
	separator_hline->setFrameShadow(QFrame::Sunken);
	dialog_layout->addWidget(separator_hline,6,0,1,2);

	button_hlayout = new QHBoxLayout();

	button_hspacer = new QSpacerItem(0,0,QSizePolicy::Expanding,QSizePolicy::Minimum);
	button_hlayout->addItem(button_hspacer);

	cancel_button = new QPushButton(this);
	cancel_button->setAutoDefault(false);
	button_hlayout->addWidget(cancel_button);

	d_apply_button = new QPushButton(this);
	d_apply_button->setAutoDefault(false);
	button_hlayout->addWidget(d_apply_button);
	dialog_layout->addLayout(button_hlayout,7,0,1,2);
	
	// defaults
	d_min = 0.0;
	d_max = 1000.0;
	
	d_pos1_edit->setReadOnly(true);
	d_pos2_edit->setReadOnly(true);
	d_pos3_edit->setReadOnly(true);
	d_pos4_edit->setReadOnly(true);
	d_pos5_edit->setReadOnly(true);
	
	QWidget::setTabOrder(d_err1_edit,d_err2_edit);
	QWidget::setTabOrder(d_err2_edit,d_err3_edit);
	QWidget::setTabOrder(d_err3_edit,d_err4_edit);
	QWidget::setTabOrder(d_err4_edit,d_err5_edit);
	QWidget::setTabOrder(d_err5_edit,d_pos1_edit);
	QWidget::setTabOrder(d_pos1_edit,d_pos2_edit);
	QWidget::setTabOrder(d_pos2_edit,d_pos3_edit);
	QWidget::setTabOrder(d_pos3_edit,d_pos4_edit);
	QWidget::setTabOrder(d_pos4_edit,d_pos5_edit);
	QWidget::setTabOrder(d_pos5_edit,d_apply_button);
	QWidget::setTabOrder(d_apply_button,cancel_button);

	this->setWindowTitle(QStringLiteral("Expression Editor"));
	
	position_label->setText(QStringLiteral("Position (mm)"));
	error_label->setText(QStringLiteral("Error (mm)"));
	d_apply_button->setText(QStringLiteral("Apply"));
	cancel_button->setText(QStringLiteral("Cancel"));
	
	connect(d_apply_button,SIGNAL(clicked(void)),this,SLOT(accept(void)));
	connect(cancel_button,SIGNAL(clicked(void)),this,SLOT(reject(void)));
}

TExpressionDialog::~TExpressionDialog(void)
{
}

QString TExpressionDialog::Expression(void) const
{
	QString								result(QStringLiteral("0.0000"));
	double								a,b,c,d,e;
	bool								valid1,valid2;
	bool								init;
	std::vector<TVector2>				pnts;
	TVector2							pnt;
	static const double					MIN_VALUE(0.000049);
	
	pnt.x = d_pos1_edit->text().toDouble(&valid1) / 1000.0;
	pnt.y = d_err1_edit->text().toDouble(&valid2);
	
	if(valid1 == false || valid2 == false)
	{
		return result;
	}
	
	pnts.push_back(pnt);
	
	pnt.x = d_pos2_edit->text().toDouble(&valid1) / 1000.0;
	pnt.y = d_err2_edit->text().toDouble(&valid2);
	
	if(valid1 == false || valid2 == false)
	{
		return result;
	}
	
	pnts.push_back(pnt);
	
	pnt.x = d_pos3_edit->text().toDouble(&valid1) / 1000.0;
	pnt.y = d_err3_edit->text().toDouble(&valid2);
	
	if(valid1 == false || valid2 == false)
	{
		return result;
	}
	
	pnts.push_back(pnt);
	
	pnt.x = d_pos4_edit->text().toDouble(&valid1) / 1000.0;
	pnt.y = d_err4_edit->text().toDouble(&valid2);
	
	if(valid1 == false || valid2 == false)
	{
		return result;
	}
	
	pnts.push_back(pnt);
		
	pnt.x = d_pos5_edit->text().toDouble(&valid1) / 1000.0;
	pnt.y = d_err5_edit->text().toDouble(&valid2);
	
	if(valid1 == false || valid2 == false)
	{
		return result;
	}
	
	pnts.push_back(pnt);

	if(LeastSquares::Fit_Curve(pnts,&e,&d,&c,&b,&a))	// ax^4+bx^3+cx^2+dx+e
	{	
		init = true;
		
		if(fabs(a) > MIN_VALUE)
		{
			if(init)
			{
				result = QString("%1").arg(a,0,'f',4);
				init = false;
			}
			else
			{
				if(a > 0.0)
				{
					result += QString("+%1").arg(a,0,'f',4);
				}
				else
				{
					result += QString("%1").arg(a,0,'f',4);
				}
			}
		}
		
		if(fabs(b) > MIN_VALUE)
		{
			if(init)
			{
				result = QString("%1*L").arg(b,0,'f',4);
				init = false;
			}
			else
			{
				if(b > 0.0)
				{
					result += QString("+%1*L").arg(b,0,'f',4);
				}
				else
				{
					result += QString("%1*L").arg(b,0,'f',4);
				}
			}
		}
		
		if(fabs(c) > MIN_VALUE)
		{
			if(init)
			{
				result = QString("%1*L^2").arg(c,0,'f',4);
				init = false;
			}
			else
			{
				if(c > 0.0)
				{
					result += QString("+%1*L^2").arg(c,0,'f',4);
				}
				else
				{
					result += QString("%1*L^2").arg(c,0,'f',4);
				}
			}
		}
		
		if(fabs(d) > MIN_VALUE)
		{
			if(init)
			{
				result = QString("%1*L^3").arg(d,0,'f',4);
				init = false;
			}
			else
			{
				if(d > 0.0)
				{
					result += QString("+%1*L^3").arg(d,0,'f',4);
				}
				else
				{
					result += QString("%1*L^3").arg(d,0,'f',4);
				}
			}
		}

		if(fabs(e) > MIN_VALUE)
		{
			if(init)
			{
				result = QString("%1*L^4").arg(e,0,'f',4);
				init = false;
			}
			else
			{
				if(e > 0.0)
				{
					result += QString("+%1*L^4").arg(e,0,'f',4);
				}
				else
				{
					result += QString("%1*L^4").arg(e,0,'f',4);
				}
			}
		}
	}
	
	return result;
}

void TExpressionDialog::Set_Expression(
	const QString						&text)
{
	TExpression							expression;
	double								pos1;
	double								pos2;
	double								pos3;
	double								dval;
	bool								valid;
	
	pos2 = (d_min + d_max) / 2.0;
	pos1 = (d_min + pos2) / 2.0;
	pos3 = (pos2 + d_max) / 2.0;
	
	expression.Add_Variable(QStringLiteral("L"),d_min / 1000.0);
	
	d_pos1_edit->setText(QString("%1").arg(d_min,0,'f',1));
	dval = expression.Evaluate(text,&valid);
	if(valid)
	{
		d_err1_edit->setText(QString("%1").arg(dval,0,'f',4));
	}
	else
	{
		d_err1_edit->setText(QStringLiteral("0.0000"));
	}
		
	expression.Add_Variable(QStringLiteral("L"),pos1 / 1000.0);

	d_pos2_edit->setText(QString("%1").arg(pos1,0,'f',1));
	dval = expression.Evaluate(text,&valid);
	if(valid)
	{
		d_err2_edit->setText(QString("%1").arg(dval,0,'f',4));
	}
	else
	{
		d_err2_edit->setText(QStringLiteral("0.0000"));
	}

	expression.Add_Variable(QStringLiteral("L"),pos2 / 1000.0);

	d_pos3_edit->setText(QString("%1").arg(pos2,0,'f',1));
	dval = expression.Evaluate(text,&valid);
	if(valid)
	{
		d_err3_edit->setText(QString("%1").arg(dval,0,'f',4));
	}
	else
	{
		d_err3_edit->setText(QStringLiteral("0.0000"));
	}
		
	expression.Add_Variable(QStringLiteral("L"),pos3 / 1000.0);

	d_pos4_edit->setText(QString("%1").arg(pos3,0,'f',1));
	dval = expression.Evaluate(text,&valid);
	if(valid)
	{
		d_err4_edit->setText(QString("%1").arg(dval,0,'f',4));
	}
	else
	{
		d_err4_edit->setText(QStringLiteral("0.0000"));
	}
	
	expression.Add_Variable(QStringLiteral("L"),d_max / 1000.0);

	d_pos5_edit->setText(QString("%1").arg(d_max,0,'f',1));
	dval = expression.Evaluate(text,&valid);
	
	if(valid)
	{
		d_err5_edit->setText(QString("%1").arg(dval,0,'f',4));
	}
	else
	{
		d_err5_edit->setText(QStringLiteral("0.0000"));
	}
}


void TExpressionDialog::closeEvent(
	QCloseEvent							*event)
{
	event->accept();
}


