/////////////////////////////////////////////////////////////////////
//
//            X   X           X
//           XXX XX         XX
//          XXXXXXXX      XXX
//         XX X XXXXXXXXXXX
//        XXXXX XXXXXXXXX
//       XXXXX XXXXXXXXXX
//            XXX XXX XXX
//           XXX XX   XX
//           X   X     X
//
//    Copyright (C) 2003-2023  Ron Jakl
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
/////////////////////////////////////////////////////////////////////

#include <assert.h>

#include "../../core/interpolation.h"
#include "../../core/expression.h"
#include "../../core/vector2.h"
#include "types.h"

#include "machine_deflectioncomp.h"

TMachineDeflectionComp::TMachineDeflectionComp(void)
:d_vertical_axis_start(0.0),
d_vertical_axis_end(0.0),
d_vertical_axis_mechanical_origin(0.0),
d_horizontal_axis_start(0.0),
d_horizontal_axis_end(0.0),
d_horizontal_axis_mechanical_origin(0.0),
d_deflection_sign(1.0),
d_deflection_value(0.0)
{
}

TMachineDeflectionComp::~TMachineDeflectionComp(void)
{
}

bool TMachineDeflectionComp::Is_Valid(void) const
{
	const double						MIN_AXIS_RANGE(1.00);

	if((d_vertical_axis_end - d_vertical_axis_start) < MIN_AXIS_RANGE ||
	   (d_horizontal_axis_end - d_horizontal_axis_start) < MIN_AXIS_RANGE)
	{
		return false;
	}
	
	if(d_vertical_axis_mechanical_origin < d_vertical_axis_start ||
	   d_vertical_axis_mechanical_origin > d_vertical_axis_end)
	{
		return false;
	}
	
	if(d_horizontal_axis_mechanical_origin < d_horizontal_axis_start ||
	   d_horizontal_axis_mechanical_origin > d_horizontal_axis_end)
	{
		return false;
	}
	

	return true;
}


double TMachineDeflectionComp::Evaluate(
	const double						&pos_vert,
	const double						&pos_horz) const
{
	double								length_vertical;
	double								length_horizontal;
	double								range_vertical;
	double								range_horizontal;
	double								scale_vertical;
	double								scale_horizontal;
	const double						ZERO_EPSILON(0.001);
	
	range_vertical = d_vertical_axis_end - d_vertical_axis_start;
	range_horizontal = d_horizontal_axis_end - d_horizontal_axis_start;
	
	if(range_vertical < ZERO_EPSILON || range_horizontal < ZERO_EPSILON)
	{
		throw TMachineAxisException(QString("Deflection axis range invalid."));
	}

	length_vertical = pos_vert - d_vertical_axis_mechanical_origin;
	length_horizontal = pos_horz - d_horizontal_axis_mechanical_origin;

	scale_vertical = length_vertical / range_vertical;
	scale_horizontal = (length_horizontal / range_horizontal) - 0.5;
	
	return d_deflection_value * scale_vertical * scale_horizontal;
}

void TMachineDeflectionComp::Set_Vertical_Axis(
	const double						&start,
	const double						&end,
	const double						&mech_origin)
{
	if(end > start)
	{
		d_vertical_axis_start = start;
		d_vertical_axis_end = end;
	}
	else
	{
		d_vertical_axis_start = end;
		d_vertical_axis_end = start;
	}
	
	d_vertical_axis_mechanical_origin = mech_origin;
}

void TMachineDeflectionComp::Set_Horizontal_Axis(
	const double						&start,
	const double						&end,
	const double						&mech_origin)
{
	if(end > start)
	{
		d_horizontal_axis_start = start;
		d_horizontal_axis_end = end;
	}
	else
	{
		d_horizontal_axis_start = end;
		d_horizontal_axis_end = start;
	}
	d_horizontal_axis_mechanical_origin = mech_origin;
}


void TMachineDeflectionComp::Initialize(
	const double						&deflection,
	const int							sign)
{
	if(sign > 0)
	{
		d_deflection_sign = 1.0;
	}
	else
	{
		d_deflection_sign = -1.0;
	}
	
	d_deflection_value = deflection;
}
