/////////////////////////////////////////////////////////////////////
//
//            X   X           X
//           XXX XX         XX
//          XXXXXXXX      XXX
//         XX X XXXXXXXXXXX
//        XXXXX XXXXXXXXX
//       XXXXX XXXXXXXXXX
//            XXX XXX XXX
//           XXX XX   XX
//           X   X     X
//
//    Copyright (C) 2003-2023  Ron Jakl
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
/////////////////////////////////////////////////////////////////////

#include <assert.h>

#include "../../core/interpolation.h"
#include "../../core/expression.h"
#include "types.h"

#include "machine_standardcomp.h"

TMachineStandardComp::TMachineStandardComp(void)
:d_axis_id(-1),
d_rx_sign(1),
d_ry_sign(1),
d_rz_sign(1),
d_scale_sign(1),
d_quad_sign(1),
d_axis_start(0.0),
d_axis_end(0.0),
d_math_origin(0.0),
d_mechanical_origin(0.0),
d_quad(0.0)
{
}

TMachineStandardComp::~TMachineStandardComp(void)
{
}

int TMachineStandardComp::Data_Size(void)
{
	return static_cast<int>(d_position_data.size());
}

void TMachineStandardComp::Data(
	const int							index,
	double								* const position,
	TVector3							* const angular,
	TVector3							* const linear) const
{
	assert(position);
	assert(angular);
	assert(linear);
	
	assert(static_cast<unsigned int>(index) < d_position_data.size());
	assert(static_cast<unsigned int>(index) < d_angular_data.size());
	assert(static_cast<unsigned int>(index) < d_linear_data.size());
	
	*position = d_position_data[index];
	*angular = d_angular_data[index];
	*linear = d_linear_data[index];
}

void TMachineStandardComp::Data(
	const double						&position,
	TVector3							* const angular,
	TVector3							* const linear) const
{
	double								dval;
	
	assert(angular);
	assert(linear);
	
	Extrapolation::Get_Value(d_rx_data,position,&dval);
	(*angular).x = dval;
	
	Extrapolation::Get_Value(d_ry_data,position,&dval);
	(*angular).y = dval;
	
	Extrapolation::Get_Value(d_rz_data,position,&dval);
	(*angular).z = dval;
	
	Extrapolation::Get_Value(d_lx_data,position,&dval);
	(*linear).x = dval;
	
	Extrapolation::Get_Value(d_ly_data,position,&dval);
	(*linear).y = dval;
	
	Extrapolation::Get_Value(d_lz_data,position,&dval);
	(*linear).z = dval;
}

void TMachineStandardComp::Initialize(
	const int							axis_id)
{
	double								position;
	double								increment;
	int									cntr;
	
	assert(TMachineStandardComp::AXIS_DATA_SIZE > 1);
	
	assert(!(d_axis_end < d_axis_start));
	assert(!(d_math_origin < d_axis_start || d_math_origin > d_axis_end));
	assert(!(d_mechanical_origin < d_axis_start || d_mechanical_origin > d_axis_end));
	
	d_position_data.clear();
	d_angular_data.clear();
	d_linear_data.clear();
	d_lx_data.clear();
	d_ly_data.clear();
	d_lz_data.clear();
	d_rx_data.clear();
	d_ry_data.clear();
	d_rz_data.clear();
	
	d_position_data.resize(TMachineStandardComp::AXIS_DATA_SIZE);
	d_angular_data.resize(TMachineStandardComp::AXIS_DATA_SIZE);
	d_linear_data.resize(TMachineStandardComp::AXIS_DATA_SIZE);
	
	d_lx_data.resize(TMachineStandardComp::AXIS_DATA_SIZE);
	d_ly_data.resize(TMachineStandardComp::AXIS_DATA_SIZE);
	d_lz_data.resize(TMachineStandardComp::AXIS_DATA_SIZE);
	d_rx_data.resize(TMachineStandardComp::AXIS_DATA_SIZE);
	d_ry_data.resize(TMachineStandardComp::AXIS_DATA_SIZE);
	d_rz_data.resize(TMachineStandardComp::AXIS_DATA_SIZE);

	position = d_axis_start;
	increment = (d_axis_end - d_axis_start) / (static_cast<double>(TMachineStandardComp::AXIS_DATA_SIZE - 1));
											   
	for(cntr=0; cntr < TMachineStandardComp::AXIS_DATA_SIZE; ++cntr)
	{
		d_position_data[cntr] = position;
		position += increment;
	}
	
	d_rx_sign = 1;
	d_ry_sign = 1;
	d_rz_sign = 1;
	d_scale_sign = 1;
	d_quad_sign = 1;
	
	d_quad = 0.0;
	d_axis_id = axis_id;
}

void TMachineStandardComp::Set_Rx(
	const QString						&expression,
	const int							sign)
{
	TExpression							expression_evaluator;
	std::vector<double>::const_iterator	pos_iter;
	std::vector<TVector3>::iterator		ang_iter;
	std::vector<TVector2>::iterator		param_iter;
	bool								valid;
	
	assert(d_position_data.size() == d_angular_data.size());
	assert(d_position_data.size() == d_rx_data.size());
	
	for(pos_iter = d_position_data.begin(),ang_iter = d_angular_data.begin(), param_iter = d_rx_data.begin();
		pos_iter != d_position_data.end();
		++pos_iter, ++ang_iter,++param_iter)
	{
		expression_evaluator.Add_Variable(QStringLiteral("L"),(*pos_iter) / 1000);	// note, L is in meters
		
		(*ang_iter).x = expression_evaluator.Evaluate(expression,&valid);
		
		if(!valid)
		{
			throw TMachineAxisException(QString("Rx expression '%1' invalid").arg(expression));
		}
		
		(*param_iter).x = (*pos_iter);
		(*param_iter).y = (*ang_iter).x;
	}
	
	d_rx_sign = sign;
}

void TMachineStandardComp::Set_Ry(
	const QString						&expression,
	const int							sign)
{
	TExpression							expression_evaluator;
	std::vector<double>::const_iterator	pos_iter;
	std::vector<TVector3>::iterator		ang_iter;
	std::vector<TVector2>::iterator		param_iter;
	bool								valid;
	
	assert(d_position_data.size() == d_angular_data.size());
	assert(d_position_data.size() == d_rx_data.size());
	
	for(pos_iter = d_position_data.begin(),ang_iter = d_angular_data.begin(), param_iter = d_ry_data.begin();
		pos_iter != d_position_data.end();
		++pos_iter, ++ang_iter,++param_iter)
	{
		expression_evaluator.Add_Variable(QStringLiteral("L"),(*pos_iter) / 1000);	// note, L is in meters
		
		(*ang_iter).y = expression_evaluator.Evaluate(expression,&valid);
		
		if(!valid)
		{
			throw TMachineAxisException(QString("Ry expression '%1' invalid").arg(expression));
		}
		
		(*param_iter).x = (*pos_iter);
		(*param_iter).y = (*ang_iter).y;
	}
	
	d_ry_sign = sign;
}

void TMachineStandardComp::Set_Rz(
	const QString						&expression,
	const int							sign)
{
	TExpression							expression_evaluator;
	std::vector<double>::const_iterator	pos_iter;
	std::vector<TVector3>::iterator		ang_iter;
	std::vector<TVector2>::iterator		param_iter;
	bool								valid;
	
	assert(d_position_data.size() == d_angular_data.size());
	assert(d_position_data.size() == d_rx_data.size());
	
	for(pos_iter = d_position_data.begin(),ang_iter = d_angular_data.begin(), param_iter = d_rz_data.begin();
		pos_iter != d_position_data.end();
		++pos_iter, ++ang_iter,++param_iter)
	{
		expression_evaluator.Add_Variable(QStringLiteral("L"),(*pos_iter) / 1000);	// note, L is in meters
		
		(*ang_iter).z = expression_evaluator.Evaluate(expression,&valid);
		
		if(!valid)
		{
			throw TMachineAxisException(QString("Rz expression '%1' invalid").arg(expression));
		}
		
		(*param_iter).x = (*pos_iter);
		(*param_iter).y = (*ang_iter).z;
	}
	
	d_rz_sign = sign;
}

void TMachineStandardComp::Set_Scale(
	const QString						&expression,
	const int							sign)
{
	TExpression							expression_evaluator;
	std::vector<double>::const_iterator	pos_iter;
	std::vector<TVector3>::iterator		lin_iter;
	int									index;
	bool								valid;
	
	assert(d_position_data.size() == d_linear_data.size());
	assert(d_position_data.size() == d_lx_data.size());
	assert(d_position_data.size() == d_ly_data.size());
	assert(d_position_data.size() == d_lz_data.size());
	assert(d_position_data.size() == d_rx_data.size());
	assert(d_position_data.size() == d_ry_data.size());
	assert(d_position_data.size() == d_rz_data.size());
	
	for(pos_iter = d_position_data.begin(),lin_iter = d_linear_data.begin(),index = 0;
		pos_iter != d_position_data.end();
		++pos_iter, ++lin_iter,++index)
	{
		expression_evaluator.Add_Variable(QStringLiteral("L"),(*pos_iter) / 1000);	// note, L is in meters
		
		if(d_axis_id == Types::AXIS_X)
		{
			(*lin_iter).x = expression_evaluator.Evaluate(expression,&valid);
			
			if(!valid)
			{
				throw TMachineAxisException(QString("X scale expression '%1' invalid").arg(expression));
			}
			
			d_lx_data[index].x = (*pos_iter);
			d_lx_data[index].y = (*lin_iter).x;
		}
		else if(d_axis_id == Types::AXIS_Y)
		{
			(*lin_iter).y = expression_evaluator.Evaluate(expression,&valid);
			
			if(!valid)
			{
				throw TMachineAxisException(QString("Y scale expression '%1' invalid").arg(expression));
			}
			
			d_ly_data[index].x = (*pos_iter);
			d_ly_data[index].y = (*lin_iter).y;
		}
		else if(d_axis_id == Types::AXIS_Y)
		{
			(*lin_iter).z = expression_evaluator.Evaluate(expression,&valid);
			
			if(!valid)
			{
				throw TMachineAxisException(QString("Z scale expression '%1' invalid").arg(expression));
			}
			
			d_lz_data[index].x = (*pos_iter);
			d_lz_data[index].y = (*lin_iter).z;
		}
	}
	
	d_scale_sign = sign;
}

void TMachineStandardComp::Set_Quad(
	const double						&value,
	const int							sign)
{
	d_quad = value;
	d_quad_sign = sign;
}

void TMachineStandardComp::Calculate_Straightness(
	const TStraightnessParameterMode			mode)
{
	TExpression							expression_evaluator;
	std::vector<double>::const_iterator	pos_iter;
	std::vector<TVector3>::iterator		lin_iter;
	std::vector<TVector3>::iterator		ang_iter;
	bool								init;
	TVector2							pnt;
	TVector2							prev_pnt;
	double								sign;
	std::vector<TVector2>				str_data;
	std::vector<TVector2>::iterator		str_data_iter;
	double								offset;
	double								slope;
	
	assert(d_position_data.size() == d_linear_data.size());
	assert(d_position_data.size() == d_lx_data.size());
	assert(d_position_data.size() == d_ly_data.size());
	assert(d_position_data.size() == d_lz_data.size());
	assert(d_position_data.size() == d_rx_data.size());
	assert(d_position_data.size() == d_ry_data.size());
	assert(d_position_data.size() == d_rz_data.size());
	
	// rotations RXY, RYZ, RZX rotate into the negative
	
	if(mode == TMachineStandardComp::X_STR_FROM_RY)
	{
		d_ry_sign > 0 ? sign = 1.0: sign = -1.0;
		
		// get shape
		init = true;
		str_data.clear();
		
		for(pos_iter = d_position_data.begin(),ang_iter = d_angular_data.begin();
			pos_iter != d_position_data.end();
			++pos_iter, ++ang_iter)
		{
			pnt.x = (*pos_iter);
			
			if(init)
			{
				pnt.y = 0;
				prev_pnt.y = 0;
				
				init = false;
			}
			else
			{
				pnt.y = (*ang_iter).y * (pnt.x - prev_pnt.x) / 1000.0 * sign;
			}
			
			pnt.y += prev_pnt.y;
			prev_pnt = pnt;
			
			str_data.push_back(pnt);
		}
		
		assert(str_data.size() == d_linear_data.size());
		
		// zero offset and remove slope
		if(Extrapolation::Get_Value(str_data,0.0,&offset))
		{
			for(str_data_iter = str_data.begin();str_data_iter != str_data.end();++str_data_iter)
			{
				(*str_data_iter).y -= offset;
			}
		}
		else
		{
			throw TMachineAxisException(QString("X straightness from Ry not solvable"));
		}
		
		if(Regression::Get_Slope_Through_Zero(str_data,&slope))
		{
			for(str_data_iter = str_data.begin();str_data_iter != str_data.end();++str_data_iter)
			{
				(*str_data_iter).y -= ((*str_data_iter).x * slope);
			}
		}
		else
		{
			throw TMachineAxisException(QString("X straightness from Ry not solvable"));
		}
		
		for(str_data_iter = str_data.begin(), lin_iter = d_linear_data.begin();
			str_data_iter != str_data.end();
			++str_data_iter,++lin_iter)
		{
			(*lin_iter).x = (*str_data_iter).y;
		}
		
		d_lx_data = str_data;
		
		return;
	}
	
	if(mode == TMachineStandardComp::X_STR_FROM_RZ)
	{
		if(d_axis_id == Types::AXIS_Y)
		{
			d_rz_sign > 0 ? sign = -1.0: sign = 1.0;
		}
		else
		{
			d_rz_sign > 0 ? sign = 1.0: sign = -1.0;
		}
		
		// get shape
		init = true;
		str_data.clear();
		
		for(pos_iter = d_position_data.begin(),ang_iter = d_angular_data.begin();
			pos_iter != d_position_data.end();
			++pos_iter, ++ang_iter)
		{
			pnt.x = (*pos_iter);
			
			if(init)
			{
				pnt.y = 0;
				prev_pnt.y = 0;
				
				init = false;
			}
			else
			{
				pnt.y = (*ang_iter).z * (pnt.x - prev_pnt.x) / 1000.0 * sign;
			}
			
			pnt.y += prev_pnt.y;
			prev_pnt = pnt;
			
			str_data.push_back(pnt);
		}
		
		assert(str_data.size() == d_linear_data.size());
		
		// zero offset and remove slope
		if(Extrapolation::Get_Value(str_data,0.0,&offset))
		{
			for(str_data_iter = str_data.begin();str_data_iter != str_data.end();++str_data_iter)
			{
				(*str_data_iter).y -= offset;
			}
			
		}
		else
		{
			throw TMachineAxisException(QString("X straightness from Rz not solvable"));
		}
		
		if(Regression::Get_Slope_Through_Zero(str_data,&slope))
		{
			for(str_data_iter = str_data.begin();str_data_iter != str_data.end();++str_data_iter)
			{
				(*str_data_iter).y -= ((*str_data_iter).x * slope);
			}
		}
		else
		{
			throw TMachineAxisException(QString("X straightness from Rz not solvable"));
		}
		
		for(str_data_iter = str_data.begin(), lin_iter = d_linear_data.begin();
			str_data_iter != str_data.end();
			++str_data_iter,++lin_iter)
		{
			(*lin_iter).x = (*str_data_iter).y;
		}
		
		d_lx_data = str_data;
		
		return;
	}
	
	if(mode == TMachineStandardComp::Y_STR_FROM_RX)
	{
		if(d_axis_id == Types::AXIS_Z)
		{
			d_rx_sign > 0 ? sign = -1.0: sign = 1.0;
		}
		else
		{
			d_rx_sign > 0 ? sign = 1.0: sign = -1.0;
		}
		
		// get shape
		init = true;
		str_data.clear();
		
		for(pos_iter = d_position_data.begin(),ang_iter = d_angular_data.begin();
			pos_iter != d_position_data.end();
			++pos_iter, ++ang_iter)
		{
			pnt.x = (*pos_iter);
			
			if(init)
			{
				pnt.y = 0;
				prev_pnt.y = 0;
				
				init = false;
			}
			else
			{
				pnt.y = (*ang_iter).x * (pnt.x - prev_pnt.x) / 1000.0 * sign;
			}
			
			pnt.y += prev_pnt.y;
			prev_pnt = pnt;
			
			str_data.push_back(pnt);
		}
		
		assert(str_data.size() == d_linear_data.size());
		
		// zero offset and remove slope
		if(Extrapolation::Get_Value(str_data,0.0,&offset))
		{
			for(str_data_iter = str_data.begin();str_data_iter != str_data.end();++str_data_iter)
			{
				(*str_data_iter).y -= offset;
			}
		}
		else
		{
			throw TMachineAxisException(QString("Y straightness from Rx not solvable"));
		}
		
		if(Regression::Get_Slope_Through_Zero(str_data,&slope))
		{
			for(str_data_iter = str_data.begin();str_data_iter != str_data.end();++str_data_iter)
			{
				(*str_data_iter).y -= ((*str_data_iter).x * slope);
			}
		}
		else
		{
			throw TMachineAxisException(QString("Y straightness from Rx not solvable"));
		}
		
		for(str_data_iter = str_data.begin(), lin_iter = d_linear_data.begin();
			str_data_iter != str_data.end();
			++str_data_iter,++lin_iter)
		{
			(*lin_iter).y = (*str_data_iter).y;
		}
		
		d_ly_data = str_data;
		
		return;
	}
	
	if(mode == TMachineStandardComp::Y_STR_FROM_RZ)
	{
		d_rz_sign > 0 ? sign = 1.0: sign = -1.0;
		
		// get shape
		init = true;
		str_data.clear();
		
		for(pos_iter = d_position_data.begin(),ang_iter = d_angular_data.begin();
			pos_iter != d_position_data.end();
			++pos_iter, ++ang_iter)
		{
			pnt.x = (*pos_iter);
			
			if(init)
			{
				pnt.y = 0;
				prev_pnt.y = 0;
				
				init = false;
			}
			else
			{
				pnt.y = (*ang_iter).z * (pnt.x - prev_pnt.x) / 1000.0 * sign;
			}
			
			pnt.y += prev_pnt.y;
			prev_pnt = pnt;
			
			str_data.push_back(pnt);
		}
		
		assert(str_data.size() == d_linear_data.size());
		
		// zero offset and remove slope
		if(Extrapolation::Get_Value(str_data,0.0,&offset))
		{
			for(str_data_iter = str_data.begin();str_data_iter != str_data.end();++str_data_iter)
			{
				(*str_data_iter).y -= offset;
			}
		}
		else
		{
			throw TMachineAxisException(QString("Y straightness from Rz not solvable"));
		}
		
		if(Regression::Get_Slope_Through_Zero(str_data,&slope))
		{
			for(str_data_iter = str_data.begin();str_data_iter != str_data.end();++str_data_iter)
			{
				(*str_data_iter).y -= ((*str_data_iter).x * slope);
			}
		}
		else
		{
			throw TMachineAxisException(QString("Y straightness from Rz not solvable"));
		}
		
		for(str_data_iter = str_data.begin(), lin_iter = d_linear_data.begin();
			str_data_iter != str_data.end();
			++str_data_iter,++lin_iter)
		{
			(*lin_iter).y = (*str_data_iter).y;
		}
		
		d_ly_data = str_data;
		
		return;
	}
	
	if(mode == TMachineStandardComp::Z_STR_FROM_RX)
	{
		d_rx_sign > 0 ? sign = 1.0: sign = -1.0;
		
		// get shape
		init = true;
		str_data.clear();
		
		for(pos_iter = d_position_data.begin(),ang_iter = d_angular_data.begin();
			pos_iter != d_position_data.end();
			++pos_iter, ++ang_iter)
		{
			pnt.x = (*pos_iter);
			
			if(init)
			{
				pnt.y = 0;
				prev_pnt.y = 0;
				
				init = false;
			}
			else
			{
				pnt.y = (*ang_iter).x * (pnt.x - prev_pnt.x) / 1000.0 * sign;
			}
			
			pnt.y += prev_pnt.y;
			prev_pnt = pnt;
			
			str_data.push_back(pnt);
		}
		
		assert(str_data.size() == d_linear_data.size());
		
		// zero offset and remove slope
		if(Extrapolation::Get_Value(str_data,0.0,&offset))
		{
			for(str_data_iter = str_data.begin();str_data_iter != str_data.end();++str_data_iter)
			{
				(*str_data_iter).y -= offset;
			}
		}
		else
		{
			throw TMachineAxisException(QString("Z straightness from Rx not solvable"));
		}
		
		if(Regression::Get_Slope_Through_Zero(str_data,&slope))
		{
			for(str_data_iter = str_data.begin();str_data_iter != str_data.end();++str_data_iter)
			{
				(*str_data_iter).y -= ((*str_data_iter).x * slope);
			}
		}
		else
		{
			throw TMachineAxisException(QString("Z straightness from Rx not solvable"));
		}
		
		for(str_data_iter = str_data.begin(), lin_iter = d_linear_data.begin();
			str_data_iter != str_data.end();
			++str_data_iter,++lin_iter)
		{
			(*lin_iter).z = (*str_data_iter).y;
		}
		
		d_lz_data = str_data;
		
		return;
	}
	
	if(mode == TMachineStandardComp::Z_STR_FROM_RY)
	{
		if(d_axis_id == Types::AXIS_X)
		{
			d_ry_sign > 0 ? sign = -1.0: sign = 1.0;
		}
		else
		{
			d_ry_sign > 0 ? sign = 1.0: sign = -1.0;
		}
		
		// get shape
		init = true;
		str_data.clear();
		
		for(pos_iter = d_position_data.begin(),ang_iter = d_angular_data.begin();
			pos_iter != d_position_data.end();
			++pos_iter, ++ang_iter)
		{
			pnt.x = (*pos_iter);
			
			if(init)
			{
				pnt.y = 0;
				prev_pnt.y = 0;
				
				init = false;
			}
			else
			{
				pnt.y = (*ang_iter).y * (pnt.x - prev_pnt.x) / 1000.0 * sign;
			}
			
			pnt.y += prev_pnt.y;
			prev_pnt = pnt;
			
			str_data.push_back(pnt);
		}
		
		assert(str_data.size() == d_linear_data.size());
		
		// zero offset and remove slope
		if(Extrapolation::Get_Value(str_data,0.0,&offset))
		{
			for(str_data_iter = str_data.begin();str_data_iter != str_data.end();++str_data_iter)
			{
				(*str_data_iter).y -= offset;
			}
		}
		else
		{
			throw TMachineAxisException(QString("Z straightness from Ry not solvable"));
		}
		
		
		if(Regression::Get_Slope_Through_Zero(str_data,&slope))
		{
			for(str_data_iter = str_data.begin();str_data_iter != str_data.end();++str_data_iter)
			{
				(*str_data_iter).y -= ((*str_data_iter).x * slope);
			}
		}
		else
		{
			throw TMachineAxisException(QString("Z straightness from Ry not solvable"));
		}
		
		for(str_data_iter = str_data.begin(), lin_iter = d_linear_data.begin();
			str_data_iter != str_data.end();
			++str_data_iter,++lin_iter)
		{
			(*lin_iter).z = (*str_data_iter).y;
		}
		
		d_lz_data = str_data;
		
		return;
	}
}

