/////////////////////////////////////////////////////////////////////
//
//            X   X           X
//           XXX XX         XX
//          XXXXXXXX      XXX
//         XX X XXXXXXXXXXX
//        XXXXX XXXXXXXXX
//       XXXXX XXXXXXXXXX
//            XXX XXX XXX
//           XXX XX   XX
//           X   X     X
//
//    Copyright (C) 2003-2023  Ron Jakl
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
/////////////////////////////////////////////////////////////////////

#ifndef MACHINE_STANDARDCOMPHEADERFILE
#define MACHINE_STANDARDCOMPHEADERFILE

#include <QString>
#include <vector>

#include "../../core/vector3.h"
#include "../../core/vector2.h"

class TMachineStandardComp
{
public:
enum TStraightnessParameterMode
{
	X_STR_FROM_RY = 0,
	X_STR_FROM_RZ,
	Y_STR_FROM_RX,
	Y_STR_FROM_RZ,
	Z_STR_FROM_RX,
	Z_STR_FROM_RY,
};
	
// CREATORS
	TMachineStandardComp(void);
	~TMachineStandardComp(void);
	
// ACCESSORS
	
	int Axis_Id(void) const {return d_axis_id;}
	
	int Data_Size(void);
	void Data(const int index, double * const position, TVector3 * const angular, TVector3 * const linear) const;
	void Data(const double &position, TVector3 * const angular, TVector3 * const linear) const;
	double Axis_Origin(void) const {return d_math_origin;}
	double Axis_Mechanical_Origin(void) const {return d_mechanical_origin;}
	
	double Rx_Rotation_Sign(void) const {return (d_rx_sign < 0) ? 1.0 : -1.0;}	// sign suitable for rotation direction
	double Ry_Rotation_Sign(void) const {return (d_ry_sign < 0) ? 1.0 : -1.0;}	// sign suitable for rotation direction
	double Rz_Rotation_Sign(void) const {return (d_rz_sign < 0) ? 1.0 : -1.0;}	// sign suitable for rotation direction
	double Scale_Sign(void) const {return (d_scale_sign < 0) ? -1.0 : 1.0;}
	double Quad_Sign(void) const {return (d_quad_sign < 0) ? -1.0 : 1.0;}
	
	double Quad(void) const {return d_quad;}
	
// MANIPULATORS
	void Set_Axis_Range(const double &start,const double &end) {d_axis_start = start;d_axis_end = end;}
	void Set_Axis_Origin(const double &math_o,const double &mech_o) {d_math_origin = math_o;d_mechanical_origin = mech_o;}
	
	void Initialize(const int axis_id);
	
	void Set_Rx(const QString &expression,const int sign);
	void Set_Ry(const QString &expression,const int sign);
	void Set_Rz(const QString &expression,const int sign);
	void Set_Scale(const QString &expression,const int sign);
	void Set_Quad(const double &value,const int sign); // squareness calculated as error @ 1000 mm from expression
	
	void Calculate_Straightness(const TStraightnessParameterMode mode);
	
private:
	
	static const int					AXIS_DATA_SIZE = 51;	// number of map steps to create.
	
	int									d_axis_id;
	
	int									d_rx_sign;
	int									d_ry_sign;
	int									d_rz_sign;
	int									d_scale_sign;
	int									d_quad_sign;
	
	double								d_axis_start;
	double								d_axis_end;
	double								d_math_origin;
	double								d_mechanical_origin;
	double								d_quad;
	
	std::vector<double>					d_position_data;
	std::vector<TVector3>				d_angular_data;
	std::vector<TVector3>				d_linear_data;
	
	std::vector<TVector2>				d_rx_data;
	std::vector<TVector2>				d_ry_data;
	std::vector<TVector2>				d_rz_data;
	std::vector<TVector2>				d_lx_data;
	std::vector<TVector2>				d_ly_data;
	std::vector<TVector2>				d_lz_data;
	

// NOT IMPLEMENTED
	TMachineStandardComp(const TMachineStandardComp&);
	TMachineStandardComp& operator=(const TMachineStandardComp&);
};

#endif
