/////////////////////////////////////////////////////////////////////
//
//            X   X           X
//           XXX XX         XX
//          XXXXXXXX      XXX
//         XX X XXXXXXXXXXX
//        XXXXX XXXXXXXXX
//       XXXXX XXXXXXXXXX
//            XXX XXX XXX
//           XXX XX   XX
//           X   X     X
//
//    Copyright (C) 2003-2023  Ron Jakl
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
/////////////////////////////////////////////////////////////////////

#include <assert.h>

#include "../../core/vector2.h"
#include "../../core/mat4.h"
#include "../../core/interpolation.h"

#include "types.h"
#include "machinecomp.h"

TMachineComp::TMachineComp(void)
{
	d_init_mode = false;
	d_valid_data = false;
}

TMachineComp::~TMachineComp(void)
{
}

QStringList TMachineComp::Map_Data_X(
	const double						&increment)
{
	QStringList							lines;
	QString								line;
	std::vector<TVector2>				rx_data;
	std::vector<TVector2>				ry_data;
	std::vector<TVector2>				rz_data;
	std::vector<TVector2>				lx_data;
	std::vector<TVector2>				ly_data;
	std::vector<TVector2>				lz_data;
	TVector2							pnt;
	TVector3							ang_pnt;
	TVector3							lin_pnt;
	double								position;
	double								rx,ry,rz,lx,ly,lz;
	double								start_position(0);
	double								end_position(0);
	int									data_size;
	int									cntr;
	
	if(d_init_mode)
	{
		return lines;
	}
	
	if(increment < 1.0)
	{
		return lines;
	}
	
	// load up the 2D arrays
	if(d_init_parameters.first_axis_id == Types::AXIS_X)
	{
		data_size = d_first_axis.Data_Size();
		
		for(cntr = 0;cntr < data_size;++cntr)
		{
			d_first_axis.Data(cntr,&position,&ang_pnt,&lin_pnt);
			
			pnt.x = position;
			
			pnt.y = ang_pnt.x;
			rx_data.push_back(pnt);
			
			pnt.y = ang_pnt.y;
			ry_data.push_back(pnt);
			
			pnt.y = ang_pnt.z;
			rz_data.push_back(pnt);
			
			pnt.y = lin_pnt.x;
			lx_data.push_back(pnt);
			
			pnt.y = lin_pnt.y;
			ly_data.push_back(pnt);
			
			pnt.y = lin_pnt.z;
			lz_data.push_back(pnt);
			
			if(cntr == 0)
			{
				start_position = position;
			}
		}
		
		end_position = position;
	}
	else if(d_init_parameters.second_axis_id == Types::AXIS_X)
	{
		data_size = d_second_axis.Data_Size();
		
		for(cntr = 0;cntr < data_size;++cntr)
		{
			d_second_axis.Data(cntr,&position,&ang_pnt,&lin_pnt);
			
			pnt.x = position;
			
			pnt.y = ang_pnt.x;
			rx_data.push_back(pnt);
			
			pnt.y = ang_pnt.y;
			ry_data.push_back(pnt);
			
			pnt.y = ang_pnt.z;
			rz_data.push_back(pnt);
			
			pnt.y = lin_pnt.x;
			lx_data.push_back(pnt);
			
			pnt.y = lin_pnt.y;
			ly_data.push_back(pnt);
			
			pnt.y = lin_pnt.z;
			lz_data.push_back(pnt);
			
			if(cntr == 0)
			{
				start_position = position;
			}
		}
		
		end_position = position;
	}
	else if(d_init_parameters.third_axis_id == Types::AXIS_X)
	{
		data_size = d_third_axis.Data_Size();
		
		for(cntr = 0;cntr < data_size;++cntr)
		{
			d_third_axis.Data(cntr,&position,&ang_pnt,&lin_pnt);
			
			pnt.x = position;
			
			pnt.y = ang_pnt.x;
			rx_data.push_back(pnt);
			
			pnt.y = ang_pnt.y;
			ry_data.push_back(pnt);
			
			pnt.y = ang_pnt.z;
			rz_data.push_back(pnt);
			
			pnt.y = lin_pnt.x;
			lx_data.push_back(pnt);
			
			pnt.y = lin_pnt.y;
			ly_data.push_back(pnt);
			
			pnt.y = lin_pnt.z;
			lz_data.push_back(pnt);
			
			if(cntr == 0)
			{
				start_position = position;
			}
		}
		
		end_position = position;
	}
	else
	{
		return lines;
	}
	
	line = QStringLiteral("X Axis Data");
	lines.push_back(line);
	
	line = QStringLiteral("------------------------------------------------------------------------------");
	lines.push_back(line);
	
	//                    0         1         2         3         4         5         6         7         8         9
	//                    0123456789012345678901234567890123456789012345678901234567890123456789012345678901234567890
	//                    |          |          |          |          |          |          |          |
	line = QStringLiteral("     Pos         Lx         Ly         Lz         Rx         Ry         Rz");
	lines.push_back(line);
	
	line = QStringLiteral("------------------------------------------------------------------------------");
	lines.push_back(line);
	
	// sort data where end closest to zero is shown first
	if(fabs(end_position) > fabs(start_position))
	{
		position = start_position;
		while(!(position > end_position))
		{
			Extrapolation::Get_Value(rx_data,position,&rx);
			Extrapolation::Get_Value(ry_data,position,&ry);
			Extrapolation::Get_Value(rz_data,position,&rz);
			Extrapolation::Get_Value(lx_data,position,&lx);
			Extrapolation::Get_Value(ly_data,position,&ly);
			Extrapolation::Get_Value(lz_data,position,&lz);
			
			line = QString("%1 %2 %3 %4 %5 %6 %7")
						.arg(position,10,'f',1)
						.arg(lx,10,'f',4)
						.arg(ly,10,'f',4)
						.arg(lz,10,'f',4)
						.arg(rx,10,'f',4)
						.arg(ry,10,'f',4)
						.arg(rz,10,'f',4);
			
			lines.push_back(line);
			
			position += increment;
		};
	}
	else
	{
		position = end_position;
		while(!(position < start_position))
		{
			Extrapolation::Get_Value(rx_data,position,&rx);
			Extrapolation::Get_Value(ry_data,position,&ry);
			Extrapolation::Get_Value(rz_data,position,&rz);
			Extrapolation::Get_Value(lx_data,position,&lx);
			Extrapolation::Get_Value(ly_data,position,&ly);
			Extrapolation::Get_Value(lz_data,position,&lz);
			
			line = QString("%1 %2 %3 %4 %5 %6 %7")
						.arg(position,10,'f',1)
						.arg(lx,10,'f',4)
						.arg(ly,10,'f',4)
						.arg(lz,10,'f',4)
						.arg(rx,10,'f',4)
						.arg(ry,10,'f',4)
						.arg(rz,10,'f',4);
			
			lines.push_back(line);
			
			position -= increment;
		};
	}
	
	return lines;
}

QStringList TMachineComp::Map_Data_Y(
	const double						&increment)
{
	QStringList							lines;
	QString								line;
	std::vector<TVector2>				rx_data;
	std::vector<TVector2>				ry_data;
	std::vector<TVector2>				rz_data;
	std::vector<TVector2>				lx_data;
	std::vector<TVector2>				ly_data;
	std::vector<TVector2>				lz_data;
	TVector2							pnt;
	TVector3							ang_pnt;
	TVector3							lin_pnt;
	double								position;
	double								rx,ry,rz,lx,ly,lz;
	double								start_position(0);
	double								end_position(0);
	int									data_size;
	int									cntr;
	
	if(d_init_mode)
	{
		return lines;
	}
	
	if(increment < 1.0)
	{
		return lines;
	}
	
	// load up the 2D arrays
	if(d_init_parameters.first_axis_id == Types::AXIS_Y)
	{
		data_size = d_first_axis.Data_Size();
		
		for(cntr = 0;cntr < data_size;++cntr)
		{
			d_first_axis.Data(cntr,&position,&ang_pnt,&lin_pnt);
			
			pnt.x = position;
			
			pnt.y = ang_pnt.x;
			rx_data.push_back(pnt);
			
			pnt.y = ang_pnt.y;
			ry_data.push_back(pnt);
			
			pnt.y = ang_pnt.z;
			rz_data.push_back(pnt);
			
			pnt.y = lin_pnt.x;
			lx_data.push_back(pnt);
			
			pnt.y = lin_pnt.y;
			ly_data.push_back(pnt);
			
			pnt.y = lin_pnt.z;
			lz_data.push_back(pnt);
			
			if(cntr == 0)
			{
				start_position = position;
			}
		}
		
		end_position = position;
	}
	else if(d_init_parameters.second_axis_id == Types::AXIS_Y)
	{
		data_size = d_second_axis.Data_Size();
		
		for(cntr = 0;cntr < data_size;++cntr)
		{
			d_second_axis.Data(cntr,&position,&ang_pnt,&lin_pnt);
			
			pnt.x = position;
			
			pnt.y = ang_pnt.x;
			rx_data.push_back(pnt);
			
			pnt.y = ang_pnt.y;
			ry_data.push_back(pnt);
			
			pnt.y = ang_pnt.z;
			rz_data.push_back(pnt);
			
			pnt.y = lin_pnt.x;
			lx_data.push_back(pnt);
			
			pnt.y = lin_pnt.y;
			ly_data.push_back(pnt);
			
			pnt.y = lin_pnt.z;
			lz_data.push_back(pnt);
			
			if(cntr == 0)
			{
				start_position = position;
			}
		}
		
		end_position = position;
	}
	else if(d_init_parameters.third_axis_id == Types::AXIS_Y)
	{
		data_size = d_third_axis.Data_Size();
		
		for(cntr = 0;cntr < data_size;++cntr)
		{
			d_third_axis.Data(cntr,&position,&ang_pnt,&lin_pnt);
			
			pnt.x = position;
			
			pnt.y = ang_pnt.x;
			rx_data.push_back(pnt);
			
			pnt.y = ang_pnt.y;
			ry_data.push_back(pnt);
			
			pnt.y = ang_pnt.z;
			rz_data.push_back(pnt);
			
			pnt.y = lin_pnt.x;
			lx_data.push_back(pnt);
			
			pnt.y = lin_pnt.y;
			ly_data.push_back(pnt);
			
			pnt.y = lin_pnt.z;
			lz_data.push_back(pnt);
			
			if(cntr == 0)
			{
				start_position = position;
			}
		}
		
		end_position = position;
	}
	else
	{
		return lines;
	}
	
	line = QStringLiteral("Y Axis Data");
	lines.push_back(line);
	
	line = QStringLiteral("------------------------------------------------------------------------------");
	lines.push_back(line);
	
	//                    0         1         2         3         4         5         6         7         8         9
	//                    0123456789012345678901234567890123456789012345678901234567890123456789012345678901234567890
	//                    |          |          |          |          |          |          |          |
	line = QStringLiteral("     Pos         Lx         Ly         Lz         Rx         Ry         Rz");
	lines.push_back(line);
	
	line = QStringLiteral("------------------------------------------------------------------------------");
	lines.push_back(line);
	
	// sort data where end closest to zero is shown first
	if(fabs(end_position) > fabs(start_position))
	{
		position = start_position;
		while(!(position > end_position))
		{
			Extrapolation::Get_Value(rx_data,position,&rx);
			Extrapolation::Get_Value(ry_data,position,&ry);
			Extrapolation::Get_Value(rz_data,position,&rz);
			Extrapolation::Get_Value(lx_data,position,&lx);
			Extrapolation::Get_Value(ly_data,position,&ly);
			Extrapolation::Get_Value(lz_data,position,&lz);
			
			line = QString("%1 %2 %3 %4 %5 %6 %7")
						.arg(position,10,'f',1)
						.arg(lx,10,'f',4)
						.arg(ly,10,'f',4)
						.arg(lz,10,'f',4)
						.arg(rx,10,'f',4)
						.arg(ry,10,'f',4)
						.arg(rz,10,'f',4);
			
			lines.push_back(line);
			
			position += increment;
		};
	}
	else
	{
		position = end_position;
		while(!(position < start_position))
		{
			Extrapolation::Get_Value(rx_data,position,&rx);
			Extrapolation::Get_Value(ry_data,position,&ry);
			Extrapolation::Get_Value(rz_data,position,&rz);
			Extrapolation::Get_Value(lx_data,position,&lx);
			Extrapolation::Get_Value(ly_data,position,&ly);
			Extrapolation::Get_Value(lz_data,position,&lz);
			
			line = QString("%1 %2 %3 %4 %5 %6 %7")
						.arg(position,10,'f',1)
						.arg(lx,10,'f',4)
						.arg(ly,10,'f',4)
						.arg(lz,10,'f',4)
						.arg(rx,10,'f',4)
						.arg(ry,10,'f',4)
						.arg(rz,10,'f',4);
			
			lines.push_back(line);
			
			position -= increment;
		};
	}
	
	return lines;
}

QStringList TMachineComp::Map_Data_Z(
	const double						&increment)
{
	QStringList							lines;
	QString								line;
	std::vector<TVector2>				rx_data;
	std::vector<TVector2>				ry_data;
	std::vector<TVector2>				rz_data;
	std::vector<TVector2>				lx_data;
	std::vector<TVector2>				ly_data;
	std::vector<TVector2>				lz_data;
	TVector2							pnt;
	TVector3							ang_pnt;
	TVector3							lin_pnt;
	double								position;
	double								rx,ry,rz,lx,ly,lz;
	double								start_position(0);
	double								end_position(0);
	int									data_size;
	int									cntr;
	
	if(d_init_mode)
	{
		return lines;
	}
	
	if(increment < 1.0)
	{
		return lines;
	}
	
	// load up the 2D arrays
	if(d_init_parameters.first_axis_id == Types::AXIS_Z)
	{
		data_size = d_first_axis.Data_Size();
		
		for(cntr = 0;cntr < data_size;++cntr)
		{
			d_first_axis.Data(cntr,&position,&ang_pnt,&lin_pnt);
			
			pnt.x = position;
			
			pnt.y = ang_pnt.x;
			rx_data.push_back(pnt);
			
			pnt.y = ang_pnt.y;
			ry_data.push_back(pnt);
			
			pnt.y = ang_pnt.z;
			rz_data.push_back(pnt);
			
			pnt.y = lin_pnt.x;
			lx_data.push_back(pnt);
			
			pnt.y = lin_pnt.y;
			ly_data.push_back(pnt);
			
			pnt.y = lin_pnt.z;
			lz_data.push_back(pnt);
			
			if(cntr == 0)
			{
				start_position = position;
			}
		}
		
		end_position = position;
	}
	else if(d_init_parameters.second_axis_id == Types::AXIS_Z)
	{
		data_size = d_second_axis.Data_Size();
		
		for(cntr = 0;cntr < data_size;++cntr)
		{
			d_second_axis.Data(cntr,&position,&ang_pnt,&lin_pnt);
			
			pnt.x = position;
			
			pnt.y = ang_pnt.x;
			rx_data.push_back(pnt);
			
			pnt.y = ang_pnt.y;
			ry_data.push_back(pnt);
			
			pnt.y = ang_pnt.z;
			rz_data.push_back(pnt);
			
			pnt.y = lin_pnt.x;
			lx_data.push_back(pnt);
			
			pnt.y = lin_pnt.y;
			ly_data.push_back(pnt);
			
			pnt.y = lin_pnt.z;
			lz_data.push_back(pnt);
			
			if(cntr == 0)
			{
				start_position = position;
			}
		}
		
		end_position = position;
	}
	else if(d_init_parameters.third_axis_id == Types::AXIS_Z)
	{
		data_size = d_third_axis.Data_Size();
		
		for(cntr = 0;cntr < data_size;++cntr)
		{
			d_third_axis.Data(cntr,&position,&ang_pnt,&lin_pnt);
			
			pnt.x = position;
			
			pnt.y = ang_pnt.x;
			rx_data.push_back(pnt);
			
			pnt.y = ang_pnt.y;
			ry_data.push_back(pnt);
			
			pnt.y = ang_pnt.z;
			rz_data.push_back(pnt);
			
			pnt.y = lin_pnt.x;
			lx_data.push_back(pnt);
			
			pnt.y = lin_pnt.y;
			ly_data.push_back(pnt);
			
			pnt.y = lin_pnt.z;
			lz_data.push_back(pnt);
			
			if(cntr == 0)
			{
				start_position = position;
			}
		}
		
		end_position = position;
	}
	else
	{
		return lines;
	}
	
	line = QStringLiteral("Z Axis Data");
	lines.push_back(line);
	
	line = QStringLiteral("------------------------------------------------------------------------------");
	lines.push_back(line);
	
	//                    0         1         2         3         4         5         6         7         8         9
	//                    0123456789012345678901234567890123456789012345678901234567890123456789012345678901234567890
	//                    |          |          |          |          |          |          |          |
	line = QStringLiteral("     Pos         Lx         Ly         Lz         Rx         Ry         Rz");
	lines.push_back(line);
	
	line = QStringLiteral("------------------------------------------------------------------------------");
	lines.push_back(line);
	
	// sort data where end closest to zero is shown first
	if(fabs(end_position) > fabs(start_position))
	{
		position = start_position;
		while(!(position > end_position))
		{
			Extrapolation::Get_Value(rx_data,position,&rx);
			Extrapolation::Get_Value(ry_data,position,&ry);
			Extrapolation::Get_Value(rz_data,position,&rz);
			Extrapolation::Get_Value(lx_data,position,&lx);
			Extrapolation::Get_Value(ly_data,position,&ly);
			Extrapolation::Get_Value(lz_data,position,&lz);
			
			line = QString("%1 %2 %3 %4 %5 %6 %7")
						.arg(position,10,'f',1)
						.arg(lx,10,'f',4)
						.arg(ly,10,'f',4)
						.arg(lz,10,'f',4)
						.arg(rx,10,'f',4)
						.arg(ry,10,'f',4)
						.arg(rz,10,'f',4);
			
			lines.push_back(line);
			
			position += increment;
		};
	}
	else
	{
		position = end_position;
		while(!(position < start_position))
		{
			Extrapolation::Get_Value(rx_data,position,&rx);
			Extrapolation::Get_Value(ry_data,position,&ry);
			Extrapolation::Get_Value(rz_data,position,&rz);
			Extrapolation::Get_Value(lx_data,position,&lx);
			Extrapolation::Get_Value(ly_data,position,&ly);
			Extrapolation::Get_Value(lz_data,position,&lz);
			
			line = QString("%1 %2 %3 %4 %5 %6 %7")
						.arg(position,10,'f',1)
						.arg(lx,10,'f',4)
						.arg(ly,10,'f',4)
						.arg(lz,10,'f',4)
						.arg(rx,10,'f',4)
						.arg(ry,10,'f',4)
						.arg(rz,10,'f',4);
			
			lines.push_back(line);
			
			position -= increment;
		};
	}
	
	return lines;
}

QStringList TMachineComp::Map_Data_D(
	const double						&increment)
{
	QStringList							lines;
	QString								line;
	double								vertical_position;
	double								horizontal_position;
	double								horizontal_start,horizontal_end;
	double								horizontal_increment;
	double								vertical_start,vertical_end;
	double								dval;
	int									cntr;
	const int							HORIZONTAL_SAMPLE_SIZE(6);
	
	if(increment < 1.0)
	{
		return lines;
	}

	if(!d_deflection_axis.Is_Valid())
	{
		d_last_error = QStringLiteral("D axis deflection expression invalid.");
		return lines;
	}
	
	// output format shows 2D grid of corrections since value is not constant
	
	horizontal_start = d_deflection_axis.Axis_Horizontal_Start();
	horizontal_end = d_deflection_axis.Axis_Horizontal_End();
	horizontal_increment = (horizontal_end - horizontal_start) / static_cast<double>(HORIZONTAL_SAMPLE_SIZE - 1);

	if(d_init_parameters.kinematic == Types::KINEMATIC_XZY)
	{
		line = QStringLiteral("D Axis Data - Rotation of tower around X axis");
		lines.push_back(line);
		
		line = QStringLiteral("------------------------------------------------------------------------------");
		lines.push_back(line);
				
		horizontal_position = horizontal_start;
		line = QStringLiteral("      Z\\Y  ");
	}
	else if(d_init_parameters.kinematic == Types::KINEMATIC_YZX)
	{
		line = QStringLiteral("D Axis Data - Rotation of tower around Y axis");
		lines.push_back(line);
		
		line = QStringLiteral("------------------------------------------------------------------------------");
		lines.push_back(line);
		
		horizontal_position = horizontal_start;
		line = QStringLiteral("      Z\\X  ");
	}

	for(cntr = 0;cntr < HORIZONTAL_SAMPLE_SIZE;++cntr)
	{
		line += QString(" %1").arg(horizontal_position,10,'f',1);
		horizontal_position += horizontal_increment;
	}
	
	lines.push_back(line);

	
	//                    0         1         2         3         4         5         6         7         8         9
	//                    0123456789012345678901234567890123456789012345678901234567890123456789012345678901234567890
	//                    |          |          |          |          |          |          |          |
	
	line = QStringLiteral("------------------------------------------------------------------------------");
	lines.push_back(line);
		
	vertical_start = d_deflection_axis.Axis_Vertical_Start();
	vertical_end = d_deflection_axis.Axis_Vertical_End();
	
	// sort data where end closest to zero is shown first
	if(fabs(vertical_end) > fabs(vertical_start))
	{
		vertical_position = vertical_start;
	
		while(!(vertical_position > vertical_end))
		{
			line = QString(" %1").arg(vertical_position,10,'f',1);

			horizontal_position = horizontal_start;
			while(!(horizontal_position > horizontal_end))
			{
				try
				{
					dval = d_deflection_axis.Evaluate(vertical_position,horizontal_position);
				}
				catch(TMachineAxisException exception)
				{
					// may do more with this exception in the future.
					// biggest issue is with probe offsets that exceed the volume of the machine.
					dval = 0.0;
				}

				line += QString(" %1").arg(dval,10,'f',4);

				horizontal_position += horizontal_increment;
			}
			
			lines.push_back(line);

			vertical_position += increment;
		}
	}
	else
	{
		vertical_position = vertical_end;
	
		while(!(vertical_position < vertical_start))
		{
			line = QString(" %1").arg(vertical_position,10,'f',1);

			horizontal_position = horizontal_start;
			while(!(horizontal_position > horizontal_end))
			{
				try
				{
					dval = d_deflection_axis.Evaluate(vertical_position,horizontal_position);
				}
				catch(TMachineAxisException exception)
				{
					// may do more with this exception in the future.
					// biggest issue is with probe offsets that exceed the volume of the machine.
					dval = 0.0;
				}

				line += QString(" %1").arg(dval,10,'f',4);

				horizontal_position += horizontal_increment;
			}
			
			lines.push_back(line);

			vertical_position -= increment;
		}
	}

	return lines;
}

QStringList TMachineComp::Map_Data_Quad(void)
{
	QStringList							lines;
	QString								line;
	double								dval;

	line = QStringLiteral("Squareness");
	lines.push_back(line);
	
	line = QStringLiteral("------------------------------------------------------------------------------");
	lines.push_back(line);
	
	dval = d_first_axis.Quad();
	
	switch(d_init_parameters.first_axis_id)
	{
		case Types::AXIS_X:
			line = QString("XY: %1").arg(dval,0,'f',4);
			lines.push_back(line);
			break;
		case Types::AXIS_Y:
			line = QString("YZ: %1").arg(dval,0,'f',4);
			lines.push_back(line);
			break;
		case Types::AXIS_Z:
			line = QString("ZX: %1").arg(dval,0,'f',4);
			lines.push_back(line);
			break;
	}
	
	dval = d_second_axis.Quad();

	switch(d_init_parameters.second_axis_id)
	{
		case Types::AXIS_X:
			line = QString("XY: %1").arg(dval,0,'f',4);
			lines.push_back(line);
			break;
		case Types::AXIS_Y:
			line = QString("YZ: %1").arg(dval,0,'f',4);
			lines.push_back(line);
			break;
		case Types::AXIS_Z:
			line = QString("ZX: %1").arg(dval,0,'f',4);
			lines.push_back(line);
			break;
	}
	
	dval = d_third_axis.Quad();
	
	switch(d_init_parameters.third_axis_id)
	{
		case Types::AXIS_X:
			line = QString("XY: %1").arg(dval,0,'f',4);
			lines.push_back(line);
			break;
		case Types::AXIS_Y:
			line = QString("YZ: %1").arg(dval,0,'f',4);
			lines.push_back(line);
			break;
		case Types::AXIS_Z:
			line = QString("ZX: %1").arg(dval,0,'f',4);
			lines.push_back(line);
			break;
	}

	return lines;
}

TVector3 TMachineComp::Map_Data_Min(void) const
{
	TVector3							pnt;
	
	switch(d_init_parameters.first_axis_id)
	{
		case Types::AXIS_X:
			pnt.x = d_init_parameters.first_axis_start;
			break;
		case Types::AXIS_Y:
			pnt.y = d_init_parameters.first_axis_start;
			break;
		case Types::AXIS_Z:
			pnt.z = d_init_parameters.first_axis_start;
			break;
	}
	
	switch(d_init_parameters.second_axis_id)
	{
		case Types::AXIS_X:
			pnt.x = d_init_parameters.second_axis_start;
			break;
		case Types::AXIS_Y:
			pnt.y = d_init_parameters.second_axis_start;
			break;
		case Types::AXIS_Z:
			pnt.z = d_init_parameters.second_axis_start;
			break;
	}
	
	switch(d_init_parameters.third_axis_id)
	{
		case Types::AXIS_X:
			pnt.x = d_init_parameters.third_axis_start;
			break;
		case Types::AXIS_Y:
			pnt.y = d_init_parameters.third_axis_start;
			break;
		case Types::AXIS_Z:
			pnt.z = d_init_parameters.third_axis_start;
			break;
	}
	
	return pnt;
}

TVector3 TMachineComp::Map_Data_Max(void) const
{
	TVector3							pnt;
	
	switch(d_init_parameters.first_axis_id)
	{
		case Types::AXIS_X:
			pnt.x = d_init_parameters.first_axis_end;
			break;
		case Types::AXIS_Y:
			pnt.y = d_init_parameters.first_axis_end;
			break;
		case Types::AXIS_Z:
			pnt.z = d_init_parameters.first_axis_end;
			break;
	}
	
	switch(d_init_parameters.second_axis_id)
	{
		case Types::AXIS_X:
			pnt.x = d_init_parameters.second_axis_end;
			break;
		case Types::AXIS_Y:
			pnt.y = d_init_parameters.second_axis_end;
			break;
		case Types::AXIS_Z:
			pnt.z = d_init_parameters.second_axis_end;
			break;
	}
	
	switch(d_init_parameters.third_axis_id)
	{
		case Types::AXIS_X:
			pnt.x = d_init_parameters.third_axis_end;
			break;
		case Types::AXIS_Y:
			pnt.y = d_init_parameters.third_axis_end;
			break;
		case Types::AXIS_Z:
			pnt.z = d_init_parameters.third_axis_end;
			break;
	}
	
	return pnt;
}

TVector3 TMachineComp::Evaluate(
	const TVector3						&position,
	const TVector3						&probe_offset,
	bool * const						valid) const
{
	TVector3							pnt;
	TVector3							eval_axis_min,eval_axis_max;
	static const double					EVALUATION_BUFFER(5.0);		// amount to allow outside of machine volume before triggering error
	
	if(valid)
	{
		*valid = false;
	}
	
	if(!d_valid_data)
	{
		d_last_error = QStringLiteral("Map data is invalid");
		return TVector3();
	}
			
	switch(d_init_parameters.kinematic)
	{
		case Types::KINEMATIC_XYZ:
			eval_axis_min.Set(d_init_parameters.first_axis_start - EVALUATION_BUFFER,
							  d_init_parameters.second_axis_start - EVALUATION_BUFFER,
							  d_init_parameters.third_axis_start - EVALUATION_BUFFER);
			
			eval_axis_max.Set(d_init_parameters.first_axis_end + EVALUATION_BUFFER,
							  d_init_parameters.second_axis_end + EVALUATION_BUFFER,
							  d_init_parameters.third_axis_end + EVALUATION_BUFFER);
			break;
			
		case Types::KINEMATIC_YXZ:
			eval_axis_min.Set(d_init_parameters.second_axis_start - EVALUATION_BUFFER,
							  d_init_parameters.first_axis_start - EVALUATION_BUFFER,
							  d_init_parameters.third_axis_start - EVALUATION_BUFFER);
			
			eval_axis_max.Set(d_init_parameters.second_axis_end + EVALUATION_BUFFER,
							  d_init_parameters.first_axis_end + EVALUATION_BUFFER,
							  d_init_parameters.third_axis_end + EVALUATION_BUFFER);
			break;
			
		case Types::KINEMATIC_XZY:
			eval_axis_min.Set(d_init_parameters.first_axis_start - EVALUATION_BUFFER,
							  d_init_parameters.third_axis_start - EVALUATION_BUFFER,
							  d_init_parameters.second_axis_start - EVALUATION_BUFFER);
			
			eval_axis_max.Set(d_init_parameters.first_axis_end + EVALUATION_BUFFER,
							  d_init_parameters.third_axis_end + EVALUATION_BUFFER,
							  d_init_parameters.second_axis_end + EVALUATION_BUFFER);
			break;
			
		case Types::KINEMATIC_YZX:
			eval_axis_min.Set(d_init_parameters.third_axis_start - EVALUATION_BUFFER,
							  d_init_parameters.first_axis_start - EVALUATION_BUFFER,
							  d_init_parameters.second_axis_start - EVALUATION_BUFFER);
			
			eval_axis_max.Set(d_init_parameters.third_axis_end + EVALUATION_BUFFER,
							  d_init_parameters.first_axis_end + EVALUATION_BUFFER,
							  d_init_parameters.second_axis_end + EVALUATION_BUFFER);
			break;
			
		default:
			d_last_error = QStringLiteral("Map data kinematic is invalid");
			return position;
	}
	
	// convert to positions along first, second, third axis.
	switch(d_init_parameters.kinematic)
	{
		case Types::KINEMATIC_XYZ:
			pnt = this->Evaluate_XYZ(position,probe_offset);
			break;
			
		case Types::KINEMATIC_YXZ:
			pnt = this->Evaluate_YXZ(position,probe_offset);
			break;
			
		case Types::KINEMATIC_XZY:
			pnt = this->Evaluate_XZY(position,probe_offset);
			break;
			
		case Types::KINEMATIC_YZX:
			pnt = this->Evaluate_YZX(position,probe_offset);
			break;
			
		default:
			d_last_error = QStringLiteral("Map data kinematic is invalid");
			return pnt;
	}
	
	if((position.x < eval_axis_min.x || position.x > eval_axis_max.x) ||
	   (position.y < eval_axis_min.y || position.y > eval_axis_max.y) ||
	   (position.z < eval_axis_min.z || position.z > eval_axis_max.z))
	{
		d_last_error = QString("Position %1 %2 %3 outside of machine volume.  Using extrapolated machine map.")
				.arg(position.x,0,'f',4)
				.arg(position.y,0,'f',4)
				.arg(position.z,0,'f',4);
	}
	else
	{
		if(valid)
		{
			*valid = true;
		}
	}

	return pnt;
}

TVector3 TMachineComp::Get_Angles_RA_RB_RC(
	const TVector3						&axis,
	const TVector3						&probe_axis,
	const TVector3						&position) const
{
	TVector3							angular;
	TVector3							linear;
	TVector3							abc;
	double								deflection;
	static const double					AXIS_ANGLE_LIMIT(0.996194698091746); // 5 degrees from axis
	static const double					PROBE_ANGLE_LIMIT(0.707106781186548); // 45 degrees from axis
	
	if(fabs(TVector3(1,0,0) ^ axis) > AXIS_ANGLE_LIMIT) // x axis
	{
		switch(d_init_parameters.kinematic)
		{
			case Types::KINEMATIC_XYZ:
				d_first_axis.Data(position.x,&angular,&linear);
				break;
				
			case Types::KINEMATIC_YXZ:
				d_second_axis.Data(position.x,&angular,&linear);
				break;
				
			case Types::KINEMATIC_XZY:
				d_first_axis.Data(position.x,&angular,&linear);
				break;
				
			case Types::KINEMATIC_YZX:
				d_third_axis.Data(position.x,&angular,&linear);
				
				try
				{
					deflection = d_deflection_axis.Evaluate(position.z,position.x);
				}
				catch(TMachineAxisException exception)
				{
					// may do more with this exception in the future.
					// biggest issue is with probe offsets that exceed the volume of the machine.
					deflection = 0.0;
				}
					
				angular.y += deflection;
				break;
		}
		
		// put values into proper order
		abc.Set(angular.z,angular.y,angular.x);
	}
	else if(fabs(TVector3(0,1,0) ^ axis) > AXIS_ANGLE_LIMIT) // y axis
	{
		switch(d_init_parameters.kinematic)
		{
			case Types::KINEMATIC_XYZ:
				d_second_axis.Data(position.y,&angular,&linear);
				break;
				
			case Types::KINEMATIC_YXZ:
				d_first_axis.Data(position.y,&angular,&linear);
				break;
				
			case Types::KINEMATIC_XZY:
				d_third_axis.Data(position.y,&angular,&linear);

				try
				{
					deflection = d_deflection_axis.Evaluate(position.z,position.y);
				}
				catch(TMachineAxisException exception)
				{
					// may do more with this exception in the future.
					// biggest issue is with probe offsets that exceed the volume of the machine.
					deflection = 0.0;
				}
					
				angular.x += deflection;
				break;
				
			case Types::KINEMATIC_YZX:
				break;
		}
		
		// put values into proper order
		abc.Set(angular.z,angular.x,angular.y);
	}
	else if(fabs(TVector3(0,0,1) ^ axis) > AXIS_ANGLE_LIMIT) // z axis
	{
		switch(d_init_parameters.kinematic)
		{
			case Types::KINEMATIC_XYZ:
				d_third_axis.Data(position.z,&angular,&linear);
				break;
				
			case Types::KINEMATIC_YXZ:
				d_third_axis.Data(position.z,&angular,&linear);
				break;
				
			case Types::KINEMATIC_XZY:
				d_second_axis.Data(position.z,&angular,&linear);
				
				try
				{
					deflection = d_deflection_axis.Evaluate(position.z,position.y);
				}
				catch(TMachineAxisException exception)
				{
					// may do more with this exception in the future.
					// biggest issue is with probe offsets that exceed the volume of the machine.
					deflection = 0.0;
				}
				
				angular.x += deflection;

				break;
				
			case Types::KINEMATIC_YZX:
				d_second_axis.Data(position.z,&angular,&linear);
				
				try
				{
					deflection = d_deflection_axis.Evaluate(position.z,position.x);
				}
				catch(TMachineAxisException exception)
				{
					// may do more with this exception in the future.
					// biggest issue is with probe offsets that exceed the volume of the machine.
					deflection = 0.0;
				}
				
				angular.y += deflection;
				
				break;
		}
		
		if(fabs(TVector3(1,0,0) ^ probe_axis) > PROBE_ANGLE_LIMIT) // x axis
		{
			// put values into proper order for probe point in X direction
			abc.Set(angular.x,angular.y,angular.z);
		}
		else // assume probe in Y direction
		{
			// put values into proper order for probe point in Y direction
			abc.Set(angular.y,angular.x,angular.z);
		}
	}

	return abc;
}

TVector3 TMachineComp::Calculate_Scale_XX_YY_Error(
	const TVector3						&axis,
	const TVector3						&probe_axis,
	const TVector3						&nominal,
	const TVector3						&actual) const
{
	TVector3							scale_xx_yy;
	TVector3							axis_xx;
	TVector3							axis_yy;
	TVector3							deviation;
	static const double					AXIS_ANGLE_LIMIT(0.996194698091746); // 5 degrees from axis
	static const double					PROBE_ANGLE_LIMIT(0.707106781186548); // 45 degrees from axis
	
	deviation = actual - nominal;
	
	if(fabs(TVector3(1,0,0) ^ axis) > AXIS_ANGLE_LIMIT) // x axis
	{
		axis_xx = TVector3(0,1,0);
		axis_yy = TVector3(0,0,1);
		
		scale_xx_yy.x = deviation ^ axis;
		scale_xx_yy.y = deviation ^ axis_xx;
		scale_xx_yy.z = deviation ^ axis_yy;
	}
	else if(fabs(TVector3(0,1,0) ^ axis) > AXIS_ANGLE_LIMIT) // y axis
	{
		axis_xx = TVector3(1,0,0);
		axis_yy = TVector3(0,0,1);
		
		scale_xx_yy.x = deviation ^ axis;
		scale_xx_yy.y = deviation ^ axis_xx;
		scale_xx_yy.z = deviation ^ axis_yy;
	}
	else if(fabs(TVector3(0,0,1) ^ axis) > AXIS_ANGLE_LIMIT) // z axis
	{
		if(fabs(TVector3(1,0,0) ^ probe_axis) > PROBE_ANGLE_LIMIT) // x axis
		{
			axis_xx = TVector3(0,1,0);
			axis_yy = TVector3(1,0,0);
		}
		else // assume probe in Y direction
		{
			axis_xx = TVector3(1,0,0);
			axis_yy = TVector3(0,1,0);
		}
		
		scale_xx_yy.x = deviation ^ axis;
		scale_xx_yy.y = deviation ^ axis_xx;
		scale_xx_yy.z = deviation ^ axis_yy;
	}
	else
	{
		scale_xx_yy.x = deviation ^ axis;
		scale_xx_yy.y = 0.0;
		scale_xx_yy.z = 0.0;
	}

	return scale_xx_yy;
}


bool TMachineComp::Initialize_Begin(
	const TInitParameters				&init_parameters)
{
	
	if(d_init_mode) // must be paired with init_begin() and init_end()
	{
		d_last_error = QStringLiteral("Incorrect mode to start initialization");
		return false;
	}
	
	// start position must be less than end position for all axis
	if(!(init_parameters.first_axis_start < init_parameters.first_axis_end))
	{
		d_last_error = QStringLiteral("first axis start and end positions not valid");
		return false;
	}
	
	if(!(init_parameters.second_axis_start < init_parameters.second_axis_end))
	{
		d_last_error = QStringLiteral("second axis start and end positions not valid");
		return false;
	}
	
	if(!(init_parameters.third_axis_start < init_parameters.third_axis_end))
	{
		d_last_error = QStringLiteral("third axis start and end positions not valid");
		return false;
	}
	
	// make sure math origin point is at or within range of map
	if(init_parameters.first_axis_math_origin < init_parameters.first_axis_start ||
	   init_parameters.first_axis_math_origin > init_parameters.first_axis_end)
	{
		d_last_error = QStringLiteral("first axis math origin not inside axis");
		return false;
	}
	
	if(init_parameters.second_axis_math_origin < init_parameters.second_axis_start ||
	   init_parameters.second_axis_math_origin > init_parameters.second_axis_end)
	{
		d_last_error = QStringLiteral("second axis math origin not inside axis");
		return false;
	}
	
	if(init_parameters.third_axis_math_origin < init_parameters.third_axis_start ||
	   init_parameters.third_axis_math_origin > init_parameters.third_axis_end)
	{
		d_last_error = QStringLiteral("third axis math origin not inside axis");
		return false;
	}
		
	// make sure machine origin point is at or within range of map
	if(init_parameters.first_axis_mechanical_origin < init_parameters.first_axis_start ||
	   init_parameters.first_axis_mechanical_origin > init_parameters.first_axis_end)
	{
		d_last_error = QStringLiteral("first axis machine origin not inside axis");
		return false;
	}
	
	if(init_parameters.second_axis_mechanical_origin < init_parameters.second_axis_start ||
	   init_parameters.second_axis_mechanical_origin > init_parameters.second_axis_end)
	{
		d_last_error = QStringLiteral("second axis machine origin not inside axis");
		return false;
	}
	
	if(init_parameters.third_axis_mechanical_origin < init_parameters.third_axis_start ||
	   init_parameters.third_axis_mechanical_origin > init_parameters.third_axis_end)
	{
		d_last_error = QStringLiteral("third axis machine origin not inside axis");
		return false;
	}

	// duplicate ids
	if(init_parameters.first_axis_id == init_parameters.second_axis_id)
	{
		d_last_error = QStringLiteral("first and second axis duplicated");
		return false;
	}
	
	if(init_parameters.first_axis_id == init_parameters.third_axis_id)
	{
		d_last_error = QStringLiteral("first and third axis duplicated");
		return false;
	}
	
	if(init_parameters.second_axis_id == init_parameters.third_axis_id)
	{
		d_last_error = QStringLiteral("second and third axis duplicated");
		return false;
	}

	d_init_parameters = init_parameters;
	
	d_first_axis.Set_Axis_Range(d_init_parameters.first_axis_start,d_init_parameters.first_axis_end);
	d_second_axis.Set_Axis_Range(d_init_parameters.second_axis_start,d_init_parameters.second_axis_end);
	d_third_axis.Set_Axis_Range(d_init_parameters.third_axis_start,d_init_parameters.third_axis_end);
		
	// deflection axis
	if(d_init_parameters.kinematic == Types::KINEMATIC_XZY)
	{
		// Z is up, Y is out
		if(d_init_parameters.first_axis_id == Types::AXIS_Y)
		{
			d_deflection_axis.Set_Horizontal_Axis(d_init_parameters.first_axis_start,d_init_parameters.first_axis_end,d_init_parameters.first_axis_mechanical_origin);
		}
		else if(d_init_parameters.second_axis_id == Types::AXIS_Y)
		{
			d_deflection_axis.Set_Horizontal_Axis(d_init_parameters.second_axis_start,d_init_parameters.second_axis_end,d_init_parameters.second_axis_mechanical_origin);
		}
		else if(d_init_parameters.third_axis_id == Types::AXIS_Y)
		{
			d_deflection_axis.Set_Horizontal_Axis(d_init_parameters.third_axis_start,d_init_parameters.third_axis_end,d_init_parameters.third_axis_mechanical_origin);
		}

		if(d_init_parameters.first_axis_id == Types::AXIS_Z)
		{
			d_deflection_axis.Set_Vertical_Axis(d_init_parameters.first_axis_start,d_init_parameters.first_axis_end,d_init_parameters.first_axis_mechanical_origin);
		}
		else if(d_init_parameters.second_axis_id == Types::AXIS_Z)
		{
			d_deflection_axis.Set_Vertical_Axis(d_init_parameters.second_axis_start,d_init_parameters.second_axis_end,d_init_parameters.second_axis_mechanical_origin);
		}
		else if(d_init_parameters.third_axis_id == Types::AXIS_Z)
		{
			d_deflection_axis.Set_Vertical_Axis(d_init_parameters.third_axis_start,d_init_parameters.third_axis_end,d_init_parameters.third_axis_mechanical_origin);
		}
	}
	else if(d_init_parameters.kinematic == Types::KINEMATIC_YZX)
	{
		// Z is up, X is out
		if(d_init_parameters.first_axis_id == Types::AXIS_X)
		{
			d_deflection_axis.Set_Horizontal_Axis(d_init_parameters.first_axis_start,d_init_parameters.first_axis_end,d_init_parameters.first_axis_mechanical_origin);
		}
		else if(d_init_parameters.second_axis_id == Types::AXIS_X)
		{
			d_deflection_axis.Set_Horizontal_Axis(d_init_parameters.second_axis_start,d_init_parameters.second_axis_end,d_init_parameters.second_axis_mechanical_origin);
		}
		else if(d_init_parameters.third_axis_id == Types::AXIS_X)
		{
			d_deflection_axis.Set_Horizontal_Axis(d_init_parameters.third_axis_start,d_init_parameters.third_axis_end,d_init_parameters.third_axis_mechanical_origin);
		}
		
		if(d_init_parameters.first_axis_id == Types::AXIS_Z)
		{
			d_deflection_axis.Set_Vertical_Axis(d_init_parameters.first_axis_start,d_init_parameters.first_axis_end,d_init_parameters.first_axis_mechanical_origin);
		}
		else if(d_init_parameters.second_axis_id == Types::AXIS_Z)
		{
			d_deflection_axis.Set_Vertical_Axis(d_init_parameters.second_axis_start,d_init_parameters.second_axis_end,d_init_parameters.second_axis_mechanical_origin);
		}
		else if(d_init_parameters.third_axis_id == Types::AXIS_Z)
		{
			d_deflection_axis.Set_Vertical_Axis(d_init_parameters.third_axis_start,d_init_parameters.third_axis_end,d_init_parameters.third_axis_mechanical_origin);
		}
	}
	else
	{
		d_deflection_axis.Set_Horizontal_Axis(0.0,0.0,0.0);
		d_deflection_axis.Set_Vertical_Axis(0.0,0.0,0.0);
	}

	d_first_axis.Set_Axis_Origin(d_init_parameters.first_axis_math_origin,d_init_parameters.first_axis_mechanical_origin);
	d_second_axis.Set_Axis_Origin(d_init_parameters.second_axis_math_origin,d_init_parameters.second_axis_mechanical_origin);
	d_third_axis.Set_Axis_Origin(d_init_parameters.third_axis_math_origin,d_init_parameters.third_axis_mechanical_origin);
	
	d_first_axis.Initialize(d_init_parameters.first_axis_id);
	d_second_axis.Initialize(d_init_parameters.second_axis_id);
	d_third_axis.Initialize(d_init_parameters.third_axis_id);
	d_deflection_axis.Initialize(0.0,1);	// clear any previous expression
	
	d_init_mode = true;
	d_valid_data = false;

	return true;
}


bool TMachineComp::Set_Rx(
	const int							axis_id,
	const QString						&expression,
	const int							sign)
{
	if(!d_init_mode)
	{
		d_last_error = QStringLiteral("Incorrect mode to set Rx parameter");
		return false;
	}
	
	if(axis_id == d_init_parameters.first_axis_id)
	{
		try
		{
			d_first_axis.Set_Rx(expression,sign);
		}
		catch(TMachineAxisException exception)
		{
			switch(d_init_parameters.first_axis_id)
			{
				case Types::AXIS_X:
					d_last_error = QString("X Axis: %1").arg(exception.Error_Text());
					break;
				case Types::AXIS_Y:
					d_last_error = QString("Y Axis: %1").arg(exception.Error_Text());
					break;
				case Types::AXIS_Z:
					d_last_error = QString("Z Axis: %1").arg(exception.Error_Text());
					break;
			}

			return false;
		}
	}
	else if(axis_id == d_init_parameters.second_axis_id)
	{
		try
		{
			d_second_axis.Set_Rx(expression,sign);
		}
		catch(TMachineAxisException exception)
		{
			switch(d_init_parameters.second_axis_id)
			{
				case Types::AXIS_X:
					d_last_error = QString("X Axis: %1").arg(exception.Error_Text());
					break;
				case Types::AXIS_Y:
					d_last_error = QString("Y Axis: %1").arg(exception.Error_Text());
					break;
				case Types::AXIS_Z:
					d_last_error = QString("Z Axis: %1").arg(exception.Error_Text());
					break;
			}
			
			return false;
		}
	}
	else if(axis_id == d_init_parameters.third_axis_id)
	{
		try
		{
			d_third_axis.Set_Rx(expression,sign);
		}
		catch(TMachineAxisException exception)
		{
			switch(d_init_parameters.third_axis_id)
			{
				case Types::AXIS_X:
					d_last_error = QString("X Axis: %1").arg(exception.Error_Text());
					break;
				case Types::AXIS_Y:
					d_last_error = QString("Y Axis: %1").arg(exception.Error_Text());
					break;
				case Types::AXIS_Z:
					d_last_error = QString("Z Axis: %1").arg(exception.Error_Text());
					break;
			}
			
			return false;
		}
	}
	
	return true;
}

bool TMachineComp::Set_Ry(
	const int							axis_id,
	const QString						&expression,
	const int							sign)
{
	if(!d_init_mode)
	{
		d_last_error = QStringLiteral("Incorrect mode to set Ry parameter");
		return false;
	}
	
	if(axis_id == d_init_parameters.first_axis_id)
	{
		try
		{
			d_first_axis.Set_Ry(expression,sign);
		}
		catch(TMachineAxisException exception)
		{
			switch(d_init_parameters.first_axis_id)
			{
				case Types::AXIS_X:
					d_last_error = QString("X Axis: %1").arg(exception.Error_Text());
					break;
				case Types::AXIS_Y:
					d_last_error = QString("Y Axis: %1").arg(exception.Error_Text());
					break;
				case Types::AXIS_Z:
					d_last_error = QString("Z Axis: %1").arg(exception.Error_Text());
					break;
			}
			
			return false;
		}
	}
	else if(axis_id == d_init_parameters.second_axis_id)
	{
		try
		{
			d_second_axis.Set_Ry(expression,sign);
		}
		catch(TMachineAxisException exception)
		{
			switch(d_init_parameters.second_axis_id)
			{
				case Types::AXIS_X:
					d_last_error = QString("X Axis: %1").arg(exception.Error_Text());
					break;
				case Types::AXIS_Y:
					d_last_error = QString("Y Axis: %1").arg(exception.Error_Text());
					break;
				case Types::AXIS_Z:
					d_last_error = QString("Z Axis: %1").arg(exception.Error_Text());
					break;
			}
			
			return false;
		}
	}
	else if(axis_id == d_init_parameters.third_axis_id)
	{
		try
		{
			d_third_axis.Set_Ry(expression,sign);
		}
		catch(TMachineAxisException exception)
		{
			switch(d_init_parameters.third_axis_id)
			{
				case Types::AXIS_X:
					d_last_error = QString("X Axis: %1").arg(exception.Error_Text());
					break;
				case Types::AXIS_Y:
					d_last_error = QString("Y Axis: %1").arg(exception.Error_Text());
					break;
				case Types::AXIS_Z:
					d_last_error = QString("Z Axis: %1").arg(exception.Error_Text());
					break;
			}
			
			return false;
		}
	}
	
	return true;
}

bool TMachineComp::Set_Rz(
	const int							axis_id,
	const QString						&expression,
	const int							sign)
{
	if(!d_init_mode)
	{
		d_last_error = QStringLiteral("Incorrect mode to set Rz parameter");
		return false;
	}
	
	if(axis_id == d_init_parameters.first_axis_id)
	{
		try
		{
			d_first_axis.Set_Rz(expression,sign);
		}
		catch(TMachineAxisException exception)
		{
			switch(d_init_parameters.first_axis_id)
			{
				case Types::AXIS_X:
					d_last_error = QString("X Axis: %1").arg(exception.Error_Text());
					break;
				case Types::AXIS_Y:
					d_last_error = QString("Y Axis: %1").arg(exception.Error_Text());
					break;
				case Types::AXIS_Z:
					d_last_error = QString("Z Axis: %1").arg(exception.Error_Text());
					break;
			}
			
			return false;
		}
	}
	else if(axis_id == d_init_parameters.second_axis_id)
	{
		try
		{
			d_second_axis.Set_Rz(expression,sign);
		}
		catch(TMachineAxisException exception)
		{
			switch(d_init_parameters.second_axis_id)
			{
				case Types::AXIS_X:
					d_last_error = QString("X Axis: %1").arg(exception.Error_Text());
					break;
				case Types::AXIS_Y:
					d_last_error = QString("Y Axis: %1").arg(exception.Error_Text());
					break;
				case Types::AXIS_Z:
					d_last_error = QString("Z Axis: %1").arg(exception.Error_Text());
					break;
			}
			
			return false;
		}
	}
	else if(axis_id == d_init_parameters.third_axis_id)
	{
		try
		{
			d_third_axis.Set_Rz(expression,sign);
		}
		catch(TMachineAxisException exception)
		{
			switch(d_init_parameters.third_axis_id)
			{
				case Types::AXIS_X:
					d_last_error = QString("X Axis: %1").arg(exception.Error_Text());
					break;
				case Types::AXIS_Y:
					d_last_error = QString("Y Axis: %1").arg(exception.Error_Text());
					break;
				case Types::AXIS_Z:
					d_last_error = QString("Z Axis: %1").arg(exception.Error_Text());
					break;
			}
			
			return false;
		}
	}
	
	return true;
}

bool TMachineComp::Set_Scale(
	const int							axis_id,
	const QString						&expression,
	const int							sign)
{
	QString								axis_name;
	
	if(!d_init_mode)
	{
		d_last_error = QStringLiteral("Incorrect mode to set Scale parameter");
		return false;
	}
	
	switch(axis_id)
	{
		case Types::AXIS_X:
			axis_name = QStringLiteral("X Axis");
			break;
			
		case Types::AXIS_Y:
			axis_name = QStringLiteral("Y Axis");
			break;
			
		case Types::AXIS_Z:
			axis_name = QStringLiteral("Z Axis");
			break;
	}
	
	if(axis_id == d_init_parameters.first_axis_id)
	{
		try
		{
			d_first_axis.Set_Scale(expression,sign);
		}
		catch(TMachineAxisException exception)
		{
			d_last_error = QString("%1: %2").arg(axis_name).arg(exception.Error_Text());
			return false;
		}
	}
	else if(axis_id == d_init_parameters.second_axis_id)
	{
		try
		{
			d_second_axis.Set_Scale(expression,sign);
		}
		catch(TMachineAxisException exception)
		{
			d_last_error = QString("%1: %2").arg(axis_name).arg(exception.Error_Text());
			return false;
		}
	}
	else if(axis_id == d_init_parameters.third_axis_id)
	{
		try
		{
			d_third_axis.Set_Scale(expression,sign);
		}
		catch(TMachineAxisException exception)
		{
			d_last_error = QString("%1: %2").arg(axis_name).arg(exception.Error_Text());
			return false;
		}
	}
	else
	{
		d_last_error = QStringLiteral("Set scale error.  Axis not specified");
		return false;
	}
	
	return true;
}

bool TMachineComp::Set_Deflection(
	const int							axis_id,
	const double						&value,
	const int							sign)
{
	if(axis_id != Types::AXIS_D)
	{
		d_last_error = QStringLiteral("Cannot set deflection for not deflection axis");
		return false;
	}
	
	d_deflection_axis.Initialize(value,sign);

	return true;
}


bool TMachineComp::Set_Quad(
	const int							axis_id,
	const double						&value,
	const int							sign)
{
	if(!d_init_mode)
	{
		d_last_error = QStringLiteral("Incorrect mode to set Squareness parameter");
		return false;
	}
	
	if(axis_id == d_init_parameters.first_axis_id)
	{
		try
		{
			d_first_axis.Set_Quad(value,sign);
		}
		catch(TMachineAxisException exception)
		{
			switch(d_init_parameters.first_axis_id)
			{
				case Types::AXIS_X:
					d_last_error = QString("X Axis: %1").arg(exception.Error_Text());
					break;
					
				case Types::AXIS_Y:
					d_last_error = QString("Y Axis: %1").arg(exception.Error_Text());
					break;
					
				case Types::AXIS_Z:
					d_last_error = QString("Z Axis: %1").arg(exception.Error_Text());
					break;
			}
			
			return false;
		}
	}
	else if(axis_id == d_init_parameters.second_axis_id)
	{
		try
		{
			d_second_axis.Set_Quad(value,sign);
		}
		
		catch(TMachineAxisException exception)
		{
			switch(d_init_parameters.second_axis_id)
			{
				case Types::AXIS_X:
					d_last_error = QString("X Axis: %1").arg(exception.Error_Text());
					break;
					
				case Types::AXIS_Y:
					d_last_error = QString("Y Axis: %1").arg(exception.Error_Text());
					break;
					
				case Types::AXIS_Z:
					d_last_error = QString("Z Axis: %1").arg(exception.Error_Text());
					break;
			}
			
			return false;
		}

	}
	else if(axis_id == d_init_parameters.third_axis_id)
	{
		try
		{
			d_third_axis.Set_Quad(value,sign);
		}
		catch(TMachineAxisException exception)
		{
			switch(d_init_parameters.third_axis_id)
			{
				case Types::AXIS_X:
					d_last_error = QString("X Axis: %1").arg(exception.Error_Text());
					break;
					
				case Types::AXIS_Y:
					d_last_error = QString("Y Axis: %1").arg(exception.Error_Text());
					break;
					
				case Types::AXIS_Z:
					d_last_error = QString("Z Axis: %1").arg(exception.Error_Text());
					break;
			}
			
			return false;
		}

	}
	
	return true;
}

bool TMachineComp::Initialize_End(void)
{
	if(!d_init_mode)
	{
		d_last_error = QStringLiteral("Incorrect mode to end initialization");
		return false;
	}
	
	switch(d_first_axis.Axis_Id())
	{
		case Types::AXIS_X:
			
			d_first_axis.Calculate_Straightness(TMachineStandardComp::Y_STR_FROM_RZ);
			d_first_axis.Calculate_Straightness(TMachineStandardComp::Z_STR_FROM_RY);
			break;
			
		case Types::AXIS_Y:

			d_first_axis.Calculate_Straightness(TMachineStandardComp::X_STR_FROM_RZ);
			d_first_axis.Calculate_Straightness(TMachineStandardComp::Z_STR_FROM_RX);
			break;
			
		case Types::AXIS_Z:
			
			d_first_axis.Calculate_Straightness(TMachineStandardComp::X_STR_FROM_RY);
			d_first_axis.Calculate_Straightness(TMachineStandardComp::Y_STR_FROM_RX);
			break;
	}
	
	switch(d_second_axis.Axis_Id())
	{
		case Types::AXIS_X:
			
			d_second_axis.Calculate_Straightness(TMachineStandardComp::Y_STR_FROM_RZ);
			d_second_axis.Calculate_Straightness(TMachineStandardComp::Z_STR_FROM_RY);
			break;
			
		case Types::AXIS_Y:
			
			d_second_axis.Calculate_Straightness(TMachineStandardComp::X_STR_FROM_RZ);
			d_second_axis.Calculate_Straightness(TMachineStandardComp::Z_STR_FROM_RX);
			break;
			
		case Types::AXIS_Z:
			
			d_second_axis.Calculate_Straightness(TMachineStandardComp::X_STR_FROM_RY);
			d_second_axis.Calculate_Straightness(TMachineStandardComp::Y_STR_FROM_RX);
			break;
	}
	
	switch(d_third_axis.Axis_Id())
	{
		case Types::AXIS_X:
			
			d_third_axis.Calculate_Straightness(TMachineStandardComp::Y_STR_FROM_RZ);
			d_third_axis.Calculate_Straightness(TMachineStandardComp::Z_STR_FROM_RY);
			break;
			
		case Types::AXIS_Y:
			
			d_third_axis.Calculate_Straightness(TMachineStandardComp::X_STR_FROM_RZ);
			d_third_axis.Calculate_Straightness(TMachineStandardComp::Z_STR_FROM_RX);
			break;
			
		case Types::AXIS_Z:
			
			d_third_axis.Calculate_Straightness(TMachineStandardComp::X_STR_FROM_RY);
			d_third_axis.Calculate_Straightness(TMachineStandardComp::Y_STR_FROM_RX);
			break;
	}
	
	d_init_mode = false;
	d_valid_data = true;	// may expand this to test for a minimum set of data

	return true;
}

TVector3 TMachineComp::Evaluate_XYZ(
	const TVector3						&position,
	const TVector3						&probe_offset) const
{
	TVector3							pnt;
	TVector3							error;
	TVector3							linear,angular;
	TVector3							map_position;
	TMat4								mat;
	double								quad;
	double								axis_origin;
	double								mechanical_origin;
	
	map_position = position; // - probe_offset;

	d_first_axis.Data(map_position.x,&angular,&linear);
	
	linear.x *= d_first_axis.Scale_Sign();
	angular /= 1000.0;		// units are mm/m, must be mm/mm

	error = linear;
	
	mat.Rotate(mat.X(),angular.x * d_first_axis.Rx_Rotation_Sign());
	mat.Rotate(mat.Y(),angular.y * d_first_axis.Ry_Rotation_Sign());
	mat.Rotate(mat.Z(),angular.z * d_first_axis.Rz_Rotation_Sign());
	
	d_second_axis.Data(map_position.y,&angular,&linear);
	
	linear.y *= d_first_axis.Scale_Sign();
	angular /= 1000.0;		// units are mm/m, must be mm/mm

	axis_origin = d_second_axis.Axis_Origin();
	mechanical_origin = d_second_axis.Axis_Mechanical_Origin();
	
	error += linear;
	error += mat * TVector3(0,map_position.y - axis_origin,0) - TVector3(0,map_position.y - axis_origin,0);
	error += mat * TVector3(0,mechanical_origin - axis_origin,0) - TVector3(0,mechanical_origin - axis_origin,0);
	
	mat.Rotate(mat.X(),angular.x * d_second_axis.Rx_Rotation_Sign());
	mat.Rotate(mat.Y(),angular.y * d_second_axis.Ry_Rotation_Sign());
	mat.Rotate(mat.Z(),angular.z * d_second_axis.Rz_Rotation_Sign());
	
	d_third_axis.Data(map_position.z,&angular,&linear);
	
	linear.z *= d_first_axis.Scale_Sign();
	angular /= 1000.0;		// units are mm/m, must be mm/mm

	axis_origin = d_third_axis.Axis_Origin();
	mechanical_origin = d_third_axis.Axis_Mechanical_Origin();

	error += linear;
	error += mat * TVector3(0,0,map_position.z - axis_origin) - TVector3(0,0,map_position.z - axis_origin);
	error += mat * TVector3(0,0,mechanical_origin - axis_origin) - TVector3(0,0,mechanical_origin - axis_origin);

	mat.Rotate(mat.X(),angular.x * d_third_axis.Rx_Rotation_Sign());
	mat.Rotate(mat.Y(),angular.y * d_third_axis.Ry_Rotation_Sign());
	mat.Rotate(mat.Z(),angular.z * d_third_axis.Rz_Rotation_Sign());
	
	error += mat * probe_offset - probe_offset;
	
	quad = d_first_axis.Quad() * d_first_axis.Quad_Sign();
	error.x += (quad * (map_position.y / 1000));
	
	quad = d_second_axis.Quad() * d_second_axis.Quad_Sign();
	error.y += (quad * (map_position.z / 1000.0));
	
	quad = d_third_axis.Quad() * d_third_axis.Quad_Sign();
	error.x += (quad * (map_position.z / 1000.0));

	pnt = position + error;

	return pnt;
}

TVector3 TMachineComp::Evaluate_YXZ(
	const TVector3						&position,
	const TVector3						&probe_offset) const
{
	TVector3							pnt;
	TVector3							error;
	TVector3							linear,angular;
	TVector3							map_position;
	TMat4								mat;
	double								quad;
	double								axis_origin;
	double								mechanical_origin;

	map_position = position; // - probe_offset;

	d_first_axis.Data(map_position.y,&angular,&linear);
	
	linear.y *= d_first_axis.Scale_Sign();
	angular /= 1000.0;		// units are mm/m, must be mm/mm

	error = linear;
	
	mat.Rotate(mat.X(),angular.x * d_first_axis.Rx_Rotation_Sign());
	mat.Rotate(mat.Y(),angular.y * d_first_axis.Ry_Rotation_Sign());
	mat.Rotate(mat.Z(),angular.z * d_first_axis.Rz_Rotation_Sign());

	d_second_axis.Data(map_position.x,&angular,&linear);
	
	linear.x *= d_first_axis.Scale_Sign();
	angular /= 1000.0;		// units are mm/m, must be mm/mm
	
	axis_origin = d_second_axis.Axis_Origin();
	mechanical_origin = d_second_axis.Axis_Mechanical_Origin();
	
	error += linear;
	error += mat * TVector3(map_position.x - axis_origin,0,0) - TVector3(map_position.x - axis_origin,0,0);
	error += mat * TVector3(mechanical_origin - axis_origin,0,0) - TVector3(mechanical_origin - axis_origin,0,0);
	
	mat.Rotate(mat.X(),angular.x * d_second_axis.Rx_Rotation_Sign());
	mat.Rotate(mat.Y(),angular.y * d_second_axis.Ry_Rotation_Sign());
	mat.Rotate(mat.Z(),angular.z * d_second_axis.Rz_Rotation_Sign());

	d_third_axis.Data(map_position.z,&angular,&linear);
	
	linear.z *= d_first_axis.Scale_Sign();
	angular /= 1000.0;		// units are mm/m, must be mm/mm
	
	axis_origin = d_third_axis.Axis_Origin();
	mechanical_origin = d_third_axis.Axis_Mechanical_Origin();
	
	error += linear;
	error += mat * TVector3(0,0,map_position.z - axis_origin) - TVector3(0,0,map_position.z - axis_origin);
	error += mat * TVector3(0,0,mechanical_origin - axis_origin) - TVector3(0,0,mechanical_origin - axis_origin);
	
	mat.Rotate(mat.X(),angular.x * d_third_axis.Rx_Rotation_Sign());
	mat.Rotate(mat.Y(),angular.y * d_third_axis.Ry_Rotation_Sign());
	mat.Rotate(mat.Z(),angular.z * d_third_axis.Rz_Rotation_Sign());

	error += mat * probe_offset - probe_offset;

	quad = d_first_axis.Quad() * d_first_axis.Quad_Sign();
	error.y += (quad * (map_position.z / 1000.0));
	
	quad = d_second_axis.Quad() * d_second_axis.Quad_Sign();
	error.y += (quad * (map_position.x / 1000.0));
	
	quad = d_third_axis.Quad() * d_third_axis.Quad_Sign();
	error.x += (quad * (map_position.z / 1000.0));

	pnt = position + error;
	
	return pnt;
}

TVector3 TMachineComp::Evaluate_XZY(
	const TVector3						&position,
	const TVector3						&probe_offset) const
{
	TVector3							pnt;
	TVector3							error;
	TVector3							linear,angular;
	TVector3							map_position;
	TMat4								mat;
	double								quad;
	double								axis_origin;
	double								mechanical_origin;
	double								deflection;
	
	map_position = position; // - probe_offset;

	// up = z, out = y
	try
	{
		deflection = d_deflection_axis.Evaluate(map_position.z,map_position.y) / 1000.0;	// units are mm/m, must be mm/mm
		deflection *= d_deflection_axis.Deflection_Sign();
	}
	catch(TMachineAxisException exception)
	{
		// may do more with this exception in the future.
		// biggest issue is with probe offsets that exceed the volume of the machine.
		deflection = 0.0;
	}
	
	d_first_axis.Data(map_position.x,&angular,&linear);
	
	linear.x *= d_first_axis.Scale_Sign();
	angular /= 1000.0;		// units are mm/m, must be mm/mm

	error = linear;
	
	mat.Rotate(mat.X(),angular.x  * d_first_axis.Rx_Rotation_Sign());
	mat.Rotate(mat.Y(),angular.y * d_first_axis.Ry_Rotation_Sign());
	mat.Rotate(mat.Z(),angular.z * d_first_axis.Rz_Rotation_Sign());
	
	d_second_axis.Data(map_position.z,&angular,&linear);
	
	linear.z *= d_first_axis.Scale_Sign();
	angular /= 1000.0;		// units are mm/m, must be mm/mm
	
	axis_origin = d_second_axis.Axis_Origin();
	mechanical_origin = d_second_axis.Axis_Mechanical_Origin();

	error += linear;
	error += mat * TVector3(0,0,map_position.z - axis_origin) - TVector3(0,0,map_position.z - axis_origin);
	error += mat * TVector3(0,0,mechanical_origin - axis_origin) - TVector3(0,0,mechanical_origin - axis_origin);
	
	mat.Rotate(mat.X(),angular.x * d_second_axis.Rx_Rotation_Sign());
	mat.Rotate(mat.Y(),angular.y * d_second_axis.Ry_Rotation_Sign());
	mat.Rotate(mat.Z(),angular.z * d_second_axis.Rz_Rotation_Sign());
	
	d_third_axis.Data(map_position.y,&angular,&linear);
	
	linear.y *= d_first_axis.Scale_Sign();
	angular /= 1000.0;		// units are mm/m, must be mm/mm
	
	axis_origin = d_third_axis.Axis_Origin();
	mechanical_origin = d_third_axis.Axis_Mechanical_Origin();

	error += linear;
	error += mat * TVector3(0,map_position.y - axis_origin,0) - TVector3(0,map_position.y - axis_origin,0);
	error += mat * TVector3(0,mechanical_origin - axis_origin,0) - TVector3(0,mechanical_origin - axis_origin,0);
	
	mat.Rotate(mat.X(),angular.x * d_third_axis.Rx_Rotation_Sign());
	mat.Rotate(mat.Y(),angular.y * d_third_axis.Ry_Rotation_Sign());
	mat.Rotate(mat.Z(),angular.z * d_third_axis.Rz_Rotation_Sign());
	
	error.y += (map_position.z - d_second_axis.Axis_Mechanical_Origin()) * deflection;
	error.z -= (map_position.y - d_third_axis.Axis_Mechanical_Origin()) * deflection;
	
	error += mat * probe_offset - probe_offset;
	
	quad = d_first_axis.Quad() * d_first_axis.Quad_Sign();
	error.x += (quad * (map_position.y / 1000.0));
	
	quad = d_second_axis.Quad() * d_second_axis.Quad_Sign();
	error.x += (quad * (map_position.z / 1000.0));
	
	quad = d_third_axis.Quad() * d_third_axis.Quad_Sign();
	error.y += (quad * (map_position.z / 1000.0));
	
	pnt = position + error;
	
	return pnt;
}

TVector3 TMachineComp::Evaluate_YZX(
	const TVector3						&position,
	const TVector3						&probe_offset) const
{
	TVector3							pnt;
	TVector3							error;
	TVector3							linear,angular;
	TVector3							map_position;
	TMat4								mat;
	double								quad;
	double								axis_origin;
	double								mechanical_origin;
	double								deflection;
	
	map_position = position; // - probe_offset;

	// up = z, out = x
	try
	{
		deflection = d_deflection_axis.Evaluate(map_position.z,map_position.x) / 1000.0;	// units are mm/m, must be mm/mm
		deflection *= d_deflection_axis.Deflection_Sign();
	}
	catch(TMachineAxisException exception)
	{
		// may do more with this exception in the future.
		// biggest issue is with probe offsets that exceed the volume of the machine.
		deflection = 0.0;
	}
	
	d_first_axis.Data(map_position.y,&angular,&linear);
	
	linear.y *= d_first_axis.Scale_Sign();
	angular /= 1000.0;		// units are mm/m, must be mm/mm

	error = linear;
	
	mat.Rotate(mat.X(),angular.x * d_first_axis.Rx_Rotation_Sign());
	mat.Rotate(mat.Y(),angular.y * d_first_axis.Ry_Rotation_Sign());
	mat.Rotate(mat.Z(),angular.z * d_first_axis.Rz_Rotation_Sign());
	
	d_second_axis.Data(map_position.z,&angular,&linear);
	
	linear.z *= d_first_axis.Scale_Sign();
	angular /= 1000.0;		// units are mm/m, must be mm/mm
	
	axis_origin = d_second_axis.Axis_Origin();
	mechanical_origin = d_second_axis.Axis_Mechanical_Origin();

	error += linear;
	error += mat * TVector3(0,0,map_position.z - axis_origin) - TVector3(0,0,map_position.z - axis_origin);
	error += mat * TVector3(0,0,mechanical_origin - axis_origin) - TVector3(0,0,mechanical_origin - axis_origin);
	
	mat.Rotate(mat.X(),angular.x * d_second_axis.Rx_Rotation_Sign());
	mat.Rotate(mat.Y(),angular.y * d_second_axis.Ry_Rotation_Sign());
	mat.Rotate(mat.Z(),angular.z * d_second_axis.Rz_Rotation_Sign());
	
	d_third_axis.Data(map_position.x,&angular,&linear);

	linear.x *= d_first_axis.Scale_Sign();
	angular /= 1000.0;		// units are mm/m, must be mm/mm
	
	axis_origin = d_third_axis.Axis_Origin();
	mechanical_origin = d_third_axis.Axis_Mechanical_Origin();

	error += linear;
	error += mat * TVector3(map_position.x - axis_origin,0,0) - TVector3(map_position.x - axis_origin,0,0);
	error += mat * TVector3(mechanical_origin - axis_origin,0,0) - TVector3(mechanical_origin - axis_origin,0,0);
	
	mat.Rotate(mat.X(),angular.x * d_third_axis.Rx_Rotation_Sign());
	mat.Rotate(mat.Y(),angular.y * d_third_axis.Ry_Rotation_Sign());
	mat.Rotate(mat.Z(),angular.z * d_third_axis.Rz_Rotation_Sign());
	
	error.x -= (map_position.z - d_second_axis.Axis_Mechanical_Origin()) * deflection;
	error.z += (map_position.x - d_third_axis.Axis_Mechanical_Origin()) * deflection;
	
	error += mat * probe_offset - probe_offset;
	
	quad = d_first_axis.Quad() * d_first_axis.Quad_Sign();
	error.y += (quad * (map_position.z / 1000.0));
	
	quad = d_second_axis.Quad() * d_second_axis.Quad_Sign();
	error.x += (quad * (map_position.z / 1000.0));
	
	quad = d_third_axis.Quad() * d_third_axis.Quad_Sign();
	error.x += (quad * (map_position.y / 1000.0));
	
	pnt = position + error;
	
	return pnt;
}

