/////////////////////////////////////////////////////////////////////
//
//            X   X           X
//           XXX XX         XX
//          XXXXXXXX      XXX
//         XX X XXXXXXXXXXX
//        XXXXX XXXXXXXXX
//       XXXXX XXXXXXXXXX
//            XXX XXX XXX
//           XXX XX   XX
//           X   X     X
//
//    Copyright (C) 2003-2023  Ron Jakl
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
/////////////////////////////////////////////////////////////////////

#ifndef MACHINEHEADERFILE
#define MACHINEHEADERFILE

#include <QStringList>

#include "../../core/vector3.h"
#include "machine_standardcomp.h"
#include "machine_deflectioncomp.h"

class TMachineComp
{
public:

struct TInitParameters
{
	int									kinematic;
	
	int									first_axis_id;
	int									second_axis_id;
	int									third_axis_id;
	
	double								first_axis_start;
	double								first_axis_end;
	double								first_axis_math_origin;
	double								first_axis_mechanical_origin;
	
	double								second_axis_start;
	double								second_axis_end;
	double								second_axis_math_origin;
	double								second_axis_mechanical_origin;
	
	double								third_axis_start;
	double								third_axis_end;
	double								third_axis_math_origin;
	double								third_axis_mechanical_origin;
};
	
// CREATORS
	TMachineComp(void);
	~TMachineComp(void);
	
// ACCESSORS
	QStringList Map_Data_X(const double &increment);
	QStringList Map_Data_Y(const double &increment);
	QStringList Map_Data_Z(const double &increment);
	QStringList Map_Data_D(const double &increment);
	QStringList Map_Data_Quad(void);
	
	TVector3 Map_Data_Min(void) const;
	TVector3 Map_Data_Max(void) const;
	
	TInitParameters Init_Parameters(void) const {return d_init_parameters;}
	
	// position in map coordinates
	TVector3 Evaluate(const TVector3 &position,const TVector3 &probe_offset,bool * const valid) const;
	
	TVector3 Get_Angles_RA_RB_RC(const TVector3 &axis,const TVector3 &probe_axis,const TVector3 &position) const;
	TVector3 Calculate_Scale_XX_YY_Error(const TVector3 &axis,const TVector3 &probe_axis,const TVector3 &nominal, const TVector3 &actual) const;
	
	QString Last_Error(void) const {return d_last_error;}
	
	
// MANIPULATORS
	bool Initialize_Begin(const TInitParameters &init_parameters);
	
	bool Set_Rx(const int axis_id,const QString &expression,const int sign);
	bool Set_Ry(const int axis_id,const QString &expression,const int sign);
	bool Set_Rz(const int axis_id,const QString &expression,const int sign);
	bool Set_Scale(const int axis_id,const QString &expression,const int sign);
	bool Set_Deflection(const int axis_id,const double &value,const int sign);
	bool Set_Quad(const int axis_id,const double &value,const int sign);

	bool Initialize_End(void);

private:
	TInitParameters						d_init_parameters;
	
	TMachineStandardComp				d_first_axis;
	TMachineStandardComp				d_second_axis;
	TMachineStandardComp				d_third_axis;
	TMachineDeflectionComp				d_deflection_axis;
	
	bool								d_init_mode;
	bool								d_valid_data;
	
	mutable QString						d_last_error;
	
	TVector3 Evaluate_XYZ(const TVector3 &position,const TVector3 &probe_offset) const;
	TVector3 Evaluate_YXZ(const TVector3 &position,const TVector3 &probe_offset) const;
	TVector3 Evaluate_XZY(const TVector3 &position,const TVector3 &probe_offset) const;
	TVector3 Evaluate_YZX(const TVector3 &position,const TVector3 &probe_offset) const;
	
// NOT IMPLEMENTED
	TMachineComp(const TMachineComp&);
	TMachineComp& operator=(const TMachineComp&);
};

#endif
