/////////////////////////////////////////////////////////////////////
//
//            X   X           X
//           XXX XX         XX
//          XXXXXXXX      XXX
//         XX X XXXXXXXXXXX
//        XXXXX XXXXXXXXX
//       XXXXX XXXXXXXXXX
//            XXX XXX XXX
//           XXX XX   XX
//           X   X     X
//
//    Copyright (C) 2003-2023  Ron Jakl
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
/////////////////////////////////////////////////////////////////////

#include <QApplication>
#include <QFrame>
#include <QGridLayout>
#include <QGroupBox>
#include <QHBoxLayout>
#include <QLabel>
#include <QLineEdit>
#include <QListWidget>
#include <QListWidgetItem>
#include <QRadioButton>
#include <QSizePolicy>
#include <QSpacerItem>
#include <QToolButton>
#include <QTreeWidget>
#include <QTreeWidgetItem>
#include <QComboBox>
#include <QVBoxLayout>
#include <QWidget>
#include <QPushButton>
#include <QRegularExpression>
#include <QRegularExpressionMatch>
#include <QDir>
#include <QList>
#include <assert.h>

#include "../../core/xmlfile.h"
#include "machineparameterwidget.h"
#include "types.h"

#include "machinedialog.h"

TMachineDialog::TMachineDialog(
	const QWidget						*parent,
	const Qt::WindowFlags				flags)
:QDialog(const_cast<QWidget*>(parent),flags)
{
	QFrame								*separator_hline;
	QGridLayout							*axis_group_layout;
	QGridLayout							*dialog_layout;
	QGroupBox							*axis_group;
	QGroupBox							*kinematic_group;
	QHBoxLayout							*button_hlayout;
	QLabel								*machine_parameter_label;
	QLabel								*max_x_label;
	QLabel								*max_y_label;
	QLabel								*max_z_label;
	QLabel								*min_x_label;
	QLabel								*min_y_label;
	QLabel								*min_z_label;
	QLabel								*math_origin_x_label;
	QLabel								*math_origin_y_label;
	QLabel								*math_origin_z_label;
	QLabel								*mechanical_origin_x_label;
	QLabel								*mechanical_origin_y_label;
	QLabel								*mechanical_origin_z_label;
	QPushButton							*accept_button;
	QPushButton							*cancel_button;
	QSpacerItem							*kinematic_vspacer;
	QSpacerItem							*button_hspacer;
	QVBoxLayout							*kinematic_group_vlayout;

	this->resize(801,801);

	dialog_layout = new QGridLayout(this);
		
	d_machine_name_label = new QLabel(this);
	dialog_layout->addWidget(d_machine_name_label,0,0,1,2);

	kinematic_group = new QGroupBox(this);

	kinematic_group_vlayout = new QVBoxLayout(kinematic_group);

	d_kinematic_xyz_radio = new QRadioButton(kinematic_group);
	kinematic_group_vlayout->addWidget(d_kinematic_xyz_radio);

	d_kinematic_yxz_radio = new QRadioButton(kinematic_group);
	kinematic_group_vlayout->addWidget(d_kinematic_yxz_radio);

	d_kinematic_xzy_radio = new QRadioButton(kinematic_group);
	kinematic_group_vlayout->addWidget(d_kinematic_xzy_radio);

	d_kinematic_yzx_radio = new QRadioButton(kinematic_group);
	kinematic_group_vlayout->addWidget(d_kinematic_yzx_radio);
	
	kinematic_vspacer = new QSpacerItem(0,20,QSizePolicy::Minimum,QSizePolicy::Preferred);
	kinematic_group_vlayout->addItem(kinematic_vspacer);
	
	dialog_layout->addWidget(kinematic_group,1,0,1,1);

	axis_group = new QGroupBox(this);

	axis_group_layout = new QGridLayout(axis_group);

	min_x_label = new QLabel(axis_group);
	min_x_label->setAlignment(Qt::AlignRight|Qt::AlignTrailing|Qt::AlignVCenter);
	axis_group_layout->addWidget(min_x_label,0,0,1,1);

	d_min_x_edit = new QLineEdit(axis_group);
	d_min_x_edit->setAlignment(Qt::AlignCenter);
	axis_group_layout->addWidget(d_min_x_edit,0,1,1,1);

	min_y_label = new QLabel(axis_group);
	min_y_label->setAlignment(Qt::AlignRight|Qt::AlignTrailing|Qt::AlignVCenter);
	axis_group_layout->addWidget(min_y_label,0,2,1,1);

	d_min_y_edit = new QLineEdit(axis_group);
	d_min_y_edit->setAlignment(Qt::AlignCenter);
	axis_group_layout->addWidget(d_min_y_edit,0,3,1,1);

	min_z_label = new QLabel(axis_group);
	min_z_label->setAlignment(Qt::AlignRight|Qt::AlignTrailing|Qt::AlignVCenter);
	axis_group_layout->addWidget(min_z_label,0,4,1,1);

	d_min_z_edit = new QLineEdit(axis_group);
	d_min_z_edit->setAlignment(Qt::AlignCenter);
	axis_group_layout->addWidget(d_min_z_edit,0,5,1,1);

	max_x_label = new QLabel(axis_group);
	max_x_label->setAlignment(Qt::AlignRight|Qt::AlignTrailing|Qt::AlignVCenter);
	axis_group_layout->addWidget(max_x_label,1,0,1,1);

	d_max_x_edit = new QLineEdit(axis_group);
	d_max_x_edit->setAlignment(Qt::AlignCenter);
	axis_group_layout->addWidget(d_max_x_edit,1,1,1,1);

	max_y_label = new QLabel(axis_group);
	max_y_label->setAlignment(Qt::AlignRight|Qt::AlignTrailing|Qt::AlignVCenter);
	axis_group_layout->addWidget(max_y_label,1,2,1,1);

	d_max_y_edit = new QLineEdit(axis_group);
	d_max_y_edit->setAlignment(Qt::AlignCenter);
	axis_group_layout->addWidget(d_max_y_edit,1,3,1,1);

	max_z_label = new QLabel(axis_group);
	max_z_label->setAlignment(Qt::AlignRight|Qt::AlignTrailing|Qt::AlignVCenter);
	axis_group_layout->addWidget(max_z_label,1,4,1,1);

	d_max_z_edit = new QLineEdit(axis_group);
	d_max_z_edit->setAlignment(Qt::AlignCenter);
	axis_group_layout->addWidget(d_max_z_edit,1,5,1,1);

	math_origin_x_label = new QLabel(axis_group);
	math_origin_x_label->setAlignment(Qt::AlignRight|Qt::AlignTrailing|Qt::AlignVCenter);
	axis_group_layout->addWidget(math_origin_x_label,2,0,1,1);
	
	d_math_origin_x_edit = new QLineEdit(axis_group);
	d_math_origin_x_edit->setAlignment(Qt::AlignCenter);
	axis_group_layout->addWidget(d_math_origin_x_edit,2,1,1,1);
	
	math_origin_y_label = new QLabel(axis_group);
	math_origin_y_label->setAlignment(Qt::AlignRight|Qt::AlignTrailing|Qt::AlignVCenter);
	axis_group_layout->addWidget(math_origin_y_label,2,2,1,1);
	
	d_math_origin_y_edit = new QLineEdit(axis_group);
	d_math_origin_y_edit->setAlignment(Qt::AlignCenter);
	axis_group_layout->addWidget(d_math_origin_y_edit,2,3,1,1);
	
	math_origin_z_label = new QLabel(axis_group);
	math_origin_z_label->setAlignment(Qt::AlignRight|Qt::AlignTrailing|Qt::AlignVCenter);
	axis_group_layout->addWidget(math_origin_z_label,2,4,1,1);
	
	d_math_origin_z_edit = new QLineEdit(axis_group);
	d_math_origin_z_edit->setAlignment(Qt::AlignCenter);
	axis_group_layout->addWidget(d_math_origin_z_edit,2,5,1,1);
	
	mechanical_origin_x_label = new QLabel(axis_group);
	mechanical_origin_x_label->setAlignment(Qt::AlignRight|Qt::AlignTrailing|Qt::AlignVCenter);
	axis_group_layout->addWidget(mechanical_origin_x_label,3,0,1,1);
	
	d_mechanical_origin_x_edit = new QLineEdit(axis_group);
	d_mechanical_origin_x_edit->setAlignment(Qt::AlignCenter);
	axis_group_layout->addWidget(d_mechanical_origin_x_edit,3,1,1,1);
	
	mechanical_origin_y_label = new QLabel(axis_group);
	mechanical_origin_y_label->setAlignment(Qt::AlignRight|Qt::AlignTrailing|Qt::AlignVCenter);
	axis_group_layout->addWidget(mechanical_origin_y_label,3,2,1,1);
	
	d_mechanical_origin_y_edit = new QLineEdit(axis_group);
	d_mechanical_origin_y_edit->setAlignment(Qt::AlignCenter);
	axis_group_layout->addWidget(d_mechanical_origin_y_edit,3,3,1,1);
	
	mechanical_origin_z_label = new QLabel(axis_group);
	mechanical_origin_z_label->setAlignment(Qt::AlignRight|Qt::AlignTrailing|Qt::AlignVCenter);
	axis_group_layout->addWidget(mechanical_origin_z_label,3,4,1,1);
	
	d_mechanical_origin_z_edit = new QLineEdit(axis_group);
	d_mechanical_origin_z_edit->setAlignment(Qt::AlignCenter);
	axis_group_layout->addWidget(d_mechanical_origin_z_edit,3,5,1,1);
	dialog_layout->addWidget(axis_group,1,1,1,1);

	machine_parameter_label = new QLabel(this);
	dialog_layout->addWidget(machine_parameter_label,2,0,1,2);

	d_machine_parameter_tree = new QTreeWidget(this);
	d_machine_parameter_tree->setRootIsDecorated(true);
	d_machine_parameter_tree->setHeaderHidden(true);
	d_machine_parameter_tree->setIndentation(12);
	d_machine_parameter_tree->setSelectionMode(QAbstractItemView::NoSelection);
	d_machine_parameter_tree->setVerticalScrollBarPolicy(Qt::ScrollBarAsNeeded);
	d_machine_parameter_tree->setHorizontalScrollBarPolicy(Qt::ScrollBarAsNeeded);
	dialog_layout->addWidget(d_machine_parameter_tree,3,0,1,2);
	
	separator_hline = new QFrame(this);
	separator_hline->setFrameShape(QFrame::HLine);
	separator_hline->setFrameShadow(QFrame::Sunken);
	dialog_layout->addWidget(separator_hline,4,0,1,2);
	
	button_hlayout = new QHBoxLayout();
	
	button_hspacer = new QSpacerItem(0,0,QSizePolicy::Expanding,QSizePolicy::Minimum);
	button_hlayout->addItem(button_hspacer);
	
	cancel_button = new QPushButton(this);
	cancel_button->setAutoDefault(false);
	button_hlayout->addWidget(cancel_button);
	
	accept_button = new QPushButton(this);
	accept_button->setAutoDefault(false);
	button_hlayout->addWidget(accept_button);
	dialog_layout->addLayout(button_hlayout,5,0,1,2);

	QWidget::setTabOrder(d_kinematic_xyz_radio,d_kinematic_yxz_radio);
	QWidget::setTabOrder(d_kinematic_yxz_radio,d_kinematic_xzy_radio);
	QWidget::setTabOrder(d_kinematic_xzy_radio,d_kinematic_yzx_radio);
	QWidget::setTabOrder(d_kinematic_yzx_radio,d_min_x_edit);
	QWidget::setTabOrder(d_min_x_edit,d_min_y_edit);
	QWidget::setTabOrder(d_min_y_edit,d_min_z_edit);
	QWidget::setTabOrder(d_min_z_edit,d_max_x_edit);
	QWidget::setTabOrder(d_max_x_edit,d_max_y_edit);
	QWidget::setTabOrder(d_max_y_edit,d_max_z_edit);
	QWidget::setTabOrder(d_max_z_edit,d_math_origin_x_edit);
	QWidget::setTabOrder(d_math_origin_x_edit,d_math_origin_y_edit);
	QWidget::setTabOrder(d_math_origin_y_edit,d_math_origin_z_edit);
	QWidget::setTabOrder(d_math_origin_z_edit,d_mechanical_origin_x_edit);
	QWidget::setTabOrder(d_mechanical_origin_x_edit,d_mechanical_origin_y_edit);
	QWidget::setTabOrder(d_mechanical_origin_y_edit,d_mechanical_origin_z_edit);
	QWidget::setTabOrder(d_mechanical_origin_z_edit,d_machine_parameter_tree);
	
	// defaults
	d_active_machine_id = -1;
	d_kinematic_yxz_radio->setChecked(true);
	
	this->Initialize_Parameter_Tree();
	this->Enable_Editing(false);
	d_machine_parameter_tree->expandAll();
	d_data_changed = false;

	kinematic_group->setTitle(QStringLiteral("Kinematic Order"));
	axis_group->setTitle(QStringLiteral("Axis Dimensions, Math, and Mechanical Origin"));
	d_kinematic_xyz_radio->setText(QStringLiteral("Axis: X-Y-Z"));
	d_kinematic_yxz_radio->setText(QStringLiteral("Axis: Y-X-Z"));
	d_kinematic_xzy_radio->setText(QStringLiteral("Axis: X-Z-Y"));
	d_kinematic_yzx_radio->setText(QStringLiteral("Axis: Y-Z-X"));
	min_x_label->setText(QStringLiteral("Axis Min X:"));
	min_y_label->setText(QStringLiteral("Y:"));
	min_z_label->setText(QStringLiteral("Z:"));
	max_x_label->setText(QStringLiteral("Axis Max X:"));
	max_y_label->setText(QStringLiteral("Y:"));
	max_z_label->setText(QStringLiteral("Z:"));
	math_origin_x_label->setText(QStringLiteral("Math Origin X:"));
	math_origin_y_label->setText(QStringLiteral("Y:"));
	math_origin_z_label->setText(QStringLiteral("Z:"));
	mechanical_origin_x_label->setText(QStringLiteral("Mechanical Origin X:"));
	mechanical_origin_y_label->setText(QStringLiteral("Y:"));
	mechanical_origin_z_label->setText(QStringLiteral("Z:"));
	cancel_button->setText(QStringLiteral("Cancel"));
	accept_button->setText(QStringLiteral("Accept"));
	
	machine_parameter_label->setText(QStringLiteral("<html><head/><body><p><span style=\" text-decoration: underline;\">Machine Error Parameters:</span></p></body></html>"));
	
	connect(cancel_button,&QPushButton::clicked,this,&TMachineDialog::reject);
	connect(accept_button,&QPushButton::clicked,this,&TMachineDialog::Accept_Clicked);
	
	connect(d_kinematic_xyz_radio,&QRadioButton::toggled,this,&TMachineDialog::Kinematic_Changed);
	connect(d_kinematic_yxz_radio,&QRadioButton::toggled,this,&TMachineDialog::Kinematic_Changed);
	connect(d_kinematic_xzy_radio,&QRadioButton::toggled,this,&TMachineDialog::Kinematic_Changed);
	connect(d_kinematic_yzx_radio,&QRadioButton::toggled,this,&TMachineDialog::Kinematic_Changed);
}

TMachineDialog::~TMachineDialog(void)
{
}

bool TMachineDialog::Save_Data(void) const
{
	if(d_data_changed)
	{
		if(this->Save())
		{
			d_data_changed = false;
			return true;
		}
		
		return false;
	}
	
	return true;
}

QStringList TMachineDialog::Machine_Names(void) const
{
	std::vector<TMachineItem>::const_iterator iter;
	QStringList							list;
	
	for(iter = d_machine_items.begin();iter != d_machine_items.end();++iter)
	{
		list.push_back((*iter).name);
	}

	return list;
}

int TMachineDialog::Machine_Id(
	const QString						&name)
{
	std::vector<TMachineItem>::const_iterator iter;
	
	for(iter = d_machine_items.begin();iter != d_machine_items.end();++iter)
	{
		if(name.compare((*iter).name,Qt::CaseInsensitive) == 0)
		{
			return (*iter).id;
		}
	}
	
	return -1;
}

QString TMachineDialog::Machine_Name(
	const int							id) const
{
	std::vector<TMachineItem>::const_iterator iter;
	
	for(iter = d_machine_items.begin();iter != d_machine_items.end();++iter)
	{
		if((*iter).id == id)
		{
			return (*iter).name;
		}
	}
	
	return QString();
}

TVector3 TMachineDialog::Machine_Axis_Start(
	const int 							id) const
{
	std::vector<TMachineItem>::const_iterator iter;
	
	for(iter = d_machine_items.begin();iter != d_machine_items.end();++iter)
	{
		if((*iter).id == id)
		{
			return (*iter).axis_min;
		}
	}
	
	return TVector3();
}

TVector3 TMachineDialog::Machine_Axis_End(
	const int 							id) const
{
	std::vector<TMachineItem>::const_iterator iter;
	
	for(iter = d_machine_items.begin();iter != d_machine_items.end();++iter)
	{
		if((*iter).id == id)
		{
			return (*iter).axis_max;
		}
	}
	
	return TVector3();
}

TVector3 TMachineDialog::Machine_Math_Origin(
	const int 							id) const
{
	std::vector<TMachineItem>::const_iterator iter;
	
	for(iter = d_machine_items.begin();iter != d_machine_items.end();++iter)
	{
		if((*iter).id == id)
		{
			return (*iter).math_origin;
		}
	}
	
	return TVector3();
}

TVector3 TMachineDialog::Machine_Mechanical_Origin(
	const int 							id) const
{
	std::vector<TMachineItem>::const_iterator iter;
	
	for(iter = d_machine_items.begin();iter != d_machine_items.end();++iter)
	{
		if((*iter).id == id)
		{
			return (*iter).mechanical_origin;
		}
	}
	
	return TVector3();
}

int TMachineDialog::Machine_Kinematic(
	const int							id) const
{
	std::vector<TMachineItem>::const_iterator iter;
	
	for(iter = d_machine_items.begin();iter != d_machine_items.end();++iter)
	{
		if((*iter).id == id)
		{
			return (*iter).kinematic;
		}
	}
	
	return Types::KINEMATIC_XYZ;
}

std::vector<TMachineDialog::TMapErrorItem> TMachineDialog::Map_Error_Items(
	const int							id) const
{
	std::vector<TMachineDialog::TMapErrorItem> items;
	std::vector<TMachineItem>::const_iterator iter;
	std::vector<TMachineAxisItem>::const_iterator axis_iter;
	std::vector<TMachineParameterItem>::const_iterator param_iter;
	TMachineDialog::TMapErrorItem		error_item;
	
	for(iter = d_machine_items.begin();iter != d_machine_items.end();++iter)
	{
		if((*iter).id == id)
		{
			for(axis_iter = (*iter).axis_items.begin();axis_iter != (*iter).axis_items.end();++axis_iter)
			{
				for(param_iter = (*axis_iter).parameter_items.begin();param_iter != (*axis_iter).parameter_items.end();++param_iter)
				{
					if((*param_iter).enabled)
					{
						error_item.axis = (*axis_iter).axis;
						
						error_item.value = 0.0;
						error_item.expression.clear();

						if((*axis_iter).axis == Types::AXIS_D || (*axis_iter).axis == Types::AXIS_XYZ)
						{
							error_item.value = (*param_iter).expression.toDouble();
						}
						else
						{
							error_item.expression = (*param_iter).expression;
						}
						
						error_item.parameter = (*param_iter).parameter;
						
						if((*param_iter).option_index == 0)
						{
							error_item.sign = 1;
						}
						else
						{
						
							error_item.sign = -1;
						}
						
						items.push_back(error_item);
					}
				}
			}
			break;
		}
	}

	return items;
}

bool TMachineDialog::Load_Data(void)
{
	TXmlFile							xmlfile;
	int									version;
	
	try
	{
		xmlfile.Set_File_Name(QString("%1/machines.xml").arg(d_data_path));
		xmlfile.Open("Machines",&version);
		
		switch(version)
		{
			case 1:
				this->Load_Version1(&xmlfile);
				break;
				
			case 2:
				this->Load_Version2(&xmlfile);
				break;
		}
		
		d_data_changed = false;
		
	}
	catch(TXmlFileException xml_exception)
	{
		d_last_error = QString("ERR:  Loading machines exception: %1").arg(xml_exception.ErrorText());
		return false;
	}
	
	return true;
}

int TMachineDialog::Create_Machine(
	const QString						&name)
{
	TMachineItem						item;
	TMachineAxisItem					axis_item;
	TMachineParameterItem				parameter_item;
	std::vector<TMachineItem>::const_iterator iter;
	
	if(d_machine_items.size())
	{
		item.id = d_machine_items[d_machine_items.size() - 1].id + 1;
	}
	else
	{
		item.id = 0;
	}
	
	item.name = name;
	
	item.kinematic = Types::KINEMATIC_YXZ;
	
	item.axis_min.Set(0,0,-1000);
	item.axis_max.Set(1000,1000,0);
	item.math_origin.Set(0,0,0);
	item.mechanical_origin.Set(0,0,0);
	
	axis_item.axis = Types::AXIS_X;
	
	parameter_item.enabled = false;
	parameter_item.expression = QStringLiteral("0.0000 * L");
	parameter_item.option_index = 0;
	
	parameter_item.parameter = Types::PARAMETER_RX;
	axis_item.parameter_items.push_back(parameter_item);
	
	parameter_item.parameter = Types::PARAMETER_RY;
	axis_item.parameter_items.push_back(parameter_item);
	
	parameter_item.parameter = Types::PARAMETER_RZ;
	axis_item.parameter_items.push_back(parameter_item);
	
	parameter_item.parameter = Types::PARAMETER_SCALE;
	axis_item.parameter_items.push_back(parameter_item);
	
	item.axis_items.push_back(axis_item);
	
	axis_item.axis = Types::AXIS_Y;
	item.axis_items.push_back(axis_item);

	axis_item.axis = Types::AXIS_Z;
	item.axis_items.push_back(axis_item);
	
	axis_item.axis = Types::AXIS_XYZ;
	parameter_item.expression = QStringLiteral("0.0000");
	axis_item.parameter_items.clear();
	
	parameter_item.parameter = Types::PARAMETER_QXY;
	axis_item.parameter_items.push_back(parameter_item);
	
	parameter_item.parameter = Types::PARAMETER_QYZ;
	axis_item.parameter_items.push_back(parameter_item);
	
	parameter_item.parameter = Types::PARAMETER_QZX;
	axis_item.parameter_items.push_back(parameter_item);
	
	item.axis_items.push_back(axis_item);
	axis_item.parameter_items.clear();
	
	axis_item.axis = Types::AXIS_D;
	
	parameter_item.expression = QStringLiteral("0.0000");

	parameter_item.parameter = Types::PARAMETER_DEFLECTION;
	axis_item.parameter_items.push_back(parameter_item);

	item.axis_items.push_back(axis_item);

	d_machine_items.push_back(item);
		
	d_data_changed = true;
	
	return item.id;
}

void TMachineDialog::Remove_Machine(
	const int							id)
{
	std::vector<TMachineItem>::iterator iter;
	
	for(iter = d_machine_items.begin();iter != d_machine_items.end();++iter)
	{
		if((*iter).id == id)
		{
			d_machine_items.erase(iter);
			
			d_data_changed = true;

			return;
		}
	}
}

QString TMachineDialog::Rename_Machine(
	const int							id,
	const QString						&new_name)
{
	std::vector<TMachineItem>::iterator iter;
	QString								actual_name(new_name);
	
	for(iter = d_machine_items.begin();iter != d_machine_items.end();++iter)
	{
		if((*iter).id == id)
		{
			actual_name = this->Generate_Unique_Name(id,new_name);
			(*iter).name = actual_name;
			break;
		}
	}
	
	d_data_changed = true;
	
	return actual_name;
}

void TMachineDialog::Reset(
	const QString						&machine_name)
{	
	d_machine_name_label->setText(QString("<html><head/><body><p><span style=\" text-decoration: underline;\">Machine: %1</span></p></body></html>").arg(machine_name));
	
	d_active_machine_id = this->Machine_Id(machine_name);
	
	if(!(d_active_machine_id < 0))
	{
		this->Enable_Editing(true);
		this->Update_Editor();
	}
}

void TMachineDialog::Kinematic_Changed(
	bool								state)
{
	QTreeWidgetItem						*tree_item;
	QTreeWidgetItem						*child_tree_item;
	TMachineParameterWidget				*parameter_widget;
	bool								is_horizontal_arm(false);
	int									count;
	int									cntr;
	int									child_count;
	int									child_cntr;
	int									axis;
	
	if(!state)
	{
		return;
	}
	
	if(d_active_machine_id < 0)
	{
		return;
	}
	
	if(d_kinematic_xzy_radio->isChecked())
	{
		is_horizontal_arm = true;
	}
	else if(d_kinematic_yzx_radio->isChecked())
	{
		is_horizontal_arm = true;
	}
	
	// parameter values
	count = d_machine_parameter_tree->topLevelItemCount();	// categories X,Y,Z,XYZ, D
	
	for(cntr = 0;cntr < count;++cntr)
	{
		tree_item = d_machine_parameter_tree->topLevelItem(cntr);
		axis = tree_item->data(0,Qt::UserRole).toInt();
		
		if(axis == Types::AXIS_D)
		{
			child_count = tree_item->childCount();
			
			for(child_cntr = 0;child_cntr < child_count;++child_cntr)
			{
				child_tree_item = tree_item->child(child_cntr);
				
				parameter_widget = qobject_cast<TMachineParameterWidget*>(d_machine_parameter_tree->itemWidget(child_tree_item,0));
			
				if(parameter_widget)
				{
					parameter_widget->setEnabled(is_horizontal_arm);
					parameter_widget->Set_Option_Enabled(is_horizontal_arm & parameter_widget->Is_Option_Enabled());
					
					if(!is_horizontal_arm)
					{
						d_machine_parameter_tree->collapseItem(tree_item);
					}
					else
					{
						d_machine_parameter_tree->expandItem(tree_item);
					}
				}
			}
			
			break;
		}
	}
}

void TMachineDialog::Accept_Clicked(void)
{
	this->Update_Machine();
	d_data_changed = true;
	
	this->accept();
}

void TMachineDialog::Enable_Editing(
	const bool							state)
{
	if(state)
	{
		d_machine_parameter_tree->setEnabled(true);
		
		d_min_x_edit->setEnabled(true);
		d_min_y_edit->setEnabled(true);
		d_min_z_edit->setEnabled(true);
		
		d_max_x_edit->setEnabled(true);
		d_max_y_edit->setEnabled(true);
		d_max_z_edit->setEnabled(true);

		d_math_origin_x_edit->setEnabled(true);
		d_math_origin_y_edit->setEnabled(true);
		d_math_origin_z_edit->setEnabled(true);

		d_mechanical_origin_x_edit->setEnabled(true);
		d_mechanical_origin_y_edit->setEnabled(true);
		d_mechanical_origin_z_edit->setEnabled(true);
		
		d_kinematic_xyz_radio->setEnabled(true);
		d_kinematic_yxz_radio->setEnabled(true);
		d_kinematic_xzy_radio->setEnabled(true);
		d_kinematic_yzx_radio->setEnabled(true);
	}
	else
	{
		d_machine_parameter_tree->setEnabled(false);
		
		d_min_x_edit->setEnabled(false);
		d_min_y_edit->setEnabled(false);
		d_min_z_edit->setEnabled(false);
		
		d_max_x_edit->setEnabled(false);
		d_max_y_edit->setEnabled(false);
		d_max_z_edit->setEnabled(false);
		
		d_math_origin_x_edit->setEnabled(false);
		d_math_origin_y_edit->setEnabled(false);
		d_math_origin_z_edit->setEnabled(false);
		
		d_mechanical_origin_x_edit->setEnabled(false);
		d_mechanical_origin_y_edit->setEnabled(false);
		d_mechanical_origin_z_edit->setEnabled(false);

		d_kinematic_xyz_radio->setEnabled(false);
		d_kinematic_yxz_radio->setEnabled(false);
		d_kinematic_xzy_radio->setEnabled(false);
		d_kinematic_yzx_radio->setEnabled(false);
	}
}

void TMachineDialog::Initialize_Parameter_Tree(void)
{
	QTreeWidgetItem						*axis_item;
	QTreeWidgetItem						*parameter_item;
	TMachineParameterWidget				*parameter_widget;
	QStringList							angular_signs;
	QStringList							scale_signs;
	QStringList							xy_squareness_signs;
	QStringList							yz_squareness_signs;
	QStringList							zx_squareness_signs;
	
	angular_signs.push_back(QStringLiteral("CCW = Pos"));
	angular_signs.push_back(QStringLiteral("CW = Pos"));
	
	scale_signs.push_back(QStringLiteral("Laser > CMM = Pos"));
	scale_signs.push_back(QStringLiteral("Laser < CMM = Pos"));

	xy_squareness_signs.push_back(QStringLiteral("X to Y > 90 = Pos"));
	xy_squareness_signs.push_back(QStringLiteral("X to Y < 90 = Pos"));
	
	yz_squareness_signs.push_back(QStringLiteral("Y to Z > 90 = Pos"));
	yz_squareness_signs.push_back(QStringLiteral("Y to Z < 90 = Pos"));
	
	zx_squareness_signs.push_back(QStringLiteral("Z to X > 90 = Pos"));
	zx_squareness_signs.push_back(QStringLiteral("Z to X < 90 = Pos"));

	d_machine_parameter_tree->clear();
	
	axis_item = new QTreeWidgetItem(d_machine_parameter_tree);
	axis_item->setText(0,QStringLiteral("X Axis"));
	axis_item->setData(0,Qt::UserRole,Types::AXIS_X);
	
	parameter_item = new QTreeWidgetItem(axis_item);
	parameter_item->setData(0,Qt::UserRole,Types::PARAMETER_RX);

	parameter_widget = new TMachineParameterWidget();
	parameter_widget->Set_Title_Text(QStringLiteral("RX"));
	parameter_widget->Set_Sign_Options(angular_signs);
	parameter_widget->Enable_Expression_Editor(true);
	parameter_widget->Enable_Expressions(true);
	d_machine_parameter_tree->setItemWidget(parameter_item,0,parameter_widget);
	
	parameter_item = new QTreeWidgetItem(axis_item);
	parameter_item->setData(0,Qt::UserRole,Types::PARAMETER_RY);
	
	parameter_widget = new TMachineParameterWidget();
	parameter_widget->Set_Title_Text(QStringLiteral("RY"));
	parameter_widget->Set_Sign_Options(angular_signs);
	parameter_widget->Enable_Expression_Editor(true);
	parameter_widget->Enable_Expressions(true);
	d_machine_parameter_tree->setItemWidget(parameter_item,0,parameter_widget);
	
	parameter_item = new QTreeWidgetItem(axis_item);
	parameter_item->setData(0,Qt::UserRole,Types::PARAMETER_RZ);
	
	parameter_widget = new TMachineParameterWidget();
	parameter_widget->Set_Title_Text(QStringLiteral("RZ"));
	parameter_widget->Set_Sign_Options(angular_signs);
	parameter_widget->Enable_Expression_Editor(true);
	parameter_widget->Enable_Expressions(true);
	d_machine_parameter_tree->setItemWidget(parameter_item,0,parameter_widget);
	
	parameter_item = new QTreeWidgetItem(axis_item);
	parameter_item->setData(0,Qt::UserRole,Types::PARAMETER_SCALE);
	
	parameter_widget = new TMachineParameterWidget();
	parameter_widget->Set_Title_Text(QStringLiteral("Scale"));
	parameter_widget->Set_Sign_Options(scale_signs);
	parameter_widget->Enable_Expression_Editor(true);
	parameter_widget->Enable_Expressions(true);
	d_machine_parameter_tree->setItemWidget(parameter_item,0,parameter_widget);
	
	axis_item = new QTreeWidgetItem(d_machine_parameter_tree);
	axis_item->setText(0,QStringLiteral("Y Axis"));
	axis_item->setData(0,Qt::UserRole,Types::AXIS_Y);

	parameter_item = new QTreeWidgetItem(axis_item);
	parameter_item->setData(0,Qt::UserRole,Types::PARAMETER_RX);
	
	parameter_widget = new TMachineParameterWidget();
	parameter_widget->Set_Title_Text(QStringLiteral("RX"));
	parameter_widget->Set_Sign_Options(angular_signs);
	parameter_widget->Enable_Expression_Editor(true);
	parameter_widget->Enable_Expressions(true);
	d_machine_parameter_tree->setItemWidget(parameter_item,0,parameter_widget);
	
	parameter_item = new QTreeWidgetItem(axis_item);
	parameter_item->setData(0,Qt::UserRole,Types::PARAMETER_RY);
	
	parameter_widget = new TMachineParameterWidget();
	parameter_widget->Set_Title_Text(QStringLiteral("RY"));
	parameter_widget->Set_Sign_Options(angular_signs);
	parameter_widget->Enable_Expression_Editor(true);
	parameter_widget->Enable_Expressions(true);
	d_machine_parameter_tree->setItemWidget(parameter_item,0,parameter_widget);
	
	parameter_item = new QTreeWidgetItem(axis_item);
	parameter_item->setData(0,Qt::UserRole,Types::PARAMETER_RZ);
	
	parameter_widget = new TMachineParameterWidget();
	parameter_widget->Set_Title_Text(QStringLiteral("RZ"));
	parameter_widget->Set_Sign_Options(angular_signs);
	parameter_widget->Enable_Expression_Editor(true);
	parameter_widget->Enable_Expressions(true);
	d_machine_parameter_tree->setItemWidget(parameter_item,0,parameter_widget);
	
	parameter_item = new QTreeWidgetItem(axis_item);
	parameter_item->setData(0,Qt::UserRole,Types::PARAMETER_SCALE);
	
	parameter_widget = new TMachineParameterWidget();
	parameter_widget->Set_Title_Text(QStringLiteral("Scale"));
	parameter_widget->Set_Sign_Options(scale_signs);
	parameter_widget->Enable_Expression_Editor(true);
	parameter_widget->Enable_Expressions(true);
	d_machine_parameter_tree->setItemWidget(parameter_item,0,parameter_widget);
	
	axis_item = new QTreeWidgetItem(d_machine_parameter_tree);
	axis_item->setText(0,QStringLiteral("Z Axis"));
	axis_item->setData(0,Qt::UserRole,Types::AXIS_Z);

	parameter_item = new QTreeWidgetItem(axis_item);
	parameter_item->setData(0,Qt::UserRole,Types::PARAMETER_RX);
	
	parameter_widget = new TMachineParameterWidget();
	parameter_widget->Set_Title_Text(QStringLiteral("RX"));
	parameter_widget->Set_Sign_Options(angular_signs);
	parameter_widget->Enable_Expression_Editor(true);
	parameter_widget->Enable_Expressions(true);
	d_machine_parameter_tree->setItemWidget(parameter_item,0,parameter_widget);
	
	parameter_item = new QTreeWidgetItem(axis_item);
	parameter_item->setData(0,Qt::UserRole,Types::PARAMETER_RY);
	
	parameter_widget = new TMachineParameterWidget();
	parameter_widget->Set_Title_Text(QStringLiteral("RY"));
	parameter_widget->Set_Sign_Options(angular_signs);
	parameter_widget->Enable_Expression_Editor(true);
	parameter_widget->Enable_Expressions(true);
	d_machine_parameter_tree->setItemWidget(parameter_item,0,parameter_widget);
	
	parameter_item = new QTreeWidgetItem(axis_item);
	parameter_item->setData(0,Qt::UserRole,Types::PARAMETER_RZ);
	
	parameter_widget = new TMachineParameterWidget();
	parameter_widget->Set_Title_Text(QStringLiteral("RZ"));
	parameter_widget->Set_Sign_Options(angular_signs);
	parameter_widget->Enable_Expression_Editor(true);
	parameter_widget->Enable_Expressions(true);
	d_machine_parameter_tree->setItemWidget(parameter_item,0,parameter_widget);
	
	parameter_item = new QTreeWidgetItem(axis_item);
	parameter_item->setData(0,Qt::UserRole,Types::PARAMETER_SCALE);
	
	parameter_widget = new TMachineParameterWidget();
	parameter_widget->Set_Title_Text(QStringLiteral("Scale"));
	parameter_widget->Set_Sign_Options(scale_signs);
	parameter_widget->Enable_Expression_Editor(true);
	parameter_widget->Enable_Expressions(true);
	d_machine_parameter_tree->setItemWidget(parameter_item,0,parameter_widget);

	axis_item = new QTreeWidgetItem(d_machine_parameter_tree);
	axis_item->setText(0,QStringLiteral("Squareness"));
	axis_item->setData(0,Qt::UserRole,Types::AXIS_XYZ);

	parameter_item = new QTreeWidgetItem(axis_item);
	parameter_item->setData(0,Qt::UserRole,Types::PARAMETER_QXY);
	
	parameter_widget = new TMachineParameterWidget();
	parameter_widget->Set_Title_Text(QStringLiteral("XY"));
	parameter_widget->Set_Sign_Options(xy_squareness_signs);
	parameter_widget->Enable_Expression_Editor(false);
	parameter_widget->Enable_Expressions(false);
	d_machine_parameter_tree->setItemWidget(parameter_item,0,parameter_widget);
	
	parameter_item = new QTreeWidgetItem(axis_item);
	parameter_item->setData(0,Qt::UserRole,Types::PARAMETER_QYZ);
	
	parameter_widget = new TMachineParameterWidget();
	parameter_widget->Set_Title_Text(QStringLiteral("YZ"));
	parameter_widget->Set_Sign_Options(yz_squareness_signs);
	parameter_widget->Enable_Expression_Editor(false);
	parameter_widget->Enable_Expressions(false);
	d_machine_parameter_tree->setItemWidget(parameter_item,0,parameter_widget);
	
	parameter_item = new QTreeWidgetItem(axis_item);
	parameter_item->setData(0,Qt::UserRole,Types::PARAMETER_QZX);
	
	parameter_widget = new TMachineParameterWidget();
	parameter_widget->Set_Title_Text(QStringLiteral("ZX"));
	parameter_widget->Set_Sign_Options(zx_squareness_signs);
	parameter_widget->Enable_Expression_Editor(false);
	parameter_widget->Enable_Expressions(false);
	d_machine_parameter_tree->setItemWidget(parameter_item,0,parameter_widget);

	
	axis_item = new QTreeWidgetItem(d_machine_parameter_tree);
	axis_item->setText(0,QStringLiteral("D Axis"));
	axis_item->setData(0,Qt::UserRole,Types::AXIS_D);
	
	parameter_item = new QTreeWidgetItem(axis_item);
	parameter_item->setData(0,Qt::UserRole,Types::PARAMETER_DEFLECTION);
	
	parameter_widget = new TMachineParameterWidget();
	parameter_widget->Set_Title_Text(QStringLiteral("Deflection"));
	parameter_widget->Set_Sign_Options(angular_signs);
	parameter_widget->Enable_Expression_Editor(false);
	parameter_widget->Enable_Expressions(false);
	d_machine_parameter_tree->setItemWidget(parameter_item,0,parameter_widget);
}

QString TMachineDialog::Generate_Unique_Name(
	const int							id,
	const QString						&suggested_name) const
{
	std::vector<TMachineItem>::const_iterator iter;
	QRegularExpression					reg_exp(QStringLiteral("^([A-Za-z_ ][A-za-z_ ]*)([0-9]{1,10})$"));
	QRegularExpressionMatch				expression_match;
	QStringList							expression_text_list;
	QString								name;
	QString								name_prefix;
	bool								unique;
	int									index(0);
	
	if(suggested_name.length())
	{
		name = suggested_name.simplified();
		
		expression_match = reg_exp.match(name);
		if(expression_match.hasMatch())
		{
			expression_text_list = expression_match.capturedTexts();
			
			if(expression_text_list.size() > 0)
			{
				name_prefix = expression_text_list[1].trimmed();
			}
			
			if(expression_text_list.size() > 1)
			{
				index = expression_text_list[2].toInt();
			}
		}
		else
		{
			name_prefix = name;
		}
	}
	else
	{
		name_prefix = QStringLiteral("Untitled");
	}
	
	do
	{
		unique = true;
		
		if(index)
		{
			name = QString("%1 %2").arg(name_prefix).arg(index);
		}
		else
		{
			name = QString("%1").arg(name_prefix);
		}
		
		for(iter = d_machine_items.begin();iter != d_machine_items.end();++iter)
		{
			if(name.compare((*iter).name,Qt::CaseInsensitive) == 0)
			{
				if((*iter).id != id)
				{
					unique = false;
					break;
				}
			}
		}
		
		index++;
	}
	while(!unique);
	
	return name;
}

bool TMachineDialog::Name_Valid(
	const QString						&name) const
{
	QString								test_name(name.simplified());
	std::vector<TMachineItem>::const_iterator iter;

	if(test_name.length() == 0)
	{
		return false;
	}
	
	for(iter = d_machine_items.begin();iter != d_machine_items.end();++iter)
	{
		if(0 == test_name.compare((*iter).name,Qt::CaseInsensitive))
		{
			return false;
		}
	}

	return true;
}

void TMachineDialog::Update_Editor(void)
{
	std::vector<TMachineItem>::iterator iter;
	std::vector<TMachineAxisItem>::iterator axis_iter;
	std::vector<TMachineParameterItem>::iterator parameter_iter;
	QTreeWidgetItem						*tree_item;
	QTreeWidgetItem						*child_tree_item;
	TMachineParameterWidget				*parameter_widget;
	bool								is_horizontal_arm(false);
	int									count;
	int									cntr;
	int									child_count;
	int									child_cntr;
	int									axis;
	int									parameter;
	
	if(d_active_machine_id < 0)
	{
		return;
	}
	
	for(iter = d_machine_items.begin();iter != d_machine_items.end();++iter)
	{
		if((*iter).id == d_active_machine_id)
		{
			switch((*iter).kinematic)
			{
				case Types::KINEMATIC_YXZ:
					d_kinematic_yxz_radio->setChecked(true);
					break;
					
				case Types::KINEMATIC_XYZ:
					d_kinematic_xyz_radio->setChecked(true);
					break;
					
				case Types::KINEMATIC_XZY:
					d_kinematic_xzy_radio->setChecked(true);
					is_horizontal_arm = true;
					break;
					
				case Types::KINEMATIC_YZX:
					d_kinematic_yzx_radio->setChecked(true);
					is_horizontal_arm = true;
					break;
			}
			
			d_min_x_edit->setText(QString("%1").arg((*iter).axis_min.x,0,'f',4));
			d_min_y_edit->setText(QString("%1").arg((*iter).axis_min.y,0,'f',4));
			d_min_z_edit->setText(QString("%1").arg((*iter).axis_min.z,0,'f',4));
			
			d_max_x_edit->setText(QString("%1").arg((*iter).axis_max.x,0,'f',4));
			d_max_y_edit->setText(QString("%1").arg((*iter).axis_max.y,0,'f',4));
			d_max_z_edit->setText(QString("%1").arg((*iter).axis_max.z,0,'f',4));
			
			d_math_origin_x_edit->setText(QString("%1").arg((*iter).math_origin.x,0,'f',4));
			d_math_origin_y_edit->setText(QString("%1").arg((*iter).math_origin.y,0,'f',4));
			d_math_origin_z_edit->setText(QString("%1").arg((*iter).math_origin.z,0,'f',4));
			
			d_mechanical_origin_x_edit->setText(QString("%1").arg((*iter).mechanical_origin.x,0,'f',4));
			d_mechanical_origin_y_edit->setText(QString("%1").arg((*iter).mechanical_origin.y,0,'f',4));
			d_mechanical_origin_z_edit->setText(QString("%1").arg((*iter).mechanical_origin.z,0,'f',4));
			
			// parameter values
			count = d_machine_parameter_tree->topLevelItemCount();
			
			for(cntr = 0;cntr < count;++cntr)
			{
				tree_item = d_machine_parameter_tree->topLevelItem(cntr);
				axis = tree_item->data(0,Qt::UserRole).toInt();
				
				child_count = tree_item->childCount();
				
				for(child_cntr = 0;child_cntr < child_count;++child_cntr)
				{
					child_tree_item = tree_item->child(child_cntr);
					parameter = child_tree_item->data(0,Qt::UserRole).toInt();
					
					parameter_widget = qobject_cast<TMachineParameterWidget*>(d_machine_parameter_tree->itemWidget(child_tree_item,0));
					
					if(parameter_widget)
					{
						for(axis_iter = (*iter).axis_items.begin();axis_iter != (*iter).axis_items.end();++axis_iter)
						{
							if((*axis_iter).axis == axis)
							{
								for(parameter_iter = (*axis_iter).parameter_items.begin();parameter_iter != (*axis_iter).parameter_items.end();++parameter_iter)
								{
									if((*parameter_iter).parameter == parameter)
									{
										parameter_widget->Set_Option_Enabled((*parameter_iter).enabled);
										parameter_widget->Set_Sign_Option_Index((*parameter_iter).option_index);
										parameter_widget->Set_Expression((*parameter_iter).expression);
										break;
									}
								}
								
								if(axis == Types::AXIS_D)
								{
									parameter_widget->setEnabled(is_horizontal_arm);
									parameter_widget->Set_Option_Enabled(is_horizontal_arm & parameter_widget->Is_Option_Enabled());
									
									if(!is_horizontal_arm)
									{
										d_machine_parameter_tree->collapseItem(tree_item);
									}
									else
									{
										d_machine_parameter_tree->expandItem(tree_item);
									}
								}
								
								break;
							}
						}
					}
				}
			}
		}
	}
}

void TMachineDialog::Update_Machine(void)
{
	std::vector<TMachineItem>::iterator iter;
	std::vector<TMachineAxisItem>::iterator axis_iter;
	std::vector<TMachineParameterItem>::iterator parameter_iter;
	QTreeWidgetItem						*tree_item;
	QTreeWidgetItem						*child_tree_item;
	TMachineParameterWidget				*parameter_widget;
	
	double								dval;
	int									count;
	int									cntr;
	int									child_count;
	int									child_cntr;
	int									axis;
	int									parameter;
	
	if(d_active_machine_id < 0)
	{
		return;
	}
	
	for(iter = d_machine_items.begin();iter != d_machine_items.end();++iter)
	{
		if((*iter).id == d_active_machine_id)
		{
			// kinematic
			if(d_kinematic_yxz_radio->isChecked())
			{
				(*iter).kinematic = Types::KINEMATIC_YXZ;
			}
			else if(d_kinematic_xyz_radio->isChecked())
			{
				(*iter).kinematic = Types::KINEMATIC_XYZ;
			}
			else if(d_kinematic_xzy_radio->isChecked())
			{
				(*iter).kinematic = Types::KINEMATIC_XZY;
			}
			else if(d_kinematic_yzx_radio->isChecked())
			{
				(*iter).kinematic = Types::KINEMATIC_YZX;
			}
						
			// axis
			(*iter).axis_min.x = d_min_x_edit->text().toDouble();
			(*iter).axis_min.y = d_min_y_edit->text().toDouble();
			(*iter).axis_min.z = d_min_z_edit->text().toDouble();
			
			(*iter).axis_max.x = d_max_x_edit->text().toDouble();
			(*iter).axis_max.y = d_max_y_edit->text().toDouble();
			(*iter).axis_max.z = d_max_z_edit->text().toDouble();
			
			(*iter).math_origin.x = d_math_origin_x_edit->text().toDouble();
			(*iter).math_origin.y = d_math_origin_y_edit->text().toDouble();
			(*iter).math_origin.z = d_math_origin_z_edit->text().toDouble();
			
			(*iter).mechanical_origin.x = d_mechanical_origin_x_edit->text().toDouble();
			(*iter).mechanical_origin.y = d_mechanical_origin_y_edit->text().toDouble();
			(*iter).mechanical_origin.z = d_mechanical_origin_z_edit->text().toDouble();
			
			// check max min values
			if((*iter).axis_max.x < (*iter).axis_min.x)
			{
				dval = (*iter).axis_max.x;
				(*iter).axis_max.x = (*iter).axis_min.x;
				(*iter).axis_min.x = dval;
			}
			
			if((*iter).axis_max.y < (*iter).axis_min.y)
			{
				dval = (*iter).axis_max.y;
				(*iter).axis_max.y = (*iter).axis_min.y;
				(*iter).axis_min.y = dval;
			}
			
			if((*iter).axis_max.z < (*iter).axis_min.z)
			{
				dval = (*iter).axis_max.z;
				(*iter).axis_max.z = (*iter).axis_min.z;
				(*iter).axis_min.z = dval;
			}

			
			// parameter values
			count = d_machine_parameter_tree->topLevelItemCount();	// categories X,Y,Z,Squareness, D
			
			for(cntr = 0;cntr < count;++cntr)
			{
				tree_item = d_machine_parameter_tree->topLevelItem(cntr);
				axis = tree_item->data(0,Qt::UserRole).toInt();

				child_count = tree_item->childCount();
				
				for(child_cntr = 0;child_cntr < child_count;++child_cntr)
				{
					child_tree_item = tree_item->child(child_cntr);
					parameter = child_tree_item->data(0,Qt::UserRole).toInt();
					
					parameter_widget = qobject_cast<TMachineParameterWidget*>(d_machine_parameter_tree->itemWidget(child_tree_item,0));
					
					if(parameter_widget)
					{
						for(axis_iter = (*iter).axis_items.begin();axis_iter != (*iter).axis_items.end();++axis_iter)
						{
							if((*axis_iter).axis == axis)
							{
								for(parameter_iter = (*axis_iter).parameter_items.begin();parameter_iter != (*axis_iter).parameter_items.end();++parameter_iter)
								{
									if((*parameter_iter).parameter == parameter)
									{
										(*parameter_iter).enabled = parameter_widget->Is_Option_Enabled();
										(*parameter_iter).option_index = parameter_widget->Sign_Option_Index();
										(*parameter_iter).expression = parameter_widget->Expression();
										break;
									}
								}
								
								break;
							}
						}
					}
				}
			}
		}
	}
}

bool TMachineDialog::Save(void) const
{
	TXmlFile							xmlfile;
	QDomElement							en1;
	QDomElement							en2;
	QDomElement							en3;
	QDomElement							en4;
	QDomElement							en5;
	QDomElement							en6;
	std::vector<TMachineItem>::const_iterator iter;
	std::vector<TMachineAxisItem>::const_iterator axis_iter;
	std::vector<TMachineParameterItem>::const_iterator parameter_iter;
	QString								text;
	
	try
	{
		xmlfile.Set_File_Name(QString("%1/machines.xml").arg(d_data_path));
		xmlfile.Reset("Machines",2);
		
		for(iter = d_machine_items.begin();iter != d_machine_items.end();++iter)
		{
			en1 = xmlfile.Create_Node(xmlfile.Root_Node(),QStringLiteral("Machine_Item"));
			
			en2 = xmlfile.Create_Node(&en1,QStringLiteral("Name"));
			xmlfile.Write_Text_Node(&en2,(*iter).name);
			
			en2 = xmlfile.Create_Node(&en1,QStringLiteral("Axis_Minimum"));
			xmlfile.Write_Text_Node(&en2,QString("%1,%2,%3").arg((*iter).axis_min.x,0,'f',4).arg((*iter).axis_min.y,0,'f',4).arg((*iter).axis_min.z,0,'f',4));
			
			en2 = xmlfile.Create_Node(&en1,QStringLiteral("Axis_Maximum"));
			xmlfile.Write_Text_Node(&en2,QString("%1,%2,%3").arg((*iter).axis_max.x,0,'f',4).arg((*iter).axis_max.y,0,'f',4).arg((*iter).axis_max.z,0,'f',4));
			
			en2 = xmlfile.Create_Node(&en1,QStringLiteral("Math_Origin"));
			xmlfile.Write_Text_Node(&en2,QString("%1,%2,%3").arg((*iter).math_origin.x,0,'f',4).arg((*iter).math_origin.y,0,'f',4).arg((*iter).math_origin.z,0,'f',4));
			
			en2 = xmlfile.Create_Node(&en1,QStringLiteral("Mechanical_Origin"));
			xmlfile.Write_Text_Node(&en2,QString("%1,%2,%3").arg((*iter).mechanical_origin.x,0,'f',4).arg((*iter).mechanical_origin.y,0,'f',4).arg((*iter).mechanical_origin.z,0,'f',4));
			
			en2 = xmlfile.Create_Node(&en1,QStringLiteral("Kinematic"));
			switch((*iter).kinematic)
			{
				case Types::KINEMATIC_XYZ:
					xmlfile.Write_Text_Node(&en2,QStringLiteral("XYZ"));
					break;
					
				case Types::KINEMATIC_YXZ:
				default:
					xmlfile.Write_Text_Node(&en2,QStringLiteral("YXZ"));
					break;
					
				case Types::KINEMATIC_XZY:
					xmlfile.Write_Text_Node(&en2,QStringLiteral("XZY"));
					break;
					
				case Types::KINEMATIC_YZX:
					xmlfile.Write_Text_Node(&en2,QStringLiteral("YZX"));
					break;
			}
			
			en2 = xmlfile.Create_Node(&en1,QStringLiteral("Machine_Parameters"));
			
			for(axis_iter = (*iter).axis_items.begin();axis_iter != (*iter).axis_items.end();++axis_iter)
			{
				switch((*axis_iter).axis)
				{
					case Types::AXIS_X:
						text = QStringLiteral("Axis_X");
						break;
						
					case Types::AXIS_Y:
						text = QStringLiteral("Axis_Y");
						break;
						
					case Types::AXIS_Z:
						text = QStringLiteral("Axis_Z");
						break;
						
					case Types::AXIS_D:
						text = QStringLiteral("Axis_D");
						break;
						
					case Types::AXIS_XYZ:
						text = QStringLiteral("Squareness");
						break;
						
					default:
						assert(false);	// should never get here
				}
								
				en3 = xmlfile.Create_Node(&en2,text);
				
				for(parameter_iter = (*axis_iter).parameter_items.begin();parameter_iter != (*axis_iter).parameter_items.end();++parameter_iter)
				{
					text.clear();
					
					switch((*parameter_iter).parameter)
					{
						case Types::PARAMETER_RX:
							text = QStringLiteral("Parameter_Rx");
							break;
							
						case Types::PARAMETER_RY:
							text = QStringLiteral("Parameter_Ry");
							break;
							
						case Types::PARAMETER_RZ:
							text = QStringLiteral("Parameter_Rz");
							break;
							
						case Types::PARAMETER_SCALE:
							text = QStringLiteral("Parameter_Scale");
							break;
							
						case Types::PARAMETER_DEFLECTION:
							text = QStringLiteral("Parameter_Deflection");
							break;
							
						case Types::PARAMETER_QXY:
							text = QStringLiteral("Parameter_QXY");
							break;
							
						case Types::PARAMETER_QYZ:
							text = QStringLiteral("Parameter_QYZ");
							break;
							
						case Types::PARAMETER_QZX:
							text = QStringLiteral("Parameter_QZX");
							break;
					}
					
					if(text.length())
					{
						en4 = xmlfile.Create_Node(&en3,text);

						en5 = xmlfile.Create_Node(&en4,QStringLiteral("Enabled"));
						if((*parameter_iter).enabled)
						{
							xmlfile.Write_Text_Node(&en5,QStringLiteral("True"));
						}
						else
						{
							xmlfile.Write_Text_Node(&en5,QStringLiteral("False"));
						}
						
						en5 = xmlfile.Create_Node(&en4,QStringLiteral("Option_State"));
						xmlfile.Write_Text_Node(&en5,QString("%1").arg((*parameter_iter).option_index));

						en5 = xmlfile.Create_Node(&en4,QStringLiteral("Expression"));
						
						text = (*parameter_iter).expression.simplified();
						
						if(text.length())
						{
							xmlfile.Write_Text_Node(&en5,text);
						}
					}
				}
			}
		}
		xmlfile.Write_File();
	}
	catch(TXmlFileException xml_exception)
	{
		d_last_error = QString("ERR:  Saving machines exception: %1").arg(xml_exception.ErrorText());
		return false;
	}
	
	return true;
}

void TMachineDialog::Load_Version1(
	TXmlFile							* const xml_file)
{
	QDomElement							en1;
	QDomElement							en2;
	QDomElement							en3;
	QDomElement							en4;
	QString								text;
	TMachineDialog::TMachineItem		machine_item;
	TMachineDialog::TMachineAxisItem	axis_item_x;
	TMachineDialog::TMachineAxisItem	axis_item_y;
	TMachineDialog::TMachineAxisItem	axis_item_z;
	TMachineDialog::TMachineAxisItem	axis_item_xyz;
	TMachineDialog::TMachineAxisItem	axis_item_d;
	TMachineDialog::TMachineParameterItem	parameter_item;
	int									active_axis;
	
	assert(xml_file);
	
	d_machine_items.clear();
	
	machine_item.id = 0;
	
	// initialize axis items
	axis_item_x.axis = Types::AXIS_X;
	axis_item_y.axis = Types::AXIS_Y;
	axis_item_z.axis = Types::AXIS_Z;
	axis_item_xyz.axis = Types::AXIS_XYZ;
	axis_item_d.axis = Types::AXIS_D;

	en1 = xml_file->Get_Node(xml_file->Root_Node(),QStringLiteral("Machine_Item"),true);
	while(!en1.isNull())
	{
		machine_item.axis_items.clear();
		
		axis_item_x.parameter_items.clear();
		axis_item_y.parameter_items.clear();
		axis_item_z.parameter_items.clear();
		axis_item_xyz.parameter_items.clear();
		axis_item_d.parameter_items.clear();

		en2 = xml_file->Get_Node(&en1,QStringLiteral("Name"));
		machine_item.name = xml_file->Read_Text_Node(&en2);

		en2 = xml_file->Get_Node(&en1,QStringLiteral("Map_Minimum_X"));
		machine_item.axis_min.x = xml_file->Read_Text_Node(&en2).toDouble();
		
		en2 = xml_file->Get_Node(&en1,QStringLiteral("Map_Minimum_Y"));
		machine_item.axis_min.y = xml_file->Read_Text_Node(&en2).toDouble();
		
		en2 = xml_file->Get_Node(&en1,QStringLiteral("Map_Minimum_Z"));
		machine_item.axis_min.z = xml_file->Read_Text_Node(&en2).toDouble();
		
		en2 = xml_file->Get_Node(&en1,QStringLiteral("Map_Maximum_X"));
		machine_item.axis_max.x = xml_file->Read_Text_Node(&en2).toDouble();
		
		en2 = xml_file->Get_Node(&en1,QStringLiteral("Map_Maximum_Y"));
		machine_item.axis_max.y = xml_file->Read_Text_Node(&en2).toDouble();
		
		en2 = xml_file->Get_Node(&en1,QStringLiteral("Map_Maximum_Z"));
		machine_item.axis_max.z = xml_file->Read_Text_Node(&en2).toDouble();
		
		en2 = xml_file->Get_Node(&en1,QStringLiteral("Map_Origin_X"));
		machine_item.math_origin.x = xml_file->Read_Text_Node(&en2).toDouble();
		
		en2 = xml_file->Get_Node(&en1,QStringLiteral("Map_Origin_Y"));
		machine_item.math_origin.y = xml_file->Read_Text_Node(&en2).toDouble();
		
		en2 = xml_file->Get_Node(&en1,QStringLiteral("Map_Origin_Z"));
		machine_item.math_origin.z = xml_file->Read_Text_Node(&en2).toDouble();
		
		machine_item.mechanical_origin.Set(0,0,0);
		
		en2 = xml_file->Get_Node(&en1,QStringLiteral("Kinematic"));
		text = xml_file->Read_Text_Node(&en2);
		
		if(0 == text.compare(QStringLiteral("XYZ"),Qt::CaseInsensitive))
		{
			machine_item.kinematic = Types::KINEMATIC_XYZ;
		}
		else if(0 == text.compare(QStringLiteral("YXZ"),Qt::CaseInsensitive))
		{
			machine_item.kinematic = Types::KINEMATIC_YXZ;
		}
		else if(0 == text.compare(QStringLiteral("XZY"),Qt::CaseInsensitive))
		{
			machine_item.kinematic = Types::KINEMATIC_XZY;
		}
		else if(0 == text.compare(QStringLiteral("YZX"),Qt::CaseInsensitive))
		{
			machine_item.kinematic = Types::KINEMATIC_YZX;
		}
		else
		{
			throw TXmlFileException(QStringLiteral("Kinematic order not recognized"));
		}
		
		en2 = xml_file->Get_Node(&en1,QStringLiteral("Machine_Parameters"));
		en3 = xml_file->Get_Node(&en2,QStringLiteral("Machine_Parameter_Entry"),true);
		
		while(!en3.isNull())
		{
			en4 = xml_file->Get_Node(&en3,QStringLiteral("Axis"));
			text = xml_file->Read_Text_Node(&en4);
			
			if(0 == text.compare(QStringLiteral("X"),Qt::CaseInsensitive))
			{
				active_axis = Types::AXIS_X;
			}
			else if(0 == text.compare(QStringLiteral("Y"),Qt::CaseInsensitive))
			{
				active_axis = Types::AXIS_Y;
			}
			else if(0 == text.compare(QStringLiteral("Z"),Qt::CaseInsensitive))
			{
				active_axis = Types::AXIS_Z;
			}
			else if(0 == text.compare(QStringLiteral("D"),Qt::CaseInsensitive))
			{
				active_axis = Types::AXIS_D;
			}
			else
			{
				throw TXmlFileException(QStringLiteral("Axis not recognized"));
			}

			en4 = xml_file->Get_Node(&en3,QStringLiteral("Parameter"));
			text = xml_file->Read_Text_Node(&en4);

			if(0 == text.compare(QStringLiteral("Rx"),Qt::CaseInsensitive))
			{
				parameter_item.parameter = Types::PARAMETER_RX;
			}
			else if(0 == text.compare(QStringLiteral("Ry"),Qt::CaseInsensitive))
			{
				parameter_item.parameter = Types::PARAMETER_RY;
			}
			else if(0 == text.compare(QStringLiteral("Rz"),Qt::CaseInsensitive))
			{
				parameter_item.parameter = Types::PARAMETER_RZ;
			}
			else if(0 == text.compare(QStringLiteral("Scale"),Qt::CaseInsensitive))
			{
				parameter_item.parameter = Types::PARAMETER_SCALE;
			}
			else if(0 == text.compare(QStringLiteral("Rt"),Qt::CaseInsensitive))
			{
				parameter_item.parameter = Types::PARAMETER_DEFLECTION;
			}
			else if(0 == text.compare(QStringLiteral("QXY"),Qt::CaseInsensitive))
			{
				parameter_item.parameter = Types::PARAMETER_QXY;
			}
			else if(0 == text.compare(QStringLiteral("QYZ"),Qt::CaseInsensitive))
			{
				parameter_item.parameter = Types::PARAMETER_QYZ;
			}
			else if(0 == text.compare(QStringLiteral("QZX"),Qt::CaseInsensitive))
			{
				parameter_item.parameter = Types::PARAMETER_QZX;
			}
			else
			{
				throw TXmlFileException(QStringLiteral("Parameter not recognized"));
			}

			en4 = xml_file->Get_Node(&en3,QStringLiteral("Enabled"));
			text = xml_file->Read_Text_Node(&en4);
			
			if(0 == text.compare(QStringLiteral("True"),Qt::CaseInsensitive))
			{
				parameter_item.enabled = true;
			}
			else
			{
				parameter_item.enabled = false;
			}
			
			en4 = xml_file->Get_Node(&en3,QStringLiteral("Option_State"));
			parameter_item.option_index = xml_file->Read_Text_Node(&en4).toInt();

			en4 = xml_file->Get_Node(&en3,QStringLiteral("Expression"),true);
			if(en4.isNull())
			{
				parameter_item.expression.clear();
			}
			else
			{
				text = xml_file->Read_Text_Node(&en4);

				if(text.length())
				{
					parameter_item.expression = xml_file->Read_Text_Node(&en4);
				}
				else
				{
					parameter_item.expression.clear();
				}
			}
			
			switch(active_axis)
			{
				case Types::AXIS_X:
					
					if(parameter_item.parameter == Types::PARAMETER_QXY)
					{
						axis_item_xyz.parameter_items.push_back(parameter_item);
					}
					else
					{
						axis_item_x.parameter_items.push_back(parameter_item);
					}
					
					break;
					
				case Types::AXIS_Y:
					
					if(parameter_item.parameter == Types::PARAMETER_QYZ)
					{
						axis_item_xyz.parameter_items.push_back(parameter_item);
					}
					else
					{
						axis_item_y.parameter_items.push_back(parameter_item);
					}
					
					break;
					
				case Types::AXIS_Z:
					
					if(parameter_item.parameter == Types::PARAMETER_QZX)
					{
						axis_item_xyz.parameter_items.push_back(parameter_item);
					}
					else
					{
						axis_item_z.parameter_items.push_back(parameter_item);
					}
					
					break;
					
				case Types::AXIS_D:
					axis_item_d.parameter_items.push_back(parameter_item);
					break;
					
			}
			
			en3 = xml_file->Get_Sibling_Node(&en3,QStringLiteral("Machine_Parameter_Entry"),true);
		}

		machine_item.axis_items.push_back(axis_item_x);
		machine_item.axis_items.push_back(axis_item_y);
		machine_item.axis_items.push_back(axis_item_z);
		machine_item.axis_items.push_back(axis_item_xyz);
		machine_item.axis_items.push_back(axis_item_d);
		
		d_machine_items.push_back(machine_item);
		
		machine_item.id++;
		
		en1 = xml_file->Get_Sibling_Node(&en1,QStringLiteral("Machine_Item"),true);
	}
 
}


void TMachineDialog::Load_Version2(
	TXmlFile							* const xml_file)
{
	QDomElement							en1;
	QDomElement							en2;
	QDomElement							en3;
	QDomElement							en4;
	QDomElement							en5;
	QStringList							list;
	QString								text;
	TMachineDialog::TMachineItem		machine_item;
	TMachineDialog::TMachineAxisItem	axis_item;
	TMachineDialog::TMachineParameterItem	parameter_item;
	
	assert(xml_file);
	
	d_machine_items.clear();
	
	machine_item.id = 0;

	en1 = xml_file->Get_Node(xml_file->Root_Node(),QStringLiteral("Machine_Item"),true);
	while(!en1.isNull())
	{
		en2 = xml_file->Get_Node(&en1,QStringLiteral("Name"));
		machine_item.name = xml_file->Read_Text_Node(&en2);

		en2 = xml_file->Get_Node(&en1,QStringLiteral("Axis_Minimum"));
		text = xml_file->Read_Text_Node(&en2);
		list = text.split(',');
		if(list.size() == 3)
		{
			machine_item.axis_min.x = list[0].toDouble();
			machine_item.axis_min.y = list[1].toDouble();
			machine_item.axis_min.z = list[2].toDouble();
		}
		else
		{
			machine_item.axis_min.Set(0,0,-1000);
		}

		en2 = xml_file->Get_Node(&en1,QStringLiteral("Axis_Maximum"));
		text = xml_file->Read_Text_Node(&en2);
		list = text.split(',');
		if(list.size() == 3)
		{
			machine_item.axis_max.x = list[0].toDouble();
			machine_item.axis_max.y = list[1].toDouble();
			machine_item.axis_max.z = list[2].toDouble();
		}
		else
		{
			machine_item.axis_max.Set(1000,1000,0);
		}

		en2 = xml_file->Get_Node(&en1,QStringLiteral("Math_Origin"));
		text = xml_file->Read_Text_Node(&en2);
		list = text.split(',');
		if(list.size() == 3)
		{
			machine_item.math_origin.x = list[0].toDouble();
			machine_item.math_origin.y = list[1].toDouble();
			machine_item.math_origin.z = list[2].toDouble();
		}
		else
		{
			machine_item.math_origin.Set(0,0,0);
		}
		
		en2 = xml_file->Get_Node(&en1,QStringLiteral("Mechanical_Origin"));
		text = xml_file->Read_Text_Node(&en2);
		list = text.split(',');
		if(list.size() == 3)
		{
			machine_item.mechanical_origin.x = list[0].toDouble();
			machine_item.mechanical_origin.y = list[1].toDouble();
			machine_item.mechanical_origin.z = list[2].toDouble();
		}
		else
		{
			machine_item.mechanical_origin.Set(0,0,0);
		}

		en2 = xml_file->Get_Node(&en1,QStringLiteral("Kinematic"));
		text = xml_file->Read_Text_Node(&en2);

		if(0 == text.compare(QStringLiteral("XYZ"),Qt::CaseInsensitive))
		{
			machine_item.kinematic = Types::KINEMATIC_XYZ;
		}
		else if(0 == text.compare(QStringLiteral("YXZ"),Qt::CaseInsensitive))
		{
			machine_item.kinematic = Types::KINEMATIC_YXZ;
		}
		else if(0 == text.compare(QStringLiteral("XZY"),Qt::CaseInsensitive))
		{
			machine_item.kinematic = Types::KINEMATIC_XZY;
		}
		else if(0 == text.compare(QStringLiteral("YZX"),Qt::CaseInsensitive))
		{
			machine_item.kinematic = Types::KINEMATIC_YZX;
		}
		else
		{
			throw TXmlFileException(QStringLiteral("Kinematic order not recognized"));
		}

		en2 = xml_file->Get_Node(&en1,QStringLiteral("Machine_Parameters"));
		machine_item.axis_items.clear();

		
// X AXIS
		en3 = xml_file->Get_Node(&en2,QStringLiteral("Axis_X"));
		 
		axis_item.parameter_items.clear();
		axis_item.axis = Types::AXIS_X;

// x axis, rx
		en4 = xml_file->Get_Node(&en3,QStringLiteral("Parameter_Rx"));
		parameter_item.parameter = Types::PARAMETER_RX;

		en5 = xml_file->Get_Node(&en4,QStringLiteral("Enabled"));
		text = xml_file->Read_Text_Node(&en5);

		if(0 == text.compare(QStringLiteral("True"),Qt::CaseInsensitive))
		{
			parameter_item.enabled = true;
		}
		else
		{
			parameter_item.enabled = false;
		}

		en5 = xml_file->Get_Node(&en4,QStringLiteral("Option_State"));
		parameter_item.option_index = xml_file->Read_Text_Node(&en5).toInt();

		en5 = xml_file->Get_Node(&en4,QStringLiteral("Expression"));
		parameter_item.expression = xml_file->Read_Text_Node(&en5);

		axis_item.parameter_items.push_back(parameter_item);

		
// x axis, ry
		en4 = xml_file->Get_Node(&en3,QStringLiteral("Parameter_Ry"));
		parameter_item.parameter = Types::PARAMETER_RY;
		
		en5 = xml_file->Get_Node(&en4,QStringLiteral("Enabled"));
		text = xml_file->Read_Text_Node(&en5);
		
		if(0 == text.compare(QStringLiteral("True"),Qt::CaseInsensitive))
		{
			parameter_item.enabled = true;
		}
		else
		{
			parameter_item.enabled = false;
		}
		
		en5 = xml_file->Get_Node(&en4,QStringLiteral("Option_State"));
		parameter_item.option_index = xml_file->Read_Text_Node(&en5).toInt();
		
		en5 = xml_file->Get_Node(&en4,QStringLiteral("Expression"));
		parameter_item.expression = xml_file->Read_Text_Node(&en5);
		
		axis_item.parameter_items.push_back(parameter_item);
		
// x axis, rz
		en4 = xml_file->Get_Node(&en3,QStringLiteral("Parameter_Rz"));
		parameter_item.parameter = Types::PARAMETER_RZ;
		
		en5 = xml_file->Get_Node(&en4,QStringLiteral("Enabled"));
		text = xml_file->Read_Text_Node(&en5);
		
		if(0 == text.compare(QStringLiteral("True"),Qt::CaseInsensitive))
		{
			parameter_item.enabled = true;
		}
		else
		{
			parameter_item.enabled = false;
		}
		
		en5 = xml_file->Get_Node(&en4,QStringLiteral("Option_State"));
		parameter_item.option_index = xml_file->Read_Text_Node(&en5).toInt();
		
		en5 = xml_file->Get_Node(&en4,QStringLiteral("Expression"));
		parameter_item.expression = xml_file->Read_Text_Node(&en5);
		
		axis_item.parameter_items.push_back(parameter_item);
		
// x axis, scale
		en4 = xml_file->Get_Node(&en3,QStringLiteral("Parameter_Scale"));
		parameter_item.parameter = Types::PARAMETER_SCALE;
		
		en5 = xml_file->Get_Node(&en4,QStringLiteral("Enabled"));
		text = xml_file->Read_Text_Node(&en5);
		
		if(0 == text.compare(QStringLiteral("True"),Qt::CaseInsensitive))
		{
			parameter_item.enabled = true;
		}
		else
		{
			parameter_item.enabled = false;
		}
		
		en5 = xml_file->Get_Node(&en4,QStringLiteral("Option_State"));
		parameter_item.option_index = xml_file->Read_Text_Node(&en5).toInt();
		
		en5 = xml_file->Get_Node(&en4,QStringLiteral("Expression"));
		parameter_item.expression = xml_file->Read_Text_Node(&en5);
		
		axis_item.parameter_items.push_back(parameter_item);
		
		machine_item.axis_items.push_back(axis_item);
		axis_item.parameter_items.clear();
		
// Y AXIS
		en3 = xml_file->Get_Node(&en2,QStringLiteral("Axis_Y"));
		
		axis_item.parameter_items.clear();
		axis_item.axis = Types::AXIS_Y;
		
// y axis, rx
		en4 = xml_file->Get_Node(&en3,QStringLiteral("Parameter_Rx"));
		parameter_item.parameter = Types::PARAMETER_RX;
		
		en5 = xml_file->Get_Node(&en4,QStringLiteral("Enabled"));
		text = xml_file->Read_Text_Node(&en5);
		
		if(0 == text.compare(QStringLiteral("True"),Qt::CaseInsensitive))
		{
			parameter_item.enabled = true;
		}
		else
		{
			parameter_item.enabled = false;
		}
		
		en5 = xml_file->Get_Node(&en4,QStringLiteral("Option_State"));
		parameter_item.option_index = xml_file->Read_Text_Node(&en5).toInt();
		
		en5 = xml_file->Get_Node(&en4,QStringLiteral("Expression"));
		parameter_item.expression = xml_file->Read_Text_Node(&en5);
		
		axis_item.parameter_items.push_back(parameter_item);
		
		
// y axis, ry
		en4 = xml_file->Get_Node(&en3,QStringLiteral("Parameter_Ry"));
		parameter_item.parameter = Types::PARAMETER_RY;
		
		en5 = xml_file->Get_Node(&en4,QStringLiteral("Enabled"));
		text = xml_file->Read_Text_Node(&en5);
		
		if(0 == text.compare(QStringLiteral("True"),Qt::CaseInsensitive))
		{
			parameter_item.enabled = true;
		}
		else
		{
			parameter_item.enabled = false;
		}
		
		en5 = xml_file->Get_Node(&en4,QStringLiteral("Option_State"));
		parameter_item.option_index = xml_file->Read_Text_Node(&en5).toInt();
		
		en5 = xml_file->Get_Node(&en4,QStringLiteral("Expression"));
		parameter_item.expression = xml_file->Read_Text_Node(&en5);
		
		axis_item.parameter_items.push_back(parameter_item);
		
// y axis, rz
		en4 = xml_file->Get_Node(&en3,QStringLiteral("Parameter_Rz"));
		parameter_item.parameter = Types::PARAMETER_RZ;
		
		en5 = xml_file->Get_Node(&en4,QStringLiteral("Enabled"));
		text = xml_file->Read_Text_Node(&en5);
		
		if(0 == text.compare(QStringLiteral("True"),Qt::CaseInsensitive))
		{
			parameter_item.enabled = true;
		}
		else
		{
			parameter_item.enabled = false;
		}
		
		en5 = xml_file->Get_Node(&en4,QStringLiteral("Option_State"));
		parameter_item.option_index = xml_file->Read_Text_Node(&en5).toInt();
		
		en5 = xml_file->Get_Node(&en4,QStringLiteral("Expression"));
		parameter_item.expression = xml_file->Read_Text_Node(&en5);
		
		axis_item.parameter_items.push_back(parameter_item);
		
// y axis, scale
		en4 = xml_file->Get_Node(&en3,QStringLiteral("Parameter_Scale"));
		parameter_item.parameter = Types::PARAMETER_SCALE;
		
		en5 = xml_file->Get_Node(&en4,QStringLiteral("Enabled"));
		text = xml_file->Read_Text_Node(&en5);
		
		if(0 == text.compare(QStringLiteral("True"),Qt::CaseInsensitive))
		{
			parameter_item.enabled = true;
		}
		else
		{
			parameter_item.enabled = false;
		}
		
		en5 = xml_file->Get_Node(&en4,QStringLiteral("Option_State"));
		parameter_item.option_index = xml_file->Read_Text_Node(&en5).toInt();
		
		en5 = xml_file->Get_Node(&en4,QStringLiteral("Expression"));
		parameter_item.expression = xml_file->Read_Text_Node(&en5);
		
		axis_item.parameter_items.push_back(parameter_item);
		
		
		machine_item.axis_items.push_back(axis_item);
		axis_item.parameter_items.clear();

// Z AXIS
		en3 = xml_file->Get_Node(&en2,QStringLiteral("Axis_Z"));
		
		axis_item.parameter_items.clear();
		axis_item.axis = Types::AXIS_Z;
		
// z axis, rx
		en4 = xml_file->Get_Node(&en3,QStringLiteral("Parameter_Rx"));
		parameter_item.parameter = Types::PARAMETER_RX;
		
		en5 = xml_file->Get_Node(&en4,QStringLiteral("Enabled"));
		text = xml_file->Read_Text_Node(&en5);
		
		if(0 == text.compare(QStringLiteral("True"),Qt::CaseInsensitive))
		{
			parameter_item.enabled = true;
		}
		else
		{
			parameter_item.enabled = false;
		}
		
		en5 = xml_file->Get_Node(&en4,QStringLiteral("Option_State"));
		parameter_item.option_index = xml_file->Read_Text_Node(&en5).toInt();
		
		en5 = xml_file->Get_Node(&en4,QStringLiteral("Expression"));
		parameter_item.expression = xml_file->Read_Text_Node(&en5);
		
		axis_item.parameter_items.push_back(parameter_item);
		
		
// z axis, ry
		en4 = xml_file->Get_Node(&en3,QStringLiteral("Parameter_Ry"));
		parameter_item.parameter = Types::PARAMETER_RY;
		
		en5 = xml_file->Get_Node(&en4,QStringLiteral("Enabled"));
		text = xml_file->Read_Text_Node(&en5);
		
		if(0 == text.compare(QStringLiteral("True"),Qt::CaseInsensitive))
		{
			parameter_item.enabled = true;
		}
		else
		{
			parameter_item.enabled = false;
		}
		
		en5 = xml_file->Get_Node(&en4,QStringLiteral("Option_State"));
		parameter_item.option_index = xml_file->Read_Text_Node(&en5).toInt();
		
		en5 = xml_file->Get_Node(&en4,QStringLiteral("Expression"));
		parameter_item.expression = xml_file->Read_Text_Node(&en5);
		
		axis_item.parameter_items.push_back(parameter_item);
		
// z axis, rz
		en4 = xml_file->Get_Node(&en3,QStringLiteral("Parameter_Rz"));
		parameter_item.parameter = Types::PARAMETER_RZ;
		
		en5 = xml_file->Get_Node(&en4,QStringLiteral("Enabled"));
		text = xml_file->Read_Text_Node(&en5);
		
		if(0 == text.compare(QStringLiteral("True"),Qt::CaseInsensitive))
		{
			parameter_item.enabled = true;
		}
		else
		{
			parameter_item.enabled = false;
		}
		
		en5 = xml_file->Get_Node(&en4,QStringLiteral("Option_State"));
		parameter_item.option_index = xml_file->Read_Text_Node(&en5).toInt();
		
		en5 = xml_file->Get_Node(&en4,QStringLiteral("Expression"));
		parameter_item.expression = xml_file->Read_Text_Node(&en5);
		
		axis_item.parameter_items.push_back(parameter_item);
		
// z axis, scale
		en4 = xml_file->Get_Node(&en3,QStringLiteral("Parameter_Scale"));
		parameter_item.parameter = Types::PARAMETER_SCALE;
		
		en5 = xml_file->Get_Node(&en4,QStringLiteral("Enabled"));
		text = xml_file->Read_Text_Node(&en5);
		
		if(0 == text.compare(QStringLiteral("True"),Qt::CaseInsensitive))
		{
			parameter_item.enabled = true;
		}
		else
		{
			parameter_item.enabled = false;
		}
		
		en5 = xml_file->Get_Node(&en4,QStringLiteral("Option_State"));
		parameter_item.option_index = xml_file->Read_Text_Node(&en5).toInt();
		
		en5 = xml_file->Get_Node(&en4,QStringLiteral("Expression"));
		parameter_item.expression = xml_file->Read_Text_Node(&en5);
		
		axis_item.parameter_items.push_back(parameter_item);
		
		machine_item.axis_items.push_back(axis_item);
		axis_item.parameter_items.clear();

		
// D AXIS
		en3 = xml_file->Get_Node(&en2,QStringLiteral("Axis_D"));
		
		axis_item.parameter_items.clear();
		axis_item.axis = Types::AXIS_D;
		
// d axis, deflection
		en4 = xml_file->Get_Node(&en3,QStringLiteral("Parameter_Deflection"));
		parameter_item.parameter = Types::PARAMETER_DEFLECTION;
		
		en5 = xml_file->Get_Node(&en4,QStringLiteral("Enabled"));
		text = xml_file->Read_Text_Node(&en5);
		
		if(0 == text.compare(QStringLiteral("True"),Qt::CaseInsensitive))
		{
			parameter_item.enabled = true;
		}
		else
		{
			parameter_item.enabled = false;
		}
		
		en5 = xml_file->Get_Node(&en4,QStringLiteral("Option_State"));
		parameter_item.option_index = xml_file->Read_Text_Node(&en5).toInt();
		
		en5 = xml_file->Get_Node(&en4,QStringLiteral("Expression"));
		parameter_item.expression = xml_file->Read_Text_Node(&en5);
		
		axis_item.parameter_items.push_back(parameter_item);
		
		machine_item.axis_items.push_back(axis_item);
		axis_item.parameter_items.clear();
		
		
// XYZ AXIS
		en3 = xml_file->Get_Node(&en2,QStringLiteral("Squareness"));
		
		axis_item.parameter_items.clear();
		axis_item.axis = Types::AXIS_XYZ;
		
// xyz axis, qxy
		en4 = xml_file->Get_Node(&en3,QStringLiteral("Parameter_QXY"));
		parameter_item.parameter = Types::PARAMETER_QXY;
		
		en5 = xml_file->Get_Node(&en4,QStringLiteral("Enabled"));
		text = xml_file->Read_Text_Node(&en5);
		
		if(0 == text.compare(QStringLiteral("True"),Qt::CaseInsensitive))
		{
			parameter_item.enabled = true;
		}
		else
		{
			parameter_item.enabled = false;
		}
		
		en5 = xml_file->Get_Node(&en4,QStringLiteral("Option_State"));
		parameter_item.option_index = xml_file->Read_Text_Node(&en5).toInt();
		
		en5 = xml_file->Get_Node(&en4,QStringLiteral("Expression"));
		parameter_item.expression = xml_file->Read_Text_Node(&en5);
		
		axis_item.parameter_items.push_back(parameter_item);
		
// xyz axis, qyz
		en4 = xml_file->Get_Node(&en3,QStringLiteral("Parameter_QYZ"));
		parameter_item.parameter = Types::PARAMETER_QYZ;
		
		en5 = xml_file->Get_Node(&en4,QStringLiteral("Enabled"));
		text = xml_file->Read_Text_Node(&en5);
		
		if(0 == text.compare(QStringLiteral("True"),Qt::CaseInsensitive))
		{
			parameter_item.enabled = true;
		}
		else
		{
			parameter_item.enabled = false;
		}
		
		en5 = xml_file->Get_Node(&en4,QStringLiteral("Option_State"));
		parameter_item.option_index = xml_file->Read_Text_Node(&en5).toInt();
		
		en5 = xml_file->Get_Node(&en4,QStringLiteral("Expression"));
		parameter_item.expression = xml_file->Read_Text_Node(&en5);
		
		axis_item.parameter_items.push_back(parameter_item);
		
// xyz axis, qzx
		en4 = xml_file->Get_Node(&en3,QStringLiteral("Parameter_QZX"));
		parameter_item.parameter = Types::PARAMETER_QZX;
		
		en5 = xml_file->Get_Node(&en4,QStringLiteral("Enabled"));
		text = xml_file->Read_Text_Node(&en5);
		
		if(0 == text.compare(QStringLiteral("True"),Qt::CaseInsensitive))
		{
			parameter_item.enabled = true;
		}
		else
		{
			parameter_item.enabled = false;
		}
		
		en5 = xml_file->Get_Node(&en4,QStringLiteral("Option_State"));
		parameter_item.option_index = xml_file->Read_Text_Node(&en5).toInt();
		
		en5 = xml_file->Get_Node(&en4,QStringLiteral("Expression"));
		parameter_item.expression = xml_file->Read_Text_Node(&en5);
		
		axis_item.parameter_items.push_back(parameter_item);

		machine_item.axis_items.push_back(axis_item);
		axis_item.parameter_items.clear();

		d_machine_items.push_back(machine_item);

		machine_item.id++;

		en1 = xml_file->Get_Sibling_Node(&en1,QStringLiteral("Machine_Item"),true);
	}
}


