/////////////////////////////////////////////////////////////////////
//
//            X   X           X
//           XXX XX         XX
//          XXXXXXXX      XXX
//         XX X XXXXXXXXXXX
//        XXXXX XXXXXXXXX
//       XXXXX XXXXXXXXXX
//            XXX XXX XXX
//           XXX XX   XX
//           X   X     X
//
//    Copyright (C) 2003-2023  Ron Jakl
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
/////////////////////////////////////////////////////////////////////

#ifndef MACHINEWIDGETHEADERFILE
#define MACHINEWIDGETHEADERFILE

#include <QDialog>
#include <vector>

#include "../../core/vector3.h"

class QLineEdit;
class QLabel;
class QRadioButton;
class QToolButton;
class QTreeWidget;

class TXmlFile;

class TMachineDialog : public QDialog
{
Q_OBJECT

public:
	
struct TMapErrorItem
{
	int									axis;
	int									parameter;
	QString								expression;
	double								value;
	int									sign;
};

// CREATORS
	TMachineDialog(const QWidget *parent=0, Qt::WindowFlags flags=Qt::WindowFlags());
	~TMachineDialog(void);

// ACCESSORS
	QString Last_Error(void) const {return d_last_error;}
	bool Save_Data(void) const;
	
	QStringList Machine_Names(void) const;	
	int Machine_Id(const QString &name);
	QString Machine_Name(const int id) const;
	
	TVector3 Machine_Axis_Start(const int id) const;
	TVector3 Machine_Axis_End(const int id) const;
	TVector3 Machine_Math_Origin(const int id) const;
	TVector3 Machine_Mechanical_Origin(const int id) const;
	int Machine_Kinematic(const int id) const;
	
	std::vector<TMapErrorItem> Map_Error_Items(const int id) const;

// MANIPULATORS
	void Set_Data_Path(const QString &s) {d_data_path = s;}
	bool Load_Data(void);
	
	int Create_Machine(const QString &name);
	void Remove_Machine(const int id);
	QString Rename_Machine(const int id, const QString &new_name);
	
	void Reset(const QString &machine_name);

signals:

private slots:
	void Kinematic_Changed(bool);
	void Accept_Clicked(void);
	
private:

struct TMachineParameterItem
{
	int									parameter;
	bool								enabled;
	int									option_index;
	QString								expression;
};
	
struct TMachineAxisItem
{
	int									axis;
	
	std::vector<TMachineParameterItem>	parameter_items;
};
	
struct TMachineItem
{
	int									id;
	QString								name;
	int									kinematic;
	
	TVector3							axis_min;
	TVector3							axis_max;
	TVector3							math_origin;
	TVector3							mechanical_origin;
	
	std::vector<TMachineAxisItem>		axis_items;
};
	
	QLabel								*d_machine_name_label;
	QLineEdit							*d_max_x_edit;
	QLineEdit							*d_max_y_edit;
	QLineEdit							*d_max_z_edit;
	QLineEdit							*d_min_x_edit;
	QLineEdit							*d_min_y_edit;
	QLineEdit							*d_min_z_edit;
	QLineEdit							*d_math_origin_x_edit;
	QLineEdit							*d_math_origin_y_edit;
	QLineEdit							*d_math_origin_z_edit;
	QLineEdit							*d_mechanical_origin_x_edit;
	QLineEdit							*d_mechanical_origin_y_edit;
	QLineEdit							*d_mechanical_origin_z_edit;
	QRadioButton						*d_kinematic_xyz_radio;
	QRadioButton						*d_kinematic_xzy_radio;
	QRadioButton						*d_kinematic_yxz_radio;
	QRadioButton						*d_kinematic_yzx_radio;
	QTreeWidget							*d_machine_parameter_tree;
	
	QString								d_data_path;
	std::vector<TMachineItem>			d_machine_items;
	int									d_active_machine_id;
	
	mutable bool						d_data_changed;
	mutable QString						d_last_error;

	void Initialize_Parameter_Tree(void);
	void Enable_Editing(const bool state);
	
	QString Generate_Unique_Name(const int id,const QString &suggested_name) const;
	bool Name_Valid(const QString &name) const;

	void Update_Editor();
	void Update_Machine();
	
	bool Save(void) const;
	void Load_Version1(TXmlFile * const xml_file);
	void Load_Version2(TXmlFile * const xml_file);

// NOT IMPLEMENTED
	TMachineDialog(const TMachineDialog&);
	TMachineDialog& operator=(const TMachineDialog&);
};

#endif
