/////////////////////////////////////////////////////////////////////
//
//            X   X           X
//           XXX XX         XX
//          XXXXXXXX      XXX
//         XX X XXXXXXXXXXX
//        XXXXX XXXXXXXXX
//       XXXXX XXXXXXXXXX
//            XXX XXX XXX
//           XXX XX   XX
//           X   X     X
//
//    Copyright (C) 2003-2023  Ron Jakl
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
/////////////////////////////////////////////////////////////////////

#include <QApplication>
#include <QCheckBox>
#include <QComboBox>
#include <QHBoxLayout>
#include <QLabel>
#include <QLineEdit>
#include <QToolButton>
#include <QStringList>

#include "expressiondialog.h"

#include "machineparameterwidget.h"

TMachineParameterWidget::TMachineParameterWidget(
	const QWidget						*parent,
	const Qt::WindowFlags				flags)
:QWidget(const_cast<QWidget*>(parent),flags)
{
	QHBoxLayout							*widget_hlayout;

	widget_hlayout = new QHBoxLayout(this);
	widget_hlayout->setContentsMargins(4,0,4,0);

	d_title_label = new QLabel(this);
	d_title_label->setAlignment(Qt::AlignRight|Qt::AlignTrailing|Qt::AlignVCenter);
	d_title_label->setFixedWidth(80);
	widget_hlayout->addWidget(d_title_label);

	d_enable_check = new QCheckBox(this);
	widget_hlayout->addWidget(d_enable_check);

	d_sign_combo = new QComboBox(this);
	d_sign_combo->setFixedWidth(180);
	widget_hlayout->addWidget(d_sign_combo);

	d_expression_label = new QLabel(this);
	d_expression_label->setAlignment(Qt::AlignRight|Qt::AlignTrailing|Qt::AlignVCenter);
	d_expression_label->setFixedWidth(100);
	widget_hlayout->addWidget(d_expression_label);

	d_expression_edit = new QLineEdit(this);
	widget_hlayout->addWidget(d_expression_edit);
	
	d_edit_button = new QToolButton(this);
	widget_hlayout->addWidget(d_edit_button);
	
	// default values
	d_sign_combo->setEnabled(false);
	d_expression_edit->setEnabled(false);
	
	d_expression_dialog = 0;
	
	d_min = 0.0;
	d_max = 1000.0;
	d_enable_expression_editor = false;
	d_enable_expressions = false;

	d_enable_check->setText(QStringLiteral("Enable"));
	d_edit_button->setText(QStringLiteral("..."));
	d_expression_label->setText(QStringLiteral("Error mm/m ="));
	
	connect(d_enable_check,&QCheckBox::toggled,this,&TMachineParameterWidget::Enable_Toggled);
	connect(d_edit_button,&QToolButton::clicked,this,&TMachineParameterWidget::Edit_Expression);
}

TMachineParameterWidget::~TMachineParameterWidget(void)
{
}

QString TMachineParameterWidget::Sign_Option(void) const
{
	return d_sign_combo->currentText();
}

int TMachineParameterWidget::Sign_Option_Index(void) const
{
	return d_sign_combo->currentIndex();
}

bool TMachineParameterWidget::Is_Option_Enabled(void) const
{
	return d_enable_check->isChecked();
}

QString TMachineParameterWidget::Expression(void) const
{
	double								dval;
	
	if(d_enable_expressions)
	{
		return d_expression_edit->text();
	}
	
	dval = d_expression_edit->text().toDouble();
	
	return QString("%1").arg(dval,0,'f',4);
}

void TMachineParameterWidget::Set_Title_Text(
	const QString						&text)
{
	d_title_label->setText(text);
}

void TMachineParameterWidget::Set_Sign_Options(
	const QStringList					&items)
{
	bool								prev_state;
	
	prev_state = d_sign_combo->blockSignals(true);
	
	d_sign_combo->clear();
	d_sign_combo->addItems(items);
	
	d_sign_combo->blockSignals(prev_state);
}

void TMachineParameterWidget::Set_Sign_Option_Index(
	const int							index)
{
	bool								prev_state;
	
	prev_state = d_sign_combo->blockSignals(true);

	d_sign_combo->setCurrentIndex(index);
	
	d_sign_combo->blockSignals(prev_state);
}

void TMachineParameterWidget::Set_Option_Enabled(
	const bool							state)
{
	bool								prev_state;
	
	prev_state = d_enable_check->blockSignals(true);

	d_enable_check->setChecked(state);
	
	d_enable_check->blockSignals(prev_state);
	
	d_sign_combo->setEnabled(state);
	d_expression_edit->setEnabled(state);
}

void TMachineParameterWidget::Set_Expression(
	const QString						&expression)
{
	double								dval;
	if(d_enable_expressions)
	{
		d_expression_edit->setText(expression);
	}
	else
	{
		dval = expression.toDouble();
		d_expression_edit->setText(QString("%1").arg(dval,0,'f',4));
	}
}

void TMachineParameterWidget::Enable_Expression_Editor(
	const bool							state)
{
	d_enable_expression_editor = state;
	d_edit_button->setEnabled(state);
}

void TMachineParameterWidget::Enable_Expressions(
	const bool							state)
{
	d_enable_expressions = state;
	
	if(state)
	{
		d_expression_label->setText(QStringLiteral("Expression ="));
	}
	else
	{
		d_expression_label->setText(QStringLiteral("Error mm/m ="));
	}
}

void TMachineParameterWidget::Enable_Toggled(
	bool								state)
{
	d_sign_combo->setEnabled(state);
	d_expression_edit->setEnabled(state);
}

void TMachineParameterWidget::Edit_Expression(void)
{
	if(d_enable_expression_editor)
	{
		if(!d_expression_dialog)
		{
			d_expression_dialog = new TExpressionDialog(this);
		}
		
		d_expression_dialog->Set_Parameter_Range(d_min,d_max);
		d_expression_dialog->Set_Expression(d_expression_edit->text());
		
		if(d_expression_dialog->exec() == QDialog::Accepted)
		{
			d_expression_edit->setText(d_expression_dialog->Expression());
		}
	}
}


