/////////////////////////////////////////////////////////////////////
//
//            X   X           X
//           XXX XX         XX
//          XXXXXXXX      XXX
//         XX X XXXXXXXXXXX
//        XXXXX XXXXXXXXX
//       XXXXX XXXXXXXXXX
//            XXX XXX XXX
//           XXX XX   XX
//           X   X     X
//
//    Copyright (C) 2003-2023  Ron Jakl
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
/////////////////////////////////////////////////////////////////////

#include <QApplication>
#include <QGridLayout>
#include <QHBoxLayout>
#include <QLabel>
#include <QLineEdit>
#include <QSizePolicy>
#include <QSpacerItem>
#include <QTabWidget>
#include <QToolButton>
#include <QPushButton>
#include <QRegularExpression>
#include <QRegularExpressionMatch>
#include <assert.h>

#include "../../core/xmlfile.h"

#include "types.h"
#include "editor_laser.h"
#include "editor_iso.h"
#include "editor_mcg.h"
#include "editor_ballbar.h"

#include "measurement_laser.h"
#include "measurement_iso.h"
#include "measurement_mcg.h"
#include "measurement_ballbar.h"

#include "measurementdialog.h"

TMeasurementDialog::TMeasurementDialog(
	const QWidget						*parent,
	const Qt::WindowFlags				flags)
:QDialog(const_cast<QWidget*>(parent),flags)
{
	QFrame								*separator_hline;
	QGridLayout							*dialog_layout;
	QHBoxLayout							*button_hlayout;
	QLabel								*help_label;
	QSpacerItem							*button_hspacer;

	this->resize(751,701);

	dialog_layout = new QGridLayout(this);

	d_measurement_tab = new QTabWidget(this);
	d_measurement_tab->setMinimumWidth(350);
	
	d_editor_laser = new TEditor_Laser();
	d_measurement_tab->addTab(d_editor_laser,QStringLiteral("Laser"));
	
	d_editor_iso = new TEditor_ISO();
	d_measurement_tab->addTab(d_editor_iso,QStringLiteral("ISO 10360-2"));
	
	d_editor_mcg = new TEditor_MCG();
	d_measurement_tab->addTab(d_editor_mcg,QStringLiteral("MCG"));

	d_editor_ballbar = new TEditor_Ballbar();
	d_measurement_tab->addTab(d_editor_ballbar,QStringLiteral("Ball Bar"));

	d_measurement_tab->setCurrentIndex(0);
	dialog_layout->addWidget(d_measurement_tab,0,0,1,1);

	help_label = new QLabel(this);
	help_label->setAlignment(Qt::AlignLeading|Qt::AlignLeft|Qt::AlignTop);
	help_label->setWordWrap(true);
	dialog_layout->addWidget(help_label,0,1,1,1);
	
	separator_hline = new QFrame(this);
	separator_hline->setFrameShape(QFrame::HLine);
	separator_hline->setFrameShadow(QFrame::Sunken);
	dialog_layout->addWidget(separator_hline,1,0,1,2);
	
	button_hlayout = new QHBoxLayout();
	
	button_hspacer = new QSpacerItem(0,0,QSizePolicy::Expanding,QSizePolicy::Minimum);
	button_hlayout->addItem(button_hspacer);
	
	d_cancel_button = new QPushButton(this);
	d_cancel_button->setAutoDefault(false);
	button_hlayout->addWidget(d_cancel_button);
	
	d_accept_button = new QPushButton(this);
	d_accept_button->setAutoDefault(false);
	button_hlayout->addWidget(d_accept_button);
	dialog_layout->addLayout(button_hlayout,2,0,1,2);
	
	// defaults
	d_data_changed = false;
	d_active_measurement_id = -1;

	this->setWindowTitle(QStringLiteral("Measurement Editor"));
	
	d_cancel_button->setText(QStringLiteral("Cancel"));
	d_accept_button->setText(QStringLiteral("Accept"));

	help_label->setText(QStringLiteral("<p><span style=\" font-weight:600; text-decoration: underline;\">Variable Help</span></p><p>Values can be entered as static constants or can be the result of an expression. The following variables are available to expressions:</p><p><span style=\" font-weight:600;\">XMin, YMin, ZMin:</span> Minimum coordinate values of the machine volume.</p><p><span style=\" font-weight:600;\">XMax, YMax, ZMax:</span> Maximum coordinate values of the machine volume.</p><p><span style=\" font-weight:600;\">XMid, YMid, ZMid:</span> Center coordinate values of the machine volume.</p><p><span style=\" font-weight:600;\">AxisMin, AxisMax:</span> Minimum and maximum axis length of machine volume.</p><p><span style=\" font-weight:600;\">L, R:</span> Measurement item nominal length or radius for MCG or ball bar.</p>   <p><span style=\" font-weight:600;\">I, J, K:</span> Nominal IJK axis for ball bar.</p>"));
	
	connect(d_cancel_button,&QPushButton::clicked,this,&TMeasurementDialog::reject);
	connect(d_accept_button,&QPushButton::clicked,this,&TMeasurementDialog::Accept_Clicked);
}

TMeasurementDialog::~TMeasurementDialog(void)
{
	std::vector<TMeasurement*>::iterator iter;
	
	for(iter = d_measurements.begin();iter != d_measurements.end();++iter)
	{
		delete (*iter);
	}
	
	d_measurements.clear();
}

bool TMeasurementDialog::Save_Data(void) const
{	
	if(d_data_changed)
	{
		if(this->Save())
		{
			d_data_changed = false;
			return true;
		}
		
		return false;
	}
	
	return true;
}

QStringList TMeasurementDialog::Measurement_Names(void) const
{
	std::vector<TMeasurement*>::const_iterator iter;
	QStringList							list;
	
	for(iter = d_measurements.begin();iter != d_measurements.end();++iter)
	{
		list.push_back((*iter)->Name());
	}
	
	return list;
}

std::vector<TMeasurement*> TMeasurementDialog::Measurements(
	const QStringList					&list)
{
	int									id;
	QList<QString>::const_iterator		iter;
	TMeasurement						*measurement;
	std::vector<TMeasurement*>			measurements;

	for(iter = list.begin();iter != list.end();++iter)
	{
		id = this->Measurement_Id(*iter);
		measurement = this->Measurement(id);
		
		if(measurement)
		{
			measurements.push_back(measurement);
		}
	}
	
	return measurements;
}

int TMeasurementDialog::Measurement_Id(
	const QString						&name)
{
	std::vector<TMeasurement*>::const_iterator iter;
	
	if(name.length())
	{
		for(iter = d_measurements.begin();iter != d_measurements.end();++iter)
		{
			if(name.compare((*iter)->Name(),Qt::CaseInsensitive) == 0)
			{
				return (*iter)->Id();
			}
		}
	}
	
	return -1;
}

TMeasurement* TMeasurementDialog::Measurement(
	const int							id)
{
	TMeasurement						*measurement(0);
	std::vector<TMeasurement*>::const_iterator iter;
	
	if(id < 0)
	{
		return measurement;
	}
	
	for(iter = d_measurements.begin();iter != d_measurements.end();++iter)
	{
		if((*iter)->Id() == id)
		{
			measurement = (*iter);
			break;
		}
	}
	
	return measurement;
}

bool TMeasurementDialog::Load_Data(void)
{
	TXmlFile							xmlfile;
	int									version;
	
	try
	{
		xmlfile.Set_File_Name(QString("%1/measurements.xml").arg(d_data_path));
		xmlfile.Open("Measurements",&version);
		
		switch(version)
		{
			case 1:
				this->Load_Version1(&xmlfile);
				break;
		}
		
	}
	catch(TXmlFileException xml_exception)
	{
		d_last_error = QString("ERR:  Loading measurements exception: %1").arg(xml_exception.ErrorText());
		return false;
	}
	
	return true;
}

void TMeasurementDialog::Reset(
	const QString						&measurement_name,
	const QString						&measurement_copy_name)
{	
	QString								name;
	int									measurement_copy_id;
	
	d_active_measurement_id = this->Measurement_Id(measurement_name);
	measurement_copy_id = this->Measurement_Id(measurement_copy_name);
		
	this->Update_Editor(measurement_copy_id);

	if(d_active_measurement_id < 0)
	{
		d_editor_iso->setEnabled(true);
		d_editor_laser->setEnabled(true);
		d_editor_mcg->setEnabled(true);
		d_editor_ballbar->setEnabled(true);

		d_editor_ballbar->Enable_Name(true);
		d_editor_iso->Enable_Name(true);
		d_editor_mcg->Enable_Name(true);
		d_editor_laser->Enable_Name(true);

		// make sure all default names are unique
		name = d_editor_ballbar->Name();
		name = this->Generate_Unique_Name(Types::MEASUREMENT_BALLBAR,name);
		d_editor_ballbar->Set_Name(name);
		
		name = d_editor_laser->Name();
		name = this->Generate_Unique_Name(Types::MEASUREMENT_LASER,name);
		d_editor_laser->Set_Name(name);
		
		name = d_editor_iso->Name();
		name = this->Generate_Unique_Name(Types::MEASUREMENT_ISO,name);
		d_editor_iso->Set_Name(name);
		
		name = d_editor_mcg->Name();
		name = this->Generate_Unique_Name(Types::MEASUREMENT_MCG,name);
		d_editor_mcg->Set_Name(name);
		
		d_cancel_button->setText(QStringLiteral("Close"));
		d_accept_button->setText(QStringLiteral("Create"));
		d_accept_button->setDefault(false);
	}
	else
	{
		d_editor_iso->setEnabled(false);
		d_editor_laser->setEnabled(false);
		d_editor_mcg->setEnabled(false);
		d_editor_ballbar->setEnabled(false);

		d_cancel_button->setText(QStringLiteral("Cancel"));
		d_accept_button->setText(QStringLiteral("Accept"));
		d_accept_button->setDefault(true);
		
		d_editor_ballbar->Enable_Name(false);
		d_editor_iso->Enable_Name(false);
		d_editor_mcg->Enable_Name(false);
		d_editor_laser->Enable_Name(false);
		
		this->Update_Editor(d_active_measurement_id);
	}
}

void TMeasurementDialog::Remove_Measurement(
	const int							id)
{
	std::vector<TMeasurement*>::iterator iter;
	
	for(iter = d_measurements.begin();iter != d_measurements.end();++iter)
	{
		if(id == (*iter)->Id())
		{
			delete (*iter);
			d_measurements.erase(iter);
			
			d_data_changed = true;
			
			emit Update_Measurement_Display();

			return;
		}
	}
}

QString TMeasurementDialog::Rename_Measurement(
	const int							id,
	const QString						&new_name)
{
	std::vector<TMeasurement*>::iterator iter;
	QString								actual_name(new_name);
	
	for(iter = d_measurements.begin();iter != d_measurements.end();++iter)
	{
		if(id == (*iter)->Id())
		{
			actual_name = this->Generate_Unique_Name((*iter)->Type(),new_name,id);
			(*iter)->Set_Name(actual_name);		
			
			break;
		}
	}
	
	d_data_changed = true;
	
	return actual_name;
}

void TMeasurementDialog::Accept_Clicked(void)
{
	QWidget								*active_widget;

	d_data_changed = true;

	if(d_active_measurement_id < 0)
	{
		active_widget = d_measurement_tab->currentWidget();
		
		if(active_widget == d_editor_laser)
		{
				this->Create_Laser_Measurement();
		}
		else if(active_widget == d_editor_iso)
		{
				this->Create_ISO_Measurement();
		}
		else if(active_widget == d_editor_mcg)
		{
				this->Create_MCG_Measurement();
		}
		else if(active_widget == d_editor_ballbar)
		{
				this->Create_Ballbar_Measurement();
		}
		
		emit Update_Measurement_Display();
	}
	else
	{
		this->Update_Measurement(d_active_measurement_id);
		this->accept();
	}
}

void TMeasurementDialog::Create_Ballbar_Measurement(void)
{
	TMeasurement						*measurement;
	QString								name;
	int									id;
	
	name = d_editor_ballbar->Name();
	name = this->Generate_Unique_Name(Types::MEASUREMENT_BALLBAR,name);
	d_editor_ballbar->Set_Name(name);
	
	measurement = new TMeasurement_Ballbar;
	
	measurement->Set_Name(name);
	id = measurement->Id();
	
	d_measurements.push_back(measurement);
	this->Update_Measurement(id);
	
	name = this->Generate_Unique_Name(Types::MEASUREMENT_BALLBAR,name);
	d_editor_ballbar->Set_Name(name);
}

void TMeasurementDialog::Create_Laser_Measurement(void)
{
	TMeasurement						*measurement;
	QString								name;
	int									id;
	
	name = d_editor_laser->Name();
	name = this->Generate_Unique_Name(Types::MEASUREMENT_LASER,name);
	d_editor_laser->Set_Name(name);

	measurement = new TMeasurement_Laser;
	
	measurement->Set_Name(name);
	id = measurement->Id();
	
	d_measurements.push_back(measurement);
	this->Update_Measurement(id);
		
	name = this->Generate_Unique_Name(Types::MEASUREMENT_LASER,name);
	d_editor_laser->Set_Name(name);
}

void TMeasurementDialog::Create_ISO_Measurement(void)
{
	TMeasurement						*measurement;
	QString								name;
	int									id;
	
	name = d_editor_iso->Name();
	name = this->Generate_Unique_Name(Types::MEASUREMENT_ISO,name);
	d_editor_iso->Set_Name(name);

	measurement = new TMeasurement_ISO;
	
	measurement->Set_Name(name);
	id = measurement->Id();
		
	d_measurements.push_back(measurement);
	this->Update_Measurement(id);
	
	name = this->Generate_Unique_Name(Types::MEASUREMENT_ISO,name);
	d_editor_iso->Set_Name(name);
}

void TMeasurementDialog::Create_MCG_Measurement(void)
{
	TMeasurement						*measurement;
	QString								name;
	int									id;
	
	name = d_editor_mcg->Name();
	name = this->Generate_Unique_Name(Types::MEASUREMENT_MCG,name);
	d_editor_mcg->Set_Name(name);

	measurement = new TMeasurement_MCG;
	
	measurement->Set_Name(name);
	id = measurement->Id();
	
	d_measurements.push_back(measurement);
	this->Update_Measurement(id);
		
	name = this->Generate_Unique_Name(Types::MEASUREMENT_MCG,name);
	d_editor_mcg->Set_Name(name);
}

QString TMeasurementDialog::Generate_Unique_Name(
	const Types::TMeasurementType		type,
	const QString						&suggested_name,
	const int							current_id) const
{
	std::vector<TMeasurement*>::const_iterator iter;
	QRegularExpression					reg_exp(QStringLiteral("^([A-Za-z_ ][A-za-z_ ]*)([0-9]{1,10})$"));
	QRegularExpressionMatch				expression_match;
	QStringList							expression_text_list;
	QString								name;
	QString								name_prefix;
	bool								unique;
	int									index(0);
	
	if(suggested_name.length())
	{
		name = suggested_name.simplified();

		expression_match = reg_exp.match(name);
		if(expression_match.hasMatch())
		{
			expression_text_list = expression_match.capturedTexts();
			
			if(expression_text_list.size() > 0)
			{
				name_prefix = expression_text_list[1].trimmed();
			}
			
			if(expression_text_list.size() > 1)
			{
				index = expression_text_list[2].toInt();
			}
		}
		else
		{
			name_prefix = name;
		}
	}
	else
	{
		switch(type)
		{
			case Types::MEASUREMENT_LASER:
				name_prefix = QStringLiteral("Laser");
				break;
			case Types::MEASUREMENT_ISO:
				name_prefix = QStringLiteral("ISO");
				break;
			case Types::MEASUREMENT_MCG:
				name_prefix = QStringLiteral("MCG");
				break;
			case Types::MEASUREMENT_BALLBAR:
				name_prefix = QStringLiteral("BallBar");
				break;
			default:
				name_prefix = QStringLiteral("Untitled");
				break;
		}
	}
	
	do
	{
		unique = true;
		
		if(index)
		{
			name = QString("%1 %2").arg(name_prefix).arg(index);
		}
		else
		{
			name = QString("%1").arg(name_prefix);
		}
		
		for(iter = d_measurements.begin();iter != d_measurements.end();++iter)
		{
			if(name.compare((*iter)->Name(),Qt::CaseInsensitive) == 0)
			{
				if((*iter)->Id() != current_id)
				{
					unique = false;
					break;
				}
			}
		}
		
		index++;
	}
	while(!unique);
	
	return name;
}

bool TMeasurementDialog::Name_Valid(
	const QString						&name) const
{
	QString								test_name(name.simplified());
	std::vector<TMeasurement*>::const_iterator iter;
	
	if(test_name.length() == 0)
	{
		return false;
	}
	
	for(iter = d_measurements.begin();iter != d_measurements.end();++iter)
	{
		if(0 == test_name.compare((*iter)->Name(),Qt::CaseInsensitive))
		{
			return false;
		}
	}
	
	return true;
}

void TMeasurementDialog::Update_Measurement(
	const int							id)
{
	std::vector<TMeasurement*>::iterator iter;
	TMeasurement_Ballbar				*bbar_measurement;
	TMeasurement_Laser					*laser_measurement;
	TMeasurement_ISO					*iso_measurement;
	TMeasurement_MCG					*mcg_measurement;

	if(id < 0)
	{
		return;
	}
		
	for(iter = d_measurements.begin();iter != d_measurements.end();++iter)
	{
		if(id == (*iter)->Id())
		{
			switch((*iter)->Type())
			{				
				case Types::MEASUREMENT_LASER:
					laser_measurement = reinterpret_cast<TMeasurement_Laser*>((*iter));
					
					laser_measurement->Set_Name(d_editor_laser->Name());
					
					laser_measurement->Set_Start_Position_X(d_editor_laser->Start_Position_X());
					laser_measurement->Set_Start_Position_Y(d_editor_laser->Start_Position_Y());
					laser_measurement->Set_Start_Position_Z(d_editor_laser->Start_Position_Z());
					
					laser_measurement->Set_End_Position_X(d_editor_laser->End_Position_X());
					laser_measurement->Set_End_Position_Y(d_editor_laser->End_Position_Y());
					laser_measurement->Set_End_Position_Z(d_editor_laser->End_Position_Z());
					
					laser_measurement->Set_Probe_Offset(d_editor_laser->Probe_Offset());
					
					laser_measurement->Set_Increment(d_editor_laser->Increment());
					break;
					
				case Types::MEASUREMENT_MCG:
					mcg_measurement = reinterpret_cast<TMeasurement_MCG*>((*iter));
					
					mcg_measurement->Set_Name(d_editor_mcg->Name());
					
					mcg_measurement->Set_Position_X(d_editor_mcg->Position_X());
					mcg_measurement->Set_Position_Y(d_editor_mcg->Position_Y());
					mcg_measurement->Set_Position_Z(d_editor_mcg->Position_Z());
					
					mcg_measurement->Set_Probe_Offset(d_editor_mcg->Probe_Offset());
					
					mcg_measurement->Set_Length(d_editor_mcg->Length());
					mcg_measurement->Set_Azimuth_Count(d_editor_mcg->Azimuth_Count());
					mcg_measurement->Set_Elevation_Count(d_editor_mcg->Elevation_Count());
					mcg_measurement->Set_Start_Elevation(d_editor_mcg->Start_Elevation());
					mcg_measurement->Set_End_Elevation(d_editor_mcg->End_Elevation());
					break;
					
				case Types::MEASUREMENT_ISO:
					iso_measurement = reinterpret_cast<TMeasurement_ISO*>((*iter));
					
					iso_measurement->Set_Name(d_editor_iso->Name());
					
					iso_measurement->Set_Start_Position_X(d_editor_iso->Start_Position_X());
					iso_measurement->Set_Start_Position_Y(d_editor_iso->Start_Position_Y());
					iso_measurement->Set_Start_Position_Z(d_editor_iso->Start_Position_Z());
					
					iso_measurement->Set_End_Position_X(d_editor_iso->End_Position_X());
					iso_measurement->Set_End_Position_Y(d_editor_iso->End_Position_Y());
					iso_measurement->Set_End_Position_Z(d_editor_iso->End_Position_Z());
					
					iso_measurement->Set_Probe_Offset(d_editor_iso->Probe_Offset());
					break;
					
				case Types::MEASUREMENT_BALLBAR:
					bbar_measurement = reinterpret_cast<TMeasurement_Ballbar*>((*iter));
					
					bbar_measurement->Set_Name(d_editor_ballbar->Name());
					
					bbar_measurement->Set_Position_X(d_editor_ballbar->Position_X());
					bbar_measurement->Set_Position_Y(d_editor_ballbar->Position_Y());
					bbar_measurement->Set_Position_Z(d_editor_ballbar->Position_Z());
					
					bbar_measurement->Set_Direction_I(d_editor_ballbar->Direction_I());
					bbar_measurement->Set_Direction_J(d_editor_ballbar->Direction_J());
					bbar_measurement->Set_Direction_K(d_editor_ballbar->Direction_K());
					
					bbar_measurement->Set_Probe_Offset(d_editor_ballbar->Probe_Offset());
					
					bbar_measurement->Set_Length(d_editor_ballbar->Length());

					break;
					
				default:
					return;
			}
			
			return;
		}
	}
}

void TMeasurementDialog::Update_Editor(
	const int							id)
{
	std::vector<TMeasurement*>::iterator iter;
	TMeasurement_Ballbar				*bbar_measurement;
	TMeasurement_Laser					*laser_measurement;
	TMeasurement_ISO					*iso_measurement;
	TMeasurement_MCG					*mcg_measurement;
	
	if(!(id < 0))
	{
		for(iter = d_measurements.begin();iter != d_measurements.end();++iter)
		{
			if(id == (*iter)->Id())
			{
				switch((*iter)->Type())
				{
					case Types::MEASUREMENT_LASER:
						laser_measurement = reinterpret_cast<TMeasurement_Laser*>((*iter));
						
						d_editor_laser->Set_Name(laser_measurement->Name());
						d_editor_mcg->Set_Name(QString());
						d_editor_iso->Set_Name(QString());
						d_editor_ballbar->Set_Name(QString());

						d_editor_laser->Set_Start_Position_X(laser_measurement->Start_Position_X());
						d_editor_laser->Set_Start_Position_Y(laser_measurement->Start_Position_Y());
						d_editor_laser->Set_Start_Position_Z(laser_measurement->Start_Position_Z());
						
						d_editor_laser->Set_End_Position_X(laser_measurement->End_Position_X());
						d_editor_laser->Set_End_Position_Y(laser_measurement->End_Position_Y());
						d_editor_laser->Set_End_Position_Z(laser_measurement->End_Position_Z());
						
						d_editor_laser->Set_Probe_Offset(laser_measurement->Probe_Offset());
						
						d_editor_laser->Set_Increment(laser_measurement->Increment());
						
						// enable this tab
						d_editor_laser->setEnabled(true);
						d_measurement_tab->setCurrentWidget(d_editor_laser);
						
						break;
						
					case Types::MEASUREMENT_MCG:
						mcg_measurement = reinterpret_cast<TMeasurement_MCG*>((*iter));
						
						d_editor_mcg->Set_Name(mcg_measurement->Name());
						d_editor_laser->Set_Name(QString());
						d_editor_iso->Set_Name(QString());
						d_editor_ballbar->Set_Name(QString());

						d_editor_mcg->Set_Position_X(mcg_measurement->Position_X());
						d_editor_mcg->Set_Position_Y(mcg_measurement->Position_Y());
						d_editor_mcg->Set_Position_Z(mcg_measurement->Position_Z());
						
						d_editor_mcg->Set_Probe_Offset(mcg_measurement->Probe_Offset());
						
						d_editor_mcg->Set_Length(mcg_measurement->Length());
						d_editor_mcg->Set_Azimuth_Count(mcg_measurement->Azimuth_Count());
						d_editor_mcg->Set_Elevation_Count(mcg_measurement->Elevation_Count());
						d_editor_mcg->Set_Start_Elevation(mcg_measurement->Start_Elevation());
						d_editor_mcg->Set_End_Elevation(mcg_measurement->End_Elevation());
						
						// enable this tab
						d_editor_mcg->setEnabled(true);
						d_measurement_tab->setCurrentWidget(d_editor_mcg);
						
						break;
						
					case Types::MEASUREMENT_ISO:
						iso_measurement = reinterpret_cast<TMeasurement_ISO*>((*iter));
						
						d_editor_iso->Set_Name(iso_measurement->Name());
						d_editor_laser->Set_Name(QString());
						d_editor_mcg->Set_Name(QString());
						d_editor_ballbar->Set_Name(QString());

						d_editor_iso->Set_Start_Position_X(iso_measurement->Start_Position_X());
						d_editor_iso->Set_Start_Position_Y(iso_measurement->Start_Position_Y());
						d_editor_iso->Set_Start_Position_Z(iso_measurement->Start_Position_Z());
						
						d_editor_iso->Set_End_Position_X(iso_measurement->End_Position_X());
						d_editor_iso->Set_End_Position_Y(iso_measurement->End_Position_Y());
						d_editor_iso->Set_End_Position_Z(iso_measurement->End_Position_Z());
						
						d_editor_iso->Set_Probe_Offset(iso_measurement->Probe_Offset());
						
						// enable this tab
						d_editor_iso->setEnabled(true);
						d_measurement_tab->setCurrentWidget(d_editor_iso);
						
						break;
						
					case Types::MEASUREMENT_BALLBAR:
						bbar_measurement = reinterpret_cast<TMeasurement_Ballbar*>((*iter));
						
						d_editor_ballbar->Set_Name(bbar_measurement->Name());
						d_editor_laser->Set_Name(QString());
						d_editor_mcg->Set_Name(QString());
						d_editor_iso->Set_Name(QString());
						
						d_editor_ballbar->Set_Position_X(bbar_measurement->Position_X());
						d_editor_ballbar->Set_Position_Y(bbar_measurement->Position_Y());
						d_editor_ballbar->Set_Position_Z(bbar_measurement->Position_Z());
						
						d_editor_ballbar->Set_Direction_I(bbar_measurement->Direction_I());
						d_editor_ballbar->Set_Direction_J(bbar_measurement->Direction_J());
						d_editor_ballbar->Set_Direction_K(bbar_measurement->Direction_K());
						
						d_editor_ballbar->Set_Probe_Offset(bbar_measurement->Probe_Offset());
						
						d_editor_ballbar->Set_Length(bbar_measurement->Length());
						
						// enable this tab
						d_editor_ballbar->setEnabled(true);
						d_measurement_tab->setCurrentWidget(d_editor_ballbar);
						
						break;
						
					default:
						return;
				}
				
				return;
			}
		}
	}
}

bool TMeasurementDialog::Save(void) const
{
	TMeasurement_Laser					*laser_measurement;
	TMeasurement_ISO					*iso_measurement;
	TMeasurement_MCG					*mcg_measurement;
	TMeasurement_Ballbar				*bbar_measurement;
	TXmlFile							xmlfile;
	QDomElement							en1;
	QDomElement							en2;
	std::vector<TMeasurement*>::const_iterator iter;
	TVector3							probe_offset;
	QString								text;
	
	try
	{
		xmlfile.Set_File_Name(QString("%1/measurements.xml").arg(d_data_path));
		xmlfile.Reset("Measurements",1);
		
		for(iter = d_measurements.begin();iter != d_measurements.end();++iter)
		{
			switch((*iter)->Type())
			{
				case Types::MEASUREMENT_LASER:
		
					laser_measurement = reinterpret_cast<TMeasurement_Laser*>(*iter);
					en1 = xmlfile.Create_Node(xmlfile.Root_Node(),QStringLiteral("Measurement_Laser"));
					
					en2 = xmlfile.Create_Node(&en1,QStringLiteral("Name"));
					xmlfile.Write_Text_Node(&en2,laser_measurement->Name());
					
					en2 = xmlfile.Create_Node(&en1,QStringLiteral("Start_Position_X"));
					xmlfile.Write_Text_Node(&en2,laser_measurement->Start_Position_X());
					
					en2 = xmlfile.Create_Node(&en1,QStringLiteral("Start_Position_Y"));
					xmlfile.Write_Text_Node(&en2,laser_measurement->Start_Position_Y());
					
					en2 = xmlfile.Create_Node(&en1,QStringLiteral("Start_Position_Z"));
					xmlfile.Write_Text_Node(&en2,laser_measurement->Start_Position_Z());
					
					en2 = xmlfile.Create_Node(&en1,QStringLiteral("End_Position_X"));
					xmlfile.Write_Text_Node(&en2,laser_measurement->End_Position_X());
					
					en2 = xmlfile.Create_Node(&en1,QStringLiteral("End_Position_Y"));
					xmlfile.Write_Text_Node(&en2,laser_measurement->End_Position_Y());
					
					en2 = xmlfile.Create_Node(&en1,QStringLiteral("End_Position_Z"));
					xmlfile.Write_Text_Node(&en2,laser_measurement->End_Position_Z());
					
					probe_offset = laser_measurement->Probe_Offset();
					
					text = QString("%1").arg(probe_offset.x,0,'f',4);
					en2 = xmlfile.Create_Node(&en1,QStringLiteral("Probe_Offset_X"));
					xmlfile.Write_Text_Node(&en2,text);
					
					text = QString("%1").arg(probe_offset.y,0,'f',4);
					en2 = xmlfile.Create_Node(&en1,QStringLiteral("Probe_Offset_Y"));
					xmlfile.Write_Text_Node(&en2,text);
					
					text = QString("%1").arg(probe_offset.z,0,'f',4);
					en2 = xmlfile.Create_Node(&en1,QStringLiteral("Probe_Offset_Z"));
					xmlfile.Write_Text_Node(&en2,text);
					
					en2 = xmlfile.Create_Node(&en1,QStringLiteral("Increment"));
					xmlfile.Write_Text_Node(&en2,laser_measurement->Increment());

					break;
					
				case Types::MEASUREMENT_ISO:
					
					iso_measurement = reinterpret_cast<TMeasurement_ISO*>(*iter);
					en1 = xmlfile.Create_Node(xmlfile.Root_Node(),QStringLiteral("Measurement_ISO"));
					
					en2 = xmlfile.Create_Node(&en1,QStringLiteral("Name"));
					xmlfile.Write_Text_Node(&en2,iso_measurement->Name());
					
					en2 = xmlfile.Create_Node(&en1,QStringLiteral("Start_Position_X"));
					xmlfile.Write_Text_Node(&en2,iso_measurement->Start_Position_X());
					
					en2 = xmlfile.Create_Node(&en1,QStringLiteral("Start_Position_Y"));
					xmlfile.Write_Text_Node(&en2,iso_measurement->Start_Position_Y());
					
					en2 = xmlfile.Create_Node(&en1,QStringLiteral("Start_Position_Z"));
					xmlfile.Write_Text_Node(&en2,iso_measurement->Start_Position_Z());
					
					en2 = xmlfile.Create_Node(&en1,QStringLiteral("End_Position_X"));
					xmlfile.Write_Text_Node(&en2,iso_measurement->End_Position_X());
					
					en2 = xmlfile.Create_Node(&en1,QStringLiteral("End_Position_Y"));
					xmlfile.Write_Text_Node(&en2,iso_measurement->End_Position_Y());
					
					en2 = xmlfile.Create_Node(&en1,QStringLiteral("End_Position_Z"));
					xmlfile.Write_Text_Node(&en2,iso_measurement->End_Position_Z());
					
					probe_offset = iso_measurement->Probe_Offset();

					text = QString("%1").arg(probe_offset.x,0,'f',4);
					en2 = xmlfile.Create_Node(&en1,QStringLiteral("Probe_Offset_X"));
					xmlfile.Write_Text_Node(&en2,text);
					
					text = QString("%1").arg(probe_offset.y,0,'f',4);
					en2 = xmlfile.Create_Node(&en1,QStringLiteral("Probe_Offset_Y"));
					xmlfile.Write_Text_Node(&en2,text);
					
					text = QString("%1").arg(probe_offset.z,0,'f',4);
					en2 = xmlfile.Create_Node(&en1,QStringLiteral("Probe_Offset_Z"));
					xmlfile.Write_Text_Node(&en2,text);

					break;
					
				case Types::MEASUREMENT_MCG:
					
					mcg_measurement = reinterpret_cast<TMeasurement_MCG*>(*iter);
					en1 = xmlfile.Create_Node(xmlfile.Root_Node(),QStringLiteral("Measurement_MCG"));
					
					en2 = xmlfile.Create_Node(&en1,QStringLiteral("Name"));
					xmlfile.Write_Text_Node(&en2,mcg_measurement->Name());
					
					en2 = xmlfile.Create_Node(&en1,QStringLiteral("Position_X"));
					xmlfile.Write_Text_Node(&en2,mcg_measurement->Position_X());
					
					en2 = xmlfile.Create_Node(&en1,QStringLiteral("Position_Y"));
					xmlfile.Write_Text_Node(&en2,mcg_measurement->Position_Y());
					
					en2 = xmlfile.Create_Node(&en1,QStringLiteral("Position_Z"));
					xmlfile.Write_Text_Node(&en2,mcg_measurement->Position_Z());

					probe_offset = mcg_measurement->Probe_Offset();
					
					text = QString("%1").arg(probe_offset.x,0,'f',4);
					en2 = xmlfile.Create_Node(&en1,QStringLiteral("Probe_Offset_X"));
					xmlfile.Write_Text_Node(&en2,text);
					
					text = QString("%1").arg(probe_offset.y,0,'f',4);
					en2 = xmlfile.Create_Node(&en1,QStringLiteral("Probe_Offset_Y"));
					xmlfile.Write_Text_Node(&en2,text);
					
					text = QString("%1").arg(probe_offset.z,0,'f',4);
					en2 = xmlfile.Create_Node(&en1,QStringLiteral("Probe_Offset_Z"));
					xmlfile.Write_Text_Node(&en2,text);
					
					en2 = xmlfile.Create_Node(&en1,QStringLiteral("Length"));
					xmlfile.Write_Text_Node(&en2,mcg_measurement->Length());
					
					en2 = xmlfile.Create_Node(&en1,QStringLiteral("Azimuth_Count"));
					xmlfile.Write_Text_Node(&en2,mcg_measurement->Azimuth_Count());
					
					en2 = xmlfile.Create_Node(&en1,QStringLiteral("Elevation_Count"));
					xmlfile.Write_Text_Node(&en2,mcg_measurement->Elevation_Count());
					
					en2 = xmlfile.Create_Node(&en1,QStringLiteral("Start_Elevation"));
					xmlfile.Write_Text_Node(&en2,mcg_measurement->Start_Elevation());
					
					en2 = xmlfile.Create_Node(&en1,QStringLiteral("End_Elevation"));
					xmlfile.Write_Text_Node(&en2,mcg_measurement->End_Elevation());
					
					break;
					
				case Types::MEASUREMENT_BALLBAR:
					
					bbar_measurement = reinterpret_cast<TMeasurement_Ballbar*>(*iter);
					en1 = xmlfile.Create_Node(xmlfile.Root_Node(),QStringLiteral("Measurement_BallBar"));
					
					en2 = xmlfile.Create_Node(&en1,QStringLiteral("Name"));
					xmlfile.Write_Text_Node(&en2,bbar_measurement->Name());
					
					en2 = xmlfile.Create_Node(&en1,QStringLiteral("Position_X"));
					xmlfile.Write_Text_Node(&en2,bbar_measurement->Position_X());
					
					en2 = xmlfile.Create_Node(&en1,QStringLiteral("Position_Y"));
					xmlfile.Write_Text_Node(&en2,bbar_measurement->Position_Y());
					
					en2 = xmlfile.Create_Node(&en1,QStringLiteral("Position_Z"));
					xmlfile.Write_Text_Node(&en2,bbar_measurement->Position_Z());
					
					en2 = xmlfile.Create_Node(&en1,QStringLiteral("Direction_I"));
					xmlfile.Write_Text_Node(&en2,bbar_measurement->Direction_I());
					
					en2 = xmlfile.Create_Node(&en1,QStringLiteral("Direction_J"));
					xmlfile.Write_Text_Node(&en2,bbar_measurement->Direction_J());
					
					en2 = xmlfile.Create_Node(&en1,QStringLiteral("Direction_K"));
					xmlfile.Write_Text_Node(&en2,bbar_measurement->Direction_K());
					
					probe_offset = bbar_measurement->Probe_Offset();

					text = QString("%1").arg(probe_offset.x,0,'f',4);
					en2 = xmlfile.Create_Node(&en1,QStringLiteral("Probe_Offset_X"));
					xmlfile.Write_Text_Node(&en2,text);
					
					text = QString("%1").arg(probe_offset.y,0,'f',4);
					en2 = xmlfile.Create_Node(&en1,QStringLiteral("Probe_Offset_Y"));
					xmlfile.Write_Text_Node(&en2,text);
					
					text = QString("%1").arg(probe_offset.z,0,'f',4);
					en2 = xmlfile.Create_Node(&en1,QStringLiteral("Probe_Offset_Z"));
					xmlfile.Write_Text_Node(&en2,text);
					
					en2 = xmlfile.Create_Node(&en1,QStringLiteral("Length"));
					xmlfile.Write_Text_Node(&en2,bbar_measurement->Length());

					break;
					
				default:
					assert(false);	// should never get here
			}
		}

		xmlfile.Write_File();
	}
	catch(TXmlFileException xml_exception)
	{
		d_last_error = QString("ERR:  Saving measurements exception: %1").arg(xml_exception.ErrorText());
		return false;
	}
	
	return true;
	
}

void TMeasurementDialog::Load_Version1(
	TXmlFile							* const xml_file)
{
	TMeasurement_Laser					*laser_measurement;
	TMeasurement_ISO					*iso_measurement;
	TMeasurement_MCG					*mcg_measurement;
	TMeasurement_Ballbar				*ballbar_measurement;
	QDomElement							en1;
	QDomElement							en2;
	QDomElement							en3;
	QDomElement							en4;
	TVector3							probe_offset;
	QString								text;
	
	assert(xml_file);
	
	en1 = xml_file->Get_Node(xml_file->Root_Node(),QStringLiteral("Measurement_Laser"),true);
	while(!en1.isNull())
	{
		laser_measurement = new TMeasurement_Laser;
		
		en2 = xml_file->Get_Node(&en1,QStringLiteral("Name"));
		laser_measurement->Set_Name(xml_file->Read_Text_Node(&en2));
		
		en2 = xml_file->Get_Node(&en1,QStringLiteral("Start_Position_X"));
		laser_measurement->Set_Start_Position_X(xml_file->Read_Text_Node(&en2));
		
		en2 = xml_file->Get_Node(&en1,QStringLiteral("Start_Position_Y"));
		laser_measurement->Set_Start_Position_Y(xml_file->Read_Text_Node(&en2));
		
		en2 = xml_file->Get_Node(&en1,QStringLiteral("Start_Position_Z"));
		laser_measurement->Set_Start_Position_Z(xml_file->Read_Text_Node(&en2));
		
		en2 = xml_file->Get_Node(&en1,QStringLiteral("End_Position_X"));
		laser_measurement->Set_End_Position_X(xml_file->Read_Text_Node(&en2));
		
		en2 = xml_file->Get_Node(&en1,QStringLiteral("End_Position_Y"));
		laser_measurement->Set_End_Position_Y(xml_file->Read_Text_Node(&en2));
		
		en2 = xml_file->Get_Node(&en1,QStringLiteral("End_Position_Z"));
		laser_measurement->Set_End_Position_Z(xml_file->Read_Text_Node(&en2));
		
		en2 = xml_file->Get_Node(&en1,QStringLiteral("Probe_Offset_X"));
		probe_offset.x = xml_file->Read_Text_Node(&en2).toDouble();
		
		en2 = xml_file->Get_Node(&en1,QStringLiteral("Probe_Offset_Y"));
		probe_offset.y = xml_file->Read_Text_Node(&en2).toDouble();
		
		en2 = xml_file->Get_Node(&en1,QStringLiteral("Probe_Offset_Z"));
		probe_offset.z = xml_file->Read_Text_Node(&en2).toDouble();
		
		laser_measurement->Set_Probe_Offset(probe_offset);
		
		en2 = xml_file->Get_Node(&en1,QStringLiteral("Increment"));
		laser_measurement->Set_Increment(xml_file->Read_Text_Node(&en2));
		
		d_measurements.push_back(reinterpret_cast<TMeasurement*>(laser_measurement));
		
		en1 = xml_file->Get_Sibling_Node(&en1,QStringLiteral("Measurement_Laser"),true);
	}
	
	en1 = xml_file->Get_Node(xml_file->Root_Node(),QStringLiteral("Measurement_ISO"),true);
	while(!en1.isNull())
	{
		iso_measurement = new TMeasurement_ISO;
		
		en2 = xml_file->Get_Node(&en1,QStringLiteral("Name"));
		iso_measurement->Set_Name(xml_file->Read_Text_Node(&en2));
		
		en2 = xml_file->Get_Node(&en1,QStringLiteral("Start_Position_X"));
		iso_measurement->Set_Start_Position_X(xml_file->Read_Text_Node(&en2));
		
		en2 = xml_file->Get_Node(&en1,QStringLiteral("Start_Position_Y"));
		iso_measurement->Set_Start_Position_Y(xml_file->Read_Text_Node(&en2));
		
		en2 = xml_file->Get_Node(&en1,QStringLiteral("Start_Position_Z"));
		iso_measurement->Set_Start_Position_Z(xml_file->Read_Text_Node(&en2));
		
		en2 = xml_file->Get_Node(&en1,QStringLiteral("End_Position_X"));
		iso_measurement->Set_End_Position_X(xml_file->Read_Text_Node(&en2));
		
		en2 = xml_file->Get_Node(&en1,QStringLiteral("End_Position_Y"));
		iso_measurement->Set_End_Position_Y(xml_file->Read_Text_Node(&en2));
		
		en2 = xml_file->Get_Node(&en1,QStringLiteral("End_Position_Z"));
		iso_measurement->Set_End_Position_Z(xml_file->Read_Text_Node(&en2));
		
		en2 = xml_file->Get_Node(&en1,QStringLiteral("Probe_Offset_X"));
		probe_offset.x = xml_file->Read_Text_Node(&en2).toDouble();
		
		en2 = xml_file->Get_Node(&en1,QStringLiteral("Probe_Offset_Y"));
		probe_offset.y = xml_file->Read_Text_Node(&en2).toDouble();
	
		en2 = xml_file->Get_Node(&en1,QStringLiteral("Probe_Offset_Z"));
		probe_offset.z = xml_file->Read_Text_Node(&en2).toDouble();
		
		iso_measurement->Set_Probe_Offset(probe_offset);

		d_measurements.push_back(reinterpret_cast<TMeasurement*>(iso_measurement));
		
		en1 = xml_file->Get_Sibling_Node(&en1,QStringLiteral("Measurement_ISO"),true);
	}
	
	en1 = xml_file->Get_Node(xml_file->Root_Node(),QStringLiteral("Measurement_MCG"),true);
	while(!en1.isNull())
	{
		mcg_measurement = new TMeasurement_MCG;
		
		en2 = xml_file->Get_Node(&en1,QStringLiteral("Name"));
		mcg_measurement->Set_Name(xml_file->Read_Text_Node(&en2));
		
		en2 = xml_file->Get_Node(&en1,QStringLiteral("Position_X"));
		mcg_measurement->Set_Position_X(xml_file->Read_Text_Node(&en2));
		
		en2 = xml_file->Get_Node(&en1,QStringLiteral("Position_Y"));
		mcg_measurement->Set_Position_Y(xml_file->Read_Text_Node(&en2));
		
		en2 = xml_file->Get_Node(&en1,QStringLiteral("Position_Z"));
		mcg_measurement->Set_Position_Z(xml_file->Read_Text_Node(&en2));
		
		en2 = xml_file->Get_Node(&en1,QStringLiteral("Probe_Offset_X"));
		probe_offset.x = xml_file->Read_Text_Node(&en2).toDouble();
		
		en2 = xml_file->Get_Node(&en1,QStringLiteral("Probe_Offset_Y"));
		probe_offset.y = xml_file->Read_Text_Node(&en2).toDouble();
		
		en2 = xml_file->Get_Node(&en1,QStringLiteral("Probe_Offset_Z"));
		probe_offset.z = xml_file->Read_Text_Node(&en2).toDouble();
		
		mcg_measurement->Set_Probe_Offset(probe_offset);
		
		en2 = xml_file->Get_Node(&en1,QStringLiteral("Length"));
		mcg_measurement->Set_Length(xml_file->Read_Text_Node(&en2));
		
		en2 = xml_file->Get_Node(&en1,QStringLiteral("Azimuth_Count"));
		mcg_measurement->Set_Azimuth_Count(xml_file->Read_Text_Node(&en2));
		
		en2 = xml_file->Get_Node(&en1,QStringLiteral("Elevation_Count"));
		mcg_measurement->Set_Elevation_Count(xml_file->Read_Text_Node(&en2));
		
		en2 = xml_file->Get_Node(&en1,QStringLiteral("Start_Elevation"));
		mcg_measurement->Set_Start_Elevation(xml_file->Read_Text_Node(&en2));
		
		en2 = xml_file->Get_Node(&en1,QStringLiteral("End_Elevation"));
		mcg_measurement->Set_End_Elevation(xml_file->Read_Text_Node(&en2));
		
		d_measurements.push_back(reinterpret_cast<TMeasurement*>(mcg_measurement));
		
		en1 = xml_file->Get_Sibling_Node(&en1,QStringLiteral("Measurement_MCG"),true);
	}

	en1 = xml_file->Get_Node(xml_file->Root_Node(),QStringLiteral("Measurement_BallBar"),true);
	while(!en1.isNull())
	{
		ballbar_measurement = new TMeasurement_Ballbar;
		
		en2 = xml_file->Get_Node(&en1,QStringLiteral("Name"));
		ballbar_measurement->Set_Name(xml_file->Read_Text_Node(&en2));
		
		en2 = xml_file->Get_Node(&en1,QStringLiteral("Position_X"));
		ballbar_measurement->Set_Position_X(xml_file->Read_Text_Node(&en2));
		
		en2 = xml_file->Get_Node(&en1,QStringLiteral("Position_Y"));
		ballbar_measurement->Set_Position_Y(xml_file->Read_Text_Node(&en2));
		
		en2 = xml_file->Get_Node(&en1,QStringLiteral("Position_Z"));
		ballbar_measurement->Set_Position_Z(xml_file->Read_Text_Node(&en2));
		
		en2 = xml_file->Get_Node(&en1,QStringLiteral("Direction_I"));
		ballbar_measurement->Set_Direction_I(xml_file->Read_Text_Node(&en2));
		
		en2 = xml_file->Get_Node(&en1,QStringLiteral("Direction_J"));
		ballbar_measurement->Set_Direction_J(xml_file->Read_Text_Node(&en2));
		
		en2 = xml_file->Get_Node(&en1,QStringLiteral("Direction_K"));
		ballbar_measurement->Set_Direction_K(xml_file->Read_Text_Node(&en2));
		
		en2 = xml_file->Get_Node(&en1,QStringLiteral("Probe_Offset_X"));
		probe_offset.x = xml_file->Read_Text_Node(&en2).toDouble();
		
		en2 = xml_file->Get_Node(&en1,QStringLiteral("Probe_Offset_Y"));
		probe_offset.y = xml_file->Read_Text_Node(&en2).toDouble();
		
		en2 = xml_file->Get_Node(&en1,QStringLiteral("Probe_Offset_Z"));
		probe_offset.z = xml_file->Read_Text_Node(&en2).toDouble();
		
		ballbar_measurement->Set_Probe_Offset(probe_offset);
		
		en2 = xml_file->Get_Node(&en1,QStringLiteral("Length"));
		ballbar_measurement->Set_Length(xml_file->Read_Text_Node(&en2));
		
		d_measurements.push_back(reinterpret_cast<TMeasurement*>(ballbar_measurement));
		
		en1 = xml_file->Get_Sibling_Node(&en1,QStringLiteral("Measurement_BallBar"),true);
	}
}
