/////////////////////////////////////////////////////////////////////
//
//            X   X           X
//           XXX XX         XX
//          XXXXXXXX      XXX
//         XX X XXXXXXXXXXX
//        XXXXX XXXXXXXXX
//       XXXXX XXXXXXXXXX
//            XXX XXX XXX
//           XXX XX   XX
//           X   X     X
//
//    Copyright (C) 2003-2023  Ron Jakl
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
/////////////////////////////////////////////////////////////////////

#include <QApplication>
#include <QGridLayout>
#include <QLabel>
#include <QSizePolicy>
#include <QToolButton>
#include <QFrame>
#include <QSpacerItem>
#include <QLabel>
#include <QIcon>

#include "../../core/mat4.h"
#include "../../core/referencesystem.h"
#include "../../core/openglgeometryfactory.h"
#include "measurement.h"

#include "opengldisplay.h"
#include "opengldisplaywidget.h"

TOpenGLDisplayWidget::TOpenGLDisplayWidget(
	const QWidget						*parent,
	const Qt::WindowFlags				flags)
:QFrame(const_cast<QWidget*>(parent),flags)
{
	QGridLayout							*widget_layout;
	QSizePolicy							size_policy(QSizePolicy::Maximum,QSizePolicy::Preferred);
	QSpacerItem							*widget_vspacer;
	QString								gl_error_text;
	QString								gl_version_text;
	QString								gl_extensions_text;
	int									gl_major_version;

	// check version of OpenGL.  Display will not work if opengl version is less than 2
	gl_major_version = TOpenGLWidget::Get_OpenGL_Version_Major(&gl_version_text,&gl_extensions_text);
	
	size_policy.setHorizontalStretch(0);
	size_policy.setVerticalStretch(0);
	
	this->setFrameShape(QFrame::Box);
	this->setStyleSheet(QStringLiteral("background-color: rgb(255, 255, 255);"));
	
	widget_layout = new QGridLayout(this);

	if(gl_major_version > 1)
	{
		d_opengl_display = new TOpenGLDisplay(this);
		widget_layout->addWidget(d_opengl_display,0,0,6,1);
		
		d_gldisplay_error_label = 0;
	}
	else
	{
		d_gldisplay_error_label = new QLabel(this);
		d_gldisplay_error_label->setWordWrap(true);
		d_gldisplay_error_label->setAlignment(Qt::AlignLeading|Qt::AlignLeft|Qt::AlignTop);
		widget_layout->addWidget(d_gldisplay_error_label,0,0,6,1);
		
		d_opengl_display = 0;
	}

	d_scale_to_fit_button = new QToolButton(this);
	d_scale_to_fit_button->setFixedSize(24,24);
	d_scale_to_fit_button->setCheckable(false);
	widget_layout->addWidget(d_scale_to_fit_button,0,1,1,1);

	d_pan_button = new QToolButton(this);
	d_pan_button->setFixedSize(24,24);
	d_pan_button->setCheckable(true);
	widget_layout->addWidget(d_pan_button,1,1,1,1);

	d_rotate_2d_button = new QToolButton(this);
	d_rotate_2d_button->setFixedSize(24,24);
	d_rotate_2d_button->setCheckable(true);
	widget_layout->addWidget(d_rotate_2d_button,2,1,1,1);

	d_rotate_3d_button = new QToolButton(this);
	d_rotate_3d_button->setFixedSize(24,24);
	d_rotate_3d_button->setCheckable(true);
	widget_layout->addWidget(d_rotate_3d_button,3,1,1,1);
	
	widget_vspacer = new QSpacerItem(0,0,QSizePolicy::Minimum,QSizePolicy::Expanding);
	widget_layout->addItem(widget_vspacer,4,1,1,1);

	d_scale_to_fit_button->setText(QStringLiteral("..."));
	d_pan_button->setText(QStringLiteral("..."));
	d_rotate_2d_button->setText(QStringLiteral("..."));
	d_rotate_3d_button->setText(QStringLiteral("..."));
	
	d_scale_to_fit_button->setIcon(QIcon(":/graphics/scale_to_fit_24x24.png"));
	d_pan_button->setIcon(QIcon(":/graphics/pan_24x24.png"));
	d_rotate_2d_button->setIcon(QIcon(":/graphics/rotate2d_24x24.png"));
	d_rotate_3d_button->setIcon(QIcon(":/graphics/rotate3d_24x24.png"));
	
	d_scale_to_fit_button->setToolTip(QStringLiteral("Scale to Fit"));
	d_pan_button->setToolTip(QStringLiteral("Pan Mode"));
	d_rotate_2d_button->setToolTip(QStringLiteral("Rotate 2D Mode"));
	d_rotate_3d_button->setToolTip(QStringLiteral("Rotate 3D Mode"));
	
	if(d_gldisplay_error_label)
	{
		gl_error_text = QStringLiteral("<html><head/><body>");
		gl_error_text += QStringLiteral("<p><span style=\" font-weight:700;\">ERR:  OpenGL version not supported.  No graphical display will be shown.</span></p>");
		gl_error_text += QStringLiteral("<p><span style=\" font-weight:500; text-decoration: underline;\">OpenGL Version Text:</span></p>");
		gl_error_text += QString("<p>%1</p>").arg(gl_version_text);
		gl_error_text += QStringLiteral("<p><span style=\" font-weight:500; text-decoration: underline;\">OpenGL Extensions:</span></p>");
		gl_error_text += QString("<p>%1</p>").arg(gl_extensions_text);
		gl_error_text += QStringLiteral("</body></html>");
		
		d_gldisplay_error_label->setText(gl_error_text);
	}
	
	d_pan_button->setChecked(true);
	
	connect(d_scale_to_fit_button,&QToolButton::clicked,this,&TOpenGLDisplayWidget::Scale_To_Fit);
	connect(d_pan_button,&QToolButton::toggled,this,&TOpenGLDisplayWidget::Pan);
	connect(d_rotate_2d_button,&QToolButton::toggled,this,&TOpenGLDisplayWidget::Rotate2d);
	connect(d_rotate_3d_button,&QToolButton::toggled,this,&TOpenGLDisplayWidget::Rotate3d);

	if(d_opengl_display)
	{
		connect(d_opengl_display,&TOpenGLDisplay::Model_Click_Point,this,&TOpenGLDisplayWidget::Model_Item_Clicked);
	}
}

TOpenGLDisplayWidget::~TOpenGLDisplayWidget(void)
{
}

void TOpenGLDisplayWidget::Define_Machine_Volume(
	const TVector3						&pnt_min,
	const TVector3						&pnt_max)
{
	double								dval;
	
	if(!d_opengl_display)
	{
		return;
	}
	
	d_max_xyz = pnt_max;
	d_min_xyz = pnt_min;
	
	if(d_min_xyz.x > d_max_xyz.x)
	{
		dval = d_max_xyz.x;
		d_max_xyz.x = d_min_xyz.x;
		d_min_xyz.x = dval;
	}
	
	if(d_min_xyz.y > d_max_xyz.y)
	{
		dval = d_max_xyz.y;
		d_max_xyz.y = d_min_xyz.y;
		d_min_xyz.y = dval;
	}
	
	if(d_min_xyz.z > d_max_xyz.z)
	{
		dval = d_max_xyz.z;
		d_max_xyz.z = d_min_xyz.z;
		d_min_xyz.z = dval;
	}
	
	d_opengl_display->Define_Machine_Volume(d_min_xyz,d_max_xyz);
}

void TOpenGLDisplayWidget::Clear_Render_Measurements(void)
{
	if(d_opengl_display)
	{
		d_opengl_display->Clear_Render_Items();
	}
}

void TOpenGLDisplayWidget::Clear_Render_Volume(void)
{
	if(d_opengl_display)
	{
		d_opengl_display->Clear_Render_Volume();
	}
}

void TOpenGLDisplayWidget::Render_Measurements(
	TMeasurement						* const measurement)
{
	std::vector<TOpenGLGeometryFactory::TOpenGLTriangleVertex> triangles;
	TOpenGLGeometryFactory				geometry;
	TOpenGLDisplay::TBallBar			ballbar;
	TOpenGLDisplay::TMCGGauge			mcg;
	TOpenGLDisplay::TISOGauge			iso;
	TOpenGLDisplay::TLaser				laser;
	TMeasurement::TMeasurementItem		meas_item;
	TVector3							center;
	TVector3							pnt,p1,p2,p3,p4;
	TVector3							axis;
	int									cntr;
	int									count;
	int									base_id;
	TReferenceSystem					refsys;
	double								offset;
	double								length,length_xy,length_z;
	const double						GAUGE_BLOCK_WIDTH(40);
	const double						GAUGE_BLOCK_THICKNESS(10);
	const double						GAUGE_BLOCK_GAP(5);
	const double						PROBE_OFFSET_DISTANCE(30);
	const double						PROBE_STYLUS_RADIUS(2.5);
	const double						SELECTION_DISTANCE(2);
	const double						ZERO_EPSILON(0.000001);
	const double						MIN_PROBE_OFFSET_FOR_KNUCKLE(10);
	
	assert(measurement);
	
	if(!d_opengl_display)
	{
		return;
	}
	
	geometry.Set_Geometry_Resolution(3);
	base_id = measurement->Id();
	
	switch(measurement->Type())
	{
		case Types::MEASUREMENT_LASER:
			
			for(cntr = 0;cntr < measurement->Measurement_Item_Count();++cntr)
			{
				laser.cube1.clear();
				laser.shaft.clear();
				laser.cube2.clear();
				
				laser.probe_z_stem.clear();
				laser.probe_joint.clear();
				laser.probe_xy_stem.clear();
				laser.probe_ruby.clear();
				
				meas_item = measurement->Measurement_Item(cntr);
				
				if(cntr == 0)
				{
					// draw starting point
					laser.cube1 = geometry.RenderPolygon(meas_item.start_actual,meas_item.axis_actual,TVector3(0.577,0.577,0.577),GAUGE_BLOCK_WIDTH,GAUGE_BLOCK_THICKNESS,4);
					
					// draw stylus
					p1 = p2 = meas_item.probe_offset;
					p1.z = 0;
					p2.x = 0;
					p2.y = 0;
					
					length_xy = p1.Length();
					length_z = p2.Length();
					
					if(length_xy > MIN_PROBE_OFFSET_FOR_KNUCKLE)
					{
						p1.Normal();
						
						p3 = meas_item.start_actual - p1 * PROBE_OFFSET_DISTANCE;
						laser.probe_ruby = geometry.RenderSphere(p3,PROBE_STYLUS_RADIUS);
						
						pnt = p3 - p1 * length_xy/2.0;
						laser.probe_xy_stem = geometry.RenderCylinder(pnt,p1,1.5,length_xy);
						
						pnt = p3 - p1 * length_xy;
						laser.probe_joint = geometry.RenderPolygon(pnt,TVector3(0,0,1),p1,5,5,8);
						
						pnt.z += length_z/2.0;
						laser.probe_z_stem = geometry.RenderCylinder(pnt,TVector3(0,0,1),1.5,length_z);

					}
					else if(length_z > ZERO_EPSILON)
					{
						p2.Normal();
						
						p3 = meas_item.start_actual - p2 * PROBE_OFFSET_DISTANCE;
						laser.probe_ruby = geometry.RenderSphere(p3,PROBE_STYLUS_RADIUS);
						
						pnt = p3 - p2 * length_z/2.0;
						laser.probe_z_stem = geometry.RenderCylinder(pnt,p2,1.5,length_z);
					}
				}
				
				laser.cube2 = geometry.RenderPolygon(meas_item.end_actual,meas_item.axis_actual,TVector3(0.577,0.577,0.577),GAUGE_BLOCK_WIDTH,GAUGE_BLOCK_THICKNESS,4);
				
				if(cntr == (measurement->Measurement_Item_Count() - 1))
				{
					// draw centerline cylinder
					center = (meas_item.start_actual + meas_item.end_actual) / 2.0;
					laser.shaft = geometry.RenderCylinder(center,meas_item.axis_actual,4,meas_item.length_actual);
				}
				
				laser.id = base_id + cntr;
				laser.selected = false;
				laser.error = measurement->Error_State();

				p1 = meas_item.axis_actual;
				p2 = TVector3(0,0,1) * p1;
				
				if(p2.Length() < ZERO_EPSILON)
				{
					p2 = p1 * TVector3(0,1,0);
				}
				
				p2.Normal();
				
				p3 = p1 * p2;
				
				laser.min_xyz = meas_item.end_actual - meas_item.axis_actual * SELECTION_DISTANCE - meas_item.axis_actual * GAUGE_BLOCK_THICKNESS / 2;
				laser.min_xyz -= p2 * (SELECTION_DISTANCE * GAUGE_BLOCK_WIDTH / 2);
				laser.min_xyz -= p3 * (SELECTION_DISTANCE * GAUGE_BLOCK_WIDTH / 2);
				
				laser.max_xyz = meas_item.end_actual + meas_item.axis_actual * SELECTION_DISTANCE + meas_item.axis_actual * GAUGE_BLOCK_THICKNESS / 2;
				laser.max_xyz += p2 * (SELECTION_DISTANCE * GAUGE_BLOCK_WIDTH / 2);
				laser.max_xyz += p3 * (SELECTION_DISTANCE * GAUGE_BLOCK_WIDTH / 2);
				
				laser.axis1 = meas_item.axis_actual;
				laser.axis2 = p2;
				laser.axis3 = p3;
				
				d_opengl_display->Add_Laser(laser);
			}
			
			break;
			
		case Types::MEASUREMENT_MCG:
			
			for(cntr = 0;cntr < measurement->Measurement_Item_Count();++cntr)
			{				
				meas_item = measurement->Measurement_Item(cntr);

				mcg.center = meas_item.start_actual;
				mcg.axis = meas_item.axis_actual;
				mcg.length = meas_item.length_nominal;
				
				mcg.id = base_id + cntr;
				mcg.selected = false;
				mcg.error = measurement->Error_State();

				mcg.min_xyz = meas_item.start_actual;
				mcg.min_xyz -= TVector3(1,0,0) * meas_item.length_actual/2;
				mcg.min_xyz -= TVector3(0,1,0) * meas_item.length_actual/2;
				mcg.min_xyz -= TVector3(0,0,1) * meas_item.length_actual/2;
				
				mcg.max_xyz = meas_item.start_actual;
				mcg.max_xyz += TVector3(1,0,0) * meas_item.length_actual/2;
				mcg.max_xyz += TVector3(0,1,0) * meas_item.length_actual/2;
				mcg.max_xyz += TVector3(0,0,1) * meas_item.length_actual/2;
				
				mcg.min_xyz = meas_item.start_actual;
				
				mcg.max_xyz = meas_item.end_actual;

				d_opengl_display->Add_MCGGauge(mcg);
			}
			
			break;
			
		case Types::MEASUREMENT_ISO:
			
			count = measurement->Measurement_Item_Count();
			
			for(cntr = 0;cntr < count;++cntr)
			{
				iso.face1.clear();
				iso.face2.clear();
				iso.face3.clear();
				iso.face4.clear();
				iso.face5.clear();
				iso.face6.clear();
				
				iso.probe_stem.clear();
				iso.probe_ruby.clear();
				
				meas_item = measurement->Measurement_Item(cntr);
				
				if(cntr == 0)
				{
					refsys.Skew1(meas_item.axis_actual,TReferenceSystem::XPLUS);

					// draw stylus
					// special case for ISO measurement.  stylus drawn as a straight line
					// from starting position to stylus
					p1 = meas_item.probe_offset;
					
					length = p1.Length();
					
					if(length > ZERO_EPSILON)
					{
						p1.Normal();
						
						p3 = meas_item.start_actual - p1 * (PROBE_OFFSET_DISTANCE + (GAUGE_BLOCK_THICKNESS + GAUGE_BLOCK_GAP) * 2);
						iso.probe_ruby = geometry.RenderSphere(p3,PROBE_STYLUS_RADIUS);
						
						pnt = p3 - p1 * length/2.0;
						iso.probe_stem = geometry.RenderCylinder(pnt,p1,1.5,length);
										
						refsys.Skew2(p1,TReferenceSystem::ZMINUS,TReferenceSystem::XPLUS);
					}
				}

				
				refsys.Set_Origin(meas_item.start_actual);
				
				// put the longest in the middle and stagger the rest
				if(cntr % 2)
				{
					offset = -static_cast<double>((count - cntr) / 2) * (GAUGE_BLOCK_THICKNESS + GAUGE_BLOCK_GAP);
				}
				else
				{
					offset = static_cast<double>((count - cntr) / 2) * (GAUGE_BLOCK_THICKNESS + GAUGE_BLOCK_GAP);
				}
				
				p1 = refsys.ToWorld(TVector3(0,-GAUGE_BLOCK_WIDTH/2,-GAUGE_BLOCK_THICKNESS/2 - offset));
				p2 = refsys.ToWorld(TVector3(0,-GAUGE_BLOCK_WIDTH/2,GAUGE_BLOCK_THICKNESS/2 - offset));
				p3 = refsys.ToWorld(TVector3(0,GAUGE_BLOCK_WIDTH/2,GAUGE_BLOCK_THICKNESS/2 - offset));
				p4 = refsys.ToWorld(TVector3(0,GAUGE_BLOCK_WIDTH/2,-GAUGE_BLOCK_THICKNESS/2 - offset));
				
				iso.face1 = geometry.RenderPlane(p1,p2,p3,p4);
				
				p1 = refsys.ToWorld(TVector3(meas_item.length_actual,-GAUGE_BLOCK_WIDTH/2,GAUGE_BLOCK_THICKNESS/2 - offset));
				p2 = refsys.ToWorld(TVector3(meas_item.length_actual,-GAUGE_BLOCK_WIDTH/2,-GAUGE_BLOCK_THICKNESS/2 - offset));
				p3 = refsys.ToWorld(TVector3(meas_item.length_actual,GAUGE_BLOCK_WIDTH/2,-GAUGE_BLOCK_THICKNESS/2 - offset));
				p4 = refsys.ToWorld(TVector3(meas_item.length_actual,GAUGE_BLOCK_WIDTH/2,GAUGE_BLOCK_THICKNESS/2 - offset));
				
				iso.face2 = geometry.RenderPlane(p1,p2,p3,p4);
				
				p1 = refsys.ToWorld(TVector3(0,-GAUGE_BLOCK_WIDTH/2,GAUGE_BLOCK_THICKNESS/2 - offset));
				p2 = refsys.ToWorld(TVector3(0,-GAUGE_BLOCK_WIDTH/2,-GAUGE_BLOCK_THICKNESS/2 - offset));
				p3 = refsys.ToWorld(TVector3(meas_item.length_actual,-GAUGE_BLOCK_WIDTH/2,-GAUGE_BLOCK_THICKNESS/2 - offset));
				p4 = refsys.ToWorld(TVector3(meas_item.length_actual,-GAUGE_BLOCK_WIDTH/2,GAUGE_BLOCK_THICKNESS/2 - offset));
				
				iso.face3 = geometry.RenderPlane(p1,p2,p3,p4);
				
				p1 = refsys.ToWorld(TVector3(0,GAUGE_BLOCK_WIDTH/2,-GAUGE_BLOCK_THICKNESS/2 - offset));
				p2 = refsys.ToWorld(TVector3(0,GAUGE_BLOCK_WIDTH/2,GAUGE_BLOCK_THICKNESS/2 - offset));
				p3 = refsys.ToWorld(TVector3(meas_item.length_actual,GAUGE_BLOCK_WIDTH/2,GAUGE_BLOCK_THICKNESS/2 - offset));
				p4 = refsys.ToWorld(TVector3(meas_item.length_actual,GAUGE_BLOCK_WIDTH/2,-GAUGE_BLOCK_THICKNESS/2 - offset));
				
				iso.face4 = geometry.RenderPlane(p1,p2,p3,p4);
				
				p1 = refsys.ToWorld(TVector3(0,GAUGE_BLOCK_WIDTH/2,GAUGE_BLOCK_THICKNESS/2 - offset));
				p2 = refsys.ToWorld(TVector3(0,-GAUGE_BLOCK_WIDTH/2,GAUGE_BLOCK_THICKNESS/2 - offset));
				p3 = refsys.ToWorld(TVector3(meas_item.length_actual,-GAUGE_BLOCK_WIDTH/2,GAUGE_BLOCK_THICKNESS/2 - offset));
				p4 = refsys.ToWorld(TVector3(meas_item.length_actual,GAUGE_BLOCK_WIDTH/2,GAUGE_BLOCK_THICKNESS/2 - offset));
				
				iso.face5 = geometry.RenderPlane(p1,p2,p3,p4);
				
				p1 = refsys.ToWorld(TVector3(0,-GAUGE_BLOCK_WIDTH/2,-GAUGE_BLOCK_THICKNESS/2 - offset));
				p2 = refsys.ToWorld(TVector3(0,GAUGE_BLOCK_WIDTH/2,-GAUGE_BLOCK_THICKNESS/2 - offset));
				p3 = refsys.ToWorld(TVector3(meas_item.length_actual,GAUGE_BLOCK_WIDTH/2,-GAUGE_BLOCK_THICKNESS/2 - offset));
				p4 = refsys.ToWorld(TVector3(meas_item.length_actual,-GAUGE_BLOCK_WIDTH/2,-GAUGE_BLOCK_THICKNESS/2 - offset));
				
				iso.face6 = geometry.RenderPlane(p1,p2,p3,p4);

				iso.id = base_id + cntr;
				iso.selected = false;
				iso.error = measurement->Error_State();

				iso.min_xyz = refsys.ToWorld(TVector3(-SELECTION_DISTANCE,
													  -GAUGE_BLOCK_WIDTH/2 - SELECTION_DISTANCE,
													  -GAUGE_BLOCK_THICKNESS/2 - SELECTION_DISTANCE - offset));
				
				iso.max_xyz = refsys.ToWorld(TVector3(meas_item.length_actual + SELECTION_DISTANCE,
													  GAUGE_BLOCK_WIDTH/2 + SELECTION_DISTANCE,
													  GAUGE_BLOCK_THICKNESS/2 + SELECTION_DISTANCE - offset));

				
				iso.axis1 = refsys.Rotation_Mat().X();
				iso.axis2 = refsys.Rotation_Mat().Y();
				iso.axis3 = refsys.Rotation_Mat().Z();
				
				d_opengl_display->Add_ISOGauge(iso);
			}
			
			break;
			
		case Types::MEASUREMENT_BALLBAR:
			
			ballbar.selected = false;
			ballbar.id = base_id;

			for(cntr = 0;cntr < measurement->Measurement_Item_Count();++cntr)
			{
				ballbar.ballbar_shaft.clear();
				ballbar.ballbar_spheres.clear();
				ballbar.probe_stem.clear();
				ballbar.probe_ruby.clear();
				
				meas_item = measurement->Measurement_Item(cntr);
				
				axis = meas_item.start_actual - meas_item.end_actual;
				
				length = axis.Length();
				
				if(!(length > ZERO_EPSILON))
				{
					break;
				}
				
				axis.Normal();
				
				center = (meas_item.start_actual + meas_item.end_actual) / 2.0;
				
				if(length > 148)	// minimum length or it won't look right
				{
					ballbar.ballbar_spheres = geometry.RenderSphere(meas_item.start_actual,19.5);
					
					pnt = meas_item.start_actual - axis * 34.75;
					triangles = geometry.RenderCylinder(pnt,axis,10,50);
					ballbar.ballbar_spheres.insert(ballbar.ballbar_spheres.end(),triangles.begin(),triangles.end());
					
					pnt = meas_item.start_actual - axis * 66.75;
					triangles = geometry.RenderPolygon(pnt,axis,TVector3(0,1,0),15,14,6);
					ballbar.ballbar_spheres.insert(ballbar.ballbar_spheres.end(),triangles.begin(),triangles.end());
					
					ballbar.ballbar_shaft = geometry.RenderCylinder(center,axis,14,length - 147.52);
					
					pnt = meas_item.end_actual + axis * 66.75;
					triangles = geometry.RenderPolygon(pnt,axis,TVector3(0,1,0),15,14,6);
					ballbar.ballbar_spheres.insert(ballbar.ballbar_spheres.end(),triangles.begin(),triangles.end());
					
					pnt = meas_item.end_actual + axis * 34.75;
					triangles = geometry.RenderCylinder(pnt,axis,10,50);
					ballbar.ballbar_spheres.insert(ballbar.ballbar_spheres.end(),triangles.begin(),triangles.end());
					
					triangles = geometry.RenderSphere(meas_item.end_actual,19.5);
					ballbar.ballbar_spheres.insert(ballbar.ballbar_spheres.end(),triangles.begin(),triangles.end());
				}
				else
				{
					// draw a shorter version
					ballbar.ballbar_spheres = geometry.RenderSphere(meas_item.start_actual,19.5);
					
					ballbar.ballbar_shaft = geometry.RenderCylinder(center,axis,10,length-15);
					
					triangles = geometry.RenderSphere(meas_item.end_actual,19.5);
					ballbar.ballbar_spheres.insert(ballbar.ballbar_spheres.end(),triangles.begin(),triangles.end());
				}
				
				// draw stylus
				p1 = p2 = meas_item.probe_offset;
				p1.z = 0;
				p2.x = 0;
				p2.y = 0;
				
				length_xy = p1.Length();
				length_z = p2.Length();
				
				if(length_xy > MIN_PROBE_OFFSET_FOR_KNUCKLE)
				{
					p1.Normal();
					
					p3 = meas_item.start_actual - p1 * PROBE_OFFSET_DISTANCE;
					ballbar.probe_ruby = geometry.RenderSphere(p3,PROBE_STYLUS_RADIUS);
					
					pnt = p3 - p1 * length_xy/2.0;
					ballbar.probe_stem = geometry.RenderCylinder(pnt,p1,1.5,length_xy);
					
					pnt = p3 - p1 * length_xy;
					triangles = geometry.RenderPolygon(pnt,TVector3(0,0,1),p1,5,5,8);
					ballbar.probe_stem.insert(ballbar.probe_stem.end(),triangles.begin(),triangles.end());
					
					pnt.z += length_z/2.0;
					triangles = geometry.RenderCylinder(pnt,TVector3(0,0,1),1.5,length_z);
					ballbar.probe_stem.insert(ballbar.probe_stem.end(),triangles.begin(),triangles.end());
					
				}
				else if(length_z > ZERO_EPSILON)
				{
					p2.Normal();
					
					p3 = meas_item.start_actual - p2 * PROBE_OFFSET_DISTANCE;
					ballbar.probe_ruby = geometry.RenderSphere(p3,PROBE_STYLUS_RADIUS);
					
					pnt = p3 - p2 * length_z/2.0;
					ballbar.probe_stem = geometry.RenderCylinder(pnt,p2,1.5,length_z);
				}
				
				
				p1 = axis;
				p2 = p1 * TVector3(0,0,1);
				
				if(p2.Length() < ZERO_EPSILON)
				{
					p2 = p1 * TVector3(0,1,0);
				}
				
				p2.Normal();
				
				p3 = p1 * p2;
				p2 = p3 * p1;
				
				ballbar.min_xyz = meas_item.start_actual + axis * 21;	// increase by just over half of sphere radius
				ballbar.max_xyz = meas_item.end_actual - axis * 21;		// increase by just over half of sphere radius
				
				ballbar.name = measurement->Name();
				
				// calculate where name and leader line is drawn
				pnt = (center + meas_item.start_actual)/2.0;
				
				// find right angle vector in XY plane
				p1 = axis * TVector3(0,0,1);
				
				if(p1.Length() < ZERO_EPSILON)
				{
					p1 = axis * TVector3(0,1,0);
				}
				
				p1.Normal();
				
				// second label is 10% of ball bar length
				length_xy = length * 0.10;
				
				if(length_xy < 20)
				{
					length_xy = 20;	// minimum 20 mm
				}
				
				ballbar.label_coordinates = pnt + p1 * length_xy;
				ballbar.error = measurement->Error_State();

				d_opengl_display->Add_BallBar(ballbar);
			}			
			break;
	}
}

void TOpenGLDisplayWidget::Update(
	const bool							scale_to_fit)
{
	if(!d_opengl_display)
	{
		return;
	}
	
	d_opengl_display->update();

	if(scale_to_fit)
	{
		d_opengl_display->Scale_To_Fit();
	}
}

void TOpenGLDisplayWidget::Clear_Selection()
{
	if(d_opengl_display)
	{
		d_opengl_display->Clear_Selection();
	}
}

void TOpenGLDisplayWidget::Set_Selection(
	const int							id)
{
	if(d_opengl_display)
	{
		d_opengl_display->Set_Selection(id);
	}
}

void TOpenGLDisplayWidget::Scale_To_Fit(void)
{
	if(d_opengl_display)
	{
		d_opengl_display->Scale_To_Fit();
	}
}

void TOpenGLDisplayWidget::Rotate2d(
	bool								toggled)
{
	bool								prev_state;
	
	if(toggled)
	{
		prev_state = d_pan_button->blockSignals(true);
		d_pan_button->setChecked(false);
		d_pan_button->blockSignals(prev_state);
	
		prev_state = d_rotate_3d_button->blockSignals(true);
		d_rotate_3d_button->setChecked(false);
		d_rotate_3d_button->blockSignals(prev_state);
		
		if(d_opengl_display)
		{
			d_opengl_display->Set_Rotate_2D();
		}
	}
	else
	{
		prev_state = d_pan_button->blockSignals(true);
		d_pan_button->setChecked(true);
		d_pan_button->blockSignals(prev_state);
		
		prev_state = d_rotate_2d_button->blockSignals(true);
		d_rotate_2d_button->setChecked(false);
		d_rotate_2d_button->blockSignals(prev_state);
		
		prev_state = d_rotate_3d_button->blockSignals(true);
		d_rotate_3d_button->setChecked(false);
		d_rotate_3d_button->blockSignals(prev_state);
		
		if(d_opengl_display)
		{
			d_opengl_display->Set_Pan();
		}
	}
}

void TOpenGLDisplayWidget::Rotate3d(
	bool								toggled)
{
	bool								prev_state;
	
	if(toggled)
	{
		prev_state = d_pan_button->blockSignals(true);
		d_pan_button->setChecked(false);
		d_pan_button->blockSignals(prev_state);
		
		prev_state = d_rotate_2d_button->blockSignals(true);
		d_rotate_2d_button->setChecked(false);
		d_rotate_2d_button->blockSignals(prev_state);
		
		if(d_opengl_display)
		{
			d_opengl_display->Set_Rotate_3D();
		}
	}
	else
	{
		prev_state = d_pan_button->blockSignals(true);
		d_pan_button->setChecked(true);
		d_pan_button->blockSignals(prev_state);
		
		prev_state = d_rotate_2d_button->blockSignals(true);
		d_rotate_2d_button->setChecked(false);
		d_rotate_2d_button->blockSignals(prev_state);
		
		prev_state = d_rotate_3d_button->blockSignals(true);
		d_rotate_3d_button->setChecked(false);
		d_rotate_3d_button->blockSignals(prev_state);
		
		if(d_opengl_display)
		{
			d_opengl_display->Set_Pan();
		}
	}

}

void TOpenGLDisplayWidget::Pan(
	bool								toggled)
{
	bool								prev_state;
	
	if(toggled)
	{
		prev_state = d_rotate_3d_button->blockSignals(true);
		d_rotate_3d_button->setChecked(false);
		d_rotate_3d_button->blockSignals(prev_state);
		
		prev_state = d_rotate_2d_button->blockSignals(true);
		d_rotate_2d_button->setChecked(false);
		d_rotate_2d_button->blockSignals(prev_state);
		
		if(d_opengl_display)
		{
			d_opengl_display->Set_Pan();
		}
	}
	else
	{
		prev_state = d_pan_button->blockSignals(true);
		d_pan_button->setChecked(true);
		d_pan_button->blockSignals(prev_state);
		
		prev_state = d_rotate_2d_button->blockSignals(true);
		d_rotate_2d_button->setChecked(false);
		d_rotate_2d_button->blockSignals(prev_state);
		
		prev_state = d_rotate_3d_button->blockSignals(true);
		d_rotate_3d_button->setChecked(false);
		d_rotate_3d_button->blockSignals(prev_state);
		
		if(d_opengl_display)
		{
			d_opengl_display->Set_Pan();
		}
	}
}

